# Copyright (c) 2016 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Streetifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
""" Beacon/Listener: UDP LAN broadcast for service connection """

import socket
from threading import Thread, enumerate, RLock
import time

from rhcert.environment import Environment

RHCERT_BEACON_PORT = 8009
RHCERT_BEACON_ADDRESS = '255.255.255.255'

class Beacon(Thread):
    def __init__(self, channel, debug=False):
        Thread.__init__(self, name=channel)
        self.daemon=True
        self.debug = debug
        self.payload = socket.gethostname() + ":" + channel
        self.broadcastAddress = RHCERT_BEACON_ADDRESS
        self.port = RHCERT_BEACON_PORT
        try:
            self.port = int(Environment().getServerListenerPort())
        except:
            self.__debugPrint("Error: could not load port from environment")
        self.period = 1 # seconds

    def __debugPrint(self, message):
        if self.debug:
            print(message)

    def isRunning(self):
        for thread in enumerate():
            if thread.name == self.name and thread.isAlive():
                return True
        return False

    def run (self):
        while True:
            try:
                self.__debugPrint('beacon flashing ' + self.payload)
                pinger = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                pinger.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                pinger.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
                pinger.sendto(self.payload, (self.broadcastAddress, self.port))
                time.sleep(self.period)
            except Exception as error:
                self.__debugPrint("Error: %s" % error)

class Antenna(Thread):
    def __init__(self, channel="omni", debug=False):
        Thread.__init__(self, name=channel)
        self.daemon=True
        self.debug = debug
        self.port = RHCERT_BEACON_PORT
        self.period = 1 # seconds
        self.cycles = 10 # try to pick up 10 messages, then quit
        self.channel = channel
        self.senders = dict()
        self.lock = RLock()

    def __debugPrint(self, message):
        if self.debug:
            print(message)

    def getSenders(self):
        self.lock.acquire(blocking=1)
        senders = self.senders.values()
        self.lock.release()
        return senders

    def run (self):
        while self.cycles:
            self.__debugPrint('antenna listening for ' + self.channel)
            listener = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            try:
                listener.bind(('', RHCERT_BEACON_PORT))
                (message, (senderAddress, senderPort)) = listener.recvfrom(1024) # <-- waiting forever
                if message:
                    (hostname, channel) = message.decode('utf-8').split(':', 1)
                    self.__debugPrint("received channel %s from host %s %s" % (channel, hostname, type(hostname)))
                    if (self.channel == "omni") or (self.channel == channel):
                        self.__addSender(senderAddress, senderPort, hostname)
                        self.__debugPrint("%s at %s sent: %s" %(hostname, senderAddress, message))
            except Exception as error:
                self.__debugPrint('failed to bind')
                self.__debugPrint(error)
                listener.close()
                time.sleep(1)
                # raise
                # listener.blocking(0)
            finally:
                listener.close()
            self.cycles = self.cycles-1

    def __addSender(self, address, port, host):
        self.lock.acquire(blocking=1)
        self.__debugPrint("Adding sender %s port %s host %s" %(address, port, host))
        self.senders[address] = (address, port, host)
        self.lock.release()

if __name__ == "__main__":
    antenna = Antenna(debug=True)
    antenna.start()
    wait = 2
    print(("Main thread waiting %s sec.", wait))
    time.sleep(wait)
    beacon = Beacon("rhcert:server", debug=True)
    beacon.start()
    wait = 60
    print(("Main Thread running for %s seconds " % wait))
    time.sleep(wait)


