# Copyright (c) 2018 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
# Client: Class to represent an rhcert client host (AKA System Under Test or SUT)
# Provides a way to interact with the client via the rhcertd daemon
import socket
import time

from rhcert.compatability import XmlrpcServerProxy
from rhcert.environment import Environment
from rhcert.model.task import Task
from rhcert.compatability import rhcert_make_unicode
from rhcert.model.resultsdocument import ResultsDocument

class RHCertClient(object):

    def __init__(self, host, timeout=1, environment=None):
        self.host = host # a rhcert.Host object
        self.environment = environment if environment else Environment()
        self.timeout = timeout # seconds
        self.socketDefaultTimeout = socket.getdefaulttimeout()
        self.clientProxy = XmlrpcServerProxy(('http://%s:%s'
                % (self.host.getConnection(), self.environment.getServerListenerPort())), allow_none=True)

    def getHost(self):
        return self.host

    def __call(self, clientProxyMethod, *argv):
        """  host is a Host object, timeout = 1 # sec """
        socket.setdefaulttimeout(self.timeout)
        try:
            return self.__decodeResponse(clientProxyMethod(*argv))

        except socket.timeout:
            return u'waiting for response'
        except Exception as e:
            print("RHCert Client Exception: ")
            print(e)
        finally:
            socket.setdefaulttimeout(self.socketDefaultTimeout)

        return u''

    def __decodeResponse(self, rawResponse):

        if isinstance(rawResponse, bool):
            return rawResponse

        if not rawResponse:
            return u''

        response = rhcert_make_unicode(rawResponse)
        # sometimes (python3 client) we end up with "'b <real stuff>'"
        if response.startswith('b\''):
             return response.strip()[2:-1] # decode bytes->unicode

        return response

    def getStatus(self):
        response = self.__call(self.clientProxy.getResult)
        # is it a Task for a prompt or a status message
        task = Task()
        task.loadString(response)
        if task.isValid():
            return ("task", task)
        return ("status", response)

    def waitUntilReady(self, timeout=120):
        wait = 3 # sec
        while timeout > 0:
            info, task = self.getStatus()
            if info == "task":
                statusString = task.getParameter("message")
            else:
                statusString = task
            ready = ("ready" or "Finished" in statusString
                    or ("clean" in statusString and "succeeded" in statusString))
            if ready:
                return True

            time.sleep(wait)
            timeout = timeout - wait
        print("Error: %s timeout exceeded!" % self.timeout)
        return False

    def getLog(self):
        log = self.__call(self.clientProxy.getCurrentLog)
        return log if log else None

    def restart(self):
        return self.__call(self.clientProxy.restart)

    def getResults(self):
        resultsXML = self.clientProxy.getCurrentResultsXML()
        try:
            resultsDocument = ResultsDocument()
            resultsDocument.loadString(rhcert_make_unicode(resultsXML))
            return resultsDocument
        except Exception as error:
            print("Error: failed to get results document:")
            print(error)
            print("rhcertXML %s" % type(resultsXML))
            print(resultsXML)
        return None

    def runCLI(self, commandLine):
        """ commandLine is a rhcert-cli/-util command line string """
        return self.__call(self.clientProxy.runCLI, commandLine)

    def runXML(self, xml):
        """ xml is a Task or ResultsDocument """
        return self.__call(self.clientProxy.run, xml)
