# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Streetifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#

import os, time, sys, syslog, socket, copy, shutil, traceback, json, ast
import textwrap

from rhcert.collabrequest import CollabRequest
from rhcert.controller import Controller
from rhcert.tags import Tags, Constants, TestTag, SystemCategories, Programs, TestResult
from rhcert.environment import Environment
from rhcert.model.resultsdocument import ResultsDocument
from rhcert.model.certificationdocument import CertificationDocument
from rhcert.model.testdocument import TestDocument
from rhcert.report import Report
from rhcert.reporthtml import ReportHTML
from rhcert.reportjunit import ReportJunit
from rhcert.planner import Planner
import rhcert.version
from rhcert.command import Command, HwCertCommandException
from rhcert.redhatrelease import RedHatRelease
from rhcert.log import Log
from rhcert.client.workflowCLI import WorkflowCLI
from rhcert.host import Host
from rhcert.localTestServer import LocalTestServer
from rhcert.services.microservices import MicroServices
from rhcert.rhcertStore import RedHatCertificationStore

class TestHarness(Controller):

    def __init__(self, options, echoResponses=True):
        self.options = options
        Controller.__init__(self, self.options.debug, echoResponses)
        Command.setDebug(self.options.debug)
        self.planner = None
        self.ui.setMode(self.options.mode)
        self.environment = Environment()
        self.rhcertStore = RedHatCertificationStore()
        self.makeDirectoryPath(self.environment.getDataDirectory())
        self.redHatRelease = RedHatRelease()
        self.workflowCLI = WorkflowCLI(self.environment, self.rhcertStore, self.options.debug)
        self.resultsDocument = None
        self.testCompleteCallback = None

    def setTestCompleteCallback(self, callback):
        self.testCompleteCallback = callback

    def _initializeHost(self):
        return Host(self.options.serverIP) if self.options.serverIP else Host(self.options.server)

    def _initializePlanner(self):
        if not self.planner:
            self.planner = Planner(self.options)
            self.planner.analyse(self.environment.getSuiteDirectory())

    def __setPlannerProgram(self, certificationDocument):
        if certificationDocument.getCertificationProgram():
            self.planner.setTestSuites(["redhat-certification-" + certificationDocument.getCertificationProgram()])
        elif self.options.debug != Constants.off:
            print("No program set for planning")

    def doVersion(self):
        print("rhcert version %s, release %s" % (rhcert.version.version, rhcert.version.release))
        return True

    def load(self):
        if not self.resultsDocument:
            try:
                self.resultsDocument = ResultsDocument(debugging=self.options.debug)
                if not self.resultsDocument.load(self.options.results or self.environment.getResultsPath()):
                    self.resultsDocument = None
            except IOError:
                self.resultsDocument = None
        return

    def loadFiltered(self):
        xslFilterPath = os.path.join(self.environment.getServerDirectory(), "xsl/default.xsl")
        xsltParameters = { "devices":"\'hide\'",
                          "device-classes": "\'hide\'",
                          "rpath":"\'%s\'" % self.environment.getResultsPath(),
                          "server-email":"\'%s\'" % self.environment.getServerEMail()}
        try:
            self.resultsDocument = ResultsDocument(debugging=self.options.debug)
            self.resultsDocument.loadFiltered(self.environment.getResultsPath(), xslFilterPath, xsltParameters)
            if not self.resultsDocument.document:
                print("Error: could not load %s filtered with %s" % (self.environment.getResultsPath(), xslFilterPath))
                self.resultsDocument = None
        except IOError as error:
            print("Error loading filtered:")
            print(error)
            self.resultsDocument = None

    def findTest(self, collabRequest):
        suitesDir = self.environment.getSuiteDirectory()
        if suitesDir not in sys.path:
            sys.path.append(suitesDir)
        try:
            trail = collabRequest.test_package_trail
            module = __import__(trail, globals(), locals(), [collabRequest.test_name])
        except ImportError:
            raise ValueError("Invalid package trail")
        if hasattr(module, collabRequest.test_name):
            return getattr(module, collabRequest.test_name)
        else:
            raise ValueError("Invalid test name: %s" % collabRequest.test_name)

    def findFunction(self, collabRequest):
        TestClass = self.findTest(collabRequest)
        # see if it was in error
        if hasattr(TestClass, "__getitem__"):
            return TestClass
        if hasattr(TestClass, collabRequest.function_name):
            return getattr(TestClass, collabRequest.function_name)
        else:
            raise ValueError("Invalid function name.  %s function not found in %s test class.\n\nHere are attributes of the class:%s" %(collabRequest.function_name, str(TestClass), dir(TestClass)))

    def doCollab(self):
        testName = self.options.testName
        testPackageTrail = self.options.testPackageTrail
        functionName = self.options.functionName
        args = self.options.args
        kwargs = ast.literal_eval(self.options.kwargs)
        collabRequest = CollabRequest(testName, testPackageTrail, functionName)
        if args:
            collabRequest.args = args
        else:
            args = []
        if kwargs:
            collabRequest.kwargs = kwargs
        else:
            akwargs = {}

        function = self.findFunction(collabRequest)
        functionResult = function(*args, **kwargs)
        return functionResult

    def doRegister(self):
        """ report certification document to the server """
        # NOTE: In case, SUT is registered after running the test, then we over-
        # write existing results file which will lead to loss of data. To avoid
        # that we need to save the existing data and register the host.
        if os.path.exists(self.environment.getResultsPath()) and self.options.mode == Constants.normal:
            self.resultsDocument = ResultsDocument()
            self.resultsDocument.load(self.environment.getResultsPath())

            # if there are test run results, and we're not running auto mode
            if self.resultsDocument.getNumberOfTestRuns():
                msg = """
                Registering the host will delete the current test results.

                Would you like to save the results to a local test server first?"""
                msg = textwrap.dedent(msg)

                if self.ui.promptConfirm(msg):
                    self.doSave()

        self.doClean(cleanAll=True, force=True)

        self.resultsDocument = ResultsDocument()
        self.resultsDocument.new()
        self.options.mode = Constants.auto

        # do discovery
        self._initializePlanner()
        self.planner.plan(None)
        self.doDiscover()
        self.planner.saveNewPlan(self.resultsDocument)
        self.resultsDocument.setPlanTime(self.getCurrentUTCTime())
        self.resultsDocument.save(self.environment.getResultsPath())
        (savedFilePath, storePath) = self.saveToStore()
        if savedFilePath:
            success = self.saveToServer(savedFilePath, self.options.serverPath)
            os.remove(savedFilePath + ".gz")
            return success
        else:
            return False

    def doDiscover(self):
        self.resultsDocument = ResultsDocument(debugging=self.options.debug)

        # try and recover old hardware and os info
        if os.path.exists(self.environment.getCertificationPath()):
            certificationDocument = CertificationDocument()
            certificationDocument.load(self.environment.getCertificationPath())
            self.resultsDocument.copy(certificationDocument)
        else:
            self.resultsDocument.new()
            self.getBiosInfo()
            print("Hardware: %s %s %s" % (self.resultsDocument.getHardware(Tags.vendor),
                                          self.resultsDocument.getHardware(Tags.make),
                                          self.resultsDocument.getHardware(Tags.model)))
            self.getOSInfo()

        if self.options.server or self.options.serverIP:
            localTestServer = LocalTestServer(self._initializeHost())
            if localTestServer.verify():
                self.localTestServer = localTestServer
                self.resultsDocument.setTestServer(self.localTestServer.getHostname())
                self.resultsDocument.setTestServerIP(self.options.serverIP)
                if localTestServer.getClientIPAddress():
                    print("Setting server IP to %s" % localTestServer.getClientIPAddress())
                    self.resultsDocument.getHostOS().setIPAddress( localTestServer.getClientIPAddress())
                else:
                    print("No server IP address")
            else:
                self.localTestServer = None

        self.resultsDocument.setDiscoverTime(self.getCurrentUTCTime())
        self.resultsDocument.save(self.environment.getResultsPath())

        # save the certification info off to certification.xml
        certificationDocument = CertificationDocument()
        certificationDocument.copy(self.resultsDocument)
        certificationDocument.save(self.environment.getCertificationPath())

        return True

    def getOSInfo(self):
        self.redHatRelease.updateProductFromService()
        hostOS = self.resultsDocument.getHostOS()
        hostOS.setCodeName(self.redHatRelease.getCodeName())
        if self.redHatRelease.getVersionPointUpdate():
            hostOS.setRelease(self.redHatRelease.getVersionPointUpdate())
        else:
            print("Error: could not determine OS release number")
            return False
        if self.redHatRelease.getProduct():
            hostOS.setProduct(self.redHatRelease.getProduct())
        else:
            print("Error: could not determine OS product name")
            return False
        print("OS: %s %s" % (hostOS.getName(), hostOS.getRelease()))
        if not hostOS.getName():
            hostname = socket.gethostname()
            if hostname:
                hostOS.setName(hostname)

        # set the certification id to blank for now
        self.resultsDocument.setCertificationID(0)
        self.resultsDocument.setBugID(0)
        self.resultsDocument.setNodeID(0)

    def checkVirtualization(self):
        # checks if we happen to be running within a para-virtualized guest
        # use the method introduced in virt-what application script -
        # http://git.annexia.org/?p=virt-what.git;a=blob;f=virt-what.in;h=d037a992f57a4154d89007f0b93538f41425639e;hb=HEAD#l231
        #
        # If its a paravirtualized VM, there should be a /proc/xen directory.
        # If /proc/xen/capabilities file exists & if it contains "control_d" then,
        # it's running under dom0 else its running on a domU (PARA VIRTUALIZATION).
        # And for hvm irrespective of above checks, need to look for the SMBIOS or DMI entry point.

        self.virtualization = Constants.paravirtualization if os.path.exists('/proc/xen') else None

    def getVirtualization(self):
        return self.virtualization

    def isRealtime(self):
        if os.uname()[3].find("RT ") == -1:
            self.realtime = False
        else:
            self.realtime = True
        return self.realtime

    def doPlan(self):
        self.ui.statusMessage("Verifying test plan")
        self.checkVirtualization()
        # if self.getVirtualization() == Constants.paravirtualization:
        #     if not self.ui.promptContinue("\n  Warning !\n\n  You're running rhcert in a Para-Virtualized guest. "
        #                                   "\n\n  Test results from a Para-Virtualized guest are not valid for "
        #                                   "certification, do you wish to continue?"):
        #         return True
        self.load()
        self._initializePlanner()

        if not self.resultsDocument:
            self.doDiscover()
            tests = self.planner.plan(self.resultsDocument)
            if not tests:
                print("No test planned. Please check if any test suite rpm is installed.")
                return False
            self.saveNewPlan()
            self.checkRequiredRPMs(tests)
            self.checkHarmfulRPMs(tests)
        else:
            self.planner.plan(self.resultsDocument)
            if not self.checkPlan():
                return False

        editPlan = self.options.mode == Constants.normal and (
            self.options.test or self.options.udi or self.options.device) and (self.options.disable or self.options.enable or self.options.add or self.options.server)

        if editPlan:
            if self.options.disable or self.options.enable:
                self.setTestsEnabled()

            elif self.options.add:
                self.addTest()

            # setting server in plan edit - allows setting different servers for different tests.
            if self.options.server or self.options.serverIP:
                localTestServer = LocalTestServer(self._initializeHost())
                if localTestServer.verify():
                    self.setServerOnTests()

        self.resultsDocument.save(self.environment.getResultsPath())
        if self.options.debug != Constants.off:
            print("saved test plan to %s" % self.environment.getResultsPath())
        return True

    def saveNewPlan(self):
        self.planner.saveNewPlan(self.resultsDocument)

        self.resultsDocument.setPlanTime(self.getCurrentUTCTime())
        print("Created a new plan with %u tests" % len(self.planner.getPlannedTests()))


    def checkPlan(self):
        if not self.resultsDocument:
            print("Error: can't check plan since no prior plan exists")
            return False
        (newTests, missingTestDocuments, restorableTestDocuments) = self.planner.checkPlan(self.resultsDocument)
        if len(newTests) > 0 or len(missingTestDocuments) > 0:

            print("\nWarning: The hardware devices have changed.")
            if len(newTests) > 0:
                print("Additional tests are required:")
                for key in newTests:
                    test = newTests[key]
                    print("%-10s %-36s" % (test.getPath(),  test.getTestParameters()))

            if len(missingTestDocuments) > 0:
                print("Previously planned tests are no longer necessary:")
                for key in missingTestDocuments:
                    testDocument = missingTestDocuments[key]
                    print("%-10s %-36s" % (testDocument.getPath(),  testDocument.getTestParameters()))
            if self.options.mode == Constants.normal and not self.ui.promptConfirm("Would you like to continue?"):
                return False

            # otherwise, change the plan
            for key in newTests:
                # restore previously-deleted tests
                if key in restorableTestDocuments:
                    restorableTestDocuments[key].unMarkDeleted()
                else:
                    self.resultsDocument.appendTest(newTests[key])
            for key in missingTestDocuments:
                missingTestDocuments[key].markDeleted()

        return True

    def checkRequiredRPMs(self, tests):
        return self.installRequiredRPMs(self.getRequiredRPMs(tests))

    def checkHarmfulRPMs(self, tests):
        return self.removeHarmfulRPMs(self.getHarmfulRPMs(tests))

    def getHarmfulRPMs(self, tests):
        if self.options.debug != Constants.off:
            print("Checking for packages that should be removed based on the test plan:")
        harmfulRPMs = dict() # test name list by package name
        for test in tests:
            rpms = test.getHarmfulRPMs()
            if len(rpms) > 0:
                testID = test.Key()
                for rpm in rpms:
                    rpm = self.__handlePythonVersion(rpm)
                    if rpm not in harmfulRPMs:
                        harmfulRPMs[rpm] = [testID]
                    else:
                        harmfulRPMs[rpm].append(testID)
                if self.options.debug != Constants.off:
                    sys.stdout.write(test.getPath() + " removes " + ", ".join(rpms))
                    sys.stdout.write("\n")
        return harmfulRPMs

    def getRequiredRPMs(self, tests):
        if self.options.debug != Constants.off:
            print("Checking for additional required packages based on the test plan:")
        requiredRPMs = dict() # test name list by package name
        for test in tests:
            rpms = test.getRequiredRPMs()
            if len(rpms) > 0:
                testID = test.Key()
                for rpm in rpms:
                    rpm = self.__handlePythonVersion(rpm)
                    if rpm not in requiredRPMs:
                        requiredRPMs[rpm] = [testID]
                    else:
                        requiredRPMs[rpm].append(testID)
                if self.options.debug != Constants.off:
                    sys.stdout.write(test.getPath() + " requires " + ", ".join(rpms))
                    sys.stdout.write("\n")
        return requiredRPMs

    def __handlePythonVersion(self, rpm):
        """ change "python-" to "python3-" for RHEL8/python3 """
        if rpm.startswith("python-") and sys.version_info.major == 3:
            return "python3-" + rpm[len("python-"):]
        return rpm

    def installRequiredRPMs(self, requiredRPMs):
        if self.options.mode == Constants.normal:
            return self.__installRequiredRPMsInteractive(requiredRPMs)
        # otherwise - auto or server mode - just do it:
        return self.__installRequiredRPMs(requiredRPMs)

    def removeHarmfulRPMs(self, harmfulRPMs):
        inessentialRPMs = list()
        print("Checking rpms that needs to be removed:")
        for rpm in harmfulRPMs.keys():
            try:
                rpmQ = Command("rpm -q --whatprovides " + rpm)
                rpmQ.echo()
                inessentialRPMs.append(rpm)
            except HwCertCommandException:
                pass
        if len(inessentialRPMs) > 0:
            print("The following packages need to be removed for testing:")
            print(", ".join(inessentialRPMs))
            if os.system("yum erase -y " + " ".join(inessentialRPMs)) != 0:
                print("Error: could not remove the packages.")
                return False
            # otherwise
            print("No harmful package is installed.")
        return True

    def __installRequiredRPMsInteractive(self, requiredRPMs):
        prompt = "Would you like to install them now?"
        while True:
            missingRPMs = list()
            if self.options.debug != Constants.off:
                print("Checking installed rpms:")
            for rpm in requiredRPMs.keys():
                try:
                    rpmQ = Command("rpm -q --whatprovides " + rpm)
                    if self.options.debug != Constants.off:
                        rpmQ.echo()
                    else:
                        rpmQ.run()
                except HwCertCommandException as exception:
                    print(exception)
                    missingRPMs.append(rpm)
            if len(missingRPMs) > 0:
                print("The following packages are required for testing:")
                print(", ".join(missingRPMs))
                unavailableRPMs = list()
                if self.ui.promptConfirm(prompt):
                    for rpm in missingRPMs:
                        if os.system("yum install -y " + rpm) != 0:
                            unavailableRPMs.append(rpm)
                    if self.options.mode != Constants.normal:
                        return False

                    # otherwise
                    print("Rechecking required rpms")
                    if len(unavailableRPMs) > 0:
                        print("Error: couldn't install all required packages")

                    prompt = "Would you like to re-try installing the missing packages?"

                else:  # user-abort
                    print("\nWarning: some tests may fail due to following missing packages")
                    for rpm in missingRPMs:
                        print("\t%s requires %s" % (", ".join(requiredRPMs[rpm]), rpm))
                    return False
            else:
                if self.options.debug != Constants.off:
                    print("All required packages installed")
                return True
        print("")

    def __installRequiredRPMs(self, requiredRPMs):
        missingRPMs = list()
        print("Checking installed rpms:")
        for rpm in requiredRPMs.keys():
            try:
                rpmQ = Command("rpm -q --whatprovides " + rpm)
                rpmQ.echo()
            except HwCertCommandException as exception:
                print(exception)
                missingRPMs.append(rpm)
        if len(missingRPMs) > 0:
            print("The following packages are required for testing:")
            print(", ".join(missingRPMs))
            unavailableRPMs = list()
            for rpm in missingRPMs:
                if os.system("yum install -y " + rpm) != 0:
                    unavailableRPMs.append(rpm)
            if len(unavailableRPMs) > 0:
                print("Error: couldn't install all required packages")
                return False
            # otherwise
            print("All required packages installed")
        return True

    def runTimeCheckRequiredRPMs(self, requiredRPMs):
        missingRPMs = list()
        print("Checking installed rpms:")
        for rpm in requiredRPMs:
            try:
                rpmQ = Command("rpm -q --whatprovides " + rpm)
                rpmQ.echoIgnoreWarnings()
            except HwCertCommandException as exception:
                print(exception)
                missingRPMs.append(rpm)
        if len(missingRPMs) > 0:
            print("Error: The following packages are required for testing:")
            print(", ".join(missingRPMs))
            return False
            # otherwise
        print("All required packages installed")
        return True

    def runTimeCheckHarmfulRPMs(self, harmfulRPMs):
        inessentialRPMs = list()
        print("Checking rpms that should not be present/installed.:")
        for rpm in harmfulRPMs:
            try:
                rpmQ = Command("rpm -q --whatprovides " + rpm)
                rpmQ.echo()
                inessentialRPMs.append(rpm)
            except HwCertCommandException as exception:
                pass
        if len(inessentialRPMs) > 0:
            print("The following packages need to be removed for testing:")
            print(", ".join(inessentialRPMs))
            return False
        return True

    def installPackages(self, requestDocument):
        """ install or upgrade packages required for the test plan """
        if not requestDocument.getCertificationProgram():
            print("Error: could not determine certification program for installing certification test packages")
            return False
        self.ui.statusMessage("Checking/installing test packages")
        packages = ["redhat-certification-" + requestDocument.getCertificationProgram()]
        # Hack to install baremetal
        if len(requestDocument.getPlanComponents()) > 0:
            for component in requestDocument.getPlanComponents():
                if component.getName() == 'BAREMETAL':
                    packages = ["redhat-certification-"+"baremetal"]
        # figure out which packages might need to be removed
        allPackages = list()
        for program in Programs.getAll():
            allPackages.append("redhat-certification-" + program)
        removePackages = set(allPackages)
        removePackages.difference_update(packages)
        if self.options.debug != Constants.off:
            print("Removing packages %s" % " ".join(removePackages))
            os.system("yum erase -y " + " ".join(removePackages))

        """ install or upgrade packages """
        result = True
        if os.system("yum install -y " + " ".join(packages)) != 0:
            result = False
        if os.system("yum update -y " + " ".join(packages)) != 0:
            result - False
        return result

    def doRun(self):
        self.__checkResultsSize()

        # Change to Non-English Locale to English for running tests
        curr_locale = os.environ.get('LANG')
        if curr_locale and not curr_locale.startswith('en_US'):
            print('Found non-english locale. Changing locale to {0} for RHCERT test run'.format(Constants.locale))
            os.environ.update({'LANG': Constants.locale})

        if not self.doPlan():
            return False

        #check tests
        if self.options.test:
            testNotFound = False
            for testName in self.options.test:
                if not self.resultsDocument.getTest(testName):
                    print("Error: No such test \"" + testName + "\"")
                    testNotFound = True
            if testNotFound:
                return False

        # otherwise
        parameters = dict()
        if self.options.device:
            parameters[Constants.device] = self.options.device
        if self.options.udi:
            parameters[Constants.udi] = self.options.udi
        if self.options.property:
            parameters.update(self.options.property)
        testDocuments = self.resultsDocument.getFilteredTests(self.options.test, self.options.tag, parameters)

        if len(testDocuments) > 0:

            testDocuments = self.addMandatoryTests(testDocuments)
            if not testDocuments:
                return False
            print("\nRunning the following tests:")
            for testDocument in testDocuments:
                print("%-10s %-36s" % (testDocument.getName(),  testDocument.getTestParameters()))

            verified = self.verify(testDocuments)
            if not verified and self.options.mode == Constants.normal:
                if not self.ui.promptConfirm("Verification failed, would you like to continue testing?"):
                    return False

            # otherwise overriding verification failure or passed verification
            return self._doRun(testDocuments)
        else:
            print("Error: no tests match the supplied options")

        # Resetting the locale back to it's original non-english locale
        if curr_locale and curr_locale.startswith('en_US'):
            print("Reverting back to non-english locale {}".format(curr_locale))
            os.environ.update({'LANG': curr_locale})

        return False

    def doContinue(self):
        self.load()
        if not self.resultsDocument:
            print("No tests planned.")
            return False
        testDocuments = self.resultsDocument.getIncompleteTestsFromLastRun()
        if len(testDocuments) is 0:
            print("All tests in the last test run are complete")
            return False
        # testDocuments = self.addMandatoryTests(testDocuments) don't re-do mandatory tests on a continue.
        self._initializePlanner()
        return self._doRun(testDocuments, continueRun=True)

    def doUpdate(self):
        """ attempted to update deferred/needs review tests via microservices """
        if self.options.results:
            try:
                if not os.path.exists(self.options.results):
                    raise Exception("Error: Results file provided to '--results' option, doesn't exists. Please check the path.")
                elif not os.path.splitext(self.options.results)[1] in [".xml", ".gz"]:
                    raise Exception("Error: Invalid results file provided to '--results' option.")
                elif not os.stat(self.options.results).st_size:
                    raise Exception("Error: Empty results file provided to '--results' option.")
                self.resultsDocument = ResultsDocument()
                print("INFO: loading %s ..." % self.options.results)
                self.resultsDocument.load(self.options.results)
            except Exception as exception:
                print("Error: could not load results file % s" % self.options.results)
                print(exception)
                return False
        else:
            print("Invalid command. Please provide a result file with --results option.")
            return False
        if not self.resultsDocument:
            print("Error: Could not load results")
            return False
        print("INFO: File loaded successfully!")
        if MicroServices().update(self.resultsDocument):
            self.resultsDocument.save()
            print("INFO: Please check updated results here: %s" % self.options.results)
        return True

    def run(self, requestDocument):

        # one run per document, clean all every request:
        self.clean()
        self.cleanAll()

        # save the request as the certification document
        requestDocument.save(self.environment.getCertificationPath())

        self.installPackages(requestDocument)
        self._initializePlanner()
        self.__setPlannerProgram(requestDocument)

        self.doPlan()

        # set server hostname and ip
        self.resultsDocument.setTestServer(requestDocument.getTestServer())
        self.resultsDocument.setTestServerIP(requestDocument.getTestServerIP())
        # set hostname per the one on the server
        self.resultsDocument.getRequiredOS().setName(requestDocument.getRequiredOS().getName())
        print("TestHarness.run set cert id: %s" % requestDocument.getCertificationID())

        testDocuments = list()
        for requestedTest in requestDocument.getTests():
            # Ignore "debug" from parameters of selected/requested test while getting the
            # actual test from the results document
            path = requestedTest.getPath()
            params = requestedTest.getTestParameters().getAll()
            if Constants.debug in params:
                del params[Constants.debug]
            testDocument = self.resultsDocument.getTest(path, params)
            if testDocument:
                testDocuments.append(testDocument)
            else:
                print("Warning: could not match test %s" % requestedTest.getPath())

        # if the test plan requested no testing, then we're done
        if not testDocuments:
            print("No tests were requested")
            self.resultsDocument.save(self.environment.getResultsPath())
            return True

        testDocuments = self.addMandatoryTests(testDocuments)

        print("\nRunning the following tests:")
        for testDocument in testDocuments:
            print("%-10s %-36s" % (testDocument.getName(),  testDocument.getTestParameters()))

        return self._doRun(testDocuments)

    def _doRun(self, testDocuments, continueRun=False):

        # don't run suspend on a continue run
        if continueRun:
            for test in testDocuments:
                if test.getName() == 'suspend':
                    testDocuments.remove(test)

        if not testDocuments:
            print("No matching tests found.")
            return False

        if self.options.debug != Constants.off:
            report = Report(self.options, self.resultsDocument)
            for test in testDocuments:
                print(report.formatTest(test))
            print("")

        # determine test server - prefer IP over hostname
        if not self.options.server:
            if self.resultsDocument.getTestServerIP():
                self.options.server = self.resultsDocument.getTestServerIP()
            elif self.resultsDocument.getTestServer():
                self.options.server = self.resultsDocument.getTestServer()


        # Set up our system logging stuff
        syslog.openlog(self.getSystemLogOpen())
        syslog.syslog("Beginning test run.")


        # create the new run, mark incomplete
        if continueRun:
            runNumber = self.resultsDocument.getNumberOfTestRuns()
            runDirectory = self.getTestRunDirectory()
        else:
            runNumber = self.resultsDocument.addTestRun()
            # Make a log dir for this testrun
            runDirectory = self.makeTestRunDirectory()

        for test in testDocuments:
            if not test.isDisabled():
                runDocument = test.getRun(runNumber)
                if not runDocument:
                    runDocument = test.newTestRun()
                    runDocument.setNumber(runNumber)
        self.resultsDocument.save(self.environment.getResultsPath())

        # Start testin'!
        combinedResult = True # assume all tests pass
        for testDocument in testDocuments:
            if testDocument.isDisabled():
                if self.options.debug != Constants.off:
                    print("skipping %s - disabled" % testDocument.getName())
                continue

            message = "Running %s" % (testDocument.getName())
            parameterString = "%s" % testDocument.getTestParameters()
            if parameterString:
                message +=  " on %s" % (parameterString)
            self.ui.statusMessage(message)
            runDocument = testDocument.getRun(runNumber)

            runDocument.setRunTime(self.getCurrentUTCTime())
            if not runDocument.getMode():
                runDocument.setMode(self.options.mode)
            runDocument.setNumber(self.resultsDocument.getNumberOfTestRuns())
            runDocument.Debugging = self.options.debug != Constants.off

            # create log dir, install test
            logDirectory=self.makeLogDirectory(runDirectory,testDocument,continueRun)
            outputFilePath=logDirectory+"/output.log" # set logfile name
            if not self.installTest(testDocument, logDirectory):
                return False

            # Here goes! Actually run the test.
            syslog.syslog("%s: begin" % testDocument.getName())
            testError = False
            returnValue = Constants.FAIL
            try:
                returnValue, _test = self.runTest(logDirectory, testDocument, runDocument, outputFilePath)
                runDocument.setSummary(returnValue)
                if returnValue == Constants.FAIL:
                    combinedResult = False
                # Required for trusted container test
                if hasattr(_test, "certified_containers"):
                    if _test.certified_containers:
                        runDocument.setCertifiedContainers(_test.certified_containers)
                if hasattr(_test, "container_projects"):
                    if _test.container_projects:
                        runDocument.setContainerProjects(_test.container_projects)
                runDocument.getResultsAttachments(self.environment, outputFilePath)

            except OSError as e:
                print("Test error: %s" % e)
                syslog.syslog("Test error: %s" % e)
                testError = True
                combinedResult = False

            # Log the end of the test.
            runDocument.setEndTime(self.getCurrentUTCTime())
            syslog.syslog("%s: end" % testDocument.getName())
            # If the test did not run correctly, don't record the results.
            if testError:
                self.removeDirectory(logDirectory)
                continue

            # Save the system log
            self.saveSystemLog(testDocument, runDocument)
            # encode return value into output.log
            self.saveOutput(runDocument, returnValue, outputFilePath)
            # print(result)
            print("Return value was %s" % returnValue)

            self.resultsDocument.save(self.environment.getResultsPath())
            if self.options.debug != Constants.off:
                print(("saved to " + self.environment.getResultsPath()))

            if self.testCompleteCallback:
                if True or self.options.debug != Constants.off:
                    print("calling testCompleteCallback")
                try:
                    self.testCompleteCallback()
                except Exception as error:
                    print("testCompleteCallback Failed: ")
                    print(error)

        syslog.syslog("Testing complete.")
        syslog.closelog()
        self.resultsDocument.save(self.environment.getResultsPath())

        return combinedResult

    def addMandatoryTests(self, requestedTestDocuments):
        if self.options.debug == Constants.off:
            requestedTestDict = dict(zip(map(TestDocument.getName, requestedTestDocuments), requestedTestDocuments))

            for mandatoryTest in self.resultsDocument.getMandatoryTests():
                try:
                    test = requestedTestDict[mandatoryTest.getName()]
                except KeyError:
                    mandatoryTest.setMode(Constants.forced)
                    if mandatoryTest.getMandatory() == Constants.pre:
                        requestedTestDocuments.insert(0,mandatoryTest)
                    elif mandatoryTest.getMandatory() == Constants.post:
                        requestedTestDocuments.append(mandatoryTest)

            # reboot test is a special case - really needs to run last
            tests = list()
            rebootTests = list()
            for test in requestedTestDocuments:
                if test.getName() != "reboot":
                    tests.append(test)
                else:
                    rebootTests.append(test)
            if rebootTests:
                tests.extend(rebootTests)
            return tests

        # otherwise - debugging - don't add any tests
        print("Warning: Test runs in debug mode are invalid for certification.")
        if self.resultsDocument.getMandatoryTests():
            # Getting all the mandatory tests to run
            mandatoryTests = list(map(TestDocument.getName, self.resultsDocument.getMandatoryTests()))
            for selectedTest in requestedTestDocuments:
                if selectedTest.getName() in mandatoryTests:
                    # removing the selected test from mandatory test list
                    mandatoryTests.remove(selectedTest.getName())

            if mandatoryTests:
                print("The following tests will be skipped: %s" % ", ".join(map(str,mandatoryTests)))
                if self.options.mode == Constants.normal and not self.ui.promptConfirm("Do you wish to continue?"):
                    return None
        return requestedTestDocuments

    def doVerify(self):
        """ run tests configuration verification code """

        if not self.doPlan():
            return False

        return self.verify(self.resultsDocument.getTests())

    def __checkResultsSize(self):
        self.load()
        if not self.resultsDocument:
            return True

        if self.resultsDocument.getLoadedFileSize() < int(self.environment.getResultsWarningSize()):
            return True

        # otherwise
        print("Warning: the test results may be too large to submit to the hardware catalog.")
        print("File: %s is %u MB" % (self.resultsDocument.getLoadedFilePath(), int(self.resultsDocument.getLoadedFileSize()/Constants.MB)))
        if self.options.mode == Constants.normal and self.ui.promptConfirm("Would you like to save the current results first?"):
            TestHarness.doSave(self) # call TestHarness.doSave as it just saves the current file without the selection dialog
            self.clean()

        # size warning was ignored (or mode auto)
        return True

    def verify(self, testDocuments):
        # just call verify on each test directly
        failures = 0

        for testDocument in testDocuments:
            if testDocument.isDisabled():
                continue
            test = self.planner.getTest(testDocument.getPath())
            self.__setTestParameters(test, testDocument)
            if not test.verify():
                failures += 1

        if failures == 0:
            print("\nTest Verification Passed")
            return True

        # otherwise:
        print("\n%u Tests Failed Verification" % failures)
        return False

    def __setTestParameters(self, test, testDocument):
        test.setTestParameters(testDocument.getTestParameters())
        # set test server 1) test-specific, or 2) command option, or 3) default from results document
        if not test.getTestParameters().get(Constants.testserver):
            # prefer IP
            if self.options.serverIP:
                if self.options.debug != Constants.off:
                    print("setting test server IP from harness options: %s " % self.options.serverIP)
                test.getTestParameters().set(Constants.testserver, self.options.serverIP)
            elif self.resultsDocument.getTestServerIP():
                if self.options.debug != Constants.off:
                    print("setting testserver to IP %s" % self.resultsDocument.getTestServerIP())
                test.getTestParameters().set(Constants.testserver, self.resultsDocument.getTestServerIP())
            elif self.options.server:
                if self.options.debug != Constants.off:
                    print("setting test server from harness options: %s " % self.options.server)
                test.getTestParameters().set(Constants.testserver, self.options.server)
            elif self.resultsDocument.getTestServer():
                if self.options.debug != Constants.off:
                    print("setting testserver to server %s" % self.resultsDocument.getTestServer())
                test.getTestParameters().set(Constants.testserver, self.resultsDocument.getTestServer())
        else:
            if self.options.debug != Constants.off:
                print("test has a specific server set:")
                print(test.getTestParameters())
        test.setMode(self.options.mode)
        test.setDebug(self.options.debug)
        lastRun = testDocument.getLastRun()
        if lastRun and (lastRun.getSummary() == Constants.INCOMPLETE):
            test.setIncomplete(True)

    def doPrint(self):
        self.load()
        format = self.options.format or 'text'

        if format == 'html':
            reportEngine = ReportHTML
        elif format == 'junit':
            reportEngine = ReportJunit
        else:
            reportEngine = Report

        report = reportEngine(self.options, self.resultsDocument)
        return report.Write()

    def printLastRunSummary(self):
        if not self.resultsDocument or self.resultsDocument.getNumberOfTestRuns() == 0:
            return False
        runSummary = True
        runNumber = self.resultsDocument.getNumberOfTestRuns()
        print("\n Test Run Summary:")
        print("--------------------------------------")
        for test in self.resultsDocument.getTests():
            runDocument = test.getRun(runNumber)
            if runDocument:
                summary = "- %s" % runDocument.getSummary()
                device = test.getTestParameters().get(Constants.device)
                if not device:
                    device = ""
                print("%-7s %-7s %s" % (test.getName(), device, summary.lower()))
                if runDocument.getSummary() != Constants.PASS:
                    runSummary = False
        return runSummary

    def doSubmit(self):
        self.load()
        return self.__submit()

    def __submit(self):

        if not self.resultsDocument or self.resultsDocument.getNumberOfTestRuns() == 0:
            print("Error: no test results to submit")
            return False

        if self.options.mode == Constants.normal:
            (submitted, success) = self.workflowCLI.submit(self.resultsDocument)
            if success and submitted and self.ui.promptConfirm("Would you like to clean current test results from this system?"):
                self.clean()

        # otherwise
        return True

    def doSave(self):
        if self.resultsDocument:
            self.resultsDocument.save() # make sure any current updates are saved
        self.load()
        if self.resultsDocument:
            (savedFilePath, storePath) = self.saveToStore()
            if storePath:
                destinationPath = os.path.join(Constants.results, storePath)
                if self.resultsDocument.getDatabasePath():
                    destinationPath = os.path.join(self.resultsDocument.getDatabasePath(), destinationPath)
                return self.saveToServer(savedFilePath, destinationPath)
            # otherwise
            return False

        #otherwise
        print("Error: no test results to save")
        return False

    def saveToStore(self):
        if self.resultsDocument:
            path = self.environment.getStoreDirectory()
            storePath = os.path.join(*self.resultsDocument.getStorePath())
            path = os.path.join(path, storePath)
            self.makeDirectoryPath(path)

            savedFilename = self.resultsDocument.getVerboseFileName()
            print("Saving current results to:")
            savedFilePath = os.path.join(path, savedFilename)
            print(savedFilePath + ".gz")
            # we didn't load the whole file, so just use file operations to copy it
            try: # first remove any old copies
                os.remove(savedFilePath)
                os.remove(savedFilePath + ".gz")
            except OSError:
                pass

            try: # copy and compress the file
                shutil.copy(self.environment.getResultsPath(), savedFilePath)
                gzip = Command("gzip -n -f \"%s\"" % savedFilePath)
                gzip.run()
            except Exception as exception:
                print("Error: could not save and compress %s" % savedFilePath)
                print(exception)
                gzip.printErrors()
            return (savedFilePath, storePath)
        return (None, None)

    def saveToServer(self, savedFilePath, storePath):
        if self.resultsDocument:
            # if it's supplied as an option, use that one but don't save it,
            # otherwise, check the certification's test server
            if not self.options.server:
                self.options.server = self.resultsDocument.getTestServer()
            if self.options.server == "unknown":
                self.options.server = None
            if not self.options.serverIP:
                self.options.serverIP = self.resultsDocument.getTestServerIP()

            if self.options.server or self.options.serverIP:
                localTestServer = LocalTestServer(self._initializeHost())
                return localTestServer.saveResults(savedFilePath+".gz", storePath)

            # otherwise
            print("Error: could not determine the local test server")
            return False

        #otherwise
        print("Error: no test results to save")
        return False

    def doClean(self, cleanAll=False, force=False):
        if not cleanAll:
            try:
                subcommand = self.args[1]
                if subcommand not in [Constants.results, Constants.all]:
                    print("Error: invalid option %s, should be %s or %s" % (subcommand, Constants.results, Constants.all))
                    return False
                cleanAll = (self.args[1] == Constants.all)
            except: # index error - never mind
                pass


        force = force or self.options.mode != Constants.normal
        if force or self.ui.promptConfirm("Are you sure you want to delete all current test results?"):
            self.clean()

        if cleanAll and (force or self.ui.promptConfirm("Also remove certification data?")):
            self.cleanAll()

        return True

    def getBiosInfo(self):
        arch = self.getArch()
        self.resultsDocument.setHardware(Tags.arch, arch)
        if self.options.debug != Constants.off:
            print("Getting Bios Info for %s" % arch)
        vendor = ""
        make = ""
        model = ""
        if arch in ['i386','i586','i686','x86_64', 'aarch64']:
            dmidecode = os.popen("/usr/sbin/dmidecode  -t 1")

            for line in dmidecode:
                if line.find(':') == -1:
                    continue
                (key,val)=[t.strip() for t in line.split(':',1) if line.find(':')]

                if key == "Manufacturer" and val:
                    vendor = val
                if key == "Product Name" and val:
                    model = val
                    # try and use the first word as the make
                    make = val.split()[0]
                if vendor and model:
                    break
            dmidecode.close()
        elif os.path.exists('/proc/device-tree/model'):
            f=open('/proc/device-tree/model')
            model = f.readline().strip()
            # if there's a comma, guess the format is <vendor>, <model>
            # failing that, if there's a space, assume it's <vendor> <model>
            try:
                for separator in [',', ' ']:
                    if separator in model:
                        vendor = model.split(separator)[0]
                        model = model.split(separator)[1]
                        # try and use the first word of model as the make
                        make = model.split()[0]
                        break
            except:
                pass

            f.close()
        # IBM-likely arches
        if arch in ['s390','s390x', "ppc", "ppc64"]:
            if not vendor:
                vendor = 'IBM'
            if not make:
                make = 'IBM'
            if not model:
                model = arch



        self.resultsDocument.setHardware(Tags.model, model)
        self.resultsDocument.setHardware(Tags.make, make)
        self.resultsDocument.setHardware(Tags.vendor, vendor)

    def getArch(self):
        system,node,release,version,machine = os.uname()
        return machine

    def makeTestRunDirectory(self):
        """Create a directory for the current test run"""
        testRunDirectory = self.getTestRunDirectory()
        self.removeDirectory(testRunDirectory)
        self.makeDirectoryPath(testRunDirectory)
        return testRunDirectory

    def getTestRunDirectory(self):
        runNumber = self.resultsDocument.getNumberOfTestRuns()
        return self.environment.getLogDirectory() + "/runs/%u" % runNumber

    def makeLogDirectory(self, directory, test, continueRun=False):
        "Create a directory for logs for the currently-running test"
        logdir = os.path.join(directory, test.getPath())
        if not continueRun:
            self.removeDirectory(logdir) # clean it, if it's there
        self.makeDirectoryPath(logdir)
        return logdir

    def installTest(self, test, scratchDirectory):
        "Install a test's files to the named tmpdir for running"
        if self.options.debug != Constants.off:
            print("installing test from %s into %s" % (test.getSourceDirectory(), scratchDirectory))
        try:
            cwd=os.getcwd()
            os.chdir(test.getSourceDirectory())
            # allow c programs via Make
            for sourceFile in os.listdir(os.getcwd()):
                (_, extension) = os.path.splitext(sourceFile)
                if sourceFile == "Makefile" or extension in [".c", ".xml", ".conf", ".sh"]:
                    try:
                        shutil.copy(sourceFile, scratchDirectory)
                    except:
                        pass
            os.chdir(cwd)
        except Exception as error:
            print("Error: Could not install test sources - %s" % error)
            return False
        return True

    def runTest(self, logDirectory, testDocument, runDocument, outputFilePath):
        """Run a test, sending output to the logfile and stdout"""
        if self.options.debug != Constants.off:
            print("HardwareTestHarness.runTest:")
            print(("directory %s    parameters: %s" % (logDirectory, testDocument.getTestParameters())))
        # change cwd to test logDirectory
        cwd=os.getcwd()
        os.chdir(logDirectory)

        # call make build to compile/chmod any called scripts
        if os.path.exists("Makefile"):
            try:
                buildCommand = Command("make build")
                buildCommand.echo()
            except HwCertCommandException as exception:
                print("Warning: test build produced errors.")
                print(exception)
                buildCommand.printErrors()

        #tee stdout and stderr to terminal and log file
        realStdout = sys.stdout
        realStderr = sys.stderr
        sys.stdout = Log(outputFilePath)
        sys.stderr = sys.stdout

        # find the test driver python and call it directly
        test = self.planner.getTest(testDocument.getPath())
        if not test:
            print("Error: test %s not found" % testDocument.getPath())
            return TestResult.FAIL, None
        # otherwise
        self.__setTestParameters(test, testDocument)
        test.getTestParameters().set(Constants.outputfile, outputFilePath)
        runDocument.setTestParameters(test.getTestParameters()) # log them in the run document
        if self.options.debug != Constants.off:
            print("Test Parameters: %s" % test.getTestParameters())

        test.setResultsDocument(self.resultsDocument)

        noMissingRPMs = self.runTimeCheckRequiredRPMs(test.getRequiredRPMs())
        noHarmfulRPMs = self.runTimeCheckHarmfulRPMs(test.getHarmfulRPMs())
        rv = False
        try:
            if noMissingRPMs and noHarmfulRPMs:
                rv = test.run()
            elif self.options.debug != Constants.off:
                print("In debug mode - running the test despite missing/harmful packages.")
                rv = False # test fails because of missing or harmful RPMs
                test.run()
        except Exception as error:
            print("Error: Exception raised in test %s:" % test.getPath())
            print(traceback.format_exc())
            sys.stdout.flush()
            return TestResult.FAIL, None
        finally:
            # restore stdout and stdin, and cwd
            sys.stdout.close()
            sys.stdout = realStdout
            sys.stderr = realStderr
            os.chdir(cwd)

        # log devices
        if not test.finish(self.planner, runDocument):
            testError = True

        if self.options.debug != Constants.off:
            print("...Done: return value = %s." % rv)
        return rv, test

    def runServer(self, test, subcommand):
        if subcommand == "start":
            return test.startServer()
        if subcommand == "stop":
            return test.stopServer()

        # otherwise
        return test.statusServer()

    def saveSystemLog(self, test, run):
        "Save a section of the system log to the named file"
        # It feels a bit wrong to have this stuff hardcoded, but it works.
        contents = self.getSystemLog(test.getName())
        run.setSystemLog(contents)

    def saveOutput(self, run, retval, outputFilePath):
        print("saveOutput: %s" % outputFilePath)
        run.setReturnValue(retval)
        if not run.setTestOutput(outputFilePath):
            print("Error: could not save test logs, marking as failed.")
            run.setSummary(Constants.FAIL)

    def _getTestsFromOptions(self):
        if self.options.device:
            return self.resultsDocument.getTestsByParameter(Constants.device, self.options.device)
        if self.options.udi:
            return self.resultsDocument.getTestsByParameter(Constants.udi, self.options.udi, partialMatch="left")
        if self.options.test:
            return self.resultsDocument.getTests(self.options.test)
        print("Error: no tests specified in options")
        return None

    def setTestsEnabled(self):
        tests =  self._getTestsFromOptions()
        if self.options.debug != Constants.off:
            print("HardwareTest.setTestsEnabled:")
        disabled = True
        output = "Disabled"
        count = 0
        if self.options.enable:
            disabled = False
            output = "Enabled"
        for test in tests:
            if disabled and test.getMandatory():
                print("The %s test must be included in every test run and can not be disabled." % test.getName())
                continue
            if test.isDisabled() != disabled:
                test.setDisabled(disabled)
                count = count + 1
            else:
                print("test %s is already %s" % (test.getName(), output))

            if self.options.debug != Constants.off:
                print((output + " test: %s %s") % (test.getName(), test.getTestParameters()))
        print((output + " %u tests") % count)

    def setServerOnTests(self):
        tests =  self._getTestsFromOptions()
        if self.options.debug != Constants.off:
            print("HardwareTest.setServerOnTests:")

        if self.options.server:
            output = "Set server to %s for " % self.options.server
            for test in tests:
                test.setServer(self.options.server)
                if self.options.debug != Constants.off:
                    print((output + " test: %s %s") % (test.getName(), test.getTestParameters()))
            print((output + " %u test(s)") % len(tests))

    def addTest(self):
        if not self.options.test:
            print("Must specify a test , use --test <test path>")
            return

        # otherwise, get the test to be added
        self._initializePlanner()
        try:
            test = self.planner.getTest(self.options.test[0], byName=True)
            if not test:
                print("No such test %s" % self.options.test)
                return
            test = copy.copy(test)
            test.add(self.planner.getPlanner(test))
            self.checkRequiredRPMs([test])
            self.checkHarmfulRPMs([test])
        except KeyError:
            print("No such test %s" % self.options.test)
            return

        test.setSource(Constants.manual)
        self.resultsDocument.appendTest(test)
        self.resultsDocument.setPlanTime(self.getCurrentUTCTime())
        print("Added test")

    def getCurrentUTCTime(self):
        return time.gmtime(time.time())

    def clean(self):
        # is the rug going to be pulled out from under us?
        if self.environment.getLogDirectory() in os.getcwd():
            os.chdir("/var/log")
        if self.options.debug != Constants.off:
            print("removing test logs (%s)..." % self.environment.getLogRunsDirectory())
        os.system("rm -rf %s" % self.environment.getLogRunsDirectory())
        if self.options.debug != Constants.off:
            print("removing test results results (%s)..." % self.environment.getResultsPath())
        os.system("rm -f %s" % self.environment.getResultsPath())
        self.resultsDocument = None

    def cleanAll(self):
        if self.options.debug != Constants.off:
            print("removing local certification document (%s)..." % self.environment.getCertificationPath())
        os.system("rm -f %s" % self.environment.getCertificationPath())

    def getLock(self):
        if os.path.exists(self.environment.getLockFile()):
            print("Error: rhcert is already running (lock file %s found)" % self.environment.getLockFile())
            if self.options.mode == Constants.normal:
                if self.ui.promptConfirm("Override?"):
                    return True
            # otherwise, can't ask about override
            return False
        # otherwise, no lock - go ahead and lock it
        lock = open(self.environment.getLockFile(), "w")
        return True

    def releaseLock(self):
        try:
            os.remove(self.environment.getLockFile())
        except OSError as e:
            print("Warning: rhcert lock file missing - other instances of rhcert may have run.")
