# Copyright (c) 2016 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
import sys, socket, time, re

from rhcert.host import Host
from rhcert.model.task import Task
from rhcert.localTestServer import LocalTestServer
from rhcert.rhcertStore import RedHatCertificationStore
from rhcert.model.resultsdocument import ResultsDocument
from rhcert.tags import Constants, Attributes
from rhcert.compatability import XmlrpcServerProxy

class Remote(object):

    def __init__(self, harness, program, sandboxID):
        self.localHarness = harness
        self.program = program
        self.sandboxID = sandboxID
        self.client = Host(self.localHarness.options.client)
        self.debug = self.localHarness.options.debug
        self.environment = self.localHarness.environment
        self.socketDefaultTimeout = socket.getdefaulttimeout()
        self.localTestServer = LocalTestServer(Host("localhost"))
        self.xmlrpcClientProxy = None
        self.rhcertStore = RedHatCertificationStore()

    def getClientStatus(self, timeout=5):
        """  host is a Host object, timeout = 5 # sec """
        socket.setdefaulttimeout(timeout)
        self.xmlrpcClientProxy = XmlrpcServerProxy(('http://%s:%s' % (self.client.getConnection(), self.environment.getServerListenerPort())), allow_none=True)
        try:
            result = self.xmlrpcClientProxy.getResult()
            socket.setdefaulttimeout(self.socketDefaultTimeout)
            if result:
                # is it a Task for a prompt?
                task = Task()
                task.loadString(result)
                if task.isValid():
                    if task.getTask() == Task.message:
                        if True: # task.getParameter(Task.message) in ["waiting for connection", "ready", "Command \"register\" succeeded"]:
                            return ("True", task.getParameter(Task.message))
                        return (False, task.getParameter(Task.message))
                    else:
                        return (False, "Waiting at prompt: %s" % task.getParameter(Task.question))
                else:
                    return (False, result)
        except socket.timeout:
            return (False, "waiting for response")
        except Exception as e:
            print("Host Status Exception: ")
            print(e)
            return (False, "waiting for connection")
        finally:
            socket.setdefaulttimeout(self.socketDefaultTimeout)
        return (True, "ready")

    def getServerStatus(self):
        """ remote requires the rhcert web server to be running the localhost """
        return (self.localTestServer.getStatus() == "running")

    def checkSystemStatus(self):
        (ready, message) =  self.getClientStatus()
        print("ready: %s  message: %s" % (ready, message))
        if not ready:
            print("Error: client %s (%s) is not ready, please check status using \"rhcertd status\" on that host" % (self.client.getHostname(), self.client.getIPAddress()))
            return False
        if not self.getServerStatus():
            print("Error: rhcert server must be running on this host, please start the server via \"systemctl start httpd\"")
            return False
        return True

    def doRegister(self):
        """ register the client with this server """
        # otherwise
        try:
            task = Task(Task.register)
            task.setParameter(Task.server, socket.gethostname())
            print("calling client to register")
            result = self.xmlrpcClientProxy.run(task.toXML())
            # wait a couple seconds in hopes that it's ready
            time.sleep(4)
            if not result:
                self.appendError("Error: Register Request Failed for client \"%s\"" % self.client.getName())
            else:
                result = self.xmlrpcClientProxy.getResult()
                if result:
                    task = Task(result)
                    if task.isValid():
                        message = task.getParameter(Task.message)
                    else:
                        message = result
                    print("Registration message: %s" % message)
                    return True

        except Exception as exception:
            print("Error: Register Request Failed: ")
            print("%s\n" % exception)
            return False

        return True

    def __beginTesting(self):
        clientPath = self.rhcertStore.getSystemPathForHost(self.client)
        if not clientPath:
            print("Error: could not find host document for %s" % self.client.getName())
            return False

        (certificationPath, self.certification) = self.rhcertStore.getCertification(self.program, self.sandboxID)
        if not self.certification:
            print("Error: could not locate %s sandbox id %s" % (self.program, self.sandboxID))
            return False

        resultsDocument = ResultsDocument()
        resultsDocument.load(clientPath)
        resultsDocument.setCertificationID(self.certification.getCertificationID())
        resultsDocument.setBugID(self.certification.getBugID())
        resultsDocument.setCertificationProgram(self.certification.getCertificationProgram())
        for vendor in self.certification.getVendors():
            resultsDocument.setVendor(vendor)
        resultsDocument.copyPlanComponents(self.certification)

        # delete all the tests - we just want the client to plan
        for testDocument in resultsDocument.getTests():
            testDocument.delete()
        self.rhcertStore.saveResultsDocument(resultsDocument)
        # set the test server to this one
        resultsDocument.setTestServer(socket.gethostname())

        try:
            print("requesting plan")
            result = self.xmlrpcClientProxy.run(resultsDocument.document.toxml())
            if not result:
                print("Error: Plan Request Failed.")
                return False
        except Exception as exception:
            print("Error: call to host %s failed:\n" % (self.testHost.getConnection()))
            print("%s\n" % exception)
            print("%s\n" % resultsDocument.document.toxml())

        return True

    def getLatestResultsDocument(self):
        # get the latest results for this host, program, cert
        (certificationPath, self.certification) = self.rhcertStore.getCertification(self.program, self.sandboxID)
        if not self.certification:
            print("Error: could not locate %s sandbox id %s" % (self.program, self.sandboxID))
            return None

        resultsDocuments = self.rhcertStore.getResultsDocumentsFor(self.certification, False, self.client)
        if not resultsDocuments:
            return None

        # get the latest/newest one - leave all the tests selected
        paths = sorted(resultsDocuments.keys(), reverse=True)
        return resultsDocuments[paths[0]]

    def __waitForTestRun(self, runNumber=1):
        time.sleep(10) # prevent race condition
        message = ""
        waiting = False
        messagePattern = re.compile("Finished test run (?P<runNumber>[0-9]+)")
        while True:
            match = messagePattern.search(message)
            if match and match.group("runNumber") >= runNumber:
                break
            (ready, message) = self.getClientStatus()
            if self.debug != Constants.off:
                print("host response: %s" % message)
            else:
                if not waiting:
                    sys.stdout.write("waiting...")
                    waiting = True
                else:
                    sys.stdout.write(".")
                sys.stdout.flush()
            time.sleep(10)
        print("\nDone!")
        return ready

    def doPlan(self):
        if not self.checkSystemStatus():
            return False
        if not self.doRegister():
            return False
        if not self.__beginTesting():
            return False
        return self.__waitForTestRun(0)


    def doRun(self):
        if not self.doPlan():
            return False

        # get the latest results for this host, program, cert
        resultsDocument = self.getLatestResultsDocument()
        if not resultsDocument:
            print("Error: could not locate plan results for host %s" % self.client.getName())
            return False

        if self.localHarness.options.test:
            # prune the tests to those requested
            for testDocument in resultsDocument.getTests():
                requested = False
                for testNameOrPath in self.localHarness.options.test:
                    if testNameOrPath == testDocument.getName() or testNameOrPath == testDocument.getPath():
                        requested = True
                        break
                if not requested:
                    testDocument.delete()

        try:
            print("requesting test run")
            result = self.xmlrpcClientProxy.run(resultsDocument.document.toxml())
            if not result:
                print("Error: Run Request Failed.")
                return False
        except Exception as exception:
            print("Error: call to run on host %s failed:\n" % (self.testHost.getConnection()))
            print("%s\n" % exception)
            print("%s\n" % resultsDocument.document.toxml())
            return False
        print("Test is running.")
        self.__waitForTestRun()
        return True



