# Copyright (c) 2015 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#

import os, sys, socket, time

from rhcert.tags import Constants, TestTag, Programs, States
from rhcert.model.resultsdocument import ResultsDocument
from rhcert.report import Report
from rhcert.client.harness import TestHarness
from rhcert.parser import RHCertOptionParser

from .remote import Remote


class RHCertCI(TestHarness):
    """ RHCertCI drives the test Harness via command-line for CI use """

    def __init__(self):
        self.commands = {
            # command, optional, required
              'plan':     self.doPlan,
              'run':      self.doRun,
              'save':     self.doSave,
              'clean':    self.doClean,
              'print':    self.doPrint,
              'version':  self.doVersion
              }
        self.remote = None # default to local
        self.key = None
        self.getOptions()
        TestHarness.__init__(self, self.options, False)

    def getOptions(self):
        usage = "usage: %prog <command> [options]\n\ncommand:\n"\
                    "\tplan \t- plan certification testing\n"\
                    "\tclean [all] - remove all results and temporary files\n"\
                    "\trun \t- run specific certification tests\n"\
                    "\tprint \t- print certification test results\n"\
                    "\tversion\t- prints the version of rhcert\n"\
                    "\tsave\t- save a copy of the current test results to a local sandbox\n"\
                    "\n\nFor help on options, use %prog <command> --help"

        parser = RHCertOptionParser(self, usage)
        if len(sys.argv) < 2:
            parser.error("No command supplied")
            exit(1)
        if sys.argv[1] not in self.commands:
            if sys.argv[1] not in ["help", "--help", "-h"]:
                parser.error("No such command: %s" % sys.argv[1])
            else:
                parser.print_usage(sys.stderr)
            exit(1)

        # otherwise
        self.command = sys.argv[1]
        parser.add_option("", "--program", action="store", type="string", dest="program", help="certification program")
        parser.add_option("", "--no-wait", action="store_true", dest="noWait", default=False, help="don't wait for results")
        parser.add_option("", "--debug", type="choice", choices=(Constants.off, Constants.low, Constants.medium, Constants.high),
                          action="store", dest="debug", default=Constants.off,
                          help="Enable debugging output")
        parser.add_option("","--server",action="store", type="string", dest="server", help="test server")
        parser.add_option("","--server-ip",action="store", type="string", dest="serverIP", help="test server ip address")
        parser.add_option("","--client",action="store", type="string", dest="client", help="test client host")
        parser.add_option("","--id",action="store", type="string", dest="id", help="sandbox ID")
        parser.add_option("","--test", action="append", type="string", dest="test",  help="Test name or path")
        parser.add_option("","--version",action="store", type="string", dest="version", help="vendor product version")
        parser.add_option("","--mode", action="store", type="choice", choices=(Constants.normal, Constants.auto), dest="mode",
                        help="Test Mode: normal, auto")
        parser.add_option("","--format", action="store", type="choice", choices=(Constants.text, Constants.html, Constants.junit),
                                        dest="format", help="Test results output format")

        # adding tag option to rhcertci run
        test_tags = TestTag.getAll()
        test_tags.remove('interactive')
        test_tags.remove('non-interactive')
        parser.add_option("", "--tag", action="append", type="choice", dest="tag", choices=test_tags)

        parser.set_defaults(tag=[TestTag.certification])
        parser.set_defaults(mode="auto")
        parser.set_defaults(results=None)
        (self.options,self.args) = parser.parse_args()

        if (len(self.args) > 2
                or (len(self.args) == 2 and not (self.args[0] == Constants.clean and self.args[1] in [Constants.results, Constants.all]))
                or (len(self.args) == 0)):
            parser.error("Invalid command options")
            exit(1)

        self.command = self.args[0]

        try:
            if self.options.help and self.command:
                self.__showHelp(self.command)
                exit(1)
        except:
            pass # --help was not set.

        self.optionParser = parser
        if self.command == "plan" and not self.options.server:
            self.setOptions({"server": None})

    def showHelp(self):
        """ format:
            command: [(
                    variant1 command spec,
                    variant1 description),
                    (variant2 command spec,
                    variant2 description)
                    ]
        """

        help = {
                "plan": [(
                    "plan [--client <name or IP address> --program (cloud|openstack) --id <local sandbox ID>]",
                    "plan certification testing")],
                "clean": [(
                    "clean [all]",
                    "remove all results and temporary files")],
                "run": [(
                "run [--test <testname or path>]"\
                        "[--client <client host name or IP address>  --program (cloud|openstack|hardware) --id <sandbox ID>]"\
                        "[--tag <test tag>]"\
                        "[--mode <Test Mode: normal, auto>]"\
                        "[--no-wait]",
                    "run all or specific certification tests")],
                "version": [(
                    "version",
                    "prints the version of rhcert")],
                "save": [(
                    "save  --program (cloud|openstack|hardware) --id <sandbox id> --server <rhcert server host> [--client <client name or IP>]",
                    "save a copy of the current test results to a local sandbox")],
                "print": [(
                    "print [--format (%s|%s|%s)]" % (Constants.text, Constants.html, Constants.junit),
                    "show a summary of the current test results")]
                }
        globalOptions = "[ --debug low|high ]"

        if len(sys.argv) > 1:
            command = help.get(sys.argv[1])
            if not command:
                print("sorry, couldn't find command \"%s\"" % command)
                return False
            for (variant, description) in command:
                print("\n%s %s" % (variant, globalOptions))
                print("\t%s" % description)
            print("")
            return True

        # else show them all
        for (command, variants) in help.items():
            for (variant, description) in variants:
                print("\n%s" % variant)
                print("\t%s" % description)
        print("")
        return True

    def setOptions(self, options):
        self.webOptions = options
        for (name, value) in options.items():
            self.options.__dict__[name] = value

    def getCurrentUTCTime(self):
        return time.gmtime(time.time())

    def __requireOptions(self, options):
        result = True
        for option in options:
            if not self.options.__dict__.get(option):
                print("Error: option %s is required." % option)
                result = False
        return result

    def do(self):
        if self.command == "clean":
            try:
                os.remove(self.environment.getLockFile())
            except OSError as e:
                pass

        result = False
        # do the command
        if self.command in self.commands:
            result = self.commands[self.command]()
        else:
            print("unknown command: " + self.command)
        return result

    def load(self):
        if self.options.client:
            if not self.__requireOptions(["id", "program"]):
                return False
            self.remote = Remote(self, self.options.program, self.options.id)
            self.resultsDocument = self.remote.getLatestResultsDocument()
        else:
            TestHarness.load(self)
            # TestHarness.certification is really a ResultsDocument
            self.resultsDocument = self.resultsDocument
        return True

    def doSave(self):
        if not self.__requireOptions(["id", "program", "server"]):
            return False

        # otherwise
        self.load()
        if not self.resultsDocument or self.resultsDocument.getNumberOfTestRuns() == 0:
            print("Error: no test results to save")
            return False

        if self.options.program not in Programs.getAll():
            print("Error: %s is not a certification program" % self.options.program)
            return False

        self.resultsDocument.setCertificationProgram(self.options.program)
        self.resultsDocument.setCertificationID(self.options.id)
        (savedFilePath, storePath) = self.saveToStore()
        if storePath:
            return self.saveToServer(savedFilePath, os.path.join(Constants.results, storePath))

        # otherwise
        return False

    def doPlan(self):

        if not self.load():
            return False

        if self.remote:
            return self.remote.doPlan()

        if not self.getLock():
            return False
        result = self.doLocalPlan()
        self.releaseLock()
        return result

    def doLocalPlan(self):
        self.ui.statusMessage("Verifying test plan")
        self.load()
        self._initializePlanner()
        tests = self.planner.plan(self.resultsDocument)
        if not self.resultsDocument:
            self.doDiscover()
            self.saveNewPlan()
            self.checkRequiredRPMs(tests)
        else:
            if not self.checkPlan():
                return False

        self.resultsDocument.save(self.environment.getResultsPath())
        if self.options.debug != Constants.off:
            print("Saved test plan to %s" % self.environment.getResultsPath())
        report = Report(self.options, self.resultsDocument)
        report.Write()
        return True

    def doRun(self):
        if not self.load():
            return False
        if self.remote:
            if not self.remote.doRun():
                return False
            return self.getClientResults()

        # otherwise, local run
        if not self.getLock():
            return False
        result = self.doLocalRun()
        self.releaseLock()
        return result

    def getClientResults(self):
        """ get client results for remote run """
        # reload results document
        self.resultsDocument =  self.remote.getLatestResultsDocument()
        result = self.printLastRunSummary()
        print("Test Results: %s" % self.__makeResultsURL())
        return result

    def __makeResultsURL(self):
        # example: http://192.168.124.145/certification/cloud/22228/virtualblue/2016-02-26%2014:59:37/results/
        if self.resultsDocument:
            return "http://%s/certification/%s/%s/%s/%s/results/" % (
                            socket.gethostname(),
                            self.resultsDocument.getCertificationProgram(),
                            self.resultsDocument.getCertificationID(),
                            self.resultsDocument.getHostname(),
                            self.resultsDocument.getPlanTime().replace(' ', "%20"))
        return "(could not find results document)"

    def doLocalRun(self):
        if not self.doLocalPlan():
            return False

        allTestDocuments = self.resultsDocument.getTests()
        testDocuments = list()
        if not self.options.test:
            testDocuments = allTestDocuments
        else:
            for testName in self.options.test:
                for testDocument in allTestDocuments:
                    if testName == testDocument.getName() or testName == testDocument.getPath():
                        testDocuments.append(testDocument)

        testDocuments = self.resultsDocument.getTaggedTests(testDocuments, self.options.tag)
        if len(testDocuments) > 0:
            testDocuments = self.addMandatoryTests(testDocuments)
            if not testDocuments:
                return False
            print("\nRunning the following tests:")
            for testDocument in testDocuments:
                print("%-10s %-36s" % (testDocument.getName(),  testDocument.getTestParameters()))
            return self._doRun(testDocuments)
        else:
            print("Error: no tests match the supplied options")
        return False

    def doCertify(self):
        if not self.getLock():
            return False
        result = self.__certify()
        self.releaseLock()
        return result
