# Copyright (c) 2016 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols

import os
import sys
import time
import textwrap

from rhcert.tags import Constants, TestTag, Attributes
from rhcert.command import Command
from rhcert.model.resultsdocument import ResultsDocument
from rhcert.client.harness import TestHarness
from rhcert.client.beacon import Antenna
from rhcert.host import Host
from rhcert.localTestServer import LocalTestServer
from rhcert.parser import RHCertOptionParser


class RHCertCLI(TestHarness):
    """ RHCertCLI drives the test Harness via command-line """

    def __init__(self, argv=sys.argv):
        self.commands = {
            'clean': self.doClean,
            'continue': self.doContinue,
            'plan': self.doPlan,
            'print': self.doPrint,
            'register': self.doRegister,
            'submit': self.doSubmit,
            'run': self.doRun,
            'save': self.doSave,
            'version': self.doVersion
        }
        self.argv = argv
        self.__getOptions()
        TestHarness.__init__(self, self.options, False)
        self.__getArgs()

    def showHelp(self):
        """ format:
            command: [(
                    variant1 command spec,
                    variant1 description),
                    (variant2 command spec,
                    variant2 description)
                    ]
        """

        help = {
            "continue": [(
                "continue",
                "continue an interrupted or incomplete test run")],
            "plan": [(
                "plan [--test <test> (--enable | --disable | --add (--device <device> | --udi <udi> | --property <key=value>))]",
                "plan certification testing")],
            "clean": [(
                "clean [all]",
                "remove all results and temporary files")],
            "run": [(
                "run [--test <testname or path>]+  [--tag <test tag>]+ [--mode <Test Mode: normal, auto>]",
                "run all or specific certification tests")],
            "version": [(
                "version",
                "prints the version of rhcert")],
            "save": [(
                "save [--server <rhcert server host>] [--server-ip <address>]",
                "save a copy of the current test results")],
            "submit": [(
                "submit [--results <test run results file path>]",
                "submit results for a certification")],
            "print": [(
                "print",
                "show a summary of the current test results")],
            "register": [(
                "register [--server <rhcert server hostname>] [--server-ip <rhcert server ip-address>]",
                "register the host with a Red Hat Certification Local Test Server")]
        }
        globalOptions = "[ --debug low|high ]"
        if len(self.argv) > 1:
            command = help.get(self.argv[1])
            if not command:
                print("sorry, couldn't find command \"%s\"" % command)
                return False
            for (variant, description) in command:
                print("\n%s %s" % (variant, globalOptions))
                print("\t%s" % description)
            print("")
            return True

        # else show them all
        for (command, variants) in help.items():
            for (variant, description) in variants:
                print("\n%s" % variant)
                print("\t%s" % description)
        print("")
        return True

    def __getOptions(self):
        usage = """
        %prog <command> [options]
        By default it executes clean, plan, run and print commands sequentially.

        command:
        clean [all] - remove all results and temporary files
        continue    - continue a test run
        plan        - plan certification testing
        print       - print certification test results
        register    - register with local test server
                      [--server <local test server hostname>]
                      [--server-ip <local test server IP>]
        run         - run specific certification tests
        save        - save a copy of the current test results
        submit      - submit certification results to Red Hat
        version     - prints the version of rhcert
        \n\nFor help on options, use %prog <command> --help
        """

        parser = RHCertOptionParser(self, usage)
        if len(self.argv) > 1:
            if self.argv[1] not in self.commands:
                if self.argv[1] not in ["--test", "-t"]:
                    if self.argv[1] not in ["help", "--help", "-h"]:
                        parser.error("No such command: %s" % self.argv[1])
                    else:
                        parser.print_usage(sys.stderr)
                    exit(1)
        self.optionParser = parser

        parser.add_option("-t", "--test",
                          action="append", type="string", dest="test", help="Test Name or Path")

        parser.add_option("", "--debug",
                          type="choice", choices=(Constants.off, Constants.low, Constants.medium, Constants.high),
                          action="store", dest="debug", default=Constants.off,
                          help="Enable debugging output")

        parser.add_option("", "--results",
                          action="store", type="string", dest="results", help="test results file path")

        parser.add_option("", "--server",
                          action="store", type="string", dest="server", help="test server")

        parser.add_option("", "--server-ip",
                          action="store", type="string", dest="serverIP", help="test server ip address")

        parser.add_option("-i", "--udi",
                          action="append", type="string", dest="udi", help="UDI - the unique device ID")

        # note that "forced" mode is not allowed on the command line, but is valid
        parser.add_option("-m", "--mode",
                          action="store", type="choice", choices=(Constants.normal, Constants.auto), dest="mode",
                          help="Test Mode: normal, auto")

        parser.add_option("-u", "--disable",
                          action="store_true", dest="disable", default=False, help="Disable a test")

        parser.add_option("-e", "--enable",
                          action="store_true", dest="enable", default=False, help="Re-enable a test")

        parser.add_option("-a", "--add",
                          action="store_true", dest="add", default=False, help="Add a test")

        parser.add_option("-v", "--device",
                          action="store", type="string", dest="device", help="The logical device to be tested.")

        # default is TestTag.certification, set explicitly below because optparse doesn't do the right thing
        parser.add_option("-g", "--tag",
                          action="append", type="choice", dest="tag", choices=TestTag.getAll())

        parser.add_option("-p", "--property",
                          action="KeyValuePair", type="string", dest="property",
                          help="Set a property for test <key=value> (repeat option to set multiple properties)")

        # harness options fixed or unused
        parser.set_defaults(mode="normal")
        parser.set_defaults(results=None)
        parser.set_defaults(udi=None)
        parser.set_defaults(device=None)
        parser.set_defaults(tag=[TestTag.certification])
        parser.set_defaults(disable=None)
        parser.set_defaults(enable=None)
        parser.set_defaults(add=None)
        parser.set_defaults(format=Constants.text)

        (self.options, self.args) = parser.parse_args(args=self.argv[1:])
        if (len(self.args) > 2
            or (len(self.args) == 2 and not (
                    self.args[0] == Constants.clean and self.args[1] in [Constants.results, Constants.all]))):
            parser.error("Invalid command options")
            exit(1)

        self.command = self.args[0] if self.args else None

        try:
            if self.options.help and self.command:
                self.showHelp(self.command)
                exit(1)
        except:
            pass  # --help was not set.

        if not self.options.server:
            self.__setOptions({"server": None})

    def __setOptions(self, options):
        self.webOptions = options
        for (name, value) in options.items():
            self.options.__dict__[name] = value

    def __getArgs(self):
        # check for valid commands
        self.command = None
        if self.args:
            self.command = self.args[0]
        else:
            self.commands['guided workflow'] = self.doGuidedWorkflow
            self.command = 'guided workflow'
        if not self.command in self.commands:
            self.optionParser.error("Unknown command \"%s\"" % self.command)
            sys.exit(1)

        return self.args

    def doGuidedWorkflow(self):
        message = """
        Performing following operations sequentially:
        clean  	- remove all results and temporary files
        plan    - plan certification testing
        run     - run specific certification tests
        print   - print certification test results
        """
        print(message)
        if self.doClean():
            if TestHarness.doPlan(self):
                self.doRun()
                self.doPrint()

    def doPlan(self):
        if TestHarness.doPlan(self):
            self.doPrint()
            return True
        return False

    def do(self):

        if not self.getLock():
            return False

        result = False
        # do the command
        if self.command in self.commands:

            # NOTE: Check whether the SUT has been registered with a LTS. In case,
            # it is not registered, we might lose the test data while registering
            # the SUT at later point of time. For such cases, display a prompt to
            # the user suggesting for registering before continuing the operation.
            isExpectedCommand = (self.command in ['register', 'clean'])
            certExists = os.path.exists(self.environment.getCertificationPath())
            isAutoMode = (self.options.mode == Constants.auto)

            if not isExpectedCommand and not certExists and not isAutoMode:
                msg = """
                    This test system is not registered with any server(LTS).
                    Do you want to register the system?"""
                msg = textwrap.dedent(msg)
                if self.ui.promptConfirm(msg):
                    self.commands['register']()

            result = self.commands[self.command]()
        else:
            print("Unknown command: " + self.command)
        self.releaseLock()
        return result

    def doRegister(self):
        """ report certification document to the server """

        self.__checkFirewall()
        if (self.options.server or self.options.serverIP or self.__autoDetectServer()):
            localTestServer = LocalTestServer(Host(self.options.server, self.options.serverIP))
            localTestServer.verify()
        else:
            while True:
                host = Host(self.ui.prompt("Please enter the server address you'd like to register:"))

                if not host.getHostname() and not host.getIPAddress():
                    if self.ui.promptConfirm("%s does not appear to be a valid hostname or IP address. retry?" % host.getName()):
                        continue
                    else:
                        return False

                localTestServer = LocalTestServer(host)

                if not localTestServer.verify():
                    if self.ui.promptConfirm(
                                    "%s is not a Red Hat Certification local test server. retry? " % host.getConnection()):
                        continue
                    else:
                        return False

                if host.getConnection():
                    self.options.server = host.getConnection()

                if host.getIPAddress():
                    self.options.serverIP = host.getConnection()

                print(("Using local test server %s" % host.getConnection()))
                break

        if not (self.options.server or self.options.serverIP):
            return False

        if localTestServer.isMultiStoreEnabled():
            username = ""
            while not username:
                username = self.workflowCLI.checkLogin()
                if not username and not self.ui.promptConfirm("Authentication failure, retry?"):
                    return False
            self.options.serverPath = os.path.join(username, localTestServer.getPhase(), Constants.systems)
        else:
            self.options.serverPath = Constants.systems

        if TestHarness.doRegister(self):
            # start listener
            # return self.doServer(["start"])
            # need to start a new shell
            return (os.system("rhcertd start") == 0)

        # otherwise
        return False

    def __checkFirewall(self):
        # skip port checking for RHEL6 - it's IP tables.
        if "Red Hat" in self.redHatRelease.getProduct() and self.redHatRelease.getVersion() < 7:
            if self.options.debug != Constants.off:
                print("Skipping firewall (iptables) checks for RHEL6")
            return True

        restartFirewallCommand = Command("systemctl restart firewalld")
        firewallPortsOpen = Command("firewall-cmd --zone=public --list-ports")
        listenerPort = self.environment.getServerListenerPort()
        try:
            print("checking ports...")
            firewallPortsOpen.run()
            firewallPortsOpen.printOutput()
        except Exception as error:
            print("Warning: Could not check port status using firewall-cmd")
            print(error)
            print("Warning: Please check that port %s is opened for both TCP and UDP" % listenerPort)
            return False

        try:
            firewallPortsOpenOutput = " ".join(firewallPortsOpen.output)
        except TypeError:
            firewallPortsOpenOutput = ""

        portsOpen = dict()
        for port in [("%s/tcp" % listenerPort), ("%s/udp" % listenerPort)]:
            portsOpen[port] = port in firewallPortsOpenOutput

        if False in portsOpen.values():
            print(("Port %s is not open for both TCP and UDP" % listenerPort))
            print(("ports open: " + " ".join(firewallPortsOpenOutput)))
            if self.ui.promptConfirm("Would you like to open the port?"):
                for (port, isOpen) in portsOpen.items():
                    if not isOpen:
                        openPortCommand = Command("firewall-cmd --zone=public --add-port=%s --permanent" % (port))
                        try:
                            print((openPortCommand.command))
                            openPortCommand.echo()
                        except Exception as error:
                            print("Error: could not open firewall ports:")
                            print(error)
                            return False
            try:
                print("restarting firewalld")
                restartFirewallCommand.run()
            except Exception as error:
                print("Warning: could not restart firewalld:")
                print(error)
                return False

        return True

    def __autoDetectServer(self):
        antenna = Antenna("rhcert:server")
        antenna.start()
        time.sleep(3)  # wait 3 seconds
        senders = antenna.getSenders()
        if senders:
            print(("Detected %s Local Test Server(s):" % len(senders)))
            if len(senders) > 1:
                for (address, _, host) in senders:
                    print(("   %s at %s" % (host, address)))
            for (address, _, host) in senders:
                if self.ui.promptConfirm("Would you like to connect to %s at %s? " % (host, address)):
                    print("setting server to %s" % address)
                    self.options.server = host
                    self.options.serverIP = address
                    return True
        return False

    def doSubmit(self):
        if not self.workflowCLI.isReachable():
            print("Error: The hardware catalog at %s is not reachable." % self.workflowCLI.getCatalogServer())
            return False

        # otherwise
        TestHarness.doSubmit(self)

    def doExtract(self):
        if self.options.results:
            try:
                self.resultsDocument = ResultsDocument()
                print("loading " + self.options.results)
                self.resultsDocument.load(self.options.results)
            except Exception as exception:
                print("Error: could not load results file % s" % self.options.results)
                print(exception)
                return False
        # otherwise
        print("loading results")
        self.load()
        if not self.resultsDocument:
            print("Error: no current test results")
            return False

        # otherwise
        testDocument = None
        if not self.options.test or len(self.options.test) > 1:
            print("Error: a test (and only one test) must be specified with --test ")
            return False
        if not self.options.run:
            print("Error: a test run must be specified via --run ")
            return False

        testDocument = self.resultsDocument.getTest(self.options.test[0])
        if not testDocument:
            print("Error: No such test \"" + self.options.test[0] + "\"")
            return False

        runDocument = testDocument.getRun(self.options.run)
        if not runDocument:
            print("Error: no such run %s for test %s" % (self.options.run, self.options.test[0]))
            return False

        # otherwise
        if not runDocument.getAttachments():
            print("No attachments in run %s for test %s" % (self.options.run, self.options.test[0]))
        for attachment in runDocument.getAttachments():
            sys.stdout.write("Extracting " + attachment.getAttribute(Attributes.name))
            attachmentFile = open(attachment.getAttribute(Attributes.name), "wb")
            # account for zero-length files
            data = ""
            if attachment.childNodes:
                data = attachment.childNodes[0].data
            attachmentFile.write(str(data).decode("base64"))
            attachmentFile.close()
            sys.stdout.write("\n")
        return True
