# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#

import sys, time
from optparse import OptionParser

from rhcert.tags import Constants, ServerConfiguration
from rhcert.listener.daemon import RedHatCertDaemon
from rhcert.listener.listen import RedHatCertificationListener
from rhcert.client.harness import TestHarness


class RHCertD(TestHarness):
    """ rhcertd: controls SUT/client services """

    Debugging = True

    def __init__(self, echoResponses=True):
        self.commands = ["start", "stop", "restart", "status", "daemon", "listener", ]
        self.getOptions()
        self.command = None
        TestHarness.__init__(self, self.options, echoResponses)

    def getOptions(self):
        usage = "usage: %prog  ({0}) [--debug {1}|{2}|{3}]".format(" | ".join(self.commands), Constants.low,
                                                                   Constants.medium, Constants.high) + \
                "\n\tstart: starts client processes" \
                "\n\tstop: stops the client processes" \
                "\n\tstatus: reports the status of client processes" \
                "\n\tdaemon: runs the daemon process in foreground" \
                "\n\tlistener: runs the listener process in the foreground" \
                "\n\n   client processes:" \
                "\n\tlistener: receives instructions and reports status to the server" \
                "\n\tdaemon: carries out instructions, typically running client testing."

        parser = OptionParser(usage)
        self.optionParser = parser

        parser.add_option("-d", "--debug", type="choice",
                          choices=(Constants.off, Constants.low, Constants.medium, Constants.high),
                          action="store", dest="debug", default=Constants.off,
                          help="Enable debugging output")

        parser.set_defaults(mode=Constants.server)
        parser.set_defaults(results=None)
        parser.set_defaults(server=None)
        parser.set_defaults(serverIP=None)

        (self.options, self.args) = parser.parse_args()
        self.__getArgs()

        if not self.args or len(self.args) > 1 or self.args[0] not in self.commands:
            parser.error("Invalid command")
            exit(1)

        self.optionParser = parser

    def setOptions(self, options, args=None):
        self.webOptions = options
        for (name, value) in options.items():
            self.options.__dict__[name] = value
        if args:
            self.args = args

    def __getArgs(self):
        # check for valid commands
        self.command = None
        if self.args:
            self.command = self.args[0]
        if not self.command in self.commands:
            self.optionParser.error("Unknown command \"%s\"" % self.command)
            sys.exit(1)
        return self.args

    def run(self, command=None):
        self.command = command
        result = True
        if not self.command:
            if len(self.args) > 0:
                self.command = self.args[0]
            else:
                self.command = "start"

        self._initializePlanner()
        daemon = RedHatCertDaemon(self.environment, self.planner, self.options)
        listener = RedHatCertificationListener(self, self.environment)

        if self.command == "restart":
            return self.restart()

        # run daemon in foreground
        if self.command == "daemon":
            if daemon.do("status"):
                print("Stopping background daemon")
                daemon.do("stop")  # kill any in the background
            print("Starting daemon")
            sys.stdout.flush()
            return daemon.run()

        # run listener in foreground
        if self.command == "listener":
            if listener.do("status"):
                print("Stopping background listener")
                listener.do("stop")  # kill any in the background
            print("Starting listener")
            sys.stdout.flush()
            return listener.run()

        if self.environment.getServerConfiguration() != ServerConfiguration.notInstalled:
            self.__runServerModeTests()

        # do the daemon
        if not daemon.do(self.command):
            result = False
        time.sleep(1)

        # do the listener
        if not listener.do(self.command):
            result = False

        # otherwise
        return result

    def __runServerModeTests(self):
        # otherwise, do test-specific services
        tests = self.planner.getServerModeTests();
        result = True
        for test in tests:
            print("%s test server %s:" % (test.getPath(), self.command))
            if not self.runServer(test, self.command):
                result = False
            print("")
        return result

    def restart(self):
        self.run("stop")
        return self.run("start")
