# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#

import sys
import os
import shutil
import pwd
import time
from optparse import OptionParser

from rhcert.tags import Constants, Attributes
from rhcert.model.resultsdocument import ResultsDocument
from rhcert.client.harness import TestHarness
from rhcert.command import Command
from rhcert.redhatrelease import RedHatRelease
try:
    import selinux
except:
    pass

class RHCertUtil(TestHarness):
    """ RHCertUtil rhcert command-line utilities

    Attachment Extraction

        rhcert-util extract --test <test path> --run <run number> [--results <results file> ]

        Extacts attachment files form the results.  
        The --test and --run options select the test run.
        The results are either the current (local) results file, or a results file passed via --results
    """

    Debugging=True

    def __init__(self, echoResponses=True):
        self.redHatRelease = RedHatRelease()
        self.getOptions()
        TestHarness.__init__(self, self.options, echoResponses)
        self.command = None
        self.commands = {
            'update': self.doUpdate,
            'extract': self.doExtract,
            'version': self.doVersion,
            'configure': self.doConfigure,
            'translate': self.doTranslate
        }
        if self.redHatRelease.getVersion() >= 7:
            self.commands['clean'] = self.doCleanLocalStorage
        self.configureSettings = ["default", "management", "testing"]

    def getOptions(self):
        clean_local_storage = str()
        if self.redHatRelease.getVersion() >= 7:
            clean_local_storage = "\tclean ([clean local storage])\n" \
                        "\t\t- reset /var/www/rhcert/store to its initial state.\n"
        usage = "usage: %prog <command> [options]\n\ncommand:\n"\
                    "\tupdate [--results <results file path>]  \n"\
                        "\t\t- update deferred tests\n"\
                    "\textract [--results <results file path>] --test <test path> --run <run number>\n"\
                        "\t\t- extract test run attachment files\n"\
                    "\tconfigure (default|management|testing)\n"\
                        "\t\t- configure the installation type.\n"\
                    "\ttranslate ([--results <results file path>] --xslt <xslt path> [--xparam <name>:<value>]*\n"\
                        "\t\t- translate results via xslt.\n" + clean_local_storage

        parser = OptionParser(usage)

        parser.add_option("-b", "--debug",
            type="choice", choices=(Constants.off, Constants.low, Constants.medium, Constants.high),
            action="store", dest="debug", default=Constants.off, help="Enable debugging output")

        # attachment extraction
        parser.add_option("-t","--test", action="append", type="string", dest="test", help="Test Name")
        parser.add_option("", "--results", action="store", type="string", dest="results", help="results file")
        parser.add_option("", "--run", action="store", type="int", dest="run", help="test run number")
        parser.add_option("", "--xslt", action="store", type="string", dest="xslt", help="XSLT file path")
        parser.add_option("","--xparam", action="append", type="string", dest="xparam", help="XSLT parameter")

        parser.set_defaults(mode="normal")
        parser.set_defaults(results=None)

        (self.options,self.args) = parser.parse_args()
        if len(self.args) > 3 or ((len(self.args) == 2 and (self.args[0] != "configure")) or (
        (len(self.args) == 3 and (' '.join(self.args) != "clean local storage")))):
            parser.error("Invalid command")
            exit(1)

        self.optionParser = parser

    def setOptions(self, options, args=None):
        self.webOptions = options
        for (name, value) in options.items():
            self.options.__dict__[name] = value
        if args:
            self.args = args

    def getArgs(self):

        # check for valid commands
        self.command = None
        if self.args:
            self.command=self.args[0]
        if not self.command in self.commands:
            self.optionParser.error( "Unknown command \"%s\"" % self.command)
            sys.exit(1)
        if self.command in "server" and len(self.args) > 1 and not self.args[1] in self.serverCommands:
            self.optionParser.error( "Unknown server command \"%s\"" % self.args[1])
            sys.exit(1)

        return self.args

    def do(self, args):
        if not self.getArgs():
            return False

        if not self.getLock():
            return False

        result = False
        # do the command
        if self.command in self.commands:
            result = self.commands[self.command]()
        else:
            print("unknown command: " + self.command)
        self.releaseLock()
        return result

    def doExtract(self):
        if self.options.results:
            try:
                self.resultsDocument = ResultsDocument()
                print("loading " + self.options.results)
                self.resultsDocument.load(self.options.results)
            except Exception as exception:
                print("Error: could not load results file % s" % self.options.results)
                print(exception)
                return False
        # otherwise
        print("loading results")
        self.load()
        if not self.resultsDocument:
            print("Error: no current test results")
            return False

        # otherwise
        if not self.options.test or len(self.options.test) >1:
            print("Error: a test (and only one test) must be specified with --test ")
            return False
        if not self.options.run:
            print("Error: a test run must be specified via --run ")
            return False

        testDocument = self.resultsDocument.getTest(self.options.test[0])
        if not testDocument:
            print("Error: No such test \"" + self.options.test[0] + "\"")
            return False

        runDocument = testDocument.getRun(self.options.run)
        if not runDocument:
            print("Error: no such run %s for test %s" % (self.options.run, self.options.test[0]))
            return False

        # otherwise
        if not runDocument.getAttachments():
            print("No attachments in run %s for test %s" % (self.options.run, self.options.test[0]))
        for attachment in runDocument.getAttachments():
            sys.stdout.write("Extracting " + attachment.getAttribute(Attributes.name))
            attachmentFile = open(attachment.getAttribute(Attributes.name), "wb")
            # account for zero-length files
            data = ""
            if attachment.childNodes:
                data = attachment.childNodes[0].data
            attachmentFile.write(str(data).decode("base64"))
            attachmentFile.close()
            sys.stdout.write("\n")
        return True

    def doTranslate(self):
        if not self.options.xslt:
            print("Error: no XSLT file supplied with --xslt option.")
            return False

        if not self.options.results:
            print("Error: no results file supplied with --results option.")
            return False
        # xslt paramters
        xsltParams = {}
        for param in self.options.xparam:
            nameValue = param.split(':')
            if len(nameValue) != 2:
                print("Error: could not parse XSLT Parameter %s expected <name>:<value>" % param)
                return False

            xsltParams[nameValue[0]] = "\'%s\'" % nameValue[1]
        print ("XSLT Parameters:")
        print (xsltParams)
        try:
            self.resultsDocument = ResultsDocument()
            sys.stdout.write("translating... ")
            start = time.time()
            self.resultsDocument.loadFiltered(self.options.results, self.options.xslt, xsltParams)
            stop = time.time()
            print("\ntranslation of %s took %s sec." % (os.path.basename(self.options.results), stop-start))
        except Exception as exception:
            print("Error: could not translate file % s" % self.options.results)
            print(exception)
            return False

        # otherwise
        (outputName, _) = os.path.splitext(os.path.basename(self.options.results))
        outputFile = outputName + "-translated.xml"
        self.resultsDocument.save(outputFile)
        print("translated file saved to %s" % outputFile)
        return True

    def doConfigure(self):
        print(("Previous Server configuration: %s" % self.environment.getServerConfiguration()))
        if len(self.args) < 2:
            print("The available configurations are: " + " ".join(self.configureSettings))
            return True
        if self.args[1] not in self.configureSettings:
            print(("Error: invalid configuration %s" % self.args[1]))
            return False
        self.environment.setServerConfiguration(self.args[1])
        self.environment.save()
        print(("Current Server configuration: %s" % self.environment.getServerConfiguration()))
        return True

    def doCleanLocalStorage(self):
        message = ("This command will delete all data stored locally, including test results and test systems \n"
                   "registrations from local storage.")
        if len(self.args) < 3:
            print("%s Please run 'clean local storage' to clean them all." % message)
            return True

        store_path = str(self.environment.getStoreDirectory())
        prompt_msg = ("%s Enter 'confirm' if you are really sure you want to do this: " % message)

        if store_path in str(os.getcwd()):
            print("Please change your current working directory. You can't delete a folder being in that folder.")
            return False

        user_choice = self.ui.prompt(prompt_msg)
        if user_choice.lower() == 'confirm':
            try:
                print("Please wait.. Stopping httpd server")
                os.system("systemctl stop httpd")
                if os.path.exists(store_path):
                    print("Cleaning RHCert local storage (%s)..." % store_path)
                    shutil.rmtree(store_path)
                print("Creating RHCert local storage (%s)..." % store_path)
                os.mkdir(store_path, 0o755)
                apache_uid = pwd.getpwnam('apache').pw_uid
                apache_gid = pwd.getpwnam('apache').pw_gid
                os.chown(store_path, apache_uid, apache_gid)
                Command("semanage fcontext -a -t httpd_sys_rw_content_t %s" % store_path).run()
                selinux.restorecon(store_path)
                print("Restarting httpd server")
                os.system("systemctl restart httpd")
                return True
            except Exception as exception:
                print("Exception occurred: %s" % exception)
        else:
            print("cleaning Abort")
        return False
