# Copyright (c) 2012 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
# WorkflowCLI is the text UI for submitting results and other interaction with CWE

from rhcert.controller import Controller
from rhcert.model.resultsdocument import ResultsDocument
from rhcert.model.certificationdocument import CertificationDocument
from rhcert.workflow.certification import Certification as CertificationWorkflow
from rhcert.workflow.config import Config

class WorkflowCLI(Controller):

    def __init__(self, environment, rhcertStore, debug):
        Controller.__init__(self, debug, False)
        self.environment = environment
        self.rhcertStore = rhcertStore
        self.username = None
        self.password = None
        self.workflow = CertificationWorkflow(Config(environment, rhcertStore, debug))

    def isReachable(self):
        return self.workflow.isReachable()

    def getCatalogServer(self):
        return self.environment.getCatalogServer()

    def checkLogin(self):
        self.__getLogin(force=True)
        if self.workflow.login(self.username, self.password):
            return self.username
        return ""

    def submit(self, resultsDocument):
        """ submit xml results to the catalog
            returns (submitted, success)
            submitted - True if results successfully submitted
            success - False if error occurred
        """
        self.__getLogin()

        if self.__getCertificationID(resultsDocument):
            # save current results/certification to preserve program and ID info
            resultsDocument.save()
            certificationDocument = CertificationDocument()
            certificationDocument.copy(resultsDocument)
            certificationDocument.save(self.environment.getCertificationPath())

            if self.__submitResults(resultsDocument):
                return (True, True)

        return (False, False)

    def __getLogin(self, force=False):
        usernamePrompt = "Red Hat SSO User Name: "
        if force or not self.password:
            self.username = self.ui.prompt(usernamePrompt)
            self.password = self.ui.promptPassword("Password: ")

    def __getCertificationID(self, resultsDocument):
        """ command-line UI for getting and existing cert """

        if not self.isReachable():
            print("Could not reach the Red Hat Hardware Catalog %s" % self.getCatalogServer())
            return False

        oldCertID = None
        if resultsDocument.getCertificationID() and str(resultsDocument.getCertificationID()) != '0':
            if self.ui.promptConfirm("Submitting to certification %s?" % resultsDocument.getCertificationID()):
                return True
        # otherwise
        print("Certification ID is now %s" % oldCertID)
        certID = self.ui.prompt("Please enter the Cert ID: ", default=oldCertID)
        resultsDocument.setCertificationID(certID)

        return True


    def __submitResults(self, resultsDocument):

        print("compressing and encoding results...")
        resultsDocument.checkRequiredOS() # CERTPX-4042: check <certification><os> entry
        compressedFilePath = resultsDocument.compressToFile(resultsDocument.getVerboseFileName())

        self.workflow.login(self.username, self.password)
        print("uploading results...")
        self.workflow.clearErrors()
        response = self.workflow.submitTestRun(compressedFilePath, resultsDocument)
        print("done.")
        if response and not self.workflow.getErrors():
            # copy over to certification.xml to save the certification program and ID(s)
            certificationDocument = CertificationDocument()
            certificationDocument.copy(resultsDocument)
            certificationDocument.save(self.environment.getCertificationPath())
            return True
        # otherwise
        for error in self.workflow.getErrors():
            print(error)
        return False

if __name__ == "__main__":

    pass
