# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#

# parse simple config files of the form:

# <parametername> <value(s)>


class ConfigFile:

    def __init__(self, filePath, separator=None):
        self.filePath = filePath
        self.parameters = dict()
        # For tests that want to track the paramaters that they adding/removing to the config file
        self.modifiedParams = {'added':list(), 'removed':list()}
        # modifiedParams ex:
        # {'added': [{'name':'a','value':'a'},{'name':'b','value':'b'}] ,'removed':[{'name':'c','value':'c'}]}
        self.__parse(separator)
        self.modified = False

    def __parse(self, separator):
        try:
            configFile = open(self.filePath)
        except IOError:
            print("Error: configuration file %s file not found" % self.filePath)
            return False
        self.lines = configFile.readlines()
        for line in self.lines:
            # skip blank and comment lines
            if not line.strip() or line.strip().startswith("#"):
                continue
            words = line.strip().split(separator)
            if words and words[0]:
                self.parameters[words[0].strip()] = " ".join(words[1:]).strip()
        configFile.close()

    def getParameter(self, name):
        if self.parameters:
            try: 
                return self.parameters[name]
            except KeyError:
                pass
        return None

    def dump(self):
        if self.parameters:
            for key in self.parameters:
                print("%s = %s" % (key, self.parameters[key]))

    def addParameter(self, name, value):
        if not self.getParameter(name):
            self.parameters[name] = value
            self.lines.append("%s %s\n" % (name, value))
            self.modified = True
            param = {'name': name, 'value': value}
            self.modifiedParams['added'].append(param)
            return True
        # otherwise
        return False

    def removeParameter(self, name):
        if self.getParameter(name):
            param = {'name': name, 'value': self.getParameter(name)}
            self.modifiedParams['removed'].append(param)
            del self.parameters[name]
            saveLines = list()
            for line in self.lines:
                # skip/preserve comment lines
                if line.strip() and line.strip()[0] == "#":
                    saveLines.append(line)
                    continue
                words = line.strip().split(" ")
                if words and words[0] == name:
                    self.modified = True
                else:
                    saveLines.append(line)
            if self.modified:
                self.lines = saveLines

    def save(self):
        if self.modified:
            configFile = open(self.filePath, "w")
            for line in self.lines:
                configFile.write(line)

if __name__ == "__main__":
    try:
        confFile = ConfigFile("qq.txt")
        confFile.removeParameter("net")
        confFile.save()
        print("------------------------")
        confFile.dump()
        confFile = ConfigFile("qq.txt")
        confFile.addParameter("net", "asd")
        confFile.save()
        print("------------------------")
        confFile.dump()
    except Exception as exception:
        print(exception)
        exit(1)
    exit(0)
