#!/usr/bin/env python
# Copyright (c) 2008 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols <gnichols@redhat.com>

import os, time, syslog, json

from rhcert.tags import Constants
from rhcert.command import Command
from rhcert.controller import Controller
from rhcert.model.documentbase import DocumentWrapper
from rhcert.redhatrelease import RedHatRelease


class Continuation(Controller):

    def __init__(self, environment):
        Controller.__init__(self)
        self.environment = environment
        self.documentWrapper = DocumentWrapper()
        self.rebootTimeLimit = self.environment.getRebootTimelimit() 
        self.method = None
        self.timestamp = None
        self.kernel = None
        self.bootPrintPath = os.path.join(self.environment.getDataDirectory(), "bootprint")
        self.redHatRelease = RedHatRelease()

    def setInitConfig(self, marker, method=None):
        if self.redHatRelease.getVersion() < 7:
            pass
        else:
            Command("systemctl restart systemd-journald").echoIgnoreErrors()
        if method:
            self.method = method
        if self.redHatRelease.getVersion() < 7:
            chkconfig = Command("chkconfig --add " + Constants.rhcertd)
            chkconfig.echo()
        else:
            Command("systemctl enable %s" % Constants.rhcertd).echoIgnoreErrors()
            Command("systemctl daemon-reload").echoIgnoreErrors()
        # get a timestamo, save it
        theTime = time.localtime(time.time())
        self.timestamp = self.documentWrapper.timeToString(theTime)
        # save it off to the side
        timestamp = open(self.bootPrintPath, "w")
        timestamp.write(self.timestamp + "\n")
        if self.method:
            timestamp.write(self.method + "\n")
            timestamp.write(RedHatRelease().getKernel() + "\n")
        timestamp.close()
        # mark the log with this run time
        markerName = "%s-%s" % (marker, self.timestamp)
        syslog.syslog(self.getSystemLogMarker(markerName, "begin", pid=False))

    def removeInitConfig(self):
        if self.redHatRelease.getVersion() < 7:
            chkconfig = Command("chkconfig --del " + Constants.rhcertd)
            chkconfig.echo()
        else:
            Command("systemctl disable %s" % Constants.rhcertd).echoIgnoreErrors()
            Command("systemctl daemon-reload").echoIgnoreErrors()

    def isInitialized(self):
        return os.path.exists(self.bootPrintPath)

    def dumpDictDataInFile(self, filename, data):
        with open(os.path.join(self.environment.getDataDirectory(), filename), 'w') as dataFile:
            json.dump(data, dataFile)

    def loadDictDataFromFile(self, filename):
        with open(os.path.join(self.environment.getDataDirectory(), filename)) as dataFile:
            return json.load(dataFile)

    def deleteDumpFile(self, filename):
        try:
            os.remove(os.path.join(self.environment.getDataDirectory(), filename))
        except OSError as exception:
            print("Could not delete " + os.path.join(self.environment.getDataDirectory(), filename))
            print(exception)

    def verify(self, marker, maxReboots=1):
        Command("systemctl restart systemd-journald").echoIgnoreErrors()
        # get the log, verify reboot happened
        try:
            timestamp = open(self.bootPrintPath)
            self.timestamp = timestamp.readline().strip()
            method = timestamp.readline()
            if method:
                self.method = method.strip()
                kernel = timestamp.readline()
                if kernel:
                    self.kernel = kernel.strip()
            timestamp.close()
            os.remove(self.bootPrintPath)

            theTime = time.localtime(time.time())
            duration = self.documentWrapper.duration(self.timestamp, self.documentWrapper.timeToString(theTime))
            print("reboot took " + duration)
            print("method: " + self.method)
            print("kernel: " + self.kernel)
            durationData = self.documentWrapper.durationData(self.timestamp, self.documentWrapper.timeToString(theTime))
            # failure if 2X reboot limit
            if durationData.seconds > self.rebootTimeLimit*60*2:
                print("Error: reboot took longer than %u minutes" % (self.rebootTimeLimit*2))
                return False
            # warn for 1X reboot limit
            elif durationData.seconds > self.rebootTimeLimit*60:
                print("Warning: reboot took longer than %u minutes" % self.rebootTimeLimit)
            kernel = RedHatRelease().getKernel()
            if self.kernel != kernel:
                print("Error: rebooted a different kernel:")
                print("    before: " + self.kernel)
                print("    after:  " + kernel)
                return False
            log = self.getSystemLog("%s-%s" % (marker, self.timestamp.strip()), pid=False)
            rebootCount = 0
            for line in log.split('\n'):
                if "kernel:" in line and self.systemLogBootMarker in line:
                    rebootCount += 1
            if rebootCount > maxReboots:
                print("Error: system rebooted %u times" % rebootCount)
                if maxReboots > 1:
                    print("Only %u reboots per test run are allowed." % maxReboots)
                return False
            if not rebootCount:
                print("Warning: could not detect reboot")

            print("reboot verified or otherwise accepted")
            # print "Reboot log:\n---------------------------------------------------------"
            # print log
        except Exception as e:
            print("Could not verify reboot")
            print(e)
            return False

        return True

    def getMethod(self):
        return self.method
    def getTimestamp(self):
        return self.timestamp
    def getDuration(self, timestamp):
        return self.documentWrapper.durationData(self.timestamp, timestamp)
