# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#

import os, errno, re, subprocess, time

from rhcert.compatability import rhcert_make_unicode
from rhcert.testUI import TestUI
from rhcert.tags import Constants


class Controller(object):

    def __init__(self, debug=Constants.off, echoResponses=True):
        self.ui = TestUI(echoResponses)
        self.systemLogMarker = "rhcert/runtests" 
        self.systemLogBootMarker = "Linux version"
        self.debug = debug

    def removeDirectory(self, directory):
        "Remove a directory (and all its contents)"
        try:
            for (root,dirs,files) in os.walk(directory, topdown=False):
                for f in files: os.unlink(os.path.join(root,f))
                for d in dirs: os.rmdir(os.path.join(root,d))
            os.rmdir(directory)
        except OSError:
            return
    def makeDirectoryPath(self, directory):
        """Emulate mkdir -p behavior: create the given directory, including all
        needed parent directories. Do not return an error if the directories 
        already exist."""
        try:
            os.makedirs(directory)
        except OSError as e:
            if e.errno != errno.EEXIST: raise

    def removeBadCharactersInName(self, value):
        """Remove bad characters []()\/&@in generated names"""
        pattern = re.compile("[^\[\]\(\)\\\/&@,]")
        goodList = pattern.findall(value)
        return "".join(goodList)

    def getSystemLogMarker(self, markerName, mark, pid=True):
        if pid:
            return self.systemLogMarker+"[%s]: %s: %s" % (os.getpid(), markerName, mark)
        #otherwise
        return self.systemLogMarker+": %s: %s" % (markerName, mark)        

    def getSystemLogOpen(self):
        return self.systemLogMarker+"[%s]" % os.getpid()

    def getSystemLog(self, markerName, pid=True):
        "Get a named section of the system log"

        log = self.getSystemLogSinceBoot()
        beginMark = self.getSystemLogMarker(markerName, "begin", pid)
        endMark = self.getSystemLogMarker(markerName, "end", pid)

        contents = []
        logiter = iter(log)

        for l in logiter:
            if beginMark in l:
                contents.append(l)
                break

        for l in logiter:
            contents.append(l)

            if endMark in l:
                break

        return ''.join(contents)

    def getSystemLogSinceBoot(self):

        try:
            # Got logs for current boot
            try:  # python2
                contents = []
                log = subprocess.Popen(['journalctl', '-b0'], stdout=subprocess.PIPE).communicate()[0]

                for l in log.split('\n'):
                    try:
                        contents.append(l.decode('utf-8') + '\n')
                    except UnicodeDecodeError as error:
                        if self.debug != Constants.off:
                            print('Warning: Unicode decode error in system log')
                            print(error)
                return contents

            except TypeError:  # python3
                log = subprocess.Popen(['journalctl', '-b0'],stdout=subprocess.PIPE, encoding='utf8').communicate()[0]
                return log.split('\n')

        except OSError:
            # If journalctl does not exist, fall back to normal log reading.
            pass
        except subprocess.CalledProcessError:
            # Try syslog if journalctl returns an error.
            pass

        syslog = '/var/log/messages'
        log = open(syslog)
        contents = list()
        marker = False
        for l in log:
            try:
                try:  # python2
                    l = l.decode("utf-8")
                except AttributeError:
                    pass  # python3 already unicode

                if l.find("kernel:") != -1 and l.find(self.systemLogBootMarker) != -1:
                    marker = True
                    # dump old contents
                    contents = list()
                if marker:
                    contents.append(l)
            except UnicodeDecodeError:
                # if we can't decode it, skip the whole line
                continue
        log.close()
        return contents

    def getRedHatRelease(self):
        return self.ui.getRedHatRelease()

    def getCurrentUTCTime(self):
        return time.gmtime(time.time())
