# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
#Author: Greg Nichols

import difflib, shutil, os
try:
    from urllib.parse import urljoin, urlparse # python3
except ImportError:
    from urlparse import urljoin, urlparse # python2

import rhcert.compatability
import re
from rhcert.tags import Tags, Attributes, Constants, ServerConfiguration
from rhcert.model.documentbase import DocumentBase


class Environment(DocumentBase):

    def __init__(self):
        DocumentBase.__init__(self)
        self.environmentFile = "/etc/rhcert.xml"
        self.defaultEnvironmentFile = "/etc/rhcert-default.xml"

        try:
            self.load(self.environmentFile)
        except IOError:
            shutil.copy(self.defaultEnvironmentFile, self.environmentFile)
            self.load(self.environmentFile)

    def showChangesFromDefault(self):
        environmentFile = open(self.environmentFile)
        defaultFile = open(self.defaultEnvironmentFile)

        diff = difflib.context_diff(environmentFile.readlines(), defaultFile.readlines())
        delta = ''.join(x for x in diff if (x.startswith('- ') or x.startswith('+ ')))
        print(delta)

        return len(delta) > 0

    def _getFilesystemElement(self):
        return self.findOrCreateElement(self.document.documentElement, Tags.file_system)

    def _getURLsElement(self):
        return self.findOrCreateElement(self.document.documentElement, Tags.urls)

    def _getTimelimitsElement(self):
        return self.findOrCreateElement(self.document.documentElement, Tags.time_limits)

    def _getServerElement(self):
        return self.findOrCreateElement(self.document.documentElement, Tags.server)

    def getDataDirectory(self):
        dataDirectoryElement = self.findOrCreateElement(self._getFilesystemElement(), Tags.data_directory)
        return self.getTextNode(dataDirectoryElement).data

    def getServerDirectory(self):
        serverDirectoryElement = self.findOrCreateElement(self._getFilesystemElement(), Tags.server_directory)
        if not self.getTextNode(serverDirectoryElement):
            return None
        return self.getTextNode(serverDirectoryElement).data

    def getServerExport(self):
        self.getServerDirectory() # so it gets created if not found
        serverDirectoryElement = self.findOrCreateElement(self._getFilesystemElement(), Tags.server_directory)
        if not self.getTextNode(serverDirectoryElement):
            return None
        return serverDirectoryElement.getAttribute(Attributes.export)

    def getServerTransfer(self):
        self.getServerDirectory() # so it gets created if not found
        serverDirectoryElement = self.findOrCreateElement(self._getFilesystemElement(), Tags.server_directory)
        if not self.getTextNode(serverDirectoryElement):
            return None
        return serverDirectoryElement.getAttribute(Attributes.transfer)

    def getStoreDirectory(self):
        storeDirectoryElement = self.findOrCreateElement(self._getFilesystemElement(), Tags.store_directory)
        return self.getTextNode(storeDirectoryElement).data

    def isMultiStoreEnabled(self):
        storeDirectoryElement = self.findOrCreateElement(self._getFilesystemElement(), Tags.store_directory)
        return storeDirectoryElement.getAttribute(Attributes.multistore_enabled) == "enabled"

    def getLibraryDirectory(self):
        libraryDirectoryElement = self.findOrCreateElement(self._getFilesystemElement(), Tags.library_directory)
        return self.getTextNode(libraryDirectoryElement).data
    def getSuiteDirectory(self):
        sourceDirectoryElement = self.findOrCreateElement(self._getFilesystemElement(), Tags.suite_directory)
        return self.getTextNode(sourceDirectoryElement).data

    def getLogDirectory(self):
        logDirectoryElement = self.findOrCreateElement(self._getFilesystemElement(), Tags.log_directory)
        return self.getTextNode(logDirectoryElement).data

    def getLogRunsDirectory(self):
        return os.path.join(self.getLogDirectory(), Constants.runs)

    def getResultsFileName(self):
        dataDirectoryElement = self.findOrCreateElement(self._getFilesystemElement(), Tags.data_directory)
        return dataDirectoryElement.getAttribute(Attributes.database_name)

    def getCertificationFileName(self):
        dataDirectoryElement = self.findOrCreateElement(self._getFilesystemElement(), Tags.data_directory)
        if not dataDirectoryElement.getAttribute(Attributes.certification_name):
            return None
        return dataDirectoryElement.getAttribute(Attributes.certification_name)

    def getLockFile(self):
        dataDirectoryElement = self.findOrCreateElement(self._getFilesystemElement(), Tags.data_directory)
        return dataDirectoryElement.getAttribute(Attributes.lock_file)

    def getTaskFile(self):
        dataDirectoryElement = self.findOrCreateElement(self._getFilesystemElement(), Tags.data_directory)
        return dataDirectoryElement.getAttribute(Attributes.task_file)

    def getResultsPath(self):
        return "%s/%s" % (self.getDataDirectory(), self.getResultsFileName())

    def getCertificationPath(self):
        return "%s/%s" % (self.getDataDirectory(), self.getCertificationFileName())

    def getTaskFilePath(self):
        return "%s/%s" % (self.getDataDirectory(), self.getTaskFile())

    def getKVMGuestImageDirectory(self):
        guestImageElement = self.findOrCreateElement(self._getFilesystemElement(), Tags.kvm_guest_image_directory)
        if not self.getTextNode(guestImageElement):
            return None
        return self.getTextNode(guestImageElement).data

    def getKVMGuestNvramDirectory(self):
        guestNvramElement = self.findOrCreateElement(self._getFilesystemElement(), Tags.kvm_guest_nvram_directory)
        if not self.getTextNode(guestNvramElement):
            return None
        return self.getTextNode(guestNvramElement).data

    def getKVMGuestConfigDirectory(self):
        guestConfigElement = self.findOrCreateElement(self._getFilesystemElement(), Tags.kvm_guest_config_directory)
        if not self.getTextNode(guestConfigElement):
            return None
        return self.getTextNode(guestConfigElement).data

    def getFvImageMountPoint(self):
        fvImageMountPoint = self.findOrCreateElement(self._getFilesystemElement(), Tags.fv_image_mountpoint)
        if not self.getTextNode(fvImageMountPoint):
            return None
        return self.getTextNode(fvImageMountPoint).data

    def getPhase(self):
        """ return the enviroment phase: qa, stage if not production """
        phase = self._getURLsElement().getAttribute(Attributes.phase)
        if phase:
            if phase == "hydrastage":
                phase = "stage"
            elif phase == "hydraqa":
                phase = "qa"
            return phase
        return "" # empty string means production

    def __getPhaseCatalogDomain(self):
        phase = self.getPhase()
        phase = phase + "." if phase else ""
        return "https://access.%sredhat.com" % phase

    def __getPhaseSSODomain(self):
        phase = self.getPhase()
        phase = phase + "." if phase else ""
        return "https://sso.%sredhat.com" % phase

    def getPartnerServerURL(self):
        element = self.findOrCreateElement(self._getURLsElement(), Tags.partner_server)
        if not self.getTextNode(element) or not self.getTextNode(element).data:
            return None
        return self.getTextNode(element).data

    def getCatalogURL(self):
        customURL = self.__getCustomCatalogURL()
        if customURL:
            return customURL

        # otherwise, normal case(s)
        if (self._getURLsElement().getAttribute(Attributes.phase) == "hydrastage" or
            self._getURLsElement().getAttribute(Attributes.phase) == "hydraqa"):
            return urljoin(self.__getPhaseCatalogDomain(), "/hydra/rest/cwe/xmlrpc/v2")
        return urljoin(self.__getPhaseCatalogDomain(), "/hydra/rest/cwe/xmlrpc/")

    def getCatalogServer(self):
        customURL = self.__getCustomCatalogURL()
        if customURL:
            url = urlparse(customURL)
            return ("%s://%s" % (url.scheme, url.netloc))

        # normal case
        return self.__getPhaseCatalogDomain()

    def __getCustomCatalogURL(self):
        # support custom dev urls
        try:
            catalogURLs = self._getURLsElement().getElementsByTagName(Tags.catalog_url)
            return self.getTextNode(catalogURLs[0]).data
        except:
            pass
        return None

    def getSSOServer(self):
        return self.__getPhaseSSODomain()

    def getServicesURL(self):
        return urljoin(self.__getPhaseCatalogDomain(), "/hydra/rest/cwe/")

    def getCatalogSearchTimeout(self):
        catalogURLElement = self.findOrCreateElement(self._getURLsElement(), Tags.catalog_url)
        if not catalogURLElement.getAttribute(Attributes.search_timeout):
            return None
        return int(catalogURLElement.getAttribute(Attributes.search_timeout))

    def getCatalogSearchRetries(self):
        catalogURLElement = self.findOrCreateElement(self._getURLsElement(), Tags.catalog_url)
        if not catalogURLElement.getAttribute(Attributes.search_retries):
            return None
        return int(catalogURLElement.getAttribute(Attributes.search_retries))

    def getMaximumAttachmentSize(self):
        return int(self._getFilesystemElement().getAttribute(Attributes.maximum_attachment_size))

    def getMaximumSosPluginSize(self):
        if not self._getFilesystemElement().getAttribute(Attributes.maximum_sos_plugin_size):
            self._getFilesystemElement().setAttribute(Attributes.maximum_sos_plugin_size, "52428800") #bytes
        return int(self._getFilesystemElement().getAttribute(Attributes.maximum_sos_plugin_size))

    def getMaximumDialogAttachmentSize(self):
        if not self._getFilesystemElement().getAttribute(Attributes.maximum_dialog_attachment_size):
            self._getFilesystemElement().setAttribute(Attributes.maximum_dialog_attachment_size, "20971520") #bytes
        return int(self._getFilesystemElement().getAttribute(Attributes.maximum_dialog_attachment_size))

    def getResultsWarningSize(self):
        if not self._getFilesystemElement().getAttribute(Attributes.results_warning_size):
            return None
        return int(self._getFilesystemElement().getAttribute(Attributes.results_warning_size))

    def getFVGuestTimelimit(self):
        if not self._getTimelimitsElement().getAttribute(Attributes.fv_guest_timelimit):
            return None
        return int(self._getTimelimitsElement().getAttribute(Attributes.fv_guest_timelimit))

    def getRebootTimelimit(self):
        if not self._getTimelimitsElement().getAttribute(Attributes.reboot_timelimit):
            return None
        return int(self._getTimelimitsElement().getAttribute(Attributes.reboot_timelimit))

    def getSSHRetryTimelimit(self):
         if not self._getTimelimitsElement().getAttribute(Attributes.ssh_retry_timeout):
             return self._getFilesystemElement().setAttribute(Attributes.ssh_retry_timeout, "20") #seconds
         return int(self._getTimelimitsElement().getAttribute(Attributes.ssh_retry_timeout))

    def getCITimelimit(self):
        if not self._getTimelimitsElement().getAttribute(Attributes.ci_timelimit):
            self._getTimelimitsElement().setAttribute(Attributes.ci_timelimit, "30") # minutes
        return int(self._getTimelimitsElement().getAttribute(Attributes.ci_timelimit))

    def getCIPoll(self):
        if not self._getTimelimitsElement().getAttribute(Attributes.ci_poll):
            self._getTimelimitsElement().setAttribute(Attributes.ci_poll, "1") # minutes
        return int(self._getTimelimitsElement().getAttribute(Attributes.ci_poll))

    def getServerEMail(self):
        serverElement = self._getServerElement()
        email = self.findOrCreateElement(serverElement, Tags.email)
        return email.getAttribute(Attributes.from_address)

    def getServerSMTPHost(self):
        serverElement = self._getServerElement()
        email = self.findOrCreateElement(serverElement, Tags.email)
        return email.getAttribute(Attributes.smtp_host)

    def getServerListenerPort(self):
        serverElement = self._getServerElement()
        if not serverElement.getAttribute(Attributes.listener_port):
            return None
        return serverElement.getAttribute(Attributes.listener_port)

    def getServerConfiguration(self):
        # is the server installed?
        if not os.path.exists(self.getServerDirectory()):
            return ServerConfiguration.notInstalled

        serverElement = self._getServerElement()
        if serverElement.getAttribute(Attributes.configuration):
            return serverElement.getAttribute(Attributes.configuration)
        return ServerConfiguration.default

    def setServerConfiguration(self, configuration):
        serverElement = self._getServerElement()
        serverElement.setAttribute(Attributes.configuration, configuration)

    def getPublicKey(self):
        keyElement = self.findOrCreateElement(self._getFilesystemElement(), Tags.public_key)
        if not self.getTextNode(keyElement):
            return None
        # otherwise
        return self.getTextNode(keyElement).data

    def getProxyURLs(self):
        proxyURLs = dict()
        for proxyURLElement in self._getURLsElement().getElementsByTagName(Tags.proxy_url):
            if not proxyURLElement.getAttribute(Attributes.protocol):
                print("Error: proxy protocol missing for %s in %s" % Tags.proxy_url, (self.environmentFile))
            elif not self.getTextNode(proxyURLElement):
                print("Error: proxy URL missing for %s in %s" % Tags.proxy_url, (self.environmentFile))
            else:
                proxyURLs[proxyURLElement.getAttribute(Attributes.protocol)] = self.getTextNode(proxyURLElement).data
        return proxyURLs

if __name__ == "__main__":
    environment = Environment()
    environment.showChangesFromDefault()
