# Copyright (c) 2015 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols <gnichols@redhat.com>
#
import socket

from rhcert.model.documentbase import makeEnumeration
from rhcert.tags import Constants


class Host(object):
    """ Host objects represent other hosts that recieve and
        respond to communication.

        The objects encapsulate information about hosts to
        streamline communication, such as host name (with
        or without DNS support), ip address, and perhaps
        connecting interfaces.

        IP Address is always from the point-of-view of the
        calling program, and may differ from the IP address
        actually configured on the host, due to routing, NAT,
        and the presence of multiple network interfaces/paths.
    """
    use = makeEnumeration("use", hostname="hostname", ipAddress="ipAddress", unknown="unknown")
    debug = Constants.off

    def __init__(self, name, ipAddress=None):
        self.name = name # the host's name - maybe not a network (DNS) name
        self.hostname = None # the hosts' network name
        self.ipAddress = None # ip address of the host
        self.port = None # host port
        self.role = None # role of the host on the test network
        self.use = Host.use.unknown
        self.__validate(ipAddress)
        # Note: Deprecated from CERTPX-2363
        # Usage: Pull the host out of a url, if that's what we got.
        # if hostname:
        #     url = urlparse.urlparse(hostname)
        #     if url.netloc:
        #         hostname = url.netloc
        #     else:
        #         pattern = re.compile("[\w\.\-]+$")
        #         if pattern.match(hostname):
        #             hostname = hostname
        
    def __resolveHostAddress(self, hostAddress):
        host, self.port = (hostAddress.split(':') + [None] * 2)[:2]
        return host

    def __validate(self, ipAddress=None):
        # check if an address was passed as a name, otherwise, assume it's a hostname
        if not ipAddress:
            try:
                self.name = self.__resolveHostAddress(self.name)
                socket.inet_aton(self.name)
                self.ipAddress = self.name
                self.use = Host.use.ipAddress
                if Host.debug != Constants.off:
                    print("Host connection is IP address %s" % self.name)
            except socket.error:
                pass
        else:# an ip address was passed - check it
            try:
                ipAddress = self.__resolveHostAddress(ipAddress)
                socket.inet_aton(ipAddress)
                self.ipAddress = ipAddress
                if self.use == Host.use.unknown:
                    self.use = Host.use.ipAddress
                    if Host.debug != Constants.off:
                        print("Host connection is IP address %s" % ipAddress)
            except socket.error:
                if Host.debug != Constants.off:
                    print("IP address %s is not valid" % ipAddress)

        # if no ip address was passed, try and resolve name as a hostname
        if self.use == Host.use.unknown:
            try:
                self.ipAddress = socket.gethostbyname(self.name)
                self.hostname = self.name
                self.use = Host.use.hostname
                if Host.debug != Constants.off:
                    print("Host connection is hostname %s" % self.name)
            except (socket.error, TypeError):
                if Host.debug != Constants.off:
                    print("Hostname %s is not valid" % self.name)

        if self.use == Host.use.unknown and Host.debug != Constants.off:
            print("Error: Host connection for %s is unknown" % self.name)

    def isValid(self):
        if self.use == Host.use.unknown:
            return False
        return True

    def getName(self):
        return self.name

    def getRole(self):
        return self.role
    def setRole(self, role):
        self.role = role

    def setHostname(self, hostname):
        self.hostname = hostname
    def getHostname(self):
        return self.hostname

    def setIPAddress(self, ipAddress):
        self.ipAddress = ipAddress
    def getIPAddress(self):
        return self.ipAddress

    def setPort(self, port):
        self.port = port
    def getPort(self):
        return self.port

    def getConnection(self):
        """ either hostname or ipAddress depending on what worked on the network"""
        if self.use == Host.use.hostname:
            return ":".join([self.hostname, self.port]) if self.port else self.hostname

        elif self.use == Host.use.ipAddress:
            return ":".join([self.ipAddress, self.port]) if self.port else self.ipAddress

        else:
            return Host.use.unknown


def unit_test():
    Host.debug = Constants.high
    testNames = ["localhost", "127.0.0.1", "192.168.1.1", "rhcert-server.app.eng.rdu2.redhat.com", "bluetwo", "invalid.host.name"]
    print("name, hostname, ip address, connection")
    for name in testNames:
        host = Host(name)
        print("%s %s %s %s" % (name, host.getHostname(), host.getIPAddress(), host.getConnection()))

if __name__ == "__main__":
    exit(unit_test())