# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#

import os, sys, time, socket

from rhcert.tags import Constants
from rhcert.command import Command
from rhcert.model.resultsdocument import ResultsDocument
from rhcert.model.task import Task
from rhcert.client.harness import TestHarness
from rhcert.client.rhcertcli import RHCertCLI
# from rhcert.client.beacon import Beacon
from rhcert.listener.queue import Queue
from rhcert.listener.serverUI import ServerUI


class RedHatCertDaemon(Queue):
    """ The RedHatCertDaemon class implements the client daemon 
    process which carries-out requests for testing and other 
    certification operations on the System Under Test.
    """

    def __init__(self, environment, planner, options):
        Queue.__init__(self, environment)
        self.options = options
        self.debug = Constants.off
        if self.options:
            self.debug = self.options.debug
        self.environment = environment
        self.planner = planner
        self.resultsDocument = None
        print("\nStarting %s at %s" % (self.__class__.__name__, time.strftime(Constants.DATETIMEFORMAT, time.gmtime(time.time()))))

    def jobResult(self, message):
        task = Task(Task.message)
        task.setParameter(Task.message, message)
        if self.resultsDocument:
            task.setParameter(Task.certificationID, self.resultsDocument.getCertificationID())
            task.setParameter(Task.certificationProgram, self.resultsDocument.getCertificationProgram())
        Queue.jobResult(self, task.toXML())

    def run(self):
        self.clearJobs()
        self.setStatusFile()
        self.checkContinue()
        # print "starting beacon"
        # beacon = Beacon("rhcert:client")
        # beacon.start()
        job = None
        while True:
            if self.options.debug != Constants.off:
                print("checking for tasks")
                sys.stdout.flush()
            try:
                job = self.getJob()
            except IOError as e:
                print("Error: could not read job from queue\n%s" % e)
            if not job or not job.strip(): # empty line?
                continue
            # otherwise:
            print("new job:")
            print("\"" + job.strip() + "\"")

            if job.strip() == "shutdown":
                message =  "rhcert daemon shutdown"
                self.jobResult(message)
                print(message)
                return 0

            # otherwise
            # check for commands - only run rhcert-util or rhcert-cli - nothing else
            words = job.split()
            if len(words) > 0 and words[0] in [Constants.rhcert_util, Constants.rhcert_cli] and ";" not in job:
                if words[0] == Constants.rhcert_util and  words[1] == "daemon" and words[2] =="run":
                    return self.serverTest(job)
                if words[0] == Constants.rhcert_cli and words[1] in ["run", "clean"]:
                    return self.runTests(job)
                print("Error: unsupported CLI command %s" % job)

            # otherwise, hope it is XML
            self.runXML(job)

    def runXML(self, xml):
        print("running xml")
        # assume it's a task
        task = Task()
        task.loadString("".join(xml))
        if task.isValid():
            return self.runTask(task)

        # otherwise
        message = "Starting test run"
        self.jobResult(message)
        self.resultsDocument = ResultsDocument()
        self.resultsDocument.loadString("".join(xml))
        serverIP = task.getParameter(Task.serverIP) # note: works even on ResultsDocument XML
        if serverIP: # if it was set by the listener, use it, because it's where the request came from
            self.resultsDocument.setTestServerIP(serverIP)
        return self.runPlan()


    def checkContinue(self):
        message = "Checking for continuation"
        print(message)
        self.jobResult(message)
        self.options.mode = Constants.server
        harness = TestHarness(self.options)
        harness.load()
        if harness.resultsDocument: # really it's the results document
            ServerUI.currentDocument = harness.resultsDocument
            self.resultsDocument = harness.resultsDocument
            if harness.resultsDocument.getIncompleteTestsFromLastRun():
                message = "Finishing incomplete tests"
                print(message)
                self.jobResult(message)
                harness.doContinue()
                harness.doSave()
        message = "ready"
        print(message)
        self.jobResult(message)

    def runPlan(self):
        # notify caller test run started
        message = "Started run %u for %s" % (int(self.resultsDocument.getNumberOfTestRuns())+1, self.resultsDocument.getCertificationID())
        print(message)
        self.jobResult(message)

        self.options.mode = Constants.server
        self.options.server = self.resultsDocument.getTestServer()
        self.options.serverIP = self.resultsDocument.getTestServerIP()
        self.options.debug = Constants.off
        if len(self.resultsDocument.getTestsByParameter(Constants.debug, "high")) > 0:
            self.options.debug = "high"
        harness = TestHarness(self.options)

        ServerUI.currentDocument = self.resultsDocument
        # lock out the commandline
        if not harness.getLock():
            message = "Error: could not lock client"
        else:
            # harness.setTestCompleteCallback(harness.doSave) # push the results to the LTS after each test is run
            harness.run(self.resultsDocument)
            harness.doSave()
            harness.releaseLock()
            message = "Finished test run %u" % (self.resultsDocument.getNumberOfTestRuns())
        self.jobResult(message)
        print(message)
        return True


    def __parseOptions(self, line, startPosition=3):
        options = dict()
        word=line.split()[startPosition:]
        i = 0
        while i+1 < len(word):
            options[word[i][2:]] = word[i+1]
            i += 2
        return options

    def openLogFile(self, args, mode):
        if not os.path.exists(self.environment.getLogDirectory()):
            os.mkdir(self.environment.getLogDirectory(), 0o666)
        logFileName = "rhcert-daemon-"
        for (key, value) in args.items():
            if key != "results":
                logFileName = logFileName + "%s=%s_" % (key, value)
        print("log file: " + logFileName)
        log = open(os.path.join(self.environment.getLogDirectory(), logFileName), mode)
        return log

    def serverTest(self, line):
        """ run server-side test code per client/SUT request """
        options = self.__parseOptions(line)

        # if it's a results request, find the log
        try:
            if len(options["results"]) > 0:
                if self.options.debug != Constants.off:
                    print("Getting test results")
                log = self.openLogFile(options, "r")
                taskOutFile = open(self.environment.getTaskFilePath()+ ".out", 'w')
                for line in log:
                    taskOutFile.write(line)
                taskOutFile.close()
                log.close()
                return True
        except Exception:
            pass

        # otherwise, queue the request
        self.jobResult("Running test " + options["test"] + "\n")
        result = True
        try:
            test = self.planner.getTest(options["test"])
            log = self.openLogFile(options, "w")
            log.write("<output>\n")
            result = Constants.FAIL
            if test.runOnServer(options, log): 
                result = Constants.PASS
            log.write("<summary>%s</summary>\n</output>\n" % result)
            log.close()
        except KeyError:
            print("Error: test not found")
            result = False
        return result

    def runTests(self, runRequest):
        """ run the requested tests specified via rhcert-cli command-line args """
        self.jobResult("Running tests...")
        rhcertCLI = RHCertCLI(argv=runRequest.split())
        result = rhcertCLI.do()
        self.jobResult("ready")
        return result

    def runTask(self, task):
        """ run a backend command """
        print("running task:")
        print(task.toXML())
        harness = TestHarness(task.getParametersDictionary())
        commands = {
              'verify': harness.doVerify,
              'save': harness.doSave,
              Task.register: harness.doRegister,
              Task.clean: harness.doClean,
              'version': harness.doVersion,
              Task.plan: harness.doPlan,
              Task.collab: harness.doCollab
              }

        if task.getTask() in commands:
            self.jobResult("Running command \"" + task.getTask() + "\"\n")
            result = commands[task.getTask()]()
        else:
            self.jobResult("Unsupported command \"" + task.getTask() + "\"\n")
            return False

        if result:
            self.jobResult("Command \"" + task.getTask() + "\" succeeded\n")
        else:
            self.jobResult("Command \"" + task.getTask() + "\"FAILED\n")

        return result

    def do(self, subcommand):
        (running, pid) = self.getStatusFile()

        if subcommand == "start":
            if running:
                print("The rhcert daemon is already started")
                return True
            print("Starting rhcert daemon")
            sys.stdout.flush()
            if os.system(Constants.rhcertd
                         + " daemon >> %s 0</dev/null 2>&1 &" % self.getLog()):
                print("Error: could not start rhcert daemon")
                return False
        elif subcommand == "stop":
            if not running:
                print("The rhcert daemon is not running")
                return True
            print("Stopping rhcert daemon")
            sys.stdout.flush()
            try:
                kill = Command("kill %s" % pid)
                kill.echo()
                self.removeStatusFile()
            except Exception as e:
                print("Error: could not stop daemon:")
                print(e)
        else: # assume status
            if running:
                print("The rhcert daemon is running")
            else:
                print("The rhcert daemon is NOT running")
                return False

        return True

