# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols <gnichols@redhat.com>
#
import sys, os, socket, cgi, time
import io

try: # python3
    from xmlrpc.server import SimpleXMLRPCServer
    from xmlrpc.server import SimpleXMLRPCRequestHandler
except: # python2
    from SimpleXMLRPCServer import SimpleXMLRPCServer
    from SimpleXMLRPCServer import SimpleXMLRPCRequestHandler


from rhcert.model.resultsdocument import ResultsDocument
from rhcert.command import Command
from rhcert.tags import Constants
from rhcert.listener.queue import Queue
from rhcert.listener.daemon import RedHatCertDaemon
from rhcert.model.task import Task
from rhcert.model.documentbase import DocumentWrapper
from rhcert.client.rhcertcli import TestHarness

class ServerMessage(Queue):

    def __init__(self, backend, environment, server):
        Queue.__init__(self, environment)
        self.options = dict()
        self.backend = backend
        self.server = server
        print("\nStarting %s at %s" % (self.__class__.__name__, time.strftime(Constants.DATETIMEFORMAT, time.gmtime(time.time()))))

    def run(self, taskXML):
        print("in ServerMessage.run")
        print("calling addJob")
        # if the caller (AKA client, AKA rhcert server) IP address is set, add it to the task XML
        if self.server.clientAddress:
            (callerID, callerPort) = self.server.clientAddress
            print("rhcert server address is:")
            print(callerID)
            task = Task()
            task.loadString(u'{0}'.format(taskXML).encode('utf-8'))
            task.setParameter(Task.serverIP, callerID)
            taskXML = task.toXML()
            print(taskXML)
            print(type(taskXML))
        if self.addJob(taskXML):
            return True
        else:
            print("Error: addJob failed!")
            return False

    def runCLI(self, commandLine):
        print("in ServerMessage.runCLI")
        print("calling addJob: %s" % commandLine)
        if self.addJob(commandLine):
            return True
        else:
            print("Error: addJob failed!")
            return False

    def getResult(self):
        return  self.checkResult()

    def restart(self):
        """ kill and restart the daemon """

        self.jobResult("Restarting...")
        daemon = RedHatCertDaemon(self.environment, None, None)

        # Get the harness for performing the clean operation with debug mode OFF
        harness = TestHarness(options=self.backend.options)

        # Stop the daemon, release lock file, perform cleaning of runs and
        # certification data, and start back daemon
        daemon.do("stop")
        harness.releaseLock()
        harness.doClean(cleanAll=True)
        daemon.do("start")

        self.jobResult("Restarted")
        return True

    def getFileContent(self, filePath):
        """ relative to /tmp directory """
        with open("/tmp/%s" % filePath, "r") as f:
            content = f.read()
            return content


    def getCurrentLog(self):
        """ fetch the current daemon log """
        # not the "real" daemon, just a temp object for the log path.
        daemon = RedHatCertDaemon(self.environment, planner=None, options=None)
        logPath = daemon.getLog()
        try:
            # limit log size:
            lines = 200
            tail = Command("tail -n %s %s" % (lines, logPath))
            tail.run()
            log = "\n".join(tail.output[-lines:])
            return DocumentWrapper().removeBadCharacters(cgi.escape(log))
        except Exception as exception:
            # error =  "Error: could not get the log:\n"
            # error += "%s" % exception
            # print error
            # return error
            return None

    def getCurrentResultsXML(self):
        """ get the current results and send them back ala submit """
        try:
            resultsXMLFile = io.open(self.environment.getResultsPath(), "rb")
            resultsXML = resultsXMLFile.read()
            resultsXMLFile.close()
            return resultsXML
        except Exception as error:
            message = "Error: could not read local test results file %s: \n%s" % (
                self.environment.getResultsPath(), error)
            print(message)
            return message

class Handler(SimpleXMLRPCRequestHandler):
    def __init__(self, request, client_address, server):
        server.clientAddress = client_address
        SimpleXMLRPCRequestHandler.__init__(self, request, client_address, server)

class RedHatCertificationListener(Queue):

    def __init__(self, backend, environment):
        Queue.__init__(self, environment)
        self.server = None
        self.environment = environment
        self.backend = backend
        self.debug = self.backend.options.debug

    def selfTest(self):
        message = ServerMessage()
        options = dict()
        options["mode"] = "auto"
        options["debug"] = "high"
        options["server"] = "localhost"
        message.setOptions(options)
        resultsDocument = ResultsDocument()
        resultsDocument.load("/var/hwcert/results.xml")
        tests = resultsDocument.getTests()
        for test in tests:
            if test.getName() != "memory":
                test.delete()
        message.run(resultsDocument.document.toxml())


    def run(self):
        self.setStatusFile()
        # hostname = "localhost"
        # hostname = socket.gethostname()
        hostname = ""
        retries = 5
        while retries:
            try:
                self.server = SimpleXMLRPCServer((hostname, int(self.environment.getServerListenerPort())),
                                    requestHandler=Handler,
                                    allow_none=True)
                self.server.register_introspection_functions()
                self.server.register_instance(ServerMessage(self.backend, self.environment, self.server))
                self.server.serve_forever()
            except Exception as e:
                if retries:
                    print("Warning: exception raised:")
                    print(e)
                    print("retrying")
                    retries = retries -1
                    time.sleep(1)
                else:
                    raise e


    def do(self, subcommand):
        (running, pid) = self.getStatusFile()

        if subcommand == "start":
            if running:
                print("The rhcert listener is already started")
                return True
            print("Starting rhcert listener")
            sys.stdout.flush()
            if os.system(Constants.rhcertd
                         + " listener >> %s 0</dev/null 2>&1 &" % self.getLog()):
                print("Error: could not start rhcert listener")
                return False
        elif subcommand == "stop":
            if not running:
                print("The rhcert listener is not running")
                return True
            print("Stopping rhcert listener")
            sys.stdout.flush()
            try:
                kill = Command("kill %s" % pid)
                kill.echo()
                self.removeStatusFile()
            except Exception as e:
                print("Error: could not stop listener:")
                print(e)
        else: # assume status
            if running:
                print("The rhcert listener is running")
            else:
                print("The rhcert listener is NOT running")
                return False

        return True
