# Copyright (c) 2014 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#

import os
import sys
import time
import re
import errno
import io

from rhcert.environment import Environment
from rhcert.command import Command
from rhcert.tags import Constants
from rhcert.commandLineUI import CommandLineUI

class Queue(object):
    """ Queue provides methods for both producer and consumer roles:
        producer: calls addJob to add a job, and checkResults to check results
        consumer: calls getJob to get a job, and jobResult to communicate results

        producer does not block for queue or response access
        consumer blocks on queue (getJob).
    """

    def __init__(self, environment):
        self.environment = environment
        self.jobID = 0
        self.idPattern = re.compile("[0-9]+")
        self.debug = Constants.off

    def debugMessage(self, message):
        if self.debug != Constants.off:
            print(message)

    def clearJobs(self):
        self.makeDirectoryPath(os.path.dirname(self.environment.getTaskFilePath()))
        for pipe in [self.environment.getTaskFilePath()+ ".in"]:
            try:
                os.remove(pipe)
            except OSError:
                pass # ignore
            mask = os.umask(0o111)
            os.mkfifo(pipe, 0o666)
        try:
            os.remove(self.environment.getTaskFilePath()+ ".out")
        except:
            pass
        outfile = io.open(self.environment.getTaskFilePath()+ ".out", 'wb')
        outfile.close()
        os.umask(mask)

    def makeDirectoryPath(self, directory):
        """Emulate mkdir -p behavior: create the given directory, including all
        needed parent directories. Do not return an error if the directories 
        already exist."""
        try:
            os.makedirs(directory)
        except OSError as e:
            if e.errno != errno.EEXIST: raise

    def getJob(self):
        """ wait and pull a message from the queue """
        taskInFile = io.open(self.environment.getTaskFilePath()+ ".in", 'r')
        lines = taskInFile.read()
        taskInFile.close()
        self.debugMessage("got a job")
        self.debugMessage(lines)
        self.jobID = self.jobID + 1
        return lines

    def jobResult(self, message):
        """ respond to a request on the queue """
        taskOutFile = os.open(self.environment.getTaskFilePath()+".out", os.O_WRONLY)
        encodedMessage = ("%s\n" % message).encode("utf-8")
        os.write(taskOutFile, encodedMessage)
        os.ftruncate(taskOutFile, len(encodedMessage))
        os.close(taskOutFile)
        return True

    def addJob(self, job):
        """ add a message to the queue (don't wait)"""
        taskInFile = os.open(self.environment.getTaskFilePath() + ".in", os.O_WRONLY)
        self.debugMessage("opened task in file")
        os.write(taskInFile, job.encode('utf-8'))
        os.close(taskInFile)
        return True

    def checkResult(self):
        """ check for job responses """
        taskOutFile = io.open(self.environment.getTaskFilePath()+".out", 'r')
        lines = taskOutFile.read()
        taskOutFile.close()
        return lines

    def getLog(self, clear=False):
        log = os.path.join(self.environment.getLogDirectory(), 
                           ("%s.log" % self.__class__.__name__))
        if not os.path.exists(self.environment.getLogDirectory()):
            os.mkdir(self.environment.getLogDirectory(), 0o666)
        elif clear and os.path.exists(log):
            os.remove(log)
        return log

    def getStatusFile(self):
        """ return running status/process id via file """
        pid = 0
        running = False
        try:
            pidFile = io.open(self.environment.getTaskFilePath()+ ("-%s.pid" % self.__class__.__name__), 'r')
            pid = pidFile.readline().strip()
            pidFile.close()
        except IOError:
            return (False, 0) # return not running, pid=0
        try:
            command = Command("pgrep -f " + Constants.rhcertd)
            command.getString(pid, singleLine=False)
            running = True
        except:
            running = False

        return (running, pid)

    def setStatusFile(self):
        """ set pid status in file """
        print("Setting %s pid to %s" % (self.__class__.__name__, os.getpid()))
        pid = io.open(self.environment.getTaskFilePath()+ ("-%s.pid" % self.__class__.__name__), "w")
        pid.write(u'%u' % os.getpid())
        pid.close()

    def removeStatusFile(self):
        statusFile = self.environment.getTaskFilePath()+ ("-%s.pid" % self.__class__.__name__)
        try:
            os.remove(statusFile)
            if self.debug != Constants.off:
                print("removed status file " + statusFile)
        except Exception as exception:
            if self.debug != Constants.off:
                print("Could not remove status file " + statusFile)
                print(exception)


    def consume(self):
        """ self-test for the consumer of the queue """
        self.ui = CommandLineUI()
        # self.clear()
        while True:
            self.debugMessage("checking for tasks")
            lines = self.getJob()
            if not lines:
                self.debugMessage("  it was empty.")
                continue

            self.debugMessage("got a message:")
            self.debugMessage(lines)
            for line in lines:
                match = self.idPattern.search(line)
                if match:
                    self.debugMessage("got job id %u" % int(match.group()))
                    self.jobID = int(match.group())
                    break
            self.debugMessage("pretending to work for 10 sec")
            time.sleep(10)
            self.debugMessage("done pretending")

            self.debugMessage("opening outfile")
            sys.stdout.flush()
            self.jobResult("did it!\n line2")
            self.debugMessage("sent response")

    def produce(self):
        """ self test for producer of the job queue """
        self.ui = CommandLineUI()
        self.debugMessage("Starting send loop")
        while True:
            if not self.ui.promptConfirm("ready?"):
                return
            self.debugMessage("Adding task XML")
            self.addJob("<tag>\n<g>%u</g>\n</tag>" % self.jobID)
            self.debugMessage("Done.")
            sys.stdout.flush()

            self.debugMessage("Reading Results:")
            while True:
                self.debugMessage("checking")
                lines = self.checkResult()

                if lines:
                    self.debugMessage("got a response:")
                    self.debugMessage("\"%s\"" % lines)
                    sys.stdout.flush()
                    for line in lines:
                        match = self.idPattern.search(line)
                        if match:
                            self.debugMessage("got job id %u" % int(match.group()))
                            jobID = int(match.group())
                            if jobID == self.jobID:
                                "matching ids!"
                            else:
                                "oops, no match"
                    self.jobID = self.jobID + 1
                    break
                else:
                    self.debugMessage("got nothin!")
                    time.sleep(4)
            self.debugMessage("done")


if __name__ == "__main__":
    environment = Environment()
    queue = Queue(environment)
    queue.debug = Constants.high
    if sys.argv[1] == "consume":
        queue.consume()
    else:
        queue.produce()

