# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
import sys

from rhcert.environment import Environment
from rhcert.model.task import Task
from rhcert.listener.queue import Queue

class ServerUI(object):
    """ ServerUI provides a way for tests to call back to the rhcert server for 
        interaction needed by the test """

    # current results document for testing
    currentDocument = None

    def __init__(self, test):
        self.test = test # the test object using the UI
        self.environment = Environment()
        self.queue = Queue(self.environment)
        self.echo = True


    def printPipe(self, pipe):
        """ just for logging, no prompting """
        while 1:
            line = pipe.readline()
            if line:
                print(line)
            else:
                return pipe.close()

    def echoResponse(self, response):
        if self.echo:
            sys.stdout.write("response: %s\n" % response)
            sys.stdout.flush()
            return response

    def __prompt(self, question, task):
        sys.stdout.write(question + '\n')
        sys.stdout.flush()

        task.setParameter(Task.test, self.test.getPath())
        task.setParameter(Task.question, question)
        if ServerUI.currentDocument:
            task.setParameter(Task.certificationID, ServerUI.currentDocument.getCertificationID())
            task.setParameter(Task.certificationProgram, ServerUI.currentDocument.getCertificationProgram())
        try:
            logFile = open(self.test.getOutputFile(), "r")
            task.setLog(logFile.read())
            logFile.close()
        except Exception:
            # maybe no log exists?
            pass

        self.queue.jobResult(task.toXML())
        # sys.stdout.write( "request sent: %s\n" % task.toXML())
        taskString = self.queue.getJob()

        continueMessageTask = Task(Task.message)
        continueMessageTask.setParameter(Task.message, "Continuing %s test" % self.test.getName())
        if ServerUI.currentDocument:
            continueMessageTask.setParameter(Task.certificationID, ServerUI.currentDocument.getCertificationID())
            continueMessageTask.setParameter(Task.certificationProgram, ServerUI.currentDocument.getCertificationProgram())
        self.queue.jobResult(continueMessageTask.toXML())

        task = Task()
        task.loadString(taskString)
        if task.isValid():
            return task.getParameter(Task.response)
        # otherwise
        print("Error: invalid server prompt response:")
        print(taskString)
        return None

    def prompt(self, message, validation=None, default=None):
        task = Task(Task.prompt)
        if default:
            task.setParameter(Task.default, default)
        return self.echoResponse(self.__prompt(message, task))

    def promptInteger(self, message, default=None):
        task = Task(Task.prompt)
        task.setParameter(Task.type, Task.type.integer)
        if default:
            task.setParameter(Task.default, default)
        return self.echoResponse(int(self.__prompt(message, task)))

    def promptConfirm(self, message):
        task = Task(Task.prompt)
        task.setParameter(Task.type, Task.type.confirm)
        return self.echoResponse("yes" == self.__prompt(message, task))

    def __select(self, taskType, message, answers, selected):
        task = Task(Task.prompt)
        task.setParameter(Task.default, selected)
        task.setParameter(Task.type, taskType)
        task.setParameter(Task.choices, answers)
        return self.__prompt(message, task)

    def select(self, message, answers, selected):
        return self.echoResponse(self.__select(Task.type.select, message, answers, selected))

    def multiSelect(self, message, answers, selected):
        selected_none = 'None of these'
        answers.insert(0, selected_none)
        selection = self.__select(Task.type.multi_select, message, answers, selected)
        selected = list()
        if selection:
            if selected_none in selection:
                selection = selection.replace(selected_none, '')
                selected = selection.split()
                if len(selected) >= 1:
                    print('Error: Please select valid options.')
                    answers.remove(selected_none)
                    return self.multiSelect(message, answers, selected)
                else:
                    return self.echoResponse([])
            else:
                selected = selection.split()
        return self.echoResponse(selected)

    def promptPassword(self, message):
        task = Task(Task.prompt)
        task.setParameter(Task.type, Task.type.password)
        return self.__prompt(message, task)

    def promptFile(self, message):
        task = Task(Task.prompt)
        task.setParameter(Task.type, Task.type.file)
        return self.echoResponse(self.__prompt(message, task))

    def statusMessage(self, message):
        sys.stdout.write(message)
        sys.stdout.write("\n")
        sys.stdout.flush()
        task = Task(Task.message) # message task type
        task.setParameter(Task.message, message)
        if ServerUI.currentDocument:
            task.setParameter(Task.certificationID, ServerUI.currentDocument.getCertificationID())
            task.setParameter(Task.certificationProgram, ServerUI.currentDocument.getCertificationProgram())
        self.queue.jobResult(task.toXML())
        return True

    def editText(self, message, default=None):
        return self.prompt(message, validation=None, default=default)

