# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
# localTestServer.py: rhcert client interface to the rhcert local test server

import re
import os
import sys
import json
import requests
from distutils.version import LooseVersion
from rhcert.tags import Constants, ServerConfiguration
from rhcert.httpUpload import HttpUpload
from rhcert.host import Host
from rhcert.environment import Environment


class LocalTestServer:
    def __init__(self, host):
        self.host = host
        self.status = None
        self.clientIPAddress = None  # the IP address of this client seen by the server
        self.errors = list()
        self.environment = Environment()
        self.phase = ""
        self.multistore = False

    def getStatus(self):
        if not self.status:
            self.verify()
        return self.status

    def getHostname(self):
        if self.host.getHostname():
            return self.host.getHostname()
        # otherwise:
        return self.host.getName()

    def getClientIPAddress(self):
        """ the IP address of this client seen by the server """
        return self.clientIPAddress

    def getIPAddress(self):
        """ the IP address of the server seen by this host """
        return self.host.getIPAddress()

    def getErrors(self):
        return self.errors

    def getPhase(self):
        return self.phase

    def isMultiStoreEnabled(self):
        return self.multistore

    def __getServerResponse(self, url):
        self.errors = list()

        try:
            hostname = self.host.getConnection()
            response = requests.get('http://%s%s' % (hostname, url))
            if response.status_code == 200:
                return response.text
        except Exception as exception:
            print("Error: Could not contact the rhcert server on %s" % hostname)
            print(exception)

        print("Warning: Could not contact the rhcert server on %s" % hostname)
        return None

    def verify(self):
        self.errors = list()
        if self.host.getHostname() and not self.host.getName():
            self.errors.append("Error: no hostname is set")
            return False

        print("\nVerifying rhcert Server %s (%s):" % (self.host.getConnection(), self.host.getIPAddress()))
        result = True
        response = self.__getServerResponse("/api/rhcert-status")

        if response:
            status = json.loads(response)
            config = status["data"]["rhcert"]
            self.phase = config.get("phase")
            self.multistore = config.get("multistore")
            print("rhcert-server Version %s Release %s" % (config.get("version"), config.get("release")))
            if LooseVersion(config.get("version") + "." + config.get("release")) < LooseVersion(Constants.hwCertServerMinimumVersion):
                message = "Error: rhcert server is from a prior release."
                print(message)
                self.errors.append(message)
                print("This is a likely cause of test failures")
                result = False

            if config.get("configuration") == ServerConfiguration.management:
                print("Server at IP address:%s is running in %s configuration" %
                              (self.host.getIPAddress(), config.get("configuration")))
                print("Error: Registering a test server to management LTS is not supported")
                try:
                    os.remove(self.environment.getLockFile())
                except OSError as e:
                    pass
                sys.exit(0)

            self.clientIPAddress = config.get("requestorIP")
            print("IP address at server: %s" % self.clientIPAddress)
            self.status = config.get("status")
            print("Status: %s" % self.status)
        else:
            message = "Error: Could not contact the rhcert server on %s (%s):" % \
                      (self.host.getConnection(), self.host.getIPAddress())
            print(message)
            self.errors.append(message)
            return False

        return result

    def saveResults(self, resultsFilePath, storePath):
        hostname = self.host.getConnection()
        print("Saving to server %s" % hostname)
        try:
            request = HttpUpload(hostname, "/save/")
            request.addField("server-path", storePath)
            return request.httpUpload(resultsFilePath)
        except Exception as e:
            print("Warning: could not save results to server %s (%s):" % (
            self.host.getConnection(), self.host.getIPAddress()))
            print(e)

        # otherwise
        print("Error: failed to contact server at hostname %s and ip address %s (port: %s)" % (
        self.getHostname(), self.getIPAddress(), self.host.getPort()))
        return False
