#!/usr/bin/env python
# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
import re
import sys

from rhcert.tags import Tags, Attributes, Constants
from rhcert.model.documentbase import DocumentBase
from rhcert.model.vendordocument import VendorDocument
from rhcert.model.plancomponent import PlanComponent
from rhcert.model.parametersdocument import ParametersDocument
from rhcert.model.hostdocument import HostDocument

class CertificationDocument(DocumentBase):
    """ CertificationDocument represents the certification in progress on the SUT """

    def __init__(self, topElement=None, document=None, debugging=False):
        DocumentBase.__init__(self)
        self.element = topElement
        self.document = document
        self.Debugging = debugging

    def new(self):
        self._new(topElement=Tags.certification_test, stylesheet="/rhcert/css/results.css")
        self.findOrCreateElement(self.document.documentElement, Tags.hardware)
        self.findOrCreateElement(self.document.documentElement, Tags.certification)

    def copy(self, original):
            self._new(topElement=Tags.certification_test, stylesheet="/rhcert/css/results.css")
            self.element = self.findOrCreateElement(self.document, Tags.certification_test)
            hardware = self.document.importNode(original.findOrCreateElement(original.document.documentElement, Tags.hardware), True)
            self.element.appendChild(hardware)
            certification = self.document.importNode(original.findOrCreateElement(original.document.documentElement, Tags.certification), True)
            self.element.appendChild(certification)

    def serialize(self):
        return {
            'id': self.getCertificationID(),
            'program': self.getCertificationProgram(),
            'isSandbox': self.isSandbox(),
            'redhatProduct': self.getProductCertification(),
        }

    def getDatabasePath(self):
        return self.document.documentElement.getAttribute(Attributes.database_path)

    def setDatabasePath(self, databasePath):
        self.document.documentElement.setAttribute(Attributes.database_path, databasePath)

    def copyTestRunSummaries(self, original):
            # copy device classes and tests, but skip logging
            self.partialCopyTree(original.document.documentElement, self.element, [Tags.device_class, Tags.test, Tags.run, Tags.summary])

    def setHardware(self, tag, value):
        hardwareElement = self.findOrCreateElement(self.document.documentElement, Tags.hardware)
        element = self.findOrCreateElement(hardwareElement, tag)
        # some hardware ppc64, ppc64le has wacky stuff in make and model names.
        value = self.removeBadCharacters(value)
        self.setTextNode(element, value)

    def getHardware(self, tag):
        element = self.__getHardwareElement(tag)
        textNode = self.getTextNode(element)
        if textNode and not textNode.data.isspace():
            return textNode.data

        # otherwise
        return ""

    def __getHardwareElement(self, tag):
        hardwareElement = self.findOrCreateElement(self.document.documentElement, Tags.hardware)
        return self.findOrCreateElement(hardwareElement, tag)

    def getHardwareValues(self):
        "return hardware element values as dict"
        values = dict()
        hardwareElement = self.findOrCreateElement(self.document.documentElement, Tags.hardware)
        for element in hardwareElement.childNodes:
            textNode = self.getTextNode(element)
            if textNode and not textNode.data.isspace():
                values[element.nodeName] = textNode.data
        return values

    def getRequiredOS(self):
        """ <certification><os>: the host OS or base OS required by the certification """
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        return HostDocument(self.findOrCreateElement(certificationElement, Tags.os), self.document)

    def getOS(self, tag):
        """ depricated: use getRequiredOS() for the certification host and OS requirements
                            <certification><os>
                        use ResultsDocument.getHostOS() for the actual host/OS info from the host
                            <hardware><os>
        """
        if self.document:
            certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
            osElement = self.findOrCreateElement(certificationElement, Tags.os)
            element = self.findOrCreateElement(osElement, tag)
            if self.getTextNode(element):
                return self.getTextNode(element).data
        #otherwise
        return None

    def getOSValues(self):
        return self.getRequiredOS().getValues()

    def setCertificationID(self, certificationID):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        certificationElement.setAttribute(Attributes.id, str(certificationID))

    def getCertificationID(self):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        try:
            return certificationElement.getAttribute(Attributes.id)
        except ValueError:
            return 0

    def getCertificationProgram(self):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        return certificationElement.getAttribute(Attributes.type)

    def setCertificationProgram(self, certificationProgram):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        certificationElement.setAttribute(Attributes.type, certificationProgram)

    def isSandbox(self):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        return (certificationElement.getAttribute(Attributes.sandbox) == Constants.TRUE)

    def setSandbox(self, value=True):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        if value:
            certificationElement.setAttribute(Attributes.sandbox, Constants.TRUE)
        else:
            certificationElement.setAttribute(Attributes.sandbox, Constants.FALSE)

    def setCertificationState(self, state):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        certificationElement.setAttribute(Attributes.state, state)
        if self.getCertificationState() == Constants.local:
            self.setSandbox(True)

    def getCertificationState(self):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        return certificationElement.getAttribute(Attributes.state)

    """ specification ID is probably not used in rhcert 2 - replaced with bug ID """
    def setSpecificationID(self, specificationID):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        certificationElement.setAttribute(Attributes.specification_id, str(specificationID))

    def getSpecificationID(self):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        try:
            return certificationElement.getAttribute(Attributes.specification_id)
        except ValueError:
            return 0

    """ bug ID is the bugzilla bug ID associated with the cert """
    def setBugID(self, bugID):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        certificationElement.setAttribute(Attributes.bug_id, str(bugID))

    def getBugID(self):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        return certificationElement.getAttribute(Attributes.bug_id)

    def setNodeID(self, nodeID):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        certificationElement.setAttribute(Attributes.node_id, str(nodeID))

    def setCreationTime(self, dateTimeString):
        # expects Constants.DATETIMEFORMAT'ed date-time
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        certificationElement.setAttribute(Attributes.created, dateTimeString)

    def getCreationTime(self):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        return certificationElement.getAttribute(Attributes.created)

    def setUpdateTime(self, dateTimeString):
        # expects Constants.DATETIMEFORMAT'ed date-time
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        certificationElement.setAttribute(Attributes.updated, dateTimeString)

    def getUpdateTime(self):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        return certificationElement.getAttribute(Attributes.updated)

    def setCertificationVersion(self, certificationProductVersion):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        certificationElement.setAttribute(Attributes.product_version, str(certificationProductVersion))

    def getCertificationVersion(self):
        """ CWE "version" field sometimes just major, somtimes major.minor"""
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        return certificationElement.getAttribute(Attributes.product_version)

    def setCertificationFullVersion(self, certificationProductVersion):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        certificationElement.setAttribute(Attributes.full_version, str(certificationProductVersion))

    def getCertificationFullVersion(self):
        """ CWE "minor_version" field  - really seems to be the full version: major.minor"""
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        return certificationElement.getAttribute(Attributes.full_version)

    def getNodeID(self):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        return certificationElement.getAttribute(Attributes.node_id)

    def getTestPlanStatus(self):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        return certificationElement.getAttribute(Attributes.test_plan_status)

    def setTestPlanStatus(self, status):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        self.setAttribute(certificationElement, Attributes.test_plan_status, status)

    def setMultiHostRunTime(self, timestamp):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        self.setAttribute(certificationElement, Attributes.multihost_run_time, self.timeToString(timestamp))

    def getMultiHostRunTime(self):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        return certificationElement.getAttribute(Attributes.multihost_run_time)

    def getPlanComponents(self):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        planComponents = list()
        for element in self._getChildElements(certificationElement, Tags.plan_component):
            planComponents.append(PlanComponent(element, self.document))
        return planComponents

    def getPlanComponent(self, id):
        for planComponent in self.getPlanComponents():
            if planComponent.getID() == id:
                return planComponent
        return None

    def getRequiredHostRoles(self):
        """ return a dict by host ID of HostDocuments for host roles defined in the test plan components  """
        required_host_roles = {}
        for planComponent in self.getPlanComponents():
            if planComponent.hasRequiredOS():
                host = planComponent.getRequiredOS()
                if host.getID():
                    required_host_roles[host.getID()] = host
        return required_host_roles

    def isMultiHost(self):
        return True if self.getRequiredHostRoles() else False

    def createPlanComponent(self, name, id):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        planComponentElement = self.document.createElement(Tags.plan_component)
        certificationElement.appendChild(planComponentElement)
        planComponent = PlanComponent(planComponentElement, self.document)
        planComponent.setName(name)
        planComponent.setID(id)
        return planComponent


    def copyPlanComponents(self, sourceCertificationDocument):
        sourceCertificationElement = self.findOrCreateElement(sourceCertificationDocument.document.documentElement, Tags.certification)
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        for sourcePlanComponentElement in self._getChildElements(sourceCertificationElement, Tags.plan_component):
            if not self.getPlanComponent(sourcePlanComponentElement.getAttribute(Attributes.id)):
                newPlanComponentElement = self.document.importNode(sourcePlanComponentElement, True) # True = deep copy
                certificationElement.appendChild(newPlanComponentElement)

    def copyPlanComponent(self, planComponent):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        newPlanComponentElement = self.document.importNode(planComponent.element, True) # True = deep copy
        certificationElement.appendChild(newPlanComponentElement)

    def replacePlanComponent(self, newPlanComponent):
        oldPlanComponent = self.getPlanComponent(newPlanComponent.getID())
        if oldPlanComponent:
            certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
            certificationElement.removeChild(oldPlanComponent.element)
            self.copyPlanComponent(newPlanComponent)


    def getOSProductShortName(self):
        longName = self.getHostOS().getProduct()
        if "Red Hat Enterprise Linux" in longName:
            return "RHEL"

        #otherwise
        return longName

    def getName(self):
        """ rhcert-<vendor>-<model>-<os><release>_<arch> """
        pattern = re.compile('\s+')
        name =  "rhcert-%s-%s-%s%s_%s" % (self.getHardware(Tags.vendor), self.getHardware(Tags.model), self.getOS(Tags.name), self.getOS(Tags.release), self.getHardware(Tags.arch))
        return pattern.sub('_', name)

    def getVersion(self):
        return self.document.documentElement.getAttribute(Attributes.rhcert_version)
    def getRelease(self):
        return self.document.documentElement.getAttribute(Attributes.rhcert_release)

    def setTestServer(self, testServer):
        if testServer:
            self.setHardware(Tags.test_server, testServer)

    def getTestServer(self):
            return self.getHardware(Tags.test_server)

    def getTestServerIP(self):
        testServerElement = self.__getHardwareElement(Tags.test_server)
        return testServerElement.getAttribute(Attributes.ip_address)

    def setTestServerIP(self, ipAddress):
        testServerElement = self.__getHardwareElement(Tags.test_server)
        testServerElement.setAttribute(Attributes.ip_address, ipAddress)

    def getStorePath(self):
        """ returns a path: vendor, make, model, arch, OS, release """
        path = []
        if self.getDatabasePath():
            path = [self.getDatabasePath()]
        for tag in Tags.vendor, Tags.make, Tags.model, Tags.arch:
            value = self.getHardware(tag)
            if len(value) > 0 and value != "unknown":
                path.append(value)
        for value in self.getOSProductShortName(), self.getHostOS().getRelease():
            if len(value) > 0:
                path.append(value)
        return path

    def getVerboseFileName(self):
        """ redhat-certification-<type>-<id>.xml """
        filename = "redhat-certification-%s-%s.xml" % (self.getCertificationProgram(), self.getCertificationID())
        return filename

    def getVendors(self, notRedHat=False):
        vendors = list()
        certificationElement = self.getChildElement(Tags.certification)
        for element in certificationElement.childNodes:
            if element.nodeName == Tags.vendor:
                vendor = VendorDocument(element, self.document)
                if not notRedHat or vendor.getName() != Constants.RedHatInc:
                    vendors.append(VendorDocument(element, self.document))
        return vendors

    def getVendor(self, id):
        for vendor in self.getVendors():
            if id == vendor.getID():
                return vendor
        return None

    def setVendor(self, vendor):
        """ vendor = VendorDocument """
        if vendor.element.nodeName != Tags.vendor:
            return None
        if self.getVendor(vendor.getID()):
            return None
        vendorElementCopy = self.mergeDocument(vendor, self.getChildElement(Tags.certification), deepCopy=False)
        return VendorDocument(vendorElementCopy, self.document)

    def setProductCertification(self, path):
        """ sets the red hat product the certification is for """
        self.setCertificationProgram(path[Tags.program].getShortName())
        vendor = self.getVendor(path[Tags.vendor].getID())
        if vendor:
            vendor.setProduct(path[Tags.product])
        else:
            vendor = self.setVendor(path[Tags.vendor])
        # delete all products not in the path
        for product in vendor.getProducts():
            if product.getID() != path[Tags.product].getID():
                product.delete()
        # copy base OS over to old attributes
        if path.get(Tags.baseRHEL):
            self.getRequiredOS().setProduct(path.get(Tags.baseRHEL).getName())
            (major, minor, _, _) = path.get(Tags.baseRHEL).getVersions()[0]
            version = "%s" % major
            if minor:
                version += ".%s" % minor
            self.getRequiredOS().setRelease(version)

    def getProductCertification(self):
        """ return the Red Hat Product(s) for the certification """
        for vendor in self.getVendors():
            if vendor.getName() == Constants.RedHatInc:
                if vendor.getProducts():
                    return {Tags.vendor: vendor, Tags.product: vendor.getProducts()[0]}
                return {Tags.vendor: vendor, Tags.product: None}
        # no Red Hat Vendor:
        return {Tags.vendor: self.__createRedHatVendor(), Tags.product: None}

    def __createRedHatVendor(self):
        certificationElement = self.getChildElement(Tags.certification)
        vendorElement = self.document.createElement(Tags.vendor)
        certificationElement.appendChild(vendorElement)
        self.setAttribute(vendorElement, Attributes.name, Constants.RedHatInc)
        self.setAttribute(vendorElement, Attributes.scratch_id, "1")
        self.setAttribute(vendorElement, Attributes.id, 0)
        return VendorDocument(vendorElement, self.document)


    def getPartnerAndProduct(self):
        """ returns the a Vendor object (Non-redhat) and product object """
        partners = self.getVendors(notRedHat=True)
        if partners:
            partner = partners[0]
            product = None
            if partner.getProducts():
                product = partner.getProducts()[0]
            return (partner, product)
        # otherwise
        return (None, None)

    def getPartnerAndProductID(self):
        """ returns the a Vendor ID (Non-redhat) and product ID """
        (vendor, product) = self.getPartnerAndProduct()
        vendorID = productID = 0
        if vendor:
            vendorID = vendor.getID()
        if product:
            productID = product.getID()
        return (vendorID, productID)

    def copyCertification(self, sourceCertificationDocument, deleteProperties=False):
        """ copy the entire <certification> element into the document, replacing the old one."""
        oldCertificationElement = self.getChildElement(Tags.certification)
        newElement = self.document.importNode(sourceCertificationDocument.getChildElement(Tags.certification), True) # True = deep copy
        if deleteProperties:
            for tag in [Tags.property, Tags.vendor]:
                for propertyElement in self._getChildElements(newElement, tag):
                    # don't delete Red Hat vendor
                    if tag == Tags.vendor and propertyElement.getAttribute(Attributes.name) == Constants.RedHatInc:
                        continue
                    newElement.removeChild(propertyElement)
        self.element.replaceChild(newElement, oldCertificationElement)

    def getCertificationParameters(self):
        certificationElement = self.findOrCreateElement(self.document.documentElement, Tags.certification)
        return ParametersDocument(certificationElement, self.document)

    def update(self, sourceCertification):
        # require IDs to match up
        if (self.getCertificationID() != sourceCertification.getCertificationID()
            or self.getCertificationProgram() != sourceCertification.getCertificationProgram()):
            return False

        # attributes
        self.setCertificationState(sourceCertification.getCertificationState())
        self.setUpdateTime(sourceCertification.getUpdateTime())

        # test plan components
        self.updatePlanComponents(sourceCertification)

        # generic parameters
        self.getCertificationParameters().replace(sourceCertification.getCertificationParameters().getAll())
        # fix up status
        if sourceCertification.getCertificationParameters().get("certification_status"):
            self.getCertificationParameters().reset("status", sourceCertification.getCertificationParameters().get("certification_status"))

        return True

    def updatePlanComponents(self, sourceCertification):
        self.setTestPlanStatus(sourceCertification.getTestPlanStatus())

        newPlanComponents = list()
        changedPlanComponents = list()
        for sourcePlanComponent in sourceCertification.getPlanComponents():
            # work-around for missing plan component ids
            componentID = sourcePlanComponent.getID()
            if not componentID:
                componentID = sourcePlanComponent.getName()
            planComponent = self.getPlanComponent(componentID)
            if not planComponent:
                planComponent = self.createPlanComponent(sourcePlanComponent.getName(), componentID)
                newPlanComponents.append(planComponent)
            elif planComponent.getStatus() != sourcePlanComponent.getStatus():
                changedPlanComponents.append(planComponent)

            planComponent.setStatus(sourcePlanComponent.getStatus())
            if sourcePlanComponent.setSupplimental():
                planComponent.setSupplimental(True)
            planComponent.setComponent(sourcePlanComponent.getComponent())
            planComponent.setDescription(sourcePlanComponent.getDescription())
            planComponent.setType(sourcePlanComponent.getType())
            planComponent.setBits(sourcePlanComponent.getBits())
            planComponent.setStatus(sourcePlanComponent.getStatus())
            planComponent.setTestID(sourcePlanComponent.getTestID())
            planComponent.setLeveragedTestID(sourcePlanComponent.getLeveragedTestID())
            planComponent.setLeveragedCertificationID(sourcePlanComponent.getLeveragedCertificationID())

        return (newPlanComponents, changedPlanComponents)

def unitTest():
    import json
    doc = CertificationDocument()
    doc.new()
    component = doc.createPlanComponent("Gadget x100", 101)
    component.addFeature(1, "20G Ethernet", "Network")
    component.addFeature(2, "CPUScaling", "System")
    component.addFeature(3, "Memory", "System")
    component.addFeature(4, "Wireless N", "Network")
    component = doc.createPlanComponent("DocMonster", 102)
    component.addFeature(1, "USB3", "USB")
    component.addFeature(2, "USB2", "USB")
    component.addFeature(3, "HDMI", "Display")
    component.addFeature(4, "SATA Controller", "System")
    component.addFeature(5, "Wireless N", "Network")
    categories = doc.getPlanFeatureCategories()

    for category, features in categories.items():
        print(category + ":")
        for feature, components in features.items():
            print("    " + feature + ":")
            for component in components:
                print("        " + component.getName())
    doc.save("certTest.xml")
    return True

if __name__ == "__main__":
    if unitTest():
        sys.exit(0)
    sys.exit(1)

