#!/usr/bin/env python
# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
import sys, os, gzip, time, datetime
import io
from rhcert.compatability import StringIO

from xml.dom.minidom import getDOMImplementation, parse, parseString
from lxml import etree

from rhcert.compatability import rhcert_make_unicode, rhcert_make_str
from rhcert.tags import Tags, Attributes, Constants
import rhcert.version as version
from rhcert.command import Command


if sys.version.startswith('2.6'):
    #Workaround for https://bugs.python.org/issue5762
    def _write_data_fix(writer, data):
        "Writes datachars to writer."
        if data:
            data = data.replace("&", "&amp;").replace("<", "&lt;"). \
                    replace("\"", "&quot;").replace(">", "&gt;")
            writer.write(data)
    from xml.dom import minidom
    minidom._write_data = _write_data_fix

def makeEnumeration(name, **enums):
    return type(name, (), enums)

class DocumentWrapper(object):
    # non-printing ascii unacceptable to xml
    asciiBadForXML = list(map(chr, (list(range(9)) + list(range(11,13)) + list(range(14,32)))))

    def __init__(self):
        self.document = None
        self.Debugging = Constants.off
        self.element = None

    def setDebugging(self, debugging):
        self.Debugging = debugging

    def getChildElement(self, tag):
        for element in self.element.childNodes:
            if element.nodeName == tag:
                return element
        # not found
        return None

    def getDescendantElement(self, path, root=None):
        parent = root
        if not parent:
            parent = self.element
        childElements = []
        for element in parent.childNodes:
            if element.nodeName == path[0]:
                childElements.append(element)
        if not childElements:
            return None
        if len(path) == 1:
            return childElements[0]
        return self.getDescendantElement(path[1:], root=childElements[0])

    def getChildElementString(self, tag):
        element = self.getChildElement(tag)
        if element and self.getTextNode(element):
            return self.getTextNode(element).data
        return ""

    def getChildElements(self, tag):
        return self._getChildElements(self.element, tag)

    def _getChildElements(self, parent, tag):
        childElements = list()
        for element in parent.childNodes:
            if element.nodeName == tag:
                childElements.append(element)
        return childElements

    def findOrCreateElement(self, parent, elementName):
        try:
            child = self._getChildElements(parent, elementName)[0]
        except IndexError:
            child = None

        if not child:
            child = self.document.createElement(elementName)
            parent.appendChild(child)
        return child

    def delete(self):
        self.element.parentNode.removeChild(self.element)

    def setChildElement(self, tag, value):
        element = self.findOrCreateElement(self.element, tag)
        self.setTextNode(element, value)

    def setTextNode(self, element, value, filterCharacters=True):
        if value:
            if filterCharacters:
                value = self.makeUnicode(value)
            # is there an existing child text node?
            textNode = self.getTextNode(element)
            if textNode:
                textNode.data = value
            else:
                # else no existing text child note, create it
                textNode = self.document.createTextNode(value)
                element.appendChild(textNode)

    def addTextNode(self, element, value):
        if value:
            value = self.makeUnicode(value)

        textNode = self.document.createTextNode(value)
        element.appendChild(textNode)

    def getTextNode(self, element):
        for node in element.childNodes:
            if node.nodeType == node.TEXT_NODE:
                return node
        return None

    def timeToString(self, theTime):
        return time.strftime(Constants.DATETIMEFORMAT, theTime)

    def durationData(self, startTime, stopTime):
        # start = datetime.datetime.strptime(startTime, Constants.DATETIMEFORMAT)
        start = datetime.datetime(*(time.strptime(startTime, Constants.DATETIMEFORMAT)[0:5]))
        # stop = datetime.datetime.strptime(stopTime, Constants.DATETIMEFORMAT)
        stop = datetime.datetime(*(time.strptime(stopTime, Constants.DATETIMEFORMAT)[0:5]))
        return stop - start

    def duration(self, startTime, stopTime):
        duration = self.durationData(startTime, stopTime)
        if duration.days > 0:
            durationString = "u% days, " % duration.days
        else:
            durationString = ""
        hours = duration.seconds/3600
        minutes = (duration.seconds - hours*3600)/60
        seconds = duration.seconds - hours*3600 - minutes*60
        durationString += "%.2u:%.2u:%.2u" % (hours, minutes, seconds)
        return durationString

    def timeStringToTimestamp(self, theTimeString):
        return theTimeString.replace(" ", "").replace(":", "").replace("-", "")


    def setAttribute(self, element, attribute, value):
        return element.setAttribute(attribute, self.makeUnicode(value))

    def removeAttribute(self, element, attribute):
        try:
            element.removeAttribute(attribute)
        except:
            pass

    def makeUnicode(self, value):
        return rhcert_make_unicode(value)

    def removeBadCharacters(self, value):
        """ note: this method should be replaced with makeUnicode as testing permits,
        as it removes valid unicode characters.

        Note: hwcert calls this directly. """
        # encoded values are of type byte in python3 hence decoding it
        isByte = False
        if type(value) is bytes:
            value = value.decode("utf-8")
            isByte = True

        try:
            unicodeType = unicode
        except NameError:
            unicodeType = str
        if sys.version.startswith("2.6"):
            filtered = ''
        else:
            filtered = u''
        filterStart = 0
        # for c in value:
        for i in range(len(value)):
            c = value[i]
            if c in self.asciiBadForXML or (type(value) != unicodeType and ord(c) >= 128):
                if self.Debugging != Constants.off:
                    print("skipping bad char \\x%.2x" % ord(c))
                if filterStart < i:
                    filtered += value[filterStart:i]
                filterStart = i+1
        filtered += value[filterStart:]
        # Encoding back to byte, if it was a byte
        if isByte:
            filtered = filtered.encode('utf-8')
        return filtered


    def appendDevice(self, device):
        """ append device information from a Device object (rhcert/rhcert/device.py). """
        deviceElement = self.document.createElement(Tags.device)
        self.setAttribute(deviceElement, Attributes.udi, device.getUDI())
        self.setAttribute(deviceElement, Attributes.product, device.getProduct())
        # self.setAttribute(deviceElement, Attributes.vendor, device.getVendor()) - hwcert needs to implement
        self.setAttribute(deviceElement, Attributes.device_bus, device.getBus())
        self.setAttribute(deviceElement, Attributes.device_class, device.getClass())
        self.setAttribute(deviceElement, Attributes.source, device.getSource())

        for (key, value) in device.properties.items():
            # only create a property element if the value is there (some string, non-zero integer, or True).
            if value:
                propertyElement = self.document.createElement(Tags.property)
                self.setAttribute(propertyElement, Tags.name, key)
                self.setTextNode(propertyElement, "%s" % value)
                deviceElement.appendChild(propertyElement)

        self.element.appendChild(deviceElement)

    def appendDevices(self, devices):
        for device in devices:
            self.appendDevice(device)

    def merge(self, documentFilePath, nodeIndex=1):
        subdocument = DocumentBase()
        subdocument.load(documentFilePath)
        try:
            root = self.document.importNode(subdocument.document.childNodes[nodeIndex], True) # True = deep copy
            self.element.appendChild(root)
        except Exception as e:
            print("Error: could not import document in " + documentFilePath)
            print(e)

    def mergeDocument(self, subdocument, parentElement, deepCopy=True):
        """ subdocument = DocumentBase """
        newElement = self.document.importNode(subdocument.element, deepCopy) # True = deep copy
        parentElement.appendChild(newElement)
        return newElement

    def insertDocument(self, subdocument, parentElement, deepCopy=True):
        """ insert subdocument as first child of parentElement, subdocument = DocumentBase """
        newElement = self.document.importNode(subdocument.element, deepCopy) # True = deep copy
        parentElement.insertBefore(newElement, parentElement.firstChild)
        return newElement

    def appendTags(self, element, tags):
        for tag in tags:
            tagElement = self.document.createElement(Tags.tag)
            tagElement.appendChild(self.document.createTextNode(self.makeUnicode(tag)))
            element.appendChild(tagElement)

    def getTags(self):
        tagElements = self.getChildElements(Tags.tag)
        tags = list()
        for tagElement in tagElements:
            tag = self.getTextNode(tagElement)
            if tag:
                tags.append(tag.data)
        return tags

    def partialCopyTree(self, sourceElement, destinationElement, tagList, indent=""):
        """ copies xml based on a list of tag names, one tag per level
            example call:
            self.partialCopyTree(original.document.documentElement, self.element,
                               [Tags.device_class, Tags.test, Tags.run, Tags.summary])
        """
        for originalElement in sourceElement.getElementsByTagName(tagList[0]):
            newElement = self.document.importNode(originalElement, False) # False = shallow copy
            destinationElement.appendChild(newElement)
            textNode = self.getTextNode(originalElement)
            if textNode and not textNode.data.isspace():
                newElement.appendChild( self.document.createTextNode(textNode.data))
            if len(tagList) > 1:
                self.partialCopyTree(originalElement, newElement, tagList[1:], indent=indent+"   ")

    def toXML(self):
        return self.document.toxml()

class DocumentBase(DocumentWrapper):

    def __init__(self):
        DocumentWrapper.__init__(self)
        self.loadedFilePath = None
        self.loadedFileSize = 0 # bytes
        self.filtered = False

    def _new(self, topElement, stylesheet):
        dom = getDOMImplementation()
        self.document = dom.createDocument(None, topElement, None)
        self.element = self.document.documentElement
        self.element.setAttribute(Attributes.rhcert_version, version.version)
        self.element.setAttribute(Attributes.rhcert_release, version.release)

        # add the style sheet processing instruction - one for hosted rhcert, a fall-back relative path
        for host in [("https://" + Constants.rhcert_hosted_on_host), ""]:
            styleRef = self.document.createProcessingInstruction("xml-stylesheet", "href=\"%s%s\" type=\"text/css\"" % (host, stylesheet))
            self.document.insertBefore(styleRef, self.element)

    def load(self, filename):
        self.loadedFilePath = filename
        self.loadedFileSize = 0
        if filename[-2:] == "gz":
            if os.path.exists(filename):
                xmlfile = gzip.open(filename, 'rb')
                self.loadedFileSize = os.path.getsize(self.loadedFilePath)
        else:
            # Note, if this raises IOError or OSError, let the caller handle it.
            xmlfile = open(filename, "r")
            self.loadedFileSize = os.path.getsize(self.loadedFilePath) # bytes

        if self.loadedFileSize:
            self.document = parse(xmlfile)
            self.element = self.document.documentElement
            xmlfile.close()
            self.filtered = False
        else:
            raise RHCertDocumentException("Error: %s has a size of 0" % filename)

    def reload(self):
        """ often the thing to do when you want to modify a document that was filtered on load """
        if not self.loadedFilePath:
            print("Error: can not reload before loading.")
            return
        self.load(self.loadedFilePath)

    def loadFiltered(self, filename, xslFilterFile, xsltParameters):
        self.loadedFilePath = filename
        self.loadedFileSize = 0
        if filename[-2:] == "gz":
            xmlfile = gzip.open(filename, 'rb')
            self.loadedFileSize = os.path.getsize(self.loadedFilePath)
        else:
            try:
                self.loadedFileSize = os.path.getsize(self.loadedFilePath) # bytes
            except OSError:
                pass
            xmlfile = open(filename, "r")

        if self.loadedFileSize:
            try:
                xslt = etree.parse(xslFilterFile)
                # we need huge_tree set as some attachments are 24MB+
                parser = etree.XMLParser(huge_tree=True)
                xmlTree = etree.parse(xmlfile, parser)
                self.filtered = True
            except etree.XMLSyntaxError as error:
                print("Error: could not parse XML file:")
                print(error)
                self.document = None
                xmlfile.close()
                return

            access = etree.XSLTAccessControl(read_file=True, write_file=False, create_dir=False, read_network=False, write_network=False)
            transform = etree.XSLT(xslt, access_control=access)
            output = transform(xmlTree, **xsltParameters)
            outputString = StringIO()
            output.write(outputString)
            self.document = parseString(outputString.getvalue())
            self.element = self.document.documentElement
        else:
            print("Error: %s has a size of 0" % filename)
            self.document = None

        xmlfile.close()

    def loadString(self, xmlString):
        self.document = parseString(xmlString)
        self.element = self.document.documentElement

    def save(self, filePath=None):
        if not filePath:
            filePath = self.loadedFilePath
        if filePath:
            name, extension = os.path.splitext(filePath)
            if extension == ".gz":
                self.compressToFile(name)
            else:
                self.__saveXML(filePath)
            self.loadedFilePath = filePath
            self.loadedFileSize = os.path.getsize(self.loadedFilePath)
        else:
            print("Error: Cannot save as the path of the file is not loaded")

    def __saveXML(self, filename):
        xmlfile = open(filename, "wb")
        try:
            xmlfile.write(self.document.toxml(encoding='utf-8'))
        except KeyboardInterrupt:
            print("\nExiting - please wait while data is saved...")
            xmlfile.close()
            xmlfile = open(filename, "wb")
            xmlfile.write(self.document.toxml(encoding='utf-8'))
            xmlfile.close()
            sys.exit(1)

        xmlfile.close()
    
    def compressNoSave(self, fileName, fileContents):
        try:
            gzippedContents = StringIO()
            with gzip.GzipFile(fileobj=gzippedContents, mode="w") as f:
                f.write(rhcert_make_str(fileContents))
            
            if os.path.splitext(fileName)[1] == ".xml":
                return (fileName+".gz", gzippedContents.getvalue())
            if os.path.splitext(fileName)[1] == ".gz":
                return (fileName, gzippedContents.getvalue())

        except Exception as exception:
            print("Could not compress file data.")
            print(exception)
            
        return None

    def compressToFile(self, filename):
        self.save(filename)
        try:
            os.remove(filename+".gz")
        except:
            pass
        try:
            gzip = Command("gzip -n \"%s\"" % filename)
            gzip.run()
            return filename+".gz"
        except Exception as exception:
            print("Error: could not compress %s" % filename)
            print(exception)

        return None

    def importElement(self, topElement):
        dom = getDOMImplementation()
        self.document = dom.createDocument(None, topElement.localName, None)
        newTopElement = self.document.importNode(topElement, True)
        oldTopElement = self.document.childNodes[0]
        self.document.replaceChild(newTopElement, oldTopElement)
        self.element = self.document.documentElement

    def getLoadedFileSize(self):
        """ return the size of the loaded xml file in bytes """
        return self.loadedFileSize

    def getLoadedFilePath(self):
        return self.loadedFilePath

class NamedDocumentWrapper(DocumentWrapper):

    def __init__(self, element, document):
        DocumentWrapper.__init__(self)
        self.element = element
        self.document = document
        # replace idrefs with full copies
        if self.element.getAttribute(Attributes.idref):
            idref = self.element.getAttribute(Attributes.idref)
            elements = self.document.getElementsByTagName(self.element.tagName)
            for element in elements:
                if element.getAttribute(Attributes.id) == idref:
                    elementCopy = self.document.importNode(element, True) # True = deep copy
                    parent = self.element.parentNode
                    parent.removeChild(self.element)
                    parent.appendChild(elementCopy)
                    self.element = elementCopy
                    return
            # otherwise, bad ID ref
            raise Exception("tag %s has invalid idref %s" % (self.element.tagName, idref))

    def getName(self):
        return self.element.getAttribute(Attributes.name)
    def setName(self, name):
        if name:
            self.setAttribute(self.element, Attributes.name, name)
    def getShortName(self):
        return self.element.getAttribute(Attributes.short_name)
    def setShortName(self, name):
        if name:
            self.setAttribute(self.element, Attributes.short_name, name)
    def getID(self):
        return self.element.getAttribute(Attributes.id)
    def setID(self, id):
        if id:
            self.element.setAttribute(Attributes.id, str(id))

class RHCertModelException(Exception):
    pass

class RHCertDocumentException(RHCertModelException):
    pass

if __name__ == '__main__':

    """ Test class for read/write of documents to xml files """
    class XmlFileTest(DocumentBase):
        def __init__(self):
            dom = getDOMImplementation()
            self.document = dom.createDocument(None, "documentbase-test", None)
            self.topElement = self.document.documentElement
            stylesheet = self.document.createProcessingInstruction("xml-stylesheet", "href=\"/rhcert/css/results.css\" type=\"text/css\"")
            self.document.insertBefore(stylesheet, self.topElement)
        def test(self, data, id):
            self.topElement.setAttribute("test", data.encode("utf-8"))
            textNode = self.document.createTextNode(data.encode("utf-8"))
            self.topElement.appendChild(textNode)
            file = open("documentbase-selftest-%s.xml" % id, "w")
            file.write(self.document.toxml(encoding='utf-8'))
            file.close()
            file = open("documentbase-selftest-%s.xml" % id, "r")
            document = parse(file)
            file.close()
            print("data parsed, written, and read from xml file")

    testData = list()
    print("\nGenerating ascii 0-127:")
    bad = '' # str, not unicode
    for i in range(128):
        bad += chr(i)
    sample = u'' + bad
    testData.append((sample, "0-127"))

    print("\n Generating unicode 128-255")
    bad = u'' # unicode
    for i in range(128,256):
        print(unichr(i))
        bad += unichr(i)
    sample = u'' + bad
    testData.append((sample, "128-255"))

    testData.append((u'\u1234\u20ac\u0000\u2332', "multibyte"))

    testData.append((u'some\nwords\nget\nchopped', 'regular'))

    # testData.append((chr(0xc2), "ascii-value"))

    wrapper = DocumentWrapper()
    wrapper.Debugging = Constants.high

    for (sample, id) in testData:
        xmlFileTest = XmlFileTest()
        print("Doing sample %s" % id)
        sample = wrapper.removeBadCharacters(sample)
        xmlFileTest.test(sample, id)

