# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
# The Image class is for capturing tracking information about virtualization images
# used in certification testing.
# 
# timestamp: the build/release timestamp
# rhcert-image-version: the version of rhcert intended to run the image
# rhcert-host-version: the version of rhcert intended to run on the host

import time
from rhcert.tags import Tags, Attributes
from rhcert.model.documentbase import DocumentBase

class Image(DocumentBase):

    def __init__(self, imageFile=None):
        self.imageFile = "rhcert-image.xml"
        if imageFile:
            self.imageFile = imageFile

        try:
            self.load(self.imageFile)
            self.element = self.document.documentElement
        except IOError:
            self.new()

    def new(self):
        self._new(topElement=Tags.rhcert_image, stylesheet="/rhcert/css/image.css")
        self.element = self.document.documentElement

    def setTime(self):
        theTime = time.gmtime(time.time())
        timestamp = self.timeToString(theTime)
        self.setAttribute(self.element, Attributes.time, timestamp)

    def getTimestamp(self):
        return self.element.getAttribute(Attributes.time)

    def setImageVersion(self, version, release):
        imageElement = self.findOrCreateElement(self.element, Tags.image)
        imageElement.setAttribute(Attributes.rhcert_version, version)
        imageElement.setAttribute(Attributes.rhcert_release, release)

    def getImageVersion(self):
        imageElement = self.findOrCreateElement(self.element, Tags.image)
        return (imageElement.getAttribute(Attributes.rhcert_version), imageElement.getAttribute(Attributes.rhcert_release))

    def setHostVersion(self, version, release):
        hostElement = self.findOrCreateElement(self.element, Tags.host)
        hostElement.setAttribute(Attributes.rhcert_version, version)
        hostElement.setAttribute(Attributes.rhcert_release, release)

    def getHostVersion(self):
        hostElement = self.findOrCreateElement(self.element, Tags.host)
        return (hostElement.getAttribute(Attributes.rhcert_version), hostElement.getAttribute(Attributes.rhcert_release))

    def __str__(self):
        hostVersion, hostRelease = self.getHostVersion()
        imageVersion, imageRelease = self.getImageVersion()
        return "rhcert image: host: rhcert-%s-%s, guest: rhcert-%s-%s  build:  %s" % ( hostVersion, hostRelease, imageVersion, imageRelease, self.getTimestamp())

def unitTest():
    print("new image file...")
    image = Image()
    image.setTime()
    image.setHostVersion("1.3", "7")
    image.setImageVersion("1.2", "25")
    print(image)

    print("saving image...")
    image.save("/tmp/rhcert-image.xml")

    print("load image file...")
    image = Image("/tmp/rhcert-image.xml")
    print(image)
if __name__ == "__main__":
    unitTest()
