# Copyright (c) 2016 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
from xml.etree import ElementTree

from rhcert.compatability import rhcert_make_unicode
from rhcert.model.documentbase import DocumentWrapper, DocumentBase
from rhcert.tags import Attributes, Tags, TestResult


""" 
OutputDocuments are nested test logging documents.
Example:

<output description="Log rhcert and OS version and release" name="Log Versions">
Tested OS: Red Hat Enterprise Linux Server 7.2 (Maipo)
Kernel RPM: kernel-3.10.0-327.el7
redhat-certification version 2.7, release 20160722
<summary>PASS</summary>
</output>

"""
class OutputDocument(DocumentWrapper):

    def __init__(self, outputElement, document):
        DocumentWrapper.__init__(self)
        self.element = outputElement
        self.document = document

    def getName(self):
        return self.element.getAttribute(Attributes.name)

    def getDescription(self):
        return self.element.getAttribute(Attributes.description)

    def setSummary(self, summary):
        summaryElement = self.findOrCreateElement(self.element, Tags.summary)
        self.setAttribute(summaryElement, Attributes.data_value, summary) # for rendering colors
        try:
            summaryElement.childNodes[0].data = str(TestResult(summary))
        except IndexError:
            textNode = self.document.createTextNode(summary)
            summaryElement.appendChild(textNode)

    def getSummary(self):
        try:
            summaryElement = self.getChildElement(Tags.summary)
            if summaryElement:
                return summaryElement.getAttribute(Attributes.data_value)
        except IndexError:
            return None
        return None

    def getSubOutputDocuments(self):
        subs = list()
        for node in self.element.childNodes:
            if node.nodeType == node.ELEMENT_NODE and node.tagName == Tags.output:
                subs.append(OutputDocument(node, self.document))
        return subs

    def getLog(self):
        output = ""
        for node in self.element.childNodes:
            if node.nodeType == node.TEXT_NODE:
                    output += node.data + "\n"
        return output

    def appendLog(self, newText):
        textNode = self.document.createTextNode(newText)
        summaryElement = self.getChildElement(Tags.summary)
        self.element.insertBefore(textNode, summaryElement)

    def getDeferredCalls(self):
        deferredCalls = list()
        summaryElement = self.getChildElement(Tags.summary)
        if not summaryElement:
            return None
        for element in summaryElement.childNodes:
            if element.nodeName == Tags.deferred:
                deferredCall = DeferredCall()
                deferredCall.load(element.toxml())
                deferredCalls.append(deferredCall)
        return deferredCalls

    def removeDeferredNode(self, test_name):
        deferredNodes = list()
        summaryElement = self.getChildElement(Tags.summary)
        if not summaryElement:
            return None
        for element in summaryElement.childNodes:
            if (element.nodeName == Tags.deferred and
                    element.getAttribute('name') == test_name):
                deferredNodes.append(element)
        for node in deferredNodes:
            summaryElement.removeChild(node)

class FeatureData():
    def __init__(self):
        self.new()

    def new(self):
        self.element = ElementTree.Element(Tags.feature)

    def loadService(self, feature_data):
        self.__setAttribute(self.element, Attributes.name, feature_data["name"])
        for feature_item in feature_data["items"]:
            itemElement = ElementTree.SubElement(self.element, Tags.feature_item)
            for prop, value in feature_item.items():
                self.__setAttribute(itemElement, prop, value)

    def __setAttribute(self, element, name, value):
        element.set(name, rhcert_make_unicode(value))

    def toXML(self):
        return convertToXML(self.element)

class DeferredCall():
    def __init__(self):
        self.new()

    def new(self):
        self.element = ElementTree.Element(Tags.deferred)

    def load(self, xmlString):
        self.element = ElementTree.fromstring(xmlString)

    def loadService(self, microservice):
        self.__setAttribute(Attributes.name, microservice.__class__.__name__)
        self.__setAttribute(Attributes.url, microservice.getURL())
        if hasattr(microservice, "getImage"):
            self.__setAttribute(Attributes.image, microservice.getImage())
        if hasattr(microservice, "getContainer"):
            self.__setAttribute(Attributes.container, microservice.getContainer())
        if microservice.getHeaders():
            headersElement = ElementTree.SubElement(self.element, Tags.headers)
            self.__appendProperties(headersElement, microservice.getHeaders())
        if microservice.getPayload():
            payloadElement = ElementTree.SubElement(self.element, Tags.payload)
            self.__appendProperties(payloadElement, microservice.getPayload())

    def getImage(self):
        return self.element.get(Attributes.image)

    def getContainer(self):
        return self.element.get(Attributes.container)

    def getURL(self):
        return self.element.get(Attributes.url)

    def getName(self):
        return self.element.get(Attributes.name)

    def getHeaders(self):
        headers = dict()
        for header in self.element.find(Tags.headers).findall(Tags.property):
            headers[header.get(Attributes.name)] = header.text
        return headers

    def getPayload(self):
        payload = dict()
        for item in self.element.find(Tags.payload).findall(Tags.property):
            if item.get(Attributes.name) in payload: # convert to list value
                if not isinstance(payload[item.get(Attributes.name)], list):
                    values = [payload[item.get(Attributes.name)]]
                    payload[item.get(Attributes.name)] = values
                if item.text:
                    payload[item.get(Attributes.name)].append(item.text)
            else:
                if item.text:
                    payload[item.get(Attributes.name)] = item.text

        return payload

    def __appendProperties(self, element, properties):
        for (name, value) in properties.items():
            if isinstance(value, list):
                for item in value:
                    self.__appendElement(element, name, item)
            else:
                self.__appendElement(element, name, value)

    def __appendElement(self, parentElement, name, value):
        element = ElementTree.SubElement(parentElement, Tags.property)
        element.set(Attributes.name, name)
        element.text = rhcert_make_unicode(value)

    def __setAttribute(self, name, value):
        self.element.set(name, rhcert_make_unicode(value))

    def toXML(self):
        return convertToXML(self.element)

def convertToXML(element):
    try: # python2
        python2test = unicode
        return ElementTree.tostring(element, encoding="utf-8")
    except NameError: # python 3
        return str(ElementTree.tostring(element))
