# Copyright (c) 2016 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
""" Parameters are a list of name-value pairs """
from rhcert.tags import Tags, Attributes
from rhcert.model.documentbase import DocumentWrapper, DocumentBase

class ParametersDocument(DocumentWrapper):

    def __init__(self, parametersElement=None, document=None):
        self.element = parametersElement
        self.document = document
        self.parameters = dict()

    def __str__(self):
        message = ""
        for (key, value) in self.getAll().items():
            message += "%s=%s " % (key, value)
        return message

    def copy(self, testParameters):
        self.element = None
        self.document = None
        self.parameters = testParameters.getAll()

    def save(self, parametersElement, document):
        self.element = parametersElement
        self.document = document
        for (key, value) in self.parameters.items():
            self.set(key, value)

    def get(self, name):
        parameters = self.getAll()
        try:
            return parameters[name]
        except KeyError:
            return None

    def set(self, name, value):
        # print("param:set: %s = %s (%s)" % (name, value, type(value))
        self.__addValue(name, value)

        if value and self.element and self.document:

            try: # python2
                stringType = basestring
                unicodeType = unicode
            except NameError: # python3
                stringType = str
                unicodeType = str

            if isinstance(value, stringType): # put strings in list
                value = [value]
            elif isinstance(value, dict): # flatten nested dict
                for (key, data) in value.items():
                    self.set(name + "_" + key, data)
                return
            else:
                # not a dictionary or a string, or a list, then put it in a list
                try:
                    iter(value)
                except TypeError:
                    value = [value]

            for item in value: # value is a list
                try: # works for xmlrpc.Binary
                    if stringType != unicodeType:
                        item = unicodeType(item.data, "utf-8")
                    else:
                        item = str(item.data)
                except:
                    pass
                if type(item) not in [str, unicodeType]:
                    item = str(item)
                propertyElement = self.document.createElement(Tags.property)
                self.setAttribute(propertyElement, Attributes.name, name)
                self.setTextNode(propertyElement, item)
                self.element.appendChild(propertyElement)

    def reset(self, name, value):
        # account for dictionary flattening
        if isinstance(value, dict):
            for (key, _) in value.items():
                self.remove(name + "_" + key)
        else:
            self.remove(name)
        self.set(name, value)
        return self.parameters

    def remove(self, name):
        propertyElements = self.getChildElements(Tags.property)
        for propertyElement in propertyElements:
            if name == propertyElement.getAttribute(Attributes.name):
                self.element.removeChild(propertyElement)
        if name in self.parameters:
            self.parameters.pop(name)

    def replace(self, parametersDictionary):
        for (name, value) in parametersDictionary.items():
            self.reset(name, value)

    def __addValue(self, name, value):
        if name not in self.parameters:
            self.parameters[name] = value 
        # convert to list if necessary
        elif type(self.parameters[name]) is not list:
            self.parameters[name] = [self.parameters[name], value]
        else: # already a list
            self.parameters[name].append(value)

    def update(self, newValuesDictionary):
        updated = False
        for (attribute, value) in newValuesDictionary.items():
            if type(value) != type(self.get(attribute)) or self.get(attribute) != value:
                self.reset(attribute, value)
                updated = True
        return updated


    def getAll(self):
        """ returns a dict of values, values are single or lists """
        if not self.element or not self.document:
            return self.parameters
        # otherwise
        propertyElements = self.getChildElements(Tags.property)
        self.parameters = dict()
        for propertyElement in propertyElements:
            value = self.getTextNode(propertyElement)
            name = propertyElement.getAttribute(Attributes.name)
            if name and value:
                self.__addValue(name, value.data)
        return self.parameters

    def getAllSorted(self):
        """ sort the parameters into a list of tuples by key name"""
        self.getAll() # make sure self.parameters is created
        sort = list()
        for key in sorted(self.parameters):
            sort.append((key, self.parameters.get(key)))
        return sort

    def getElement(self):
        return self.element

    def dump(self):
        propertyElements = self.getChildElements(Tags.property)
        for propertyElement in propertyElements:
            print(propertyElement.toxml())

class UnitTestDocument(DocumentBase):

    def __init__(self):
        DocumentBase.__init__(self)

    def new(self):
        self._new(topElement=Tags.certification, stylesheet="/rhcert/css/certification.css")

def unitTest():
    unitTestDocument = UnitTestDocument()
    unitTestDocument.new()
    parameters = ParametersDocument(unitTestDocument.element, unitTestDocument.document)
    parameters.set("color", "red")
    parameters.set("sizes", "small")
    parameters.set("sizes", "medium")
    parameters.set("sizes", "large")
    parameters.set("lengths", ["1","2","3"])
    parameters.replace({"lengths": ["4", "5", "6"]})

    parameters.dump()
    print("\nupdate")
    parameters.update({"sizes": {"green": ["small", "medium", "large"]}})
    parameters.dump()
    print(parameters.parameters)
    print("\nupdate")
    parameters.update({"sizes": {"green": ["small", "medium", "large"]}})
    parameters.dump()
    print(parameters.parameters)

if __name__ == "__main__":
    unitTest()

