#!/usr/bin/env python
# Copyright (c) 2015 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#

from rhcert.tags import Tags, Attributes
from rhcert.model.documentbase import NamedDocumentWrapper
from rhcert.model.hostdocument import HostDocument
from rhcert.model.featuredocument import Feature

class PlanComponent(NamedDocumentWrapper):

    def __init__(self, element, document):
        NamedDocumentWrapper.__init__(self, element, document)

    def isSupplemental(self):
        self.element.getAttribute(Attributes.is_supplemental)
    def setSupplimental(self, value):
        if value:
            self.setAttribute(self.element, Attributes.is_supplemental, True)
        else:
            self.setAttribute(self.element, Attributes.is_supplemental, False)

    def getDescription(self):
        return self.getChildElementString(Tags.description)
    def setDescription(self, description):
        self.setChildElement(Tags.description, description)

    def getComponent(self):
        return self.element.getAttribute(Attributes.component)
    def setComponent(self, value):
        return self.setAttribute(self.element, Attributes.component, value)

    def getType(self):
        return self.element.getAttribute(Attributes.type)
    def setType(self, value):
        return self.setAttribute(self.element, Attributes.type, value)

    def getStatus(self):
        return self.element.getAttribute(Attributes.status)
    def setStatus(self, value):
        return self.setAttribute(self.element, Attributes.status, value)

    def getBits(self):
        return self.element.getAttribute(Attributes.bits)
    def setBits(self, value):
        return self.setAttribute(self.element, Attributes.bits, value)

    def getTestID(self):
        return self.element.getAttribute(Attributes.test_id)
    def setTestID(self, value):
        if value:
            return self.setAttribute(self.element, Attributes.test_id, str(value))
        # otherwse
        self.removeAttribute(self.element,Attributes.test_id)

    def getLeveragedTestID(self):
        return self.element.getAttribute(Attributes.leveraged_test_id)
    def setLeveragedTestID(self, value):
        if value:
            return self.setAttribute(self.element, Attributes.leveraged_test_id, str(value))
        # otherwse
        self.removeAttribute(self.element,Attributes.leveraged_test_id)

    def getLeveragedCertificationID(self):
        return self.element.getAttribute(Attributes.leveraged_cert_id)
    def setLeveragedCertificationID(self, value):
        if value:
            return self.setAttribute(self.element, Attributes.leveraged_cert_id, str(value))
        # otherwise
        self.removeAttribute(self.element,Attributes.leveraged_cert_id)

    def getRequiredOS(self):
        return HostDocument(self.findOrCreateElement(self.element, Tags.os), self.document)

    def hasRequiredOS(self):
        if self.getChildElements(Tags.os):
            return self.getRequiredOS()
        return None

    def getFeatures(self):
        features = []
        for featureElement in self.getChildElements(Tags.feature):
            features.append(Feature(featureElement, self.document))
        return features

    def getFeatureCategories(self):
        """ returns a dict[category] -> list(Feature) """
        featuresDict = {}
        for feature in self.getFeatures():
            try:
                featuresDict[feature.getCategory()].append(feature)
            except KeyError:
                featuresDict[feature.getCategory()] = [feature]
        return featuresDict

    def addFeature(self, id, name, category):
        featureElement = self.document.createElement(Tags.feature)
        self.element.appendChild(featureElement)
        feature = Feature(featureElement, self.document)
        feature.setID(id)
        feature.setName(name)
        feature.setCategory(category)
        return feature

    def removeFeature(self, id):
        for featureElement in self.getChildElements(Tags.feature):
            if str(id) == Feature(featureElement, self.document).getID():
                featureElement.parentNode.removeChild(featureElement)


if __name__ == "__main__":
    from rhcert.model.certificationdocument import CertificationDocument
    from rhcert.model.featuredocument import Feature
    doc = CertificationDocument()
    doc.new()
    componentElement = doc.findOrCreateElement(doc.element, Tags.plan_component)
    component = PlanComponent(componentElement, doc.document)
    component.addFeature(1, "20G Ethernet", "Network")
    component.addFeature(2, "CPUScaling", "System")
    component.addFeature(3, "Memory", "System")
    component.addFeature(4, "Wireless N", "Network")
    for feature in component.getFeatures():
        print("Feature: %s %s %s" % (feature.getID(), feature.getName(), feature.getCategory()))
    print("\n")
    for category, features in component.getFeatureCategories().items():
        print("category: %s:" % category)
        for feature in features:
            print("   Feature %s %s" % (feature.getID(), feature.getName()))
    doc.save("plancomponent.xml")


