#!/usr/bin/env python
# Copyright (c) 2015 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#

from rhcert.tags import Tags, Attributes
from rhcert.model.documentbase import NamedDocumentWrapper


class ProductDocument(NamedDocumentWrapper):

    def __init__(self, element, document):
        NamedDocumentWrapper.__init__(self, element, document)

    def getMake(self):
        return self.getChildElementString(Tags.make)
    def setMake(self, make):
        self.setChildElement(Tags.make, make)
    def getDescription(self):
        return self.getChildElementString(Tags.description)
    def setDescription(self, description):
        self.setChildElement(Tags.make, description)
    def getPlatform(self):
        return self.element.getAttribute(Attributes.platform)
    def setPlatform(self, platform):
        self.setAttribute(self.element, Attributes.platform, platform)
    def getCategory(self):
        return self.element.getAttribute(Attributes.category)
    def setCategory(self, category):
        self.setAttribute(self.element, Attributes.category, category)
    """ some cert IDs and program come in from CWE, so we capture that here """

    def serialize(self):
        return {
            'id': self.getID(),
            'name': self.getName(),
            'make': self.getMake(),
            'description': self.getDescription(),
            'platform': self.getPlatform(),
            'category': self.getCategory(),
            'program': self.getProgram(),
            'version': self.getVersion()
        }

    def getCertificationID(self):
        return self.element.getAttribute(Attributes.certification_id)
    def setCertificationID(self, id):
        if id:
            self.setAttribute(self.element, Attributes.certification_id, id)

    def getNodeID(self):
        return self.element.getAttribute(Attributes.node_id)
    def setNodeID(self, id):
        if id:
            self.element.setAttribute(Attributes.node_id, str(id))

    def getWorkflow(self):
        return self.element.getAttribute(Attributes.workflow)
    def setWorkflow(self, workflow):
        if workflow:
            self.setAttribute(self.element, Attributes.workflow, workflow)

    def getProgram(self):
        return self.element.getAttribute(Attributes.type)
    def setProgram(self, program):
        if program:
            self.setAttribute(self.element, Attributes.type, program)

    def getVersionString(self):
        version = self.getVersion()
        if version:
            return version.getVersion()
        return None

    def createVersion(self):
        versionElement = self.document.createElement(Tags.version)
        self.element.appendChild(versionElement)
        productVersion = ProductVersionDocument(versionElement, self.document)
        return productVersion

    def getVersions(self):
        versions = list()
        for element in self.getChildElements(Tags.version):
            versions.append(ProductVersionDocument(element, self.document))
        return versions

    def getVersion(self):
        versions = self.getVersions()
        if versions:
            return versions[0]

    def getURLs(self):
        """ return a dictionary of urls by name """
        urls = dict()
        for urlElement in self.getChildElements(Tags.url):
            urls[urlElement.getAttribute(Attributes.name)] = self.getTextNode(urlElement).data
        return urls
    def __getURLElement(self, name):
        for urlElement in self.getChildElements(Tags.url):
            if urlElement.getAttribute(Attributes.name) == name:
                return urlElement
        return None
    def setURL(self, name, value):
        urlElement = self.__getURLElement(name)
        if not urlElement:
            urlElement = self.document.createElement(Tags.url)
            self.appendChild(urlElement)
            urlElement.setAttribute(Attributes.name, name)
        self.setTextNode(urlElement, value)

    def getProducts(self):
        products = list()
        for element in self.getChildElements(Tags.product):
            products.append(ProductDocument(element, self.document))
        return products

    def isPassThrough(self):
        if self.getBaseProductID():
            return True
        return False

    def getBaseProductID(self):
        return self.element.getAttribute(Attributes.base_product_id)

    def setBaseProductID(self, baseProductID):
        if baseProductID:
            self.setAttribute(self.element, Attributes.base_product_id, baseProductID)

    def getPassThroughProductIDs(self):
        passThoughProductIDs = list()
        for element in self.getChildElements(Tags.pass_though_product):
            passThoughProductIDs.append(self.getTextNode(element).data)
        return passThoughProductIDs

    def setPassThroughProductIDs(self, passThroughProductIDs):
        for element in self.getChildElements(Tags.pass_though_product):
            self.element.removeChild(element)
        if not passThroughProductIDs:
            return None
        for passThroughID in passThroughProductIDs:
            element = self.document.createElement(Tags.pass_though_product)
            self.setTextNode(element, passThroughID)
            self.element.appendChild(element)

class ProductVersionDocument(NamedDocumentWrapper):

    def __init__(self, element, document):
        NamedDocumentWrapper.__init__(self, element, document)

    def serialize(self):
        return {
            "id": self.getID(),
            "name": self.getName(),
            "platform": self.getPlatform(),
            "version": self.getVersion()
        }

    def getPlatform(self):
        return self.element.getAttribute(Attributes.platform)
    def setPlatform(self, platform):
        self.setAttribute(self.element, Attributes.platform, platform)
    def getVersion(self):
        return self.element.getAttribute(Attributes.version)
    def setVersion(self, version):
        self.setAttribute(self.element, Attributes.version, version)

class PackageDocument(NamedDocumentWrapper):
    def __init__(self, documentElement, document):
        self.element = documentElement
        self.document = document

    def getName(self):
        if self.getTextNode(self.element):
            return self.getTextNode(self.element).data
        return None

    def getInstalled(self):
        return self.element.getAttribute(Attributes.installed)
    def setInstalled(self, installed):
        return self.setAttribute(self.element, Attributes.installed, installed)


if __name__ == "__main__":
    product = ProductDocument()
    product.new()
    productFile=open("product.xml","w")
    product.save(ProductDocument)


