#!/usr/bin/env python
# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
import re, time

from rhcert.tags import Tags, Attributes, Constants
from rhcert.model.documentbase import DocumentBase, RHCertDocumentException
from rhcert.model.testdocument import TestDocument
from rhcert.model.certificationdocument import CertificationDocument
from rhcert.model.hostdocument import HostDocument

class ResultsDocument(CertificationDocument):
    """ ResultsDocument represents the certification test results """

    def __init__(self, topElement=None, document=None, debugging=False):
        CertificationDocument.__init__(self, topElement, document, debugging)
        self.currentTestElement = None
        if self.element and self.document:
            self._wrap()

    def new(self):
        self._new(topElement=Tags.certification_test, stylesheet="/rhcert/css/results.css")
        self.findOrCreateElement(self.document.documentElement, Tags.hardware)
        self.findOrCreateElement(self.document.documentElement, Tags.certification)

    def _wrap(self):
        self.findOrCreateElement(self.document.documentElement, Tags.hardware)
        self.findOrCreateElement(self.document.documentElement, Tags.certification)


    def copy(self, orginal):
        CertificationDocument.copy(self, orginal)

    def copyCertification(self, sourceCertificationDocument, deleteProperties=False):
        CertificationDocument.copyCertification(self, sourceCertificationDocument, deleteProperties)
        self.assignComponents()

    def load(self, filename):
        DocumentBase.load(self, filename)
        if self.document:
            self.__hostOSFixup()
            return True
        return False

    def save(self, filePath=None):
        # pre-save validation
        self.validate()
        # actually save the file
        super(ResultsDocument, self).save(filePath)

    def validate(self):
        if self.filtered: raise RHCertDocumentException(
            "Error: Attempt to save filtered results may overwrite original data"
        )

    def __hostOSFixup(self):
        """ in rhcert-6.3, the actual OS detected in the host was moved from
            <certification><os> to <hardware><os> So we need to move it if found
            in the wrong place
        """
        certificationOS = self.getDescendantElement([Tags.certification, Tags.os])
        hardwareOS = self.getDescendantElement([Tags.hardware, Tags.os])
        if not hardwareOS and certificationOS:
            self.mergeDocument(self.getRequiredOS(), self.__getHardwareElement())

    def checkRequiredOS(self):
        """ reverse of __hostOSFixup for CERTPX-4042: fill in <certification><os> if missing """
        certificationOS = self.getDescendantElement([Tags.certification, Tags.os])
        hardwareOS = self.getDescendantElement([Tags.hardware, Tags.os])
        if not certificationOS and hardwareOS:
            self.mergeDocument(self.getHostOS(), self.__getCertificationElement())


    def importElement(self, topElement):
        DocumentBase.importElement(self, topElement)
        self._wrap()

    def getDiscoverTime(self):
        return self.document.documentElement.getAttribute(Attributes.discover_time)

    def setDiscoverTime(self, discoverTime):
        self.document.documentElement.setAttribute(Attributes.discover_time, self.timeToString(discoverTime))

    def getPlanTime(self):
        return self.document.documentElement.getAttribute(Attributes.plan_time)

    def setPlanTime(self, planTime):
        self.document.documentElement.setAttribute(Attributes.plan_time, self.timeToString(planTime ))

    def getSubmitTime(self):
        return self.document.documentElement.getAttribute(Attributes.submit_time)

    def setSubmitTime(self, submittedTime):
        self.document.documentElement.setAttribute(Attributes.submit_time, self.timeToString(submittedTime ))

    def getValidationJobID(self):
        return self.document.documentElement.getAttribute(Attributes.validation_job_id)

    def setValidationJobID(self, id):
        self.document.documentElement.setAttribute(Attributes.validation_job_id, str(id))

    def getState(self):
        return self.document.documentElement.getAttribute(Attributes.state)

    def setState(self, state):
        self.document.documentElement.setAttribute(Attributes.state, state)

    def getHostOS(self):
        """ returns <hardware><os>: the actual RHEL OS and host information """
        return HostDocument(self.findOrCreateElement(self.__getHardwareElement(), Tags.os), self.document)

    def getHostOSValues(self):
        return self.getHostOS().getValues()

    def getHostname(self):
        return self.getHostOS().getName()

    def __getHardwareElement(self):
        return self.findOrCreateElement(self.document.documentElement, Tags.hardware)
    def __getCertificationElement(self):
        return self.findOrCreateElement(self.document.documentElement, Tags.certification)


    def getRunTime(self):
        runNumber= 1
        while runNumber <= self.getNumberOfTestRuns():
            for test in self.getTests():
                if test.wasRun(runNumber):
                    run = test.getRun(runNumber)
                    if run and run.getRunTime():
                        return run.getRunTime()
            runNumber = runNumber + 1
        return None

    def addTestRun(self):
        numberOfRuns = self.getNumberOfTestRuns() + 1
        self.document.documentElement.setAttribute(Attributes.number_of_runs, "%u" % numberOfRuns)
        return numberOfRuns

    def getNumberOfTestRuns(self):
        numberOfRunsString = self.document.documentElement.getAttribute(Attributes.number_of_runs)
        if numberOfRunsString:
            return int(numberOfRunsString)
        else:
            return 0

    def appendTest(self, test):
        testElement = self.document.createElement(Tags.test)
        self.setAttribute(testElement, Attributes.name, test.getName())
        self.setAttribute(testElement, Attributes.path, test.getPath())
        self.setAttribute(testElement, Attributes.priority, test.getPriority())
        self.setAttribute(testElement, Attributes.mandatory, test.getMandatory())
        self.setAttribute(testElement, Attributes.source, test.getSource())
        self.setAttribute(testElement, Attributes.source_directory, test.getSourceDirectory())
        self.setAttribute(testElement, Attributes.description, test.getDescription())
        self.element.appendChild(testElement)
        self.appendTags(testElement, test.tags())
        self.appendTestParameters(testElement, test.getTestParameters())
        self.__assignComponent(testElement)

    def appendTestParameters(self, testElement, testParameters):
        parametersElement = self.findOrCreateElement(testElement, Tags.parameters)
        for (name, value) in testParameters.getAll().items():
            propertyElement = self.document.createElement(Tags.property)
            propertyElement.setAttribute(Attributes.name, name)
            propertyElement.appendChild(self.document.createTextNode(self.makeUnicode(value)))
            parametersElement.appendChild(propertyElement)
 
    def getTests(self, names=None, onlyFailures=False, recursive=False):
        tests = list()
        if recursive:
            testElements = self.document.getElementsByTagName(Tags.test)
            for testElement in testElements:
                test = TestDocument(testElement, self.document)
                if not names or (test.getName() in names) or (test.getPath() in names):
                    if not onlyFailures or (onlyFailures and test.hasFailures()):
                        tests.append(test)
        else:
            for element in self.element.childNodes:
                if element.nodeName == Tags.test:
                    test = TestDocument(element, self.document)
                    if not names or (test.getName() in names) or (test.getPath() in names):
                        if not onlyFailures or (onlyFailures and test.hasFailures()):
                            tests.append(test)
        return tests

    def getTestResults(self):
        """ returns a tuple of lists: (passed, failed, notrun)"""
        passedTests = list()
        failedTests = list()
        notRunTests = list()
        for element in self.element.childNodes:
            if element.nodeName == Tags.test:
                test = TestDocument(element, self.document)
                if test.hasFailures():
                    failedTests.append(test)
                if test.hasPassed():
                    passedTests.append(test)
                if not test.wasRun():
                    notRunTests.append(test)
        return (passedTests, failedTests, notRunTests)

    def getFilteredTests(self, testNames, tags, parameters):
        # filter the list by "anding" different options, and "or-ing" multiple instances of options
        # example: --test a --test b --tag interactive --tag virtualization
        # results in: tests named(a or b) and tagged(interactive or virtualization)
        if testNames:
            tests = self.getTests(testNames)
        else:
            tests = self.getTests()

        filteredTests = list()
        for test in tests:
            if not test.isDisabled():
                filteredTests.append(test)
        tests = filteredTests

        tests = self.filterTestsByParameters(tests, parameters)

        if tags:
            tests = self.getTaggedTests(tests, tags)

        return tests

    def getTaggedTests(self, tests, tags):
        filteredTests = list()
        # the test must have all the requested tags
        for test in tests:
            if test.getTags() and set(tags).intersection(test.getTags()) == set(tags):
                filteredTests.append(test)
        return filteredTests

    def getIncompleteTestsFromLastRun(self):
        lastRunNumber = self.getNumberOfTestRuns()
        incompleteTests = list()
        postTests = list()
        if lastRunNumber > 0:
            for test in self.getTestsRun(lastRunNumber):
                run = test.getRun(lastRunNumber)
                if run.getSummary() == Constants.INCOMPLETE:
                    if test.getMandatory() == Constants.post:
                        postTests.append(test)
                    else:
                        incompleteTests.append(test)
        incompleteTests.extend(postTests)
        return incompleteTests

    def getTestsGroupedByName(self):
        testsByName = dict()
        for test in self.getTests():
            if test.getName() not in testsByName:
                testsByName[test.getName()] =  list()
            testsByName[test.getName()].append(test)
        return testsByName

    def getTestsByParameter(self, parameter, value, partialMatch=None):
        """ partialMatch: ["in", left", "right", None] """
        tests = list()
        for test in self.getTests():
            if ((not partialMatch and test.getTestParameters().get(parameter) == value)
                or (partialMatch == "in" and value in test.getTestParameters().get(parameter))
                or (partialMatch == "left" and test.getTestParameters().get(parameter).find(value) ==  0)
                or (partialMatch == "right" and test.getTestParameters().get(parameter).endswith(value))):
                    tests.append(test)
        return tests

    def filterTestsByParameters(self, tests, parameters):
        """ filter tests by parameters """
        filteredTests = list()
        for test in tests:
            match = True
            for (parameter, value) in parameters.items():
                if test.getTestParameters().get(parameter) != value:
                    match  = False
                    break
            if match:
                filteredTests.append(test)
        return filteredTests

    def getMandatoryTests(self):
        tests = list()
        for test in self.getTests():
            if test.getMandatory():
                tests.append(test)
        return tests

    def getTest(self, nameOrPath, parameters=None):
        """ return a path match in preference to a name match 
            if parameters are supplied, all must match  """
        nameMatch = None
        for test in self.getTests():
            # if parameters are supplied, and they're not a subset of the test, skip it.
            if parameters and not (set(parameters.items()).issubset(set(test.getTestParameters().getAll().items()))):
                continue
            if test.getName() == nameOrPath:
                nameMatch = test
            if test.getPath() == nameOrPath:
                return test

        # otherwise
        return nameMatch

    def getTestMatching(self, otherTest):
        """ find a test that is a best match otherTest's important attributes - class, udi, device """
        bestMatchScore = 0
        bestMatch = None
        for testDocument in self.getTests():
            if testDocument.getPath() == otherTest.getPath():
                matchScore = 0
                for (parameter, value) in otherTest.getTestParameters().getAll().items():
                    if testDocument.getTestParameters().get(parameter) == value:
                        matchScore += 1
                if matchScore > bestMatchScore:
                    bestMatch = testDocument
                    bestMatchScore = matchScore
        return bestMatch

    def getTestCounts(self, runNumber):
        tests = self.getTests()
        count = 0
        runCount = 0
        passed = 0
        failed = 0
        start = None
        for test in tests:            
            count = count + 1
            if test.wasRun(runNumber):
                run = test.getRun(runNumber)   
                runCount = runCount + 1
                if run.passed():
                    passed = passed + 1
                else:
                    failed = failed + 1
                if not start:
                    time = run.getRunTime()
                    if time:
                        start = time

        return {'count':count,'run':runCount, 'passed':passed, 'failed':failed, 'start':start}

    def getTestsRun(self, runNumber):
        tests = list()
        for test in self.getTests():   
            if test.wasRun(runNumber):
                tests.append(test)

        return tests

    def getCombinedTestCounts(self):
        tests = self.getTests()
        count = 0
        runCount = 0
        neverPassedCount = 0
        neverFailedCount = 0
        start = None
        for test in tests:
            count = count + 1
            wasRun = False
            passed = False
            failed = False
            for run in test.getRuns():
                if run.wasRun():
                    wasRun = True
                    if run.passed():
                        passed = True
                    else:
                        failed = True

                    if not start:
                        time = run.getRunTime()
                        if time:
                            start = time
            if wasRun: 
                runCount = runCount +1
                if not passed: neverPassedCount = neverPassedCount +1
                if not failed: neverFailedCount = neverFailedCount +1

        return {'count':count,'run':runCount, 'neverPassed':neverPassedCount, 'neverFailed':neverFailedCount, 'start':start}

    def getSubTestsGroupedByNameAndParent(self):
        subTestsByNameAndParent = list()
        for test in self.getTests():
            for run in test.getRuns():
                subtests = self.getSubTestsGroupedByName(run)
                if len(subtests) > 0:
                    subTestsByNameAndParent.append(subtests)
        return subTestsByNameAndParent

    def getDeferredTests(self):
        """ returns a dict[run number] of (testDocument, outputDocument) """
        deferredTests = dict()
        for testDocument in self.getTests():
            for run in testDocument.getRuns():
                if not testDocument.wasRun():
                    continue

                subtest_output_documents = run.getSubTestOutputDocuments()
                if not subtest_output_documents:
                    continue

                for outputDocument in subtest_output_documents:
                    if outputDocument.getDeferredCalls():
                        if not deferredTests.get(run.getNumber()):
                            deferredTests[run.getNumber()] = list()
                        deferredTests[run.getNumber()].append((testDocument, outputDocument))
        return deferredTests

    def getSubCertifications(self, run):
        subCertifications = list()
        for subCertificationElement in run.getChildElements(Tags.certification_test):
            subCertifications.append(ResultsDocument(subCertificationElement, run.document))
        return subCertifications

    def getSubTests(self, subCertifications):
        # just go one level down
        subTests = list()
        for subCertification in subCertifications:
            subTests.extend(subCertification.getTests())
        return subTests

    def getSubTestsGroupedByName(self, run):
        subtestsByName = dict()
        for subCertification in self.getSubCertifications(run):
            for subtest in subCertification.getTests():
                if subtest.getName() not in subtestsByName:
                    subtestsByName[subtest.getName()] =  list()
                subtestsByName[subtest.getName()].append(subtest)
        return subtestsByName

    def deleteAttachments(self):
        for attachmentElement in self.document.getElementsByTagName(Tags.attachment):
            attachmentElement.parentNode.removeChild(attachmentElement)

    def getVerboseFileName(self):
        """ rhcert-results-<hostname>-<run timestamp>.xml """
        prefix = "rhcert-results-"
        # timestamp
        runTime = self.getRunTime()
        if not runTime:
            runTime = self.getPlanTime()
        if runTime:
            timestamp = self.timeStringToTimestamp(runTime)
        else:
            timestamp = ""
        # hostname portion
        hostname = self.getHostOS().getName()
        ipAddress = self.getHostOS().getIPAddress()
        host =  hostname
        # if "localhost" and we have an IP address, use IP address
        if (hostname and "localhost" in hostname) and ipAddress:
            host = ipAddress
        # limit to 80 char for Hydra/SFDC by truncating hostname:
        nonHostnameLength = len(prefix) + len('-') + len(timestamp) + len(".xml.gz")
        if len(host)+nonHostnameLength > Constants.maximumResultsFileNameLength:
            host = host[:Constants.maximumResultsFileNameLength-nonHostnameLength]
        return prefix + host + "-" + timestamp + ".xml"

    def hasMultiHostResults(self):
        """ is this a multi-host results file (should use rhcert-multi- prefix) """
        if not self.isMultiHost():
            return False
        # otherwise: does this contain tests from multiple hosts?
        firstRoleID = 0
        for test in self.getTests():
            roleID = test.getHostOS().getID()
            if roleID:
                if not firstRoleID:
                    firstRoleID = roleID
                elif roleID != firstRoleID:
                    return True
        # a results doc with a multi-host test plan but no multi-host runs
        return False

    def getMultiHostFileName(self):
        """ rhcert-multi-<program>-<certID>-<run timestamp>.xml """

        # timestamp
        runTime = self.getMultiHostRunTime()
        if runTime:
            timestamp = self.timeStringToTimestamp(runTime)
        else:
            timestamp = ""
        program = self.getCertificationProgram()
        certID = self.getCertificationID()
        return "-".join(["rhcert-multi", program, certID, timestamp]) + ".xml"

    def assignComponents(self):
        updated = False
        for test in self.getTests():
            if test.getPlanComponentID(): # if already assigned, skip
                continue
            if self.__assignComponent(test.element):
                updated = True
        return updated

    def __assignComponent(self, testElement):
        test = TestDocument(testElement, self.document)
        bestMatch = None
        for planComponent in self.getPlanComponents():
            # first, try and match on test name vs. component class
            if planComponent.getName().lower() == test.getName().lower():
                # next see if device parameter exists for test
                if test.getTestParameters().get(Constants.device):
                    # check if component contains device info
                    if test.getTestParameters().get(Constants.device).lower() in planComponent.getComponent().lower().split():
                        bestMatch = planComponent
                        break
                else:
                    bestMatch = planComponent
        if bestMatch:
            test.setPlanComponentID(bestMatch.getID())
            return True
        return False

    def mergeMultiHost(self, otherHostDocument):
        """ merge multi host results from the list otherHostDocuments into this one """
        otherHostDocument.setActualHostRoles()
        # add the other host actual host/OS info the <hardware> element
        hardwareElement = self.findOrCreateElement(self.document.documentElement, Tags.hardware)
        self.mergeDocument(otherHostDocument.getHostOS(), hardwareElement)

        # merge tests with actual host/host data
        for test in otherHostDocument.getTests():
            self.mergeDocument(test, self.element)

    def setActualHostRoles(self):
        """ set actual host information in each test element """
        # set role ID and role into actual host info
        self.getHostOS().setID(self.getRequiredOS().getID())
        self.getHostOS().setRole(self.getRequiredOS().getRole())

        # for each test, set actual host/os info
        for test in self.getTests():
            self.insertDocument(self.getHostOS(), test.element)

def unitTest():
    results = ResultsDocument()
    results.load("/var/rhcert/results.xml")
    print("loaded results")
    certification = CertificationDocument()
    certification.copy(results)
    certification.copyTestRunSummaries(results)
    print("copied")
    certification.save("qq.xml")

    # test long hostname
    print("results file name %s" % results.getVerboseFileName())
    results.getRequiredOS().setName("very.very.long.hostname.from.somewhere.in.redhat.com")
    print("results file name %s" % results.getVerboseFileName())

if __name__ == "__main__":
    unitTest()
    print("saved qq.xml")
