#!/usr/bin/env python
# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
import os, re
import hashlib
import mimetypes
import io
from xml.dom.minidom import parse

from rhcert.tags import Tags, Attributes, Constants, TestResult, Regex
from rhcert.model.documentbase import DocumentWrapper, RHCertDocumentException
import rhcert.model.resultsdocument
from rhcert.model.testparameters import TestParameters
from rhcert.model.outputdocument import OutputDocument
from rhcert.model.hostdocument import HostDocument
from rhcert.compatability import rhcert_base64_encode, rhcert_base64_encode_and_md5, rhcert_base64_decode


class RunDocument(DocumentWrapper):

    def __init__(self, runElement, document):
        DocumentWrapper.__init__(self)
        self.element = runElement
        self.document = document

    def getNumber(self):
        return int(self.element.getAttribute(Attributes.number))

    def setNumber(self, number):
        self.element.setAttribute(Attributes.number, "%u" % number)

    def getMode(self):
        return self.element.getAttribute(Attributes.mode)

    def setMode(self, mode):
        self.element.setAttribute(Attributes.mode, mode)

    def getRunTime(self):
        return self.element.getAttribute(Attributes.run_time)

    def setRunTime(self, runTime):
        self.setRunTimeString(self.timeToString(runTime ))

    def setRunTimeString(self, runTimeString):
        self.element.setAttribute(Attributes.run_time, runTimeString)

    def getEndTime(self):
        return self.element.getAttribute(Attributes.end_time)

    def setEndTime(self, endTime):
        self.setEndTimeString(self.timeToString(endTime ))

    def setEndTimeString(self, endTimeString):
        self.element.setAttribute(Attributes.end_time, endTimeString)

    def getReturnValue(self):
        return self.element.getAttribute(Attributes.return_value)

    def setReturnValue(self, returnValue):
        try:
            self.setReturnValueString("%s" % returnValue)
        except TypeError:
            print("Error: %s return value, should be a boolean" % returnValue)
            self.setReturnValueString(False)

    def setReturnValueString(self, returnValueString):
        self.element.setAttribute(Attributes.return_value, returnValueString)

    def getSummary(self):
        try:
            summaryElement = self.getChildElement(Tags.summary)
            if summaryElement:
                return summaryElement.childNodes[0].data
        except IndexError:
            return None
        return None

    def setSummary(self, summary):
        summaryElement = self.findOrCreateElement(self.element, Tags.summary)
        summary = str(TestResult(summary))
        try:
            summaryElement.childNodes[0].data = summary
        except IndexError:
            textNode = self.document.createTextNode(summary)
            summaryElement.appendChild(textNode)
        self.setAttribute(summaryElement, Attributes.data_value, summary) # for rendering colors

    def setContainerProjects(self, container_projects):
        containerProjectElement = self.findOrCreateElement(self.element, Tags.container_projects)
        for container, project_ids in container_projects.items():
            for project_id in project_ids:
                containerProjectSubElement = self.document.createElement(Tags.container_project)
                containerProjectElement.appendChild(containerProjectSubElement)
                containerProjectSubElement.setAttribute(Attributes.name, container)
                containerProjectSubElement.setAttribute(Attributes.id, project_id)

    def setCertifiedContainers(self, certified_containers):
        certifiedContainerElement = self.findOrCreateElement(self.element, Tags.certified_containers)
        for container, (image_id, image_metadata) in certified_containers.items():
            certifiedContainerSubElement = self.document.createElement(Tags.certified_container)
            certifiedContainerElement.appendChild(certifiedContainerSubElement)
            certifiedContainerSubElement.setAttribute(Attributes.name, container)
            certifiedContainerSubElement.setAttribute(Attributes.image_id, image_id)
            certifiedContainerSubElement.setAttribute(Attributes.url, image_metadata['url'])
            imageSHAsElement = self.document.createElement(Tags.image_SHAs)
            certifiedContainerSubElement.appendChild(imageSHAsElement)
            for layer in image_metadata['RootFSLayers']:
                imageSHASubElement = self.document.createElement(Tags.image_SHA)
                imageSHAsElement.appendChild(imageSHASubElement)
                imageSHAValue = self.document.createTextNode(layer)
                imageSHASubElement.appendChild(imageSHAValue)


    def updateSummary(self):
        result = TestResult()
        for output in self.getSubTestOutputDocuments():
            result.combine(output.getSummary())
        self.setSummary(result)

    def setTestParameters(self, testParameters):
        if testParameters.getElement():
            self.partialCopyTree(testParameters.getElement(), self.element, [Tags.parameters, Tags.property])
        else:
            newTestParameters = TestParameters(self.findOrCreateElement(self.element, Tags.parameters), self.document)
            newTestParameters.copy(testParameters)


    def wasRun(self):
        return self.getSummary() != None

    def passed(self):
        if not self.wasRun():
            return False
        return self.getSummary() == "PASS"

    def getSystemLog(self):
        try:
            systemElement = self.getChildElement(Tags.system)
            if systemElement:
                return systemElement.childNodes[0].data
            #otherwise
            return None
        except IndexError:
            return None  

    def setSystemLog(self, log):
        if log:
            systemElement = self.findOrCreateElement(self.element, Tags.system)
            self.setTextNode(systemElement, log)

    def getTestOutput(self):
        return self.__getTestOutput(self.getChildElement(Tags.output))

    def __getTestOutput(self, outputElement):
        output = ""
        if outputElement:
            for node in outputElement.childNodes:
                if node.nodeType == node.ELEMENT_NODE and node.tagName == Tags.output:
                    output += "\nSection %s " % node.getAttribute(Attributes.name)
                    if node.getAttribute(Attributes.description):
                        output += " - " + node.getAttribute(Attributes.description)
                    output += "\n"
                    output += self.__getTestOutput(node)
                    output += "End Section\n"
                elif node.nodeType == node.TEXT_NODE:
                    output += node.data
            output += "\n"
        return output

    def getOutputDocument(self):
        if self.getChildElement(Tags.output):
            return OutputDocument(self.getChildElement(Tags.output), self.document)
        return None

    def getSubTestOutputDocuments(self):
        if self.getOutputDocument():
            return self.getOutputDocument().getSubOutputDocuments()
        return None

    def setTestOutput(self, outputFile):
        errors = list()
        try:
            self.merge(outputFile, nodeIndex=0)
            return True
        except Exception as e:
            errors.append("Warning: could not merge output XML, reading as text file.")
            errors.append("%s" % e)
            for error in errors:
                print(error)

        # parser failed, just read the file
        try:
            file = open(outputFile)
            output = file.readlines()
            output.extend(errors)
            errors = list()
            if output and len(output) > 0:
                outputElement = self.findOrCreateElement(self.element, Tags.output)
                textNode = self.setTextNode(outputElement, "".join(output))
                return True
            else:
                errors.append("Error: log file is empty or missing.")

        except Exception as e:
            errors.append("Error: could not read log file " + outputFile )
            errors.append("%s" % e)

        # otherwise - couldn't read log, or empty log - print them, and copy them to run output
        if errors:
            for error in errors:
                print(error)
            outputElement = self.findOrCreateElement(self.element, Tags.output)
            textNode = self.setTextNode(outputElement, "".join(errors))

        return False

    def getResults(self, environment, resultsFile):
        try:
            file = open(resultsFile, "r")
        except IOError: # no results file?
            self.setSummary(Constants.ABORT)
            return
        resultDocument = parse(resultsFile)
        file.close()
        resultsElement = resultDocument.documentElement
        name = resultsElement.getAttribute(Attributes.name)
        # should we check if name == self.getName():
        # metric =  topElement.getAttribute(Attributes.metric)
        summaryElement = resultsElement.getElementsByTagName(Tags.summary)[0]
        result = summaryElement.childNodes[0].data
        self.setSummary(result)  
        self.getResultsAttachments(environment, resultsFile)

    def getResultsAttachments(self, environment, resultsFile):
        if self.Debugging != Constants.off:
            print("copying attachments...")
        # look for attachments in test log output directories
        testOutputDir = os.path.dirname(resultsFile)
        resultsFilePattern = re.compile(Regex.resultsFileName)
        if self.Debugging != Constants.off:
            print("checking directory %s " % testOutputDir)
        try:
            for file in os.listdir(testOutputDir):

                # skip files with no extension or .o - likely to be executables or object files
                (_, fileExtension) = os.path.splitext(file)
                if not fileExtension or fileExtension == ".o":
                    if self.Debugging != Constants.off:
                        print("Skipping %s (exec or object?)" % file)
                    continue

                filePath = os.path.join(testOutputDir, file)

                if os.path.isdir(filePath): # skip directories
                    continue

                if os.path.getsize(filePath) >  environment.getMaximumAttachmentSize():
                    print("Warning: %s size: %u is too big to attach to the results." % (file, os.path.getsize(filePath)))
                    continue

                # results XML file from FV or peer test?
                if (file == environment.getResultsFileName() or resultsFilePattern.match(file)):
                    print("Merging results XML file %s" % file)
                    self.merge(filePath)
                    continue

                # attach any other file not part of the test
                if file != "output.log":
                    self.addAttachment(filePath)
                    if self.Debugging != Constants.off:
                        print("copied attachment file %s" % file)

        except OSError as e:
            print("Error: could not check test run log directory \"%s\" for test result attachments" % testOutputDir)
            print(e)

    def addAttachment(self, attachment):
            if self.Debugging != Constants.off:
                print("Adding " + attachment)
            mimetype, encoding = mimetypes.guess_type(attachment)
            if not encoding:
                encoding = ""
            (contentsBase64, md5sum) = rhcert_base64_encode_and_md5(attachment)
            attachmentElement = self.document.createElement(Tags.attachment)
            self.element.appendChild(attachmentElement)
            attachmentElement.setAttribute(Attributes.name, os.path.basename(attachment))
            attachmentElement.setAttribute(Attributes.md5sum, md5sum)
            attachmentElement.setAttribute(Attributes.mime_type, mimetype)
            attachmentElement.setAttribute(Attributes.encoding, encoding)
            self.setTextNode(attachmentElement, contentsBase64, filterCharacters=False)

    def getAttachments(self):
        attachments = list()
        for element in self.element.childNodes:
            if element.nodeName == Tags.attachment:
                attachments.append(element)
        return attachments

    def getAttachmentsAttributes(self):
        attachments = list()
        for attachmentElement in self.getAttachments():
            attributes = dict()
            for attribute in [Attributes.name, Attributes.md5sum, Attributes.mime_type,Attributes.encoding]:
                if attachmentElement.getAttribute(attribute):
                    attributes[attribute.replace("-", "_")] = attachmentElement.getAttribute(attribute)
            attachments.append(attributes)
        return attachments

    def writeAttachments(self, destinationDirectory):
        for attachment in self.getAttachments():
            try:
                base64data = attachment.childNodes[0].data
                if base64data:
                    file = open(destinationDirectory + "/" + attachment.getAttribute(Attributes.name), "w")
                    file.write(str(base64data).decode("base64"))
                    file.close()
            except Exception as e:
                print("Error: ")
                print(e)

    def deleteAttachments(self):
        for attachment in self.getAttachments():
            self.runElement.removeChild(attachment)

    def getSubCertification(self):
        subCertificationElement = self.getChildElement(Tags.certification_test)
        if subCertificationElement:
            subCertification = certificationtest.ResultsDocument()
            subCertification.importElement(subCertificationElement)
            return subCertification
        # otherwise
        return None

    def getHostOS(self):
        """ <test><run><os>: the actual host OS and associated information of the test run """
        return HostDocument(self.findOrCreateElement(self.element, Tags.os), self.document)



if __name__ == "__main__":
    
    print("nothing to see here")

