# Copyright (c) 2014 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols <gnichols@redhat.com>
"""
<!-- task queue command format -->
<rhcert-task command="register">
    <parameters>
        <property name="server">hwcert-server.bos.redhat.com</property>
    </parameters>
</rhcert-task>
<rhcert-task command="plan">
    <parameters>
        <property name="server">hwcert-server.bos.redhat.com</property>
        <property name="add" />
        <property name="test">storage</property>
        <property name="device">sda</property>
    </parameters>
</rhcert-task>
"""
from rhcert.model.documentbase import DocumentBase, makeEnumeration
from rhcert.tags import Tags, Attributes


class ParameterDictionary(dict):
    def __getattr__(self, attr):
        return self.get(attr)

class Task(DocumentBase):
    # tasks
    register="register"
    run="run"
    clean="clean"
    prompt="prompt"
    answer="answer"
    message="message"
    plan="plan"
    collab="collab"

    # parameters AKA options
    mode = makeEnumeration("mode", auto="auto", normal="normal", server="server")
    debug = makeEnumeration("debug", off="off", low="low", normal="normal")
    # prompt sub-types
    type = makeEnumeration("type", confirm="confirm", integer="integer", select="select", multi_select="multi-select",password="password", file="file")
    server = "server"
    serverPath = "serverPath"
    test = "test"
    question = "question"
    response = "response"
    choices = "choices"
    default = "default"
    certificationID = "certificationID"
    certificationProgram = "certificationProgram"
    serverIP = "serverIP" # IP address of the server as seen by the client (listener)
    # collab
    testName = "testName"
    testPackageTrail = "testPackageTrail"
    functionName = "functionName"
    args = "args"
    kwargs = "kwargs"

    def __init__(self, task=None):
        DocumentBase.__init__(self)
        self._new(topElement=Tags.rhcert_task, stylesheet="/rhcert/css/task.css")
        self.parameters = dict()
        if task:
            self.setTask(task)
            self.__setDefaults()
        self.valid = True

    def __setDefaults(self):
        defaults = [(Task.mode, Task.mode.auto),
                    (Task.debug, Task.debug.off)]
        for (param, value) in defaults:
            if not self.__getParameterElement(param, create=False):
                self.setParameter(param, value)

    def isValid(self):
        if self.valid and self.element and self.element.nodeName == Tags.rhcert_task:
            return True
        return False

    def __getParameterElement(self, name, create=True):
        if type(name) is type:
            name = name.__name__

        propertyElements = self.getChildElements(Tags.property)
        for propertyElement in propertyElements:
            if name == propertyElement.getAttribute(Attributes.name):
                return propertyElement

        # if not auto-create mode, return not found
        if not create:
            return None

        # otherwise, add the property
        return self.__addParameterElement(name)

    def __addParameterElement(self, name):
        propertyElement = self.document.createElement(Tags.property)
        propertyElement.setAttribute(Attributes.name, name)
        self.element.appendChild(propertyElement)
        return propertyElement

    def setTask(self, task):
        self.setAttribute(self.element, Attributes.task, str(task))

    def getTask(self):
        return self.element.getAttribute(Attributes.task)

    def setParameter(self, name, value):
        if type(value) is list:
            for item in value:
                element = self.__addParameterElement(name)
                self.setTextNode(element, "%s" % item)
        else:
            propertyElement = self.__getParameterElement(name)
            self.setTextNode(propertyElement, "%s" % value)

    def getParameter(self, name):
        propertyElement = self.__getParameterElement(name, create=False)
        if propertyElement and self.getTextNode(propertyElement):
            return self.getTextNode(propertyElement).data
        # otherwise
        return ""

    def getListParameter(self, name):
        values = list()
        propertyElements = self.getChildElements(Tags.property)
        for propertyElement in propertyElements:
            if name == propertyElement.getAttribute(Attributes.name):
                values.append(self.getTextNode(propertyElement).data)
        return values

    def getParametersDictionary(self):
        """ return a optParse-style dictionary """
        parameters = ParameterDictionary()
        propertyElements = self.getChildElements(Tags.property)
        for propertyElement in propertyElements:
            # already have a value, make it a list
            if (propertyElement.getAttribute(Attributes.name) in parameters) and self.getTextNode(propertyElement).data:
                if not type(parameters[propertyElement.getAttribute(Attributes.name)]) is list:
                    valuesList = [ parameters[propertyElement.getAttribute(Attributes.name)] ]
                    valuesList.append(self.getTextNode(propertyElement).data)
                    parameters[propertyElement.getAttribute(Attributes.name)] = valuesList
                else:
                    parameters[propertyElement.getAttribute(Attributes.name)].append(self.getTextNode(propertyElement).data)
                continue

            if self.getTextNode(propertyElement):
                if propertyElement.getAttribute(Attributes.name) == Task.choices:
                    parameters[propertyElement.getAttribute(Attributes.name)] = [self.getTextNode(propertyElement).data]
                else:
                    parameters[propertyElement.getAttribute(Attributes.name)] = self.getTextNode(propertyElement).data
            else:
                parameters[propertyElement.getAttribute(Attributes.name)] = ""
        return parameters

    def loadString(self, xmlString):
        try:
            DocumentBase.loadString(self, xmlString)
        except:
            self.valid = False
        self.__setDefaults()

    def setLog(self, log):
        logElement = self.findOrCreateElement(self.element, Tags.log)
        self.setTextNode(logElement, log)
    def getLog(self):
        logElement = self.findOrCreateElement(self.element, Tags.log)
        if self.getTextNode(logElement):
            return self.getTextNode(logElement).data
        return None
