#!/usr/bin/env python
# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#

from rhcert.tags import Tags, Attributes, Constants, TestTag, TestResult
from rhcert.model.documentbase import DocumentWrapper
from rhcert.model.rundocument import RunDocument
from rhcert.model.testparameters import TestParameters
from rhcert.model.hostdocument import HostDocument

class TestDocument(DocumentWrapper):

    def __init__(self, testElement, document):
        self.element = testElement
        self.document = document

    def __cmp__(self, other):
        if TestTag.interactive in self.getTags() and not TestTag.interactive in other.getTags(): 
            return -1
        if not TestTag.interactive in self.getTags() and  TestTag.interactive in other.getTags():
            return 1
        if self.getPriority() < other.getPriority(): 
            return -1
        if self.getPriority() > other.getPriority(): 
            return 1
        return 0

    def getKey(self):
        return "%s %s" % (self.getPath(), self.getTestParameters())
    def Key(self): return self.getKey()

    def getName(self):
        return self.element.getAttribute(Attributes.name)
    def setName(self, name):
        if name:
            self.element.setAttribute(Attributes.name, name)
    def getMode(self):
        return self.element.getAttribute(Attributes.mode)
    def setMode(self, mode):
        if mode:
            self.element.setAttribute(Attributes.mode, mode)
    def getPath(self):
        return self.element.getAttribute(Attributes.path)
    def setPath(self, path):
        if path:
            self.element.setAttribute(Attributes.name, path)
    def getSourceDirectory(self):
        return self.element.getAttribute(Attributes.source_directory)
    def getPriority(self):
        priority = self.element.getAttribute(Attributes.priority)
        if priority:
            return int(priority)
        return 999

    def setPriority(self, priority):
        if priority:
            self.element.setAttribute(Attributes.priority, priority)

    def getLabel(self):
        return self.element.getAttribute(Attributes.label)
    def setLabel(self, label):
        if label:
            self.element.setAttribute(Attributes.label, label)

    def getServer(self):
        return self.element.getAttribute(Attributes.test_server)
    def setServer(self, server):
        if server:
            self.element.setAttribute(Attributes.test_server, server)

    def isDisabled(self):
        return (Constants.disabled in self.element.getAttribute(Attributes.status))

    def setDisabled(self, disabled):
        if disabled:
            value = Constants.disabled
        else:
            value = ""
        self.element.setAttribute(Attributes.status, value)

    def getStatus(self):
        return self.element.getAttribute(Attributes.status)

    def setStatus(self, value):
        if value != "" and value not in [Constants.disabled, Constants.deleted]:
            print("Error: rhcert.certificationtest.Test status can not be set to %s" % value)
            return
        self.element.setAttribute(Attributes.status, value)

    def markDeleted(self):
        self.setStatus(Constants.deleted)

    def unMarkDeleted(self):
        self.setStatus("")

    def isDeleted(self):
        return (Constants.deleted in self.element.getAttribute(Attributes.status))

    def setSource(self, value):
        self.element.setAttribute(Attributes.source, value)

    def getSource(self):
        return self.element.getAttribute(Attributes.source)

    def setPlanComponentID(self, value):
        self.element.setAttribute(Attributes.component_id, value)

    def getPlanComponentID(self):
        return self.element.getAttribute(Attributes.component_id)

    def isInteractive(self):
        return (TestTag.interactive in self.getTags())

    def getMandatory(self):
        return self.element.getAttribute(Attributes.mandatory)

    def newTestRun(self):
        runElement = self.document.createElement(Tags.run)
        self.element.appendChild(runElement)
        run = RunDocument(runElement, self.document)
        run.setSummary(Constants.INCOMPLETE)
        if self.getMode():
            run.setMode(self.getMode())
        return run

    def getRuns(self, number=-1, latest=False):
        runElements = list()
        for element in self.element.childNodes:
            if element.nodeName == Tags.run:
                runElements.append(element)
        runs = list()
        lastRun = None
        for runElement in runElements:
            run = RunDocument(runElement, self.document)
            if number==-1 or number == run.getNumber():
                if not latest:
                    runs.append(run)
                else: 
                    lastRun = run

        if latest and lastRun:
            runs.append(lastRun)
        return runs

    def deleteRuns(self):
        for element in self.element.childNodes:
            if element.nodeName == Tags.run:
                self.element.removeChild(element)

    def getRun(self, number):
        runs = self.getRuns(number)
        try:
            return runs[0]
        except IndexError: 
            pass
        return None

    def getLastRun(self):
        runs = self.getRuns(latest=True)
        if len(runs):
            return runs[0]
        return None

    def wasRun(self, runNumber=-1):
        if runNumber >=0:
            run = self.getRun(runNumber)
            if run:
                return run.wasRun()
        else:
            for run in self.getRuns():
                if run.wasRun():
                    return True

        #otherwise
        return False

    def hasFailures(self):
        for run in self.getRuns():
            if run.wasRun() and run.getSummary() == TestResult.FAIL:
                return True
        return False

    def hasPassed(self):
        for run in self.getRuns():
            if run.wasRun() and run.getSummary() == TestResult.PASS:
                return True
        return False

    def bestRun(self):
        if not self.wasRun():
            return None
        nextResult = None
        bestRun = None
        for run in self.getRuns():
            if not bestRun:
                bestRun = run
                nextResult = TestResult(run.getSummary())
            if nextResult.better(run.getSummary()):
                bestRun = run
        return bestRun

    def getTestParameters(self):
        return TestParameters(self.findOrCreateElement(self.element, Tags.parameters), self.document)

    def setTestParameters(self, testParameters):
        parametersElement = self.findOrCreateElement(self.element, Tags.parameters)
        testParameters.save(parametersElement, self.document)

    def getHostOS(self):
        return HostDocument(self.findOrCreateElement(self.element, Tags.os), self.document)

if __name__ == "__main__":
    test = TestDocument()
    test.new()
    testFile=open("test.xml","w")
    test.save(testFile)


