#!/usr/bin/env python
# Copyright (c) 2015 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
import sys

from rhcert.tags import Tags, Attributes, Constants
from rhcert.model.documentbase import DocumentBase
from rhcert.model.vendordocument import VendorDocument
from rhcert.model.productdocument import ProductDocument

class VendorProductsDocument(DocumentBase):
    """ VendorProductsDocuments represents a collection of the vendor's products """

    def __init__(self, topElement=None, document=None):
        DocumentBase.__init__(self)
        self.element = topElement
        self.document = document
        self.vendorIndex = None
        self.productIndex = None

    def new(self):
        self._new(topElement=Tags.vendor_products, stylesheet="/rhcert/css/products.css")
        self.setAttribute(self.element, Attributes.scratch_id, 1)

    def index(self):
        if self.vendorIndex and self.productIndex:
            return
        sys.stdout.flush()
        vendors = self.getVendors()
        self.vendorIndex = dict(zip(map(VendorDocument.getID, vendors), vendors))
        self.productIndex = dict()
        for vendor in self.getVendors():
            products = vendor.getProducts()
            self.productIndex[vendor.getID()] = dict(zip(map(ProductDocument.getID, products), products))

    def __unIndex(self):
        self.vendorIndex = None
        self.productIndex = None

    def getVendors(self):
        vendors = list()
        for vendorElement in self.getChildElements(Tags.vendor):
            vendors.append(VendorDocument(vendorElement, self.document))
        return vendors

    def getVendorByName(self, name):
        # substring name match
        for vendor in self.getVendors():
            if name in vendor.getName():
                return vendor
        return None

    def getVendorsByName(self, name):
        vendors = list()
        for vendor in self.getVendors():
            if name == vendor.getName():
                vendors.append(vendor)
        return vendors

    def getVendorByID(self, vendorID):
        vendorID = str(vendorID)
        if self.vendorIndex:
            return self.vendorIndex.get(vendorID)
        for vendor in self.getVendors():
            if vendor.getID() == vendorID:
                return vendor
        return None

    def getVendorProductByID(self, vendorID, productID):
        self.index()
        vendorID = str(vendorID)
        productID = str(productID)
        vendor = self.getVendorByID(vendorID)
        if not vendor:
            return (None, None)
        return (vendor, self.productIndex[vendorID].get(productID))

    def getProductByID(self, productID):
        for vendor in self.getVendors():
            vendor, product = self.getVendorProductByID(vendor.getID(), productID)
            if product:
                return (vendor, product)
        return (None, None)


    def getRecentLocalVendor(self):
        for vendor in reversed(self.getVendors()):
            if vendor.getWorkflow() == Constants.local:
                return vendor
        return None

    def createVendor(self, name, id=0):
        vendorElement = self.document.createElement(Tags.vendor)
        self.element.appendChild(vendorElement)
        self.setAttribute(vendorElement, Attributes.name, name)
        self.setAttribute(vendorElement, Attributes.scratch_id, "1")
        if not id:
            id = self.__getNextScratchID()
        self.setAttribute(vendorElement, Attributes.id, id)
        vendor = VendorDocument(vendorElement, self.document)
        self.__unIndex()
        return vendor

    def __getNextScratchID(self):
        scratchID = self.element.getAttribute(Attributes.scratch_id)
        self.setAttribute(self.element, Attributes.scratch_id, int(scratchID)+1)
        return scratchID

    def copyVendor(self, originalVendorDocument):
        scratchID = None
        storedVendor = None
        vendor = self.getVendorByID(originalVendorDocument.getID())
        if vendor:
            if vendor.getName() == originalVendorDocument.getName():
                storedVendor = vendor
            else: # local ID collision
                scratchID = self.__getNextScratchID()

        # not found, make a copy
        if not storedVendor:
            vendorElementCopy = self.document.importNode(originalVendorDocument.element, True) # True = deep copy
            self.element.appendChild(vendorElementCopy)
            storedVendor = VendorDocument(vendorElementCopy, self.document)
        if scratchID:
            storedVendor.setID(scratchID)

        for product in originalVendorDocument.getProducts():
            storedVendor.copyProduct(product)

        return storedVendor

def unitTest():
    products = VendorProductsDocument()
    products.load("../info/products.xml")
    vendor = products.createVendor("booss")
    vendor.setID("34")
    vendor.createProduct("smoot")
    products.copyVendor(vendor)
    products.save("qq.xml")

if __name__ == "__main__":
    unitTest()
    print("saved qq.xml")

