# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
# Planner: a class to plan HW tests.
#
import os, sys, inspect
from six import with_metaclass

from rhcert.model.testdocument import TestDocument

from rhcert.tags import Constants
from rhcert.test import Test

class PlannerMeta(type):
    def __init__(cls, name, bases, nmspc):
        super(PlannerMeta, cls).__init__(name, bases, nmspc)
        if not hasattr(cls, 'registry'):
            cls.registry = dict()
        source = os.path.dirname(sys.modules[cls.__module__].__file__)
        cls.registry[source] = cls

    def getClasses(cls):
        return cls.registry

class Planner(with_metaclass(PlannerMeta, object)):
    """The Planner class analyses a directory of tests, looking
    for rhcert.Test subclasses.  If found, it adds them to it's
    dictionaries, and provides lists of test(s) by various criteria
    """
    #__metaclass__ = PlannerMeta

    def __init__(self, options):
        self.options = options
        self.tests = list()
        self.planners = list()
        self.testsByPath = dict()
        self.serverTests = list()
        self.plannedTests = list()
        self.testSuites = None
        self.resultsDocument = None

    def setResultsDocument(self, resultsDocument):
        self.resultsDocument = resultsDocument

    def getResultsDocument(self):
        return self.resultsDocument

    def copy(self, fromOriginal):
        self.options = fromOriginal.options
        self.tests = fromOriginal.tests
        self.testsByPath = fromOriginal.testsByPath
        self.serverTests = fromOriginal.serverTests
        self.plannedTests = fromOriginal.plannedTests


    def setSourceDirectory(self, directory):
        self.sourceDirectory = directory

    def getSourceDirectory(self):
        return self.sourceDirectory

    def setTestSuites(self, suites):
        self.testSuites = suites

    def analyse(self, testDirectory):
        """Walk through the test directory looking for test suites"""
        self.testDirectory = testDirectory
        sys.path.append(self.testDirectory)

        if self.options.debug != Constants.off:
            print("Searching %s..." % self.testDirectory)
            print("sys.path:")
            print(sys.path)

        for (root,dirs,files) in os.walk(self.testDirectory):

            # if we're in suites/<suite name>
            if os.path.dirname(root) == self.testDirectory:
                if self.options.debug != Constants.off:
                    print("importing suite %s as module" % os.path.basename(root))
                try:
                    module = __import__(os.path.basename(root), globals(), locals(), [])
                except ImportError as exception:
                    if self.options.debug != Constants.off:
                        print("Warning: Could not import %s " % os.path.basename(root))
                        print(exception)

        self.__instantiateTestsAndPlanners()
        return True

    def __instantiateTestsAndPlanners(self):
        self.tests = list()
        for (testClass, source) in Test.getClasses():
            # only plan classes that can be intantiated with no parameters
            if len(inspect.getargspec(testClass.__init__).args) == 1:
                test = testClass()
                test.setSourceDirectory(source)
                self.tests.append(test)
                if test.getServerMode() in Constants.yes:
                    self.serverTests.append(test)
                self.testsByPath[test.getPath()] = test
                if self.options.debug != Constants.off:
                    print("Instantiated test class %s" % testClass)
            elif self.options.debug != Constants.off:
                print("Skipping class %s - constructor takes %s args" % (testClass, len(inspect.getargspec(testClass.__init__).args)))
        for (source, plannerClass) in Planner.getClasses().items():
            if len(inspect.getargspec(plannerClass.__init__).args) != 2: # can be instantiated with options paramter?
                print("Error: found Planner class %s, but only one parameter is allowed for the constructor." % plannerClass)
                continue
            planner = plannerClass(self.options)
            planner.setSourceDirectory(source)
            self.planners.append(planner)

    def __testIsInSuites(self, test):
        # disable this check for now
        return True
        if not self.testSuites: # none set so match them all:
            return True
        for suite in self.testSuites:
            if test.getSuiteName() == suite.getName():
                return True
        return False

    def getPlanner(self, test):
        """ get the best-matching Planner subclass for the test"""
        match = 0
        for planner in self.planners:
            prefix = os.path.commonprefix([planner.getSourceDirectory(), test.getSourceDirectory()])
            if len(prefix) > match:
                matchingPlanner = planner
                match = len(prefix)
        matchingPlanner.setResultsDocument(self.resultsDocument)
        return matchingPlanner


    def planTests(self):
        tests = list()

        if not self.tests:
            print("Error: No Tests Found !")
        for test in self.tests:
            if not self.__testIsInSuites(test):
                if self.options.debug != Constants.off:
                    print("Skipping test not in current test suites: %s" % test.getPath())
                continue

            if self.options.debug != Constants.off: print("Planning test %s" % test.getPath())
            test.setDebug(self.options.debug)
            test.setMode(self.options.mode)

            planner = self.getPlanner(test)
            planner.allTestsSoFar = tests
            someTests = test.plan(self.getPlanner(test))
            if someTests:
                if self.options.debug != Constants.off:
                    print("planned:")
                    for test in someTests:
                        print("%s %s" % (test.getPath(), test.getTestParameters()))
                tests.extend(someTests)

        # sort by TestMetaData comparison: interactive 1st, then by priority
        tests.sort()
        if self.options.debug != Constants.off:
            print("After Sorting:")
            for test in tests:
                print("%s %s" % (test.getPath(), test.getTestParameters()))
            print("---")
        return tests

    def plan(self, resultsDocument):
        self.resultsDocument = resultsDocument
        self.plannedTests = self.planTests()
        return self.plannedTests

    def saveNewPlan(self, resultsDocument):
        for test in self.plannedTests:
            if self.options.debug != Constants.off:
                print("Planned %s " % test.getPath())
            resultsDocument.appendTest(test)

    def getServerModeTests(self):
        return self.serverTests

    def getTest(self, path, byName=False):
        if path in self.testsByPath:
            return self.testsByPath[path]
        if byName:
            for test in self.tests:
                if test.getName() == path:
                    return test
        return None

    def getPlannedTests(self):
        return self.plannedTests

    def checkPlan(self, certification):
        # compare current test plan to previous
        newTests = dict(zip(map(Test.Key, self.plannedTests), self.plannedTests))
        priorTestDocuments = certification.getTests()
        missingTestDocuments = dict(zip(map(TestDocument.getKey, priorTestDocuments), priorTestDocuments))
        restorableTestDocuments = dict()
        if self.options.debug == Constants.high:
            print("new:")
            for k in newTests:
                print(k)
            print("prior:")
            for k in missingTestDocuments:
                print(k)

        for test in self.plannedTests:
            key = test.Key()
            if self.options.debug == Constants.high:
                print(("checking new test %s" % key))
            if key in newTests and key in missingTestDocuments:
                if self.options.debug == Constants.high:
                    print(("Found %s in both new tests and old plan" % key))
                if missingTestDocuments[key].isDeleted():
                    restorableTestDocuments[key] = missingTestDocuments[key]
                else:
                    if self.options.debug == Constants.high:
                        print(("Removing %s from new test list" % key))
                    del newTests[test.Key()]
                del missingTestDocuments[test.Key()]

        # remove deleted or manually-added tests from the missing list
        deletedKeys = list()
        for key in missingTestDocuments:
            if missingTestDocuments[key].isDeleted() or missingTestDocuments[key].getSource() == Constants.manual:
                deletedKeys.append(key)
        for key in deletedKeys:
            del missingTestDocuments[key]

        return (newTests, missingTestDocuments, restorableTestDocuments)

    def getTests(self):
        return self.tests
