# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
import os, re

from rhcert.services.productnameservice import ProductNameService
from rhcert.tags import Constants, Programs
from rhcert.command import Command, HwCertCommandException


class Release:
    def __init__(self, file):

        self.product = None
        self.productShortName = None
        self.version = None
        self.update = None
        self.codeName = None
        self.candidate = None
        self.valid = False
        self.text = None
        self.debug = False

        self.__read(file)
        self.__parse()

    def __read(self, file):
        try:
            f=open(file)
            self.text=f.readline().strip('\n')
            f.close()
        except:
            pass

    def __parse(self):
        self.product = None
        self.version = None
        self.update = None
        self.codeName = None
        self.candidate = None 
        if self.text:
            # look for <product> release <number>
            pattern = re.compile("^(?P<product>[a-zA-Z\ ]+)release\s+(?P<number>[0-9\.]+)\s*(?P<label>Beta?)*")
            # development preview: "Red Hat Enterprise Linux Server for ARM (Development Preview release 1.5)"
            alternatePattern = re.compile("^(?P<product>[a-zA-Z\ ]+)\([a-zA-Z\ ]+release\s+(?P<number>[0-9\.]+)\)")
            match = pattern.match(self.text)
            if not match:
                match = alternatePattern.match(self.text)
            if not match:
                self.valid = False
                if self.debug:
                    print("\"%s\" is Not Valid" % self.text)
                return
            # otherwise
            self.valid = True
            if match.group("product"):
                self.product = match.group("product").strip()

            if match.group("number"):
                majorMinor = match.group("number").split('.')
                if len(majorMinor)>0:
                    self.version = int(majorMinor[0])
                    if len(majorMinor)>1:
                        self.update = int(majorMinor[1])
                    elif self.codeName:
                        nameParts =  self.codeName.split(' ')
                        if len(nameParts) == 3:
                            self.update = int(nameParts[2])
            try:
                self.label = match.group("label").strip()
            except (IndexError,  AttributeError):
                # no label group
                self.label = ""

            # try for other info after the release number
            text = self.text[match.end("number"):].strip()
            pattern = re.compile("(?P<candidate>[^\(\)]*)(\((?P<name>[a-zA-Z0-9\ ]+)\))?")
            match = pattern.match(text)
            if match:
                if match.group("name"):
                    self.codeName = match.group("name").strip()
                if match.group("candidate"):
                    self.candidate = match.group("candidate").strip()

    def isValid(self):
        return self.valid and self.product and self.version

    def getProduct(self): return self.product

    def getProductShortName(self): return self.productShortName

    def getVersion(self): return self.version

    def getUpdate(self):
        if self.update:
            return self.update
        return "0"

    def getLabel(self): return self.label

    def getCodeName(self): return self.codeName

    def getCandidate(self): return self.candidate

    def getVersionPointUpdate(self):
        if self.update != None:
            return "%u.%u" % (self.version, self.update)
        # otherwise no minor AKA update number as in Fedora
        return "%u" % self.version

    def dump(self):
        print("")
        print(self.text)
        print("-------------------------------")
        print("Product: \"" + self.getProduct() + "\"")
        print("Version: \"" + str(self.getVersion()) + "\"")
        print("Update: \"" + str(self.getUpdate()) + "\"")
        print("Version.Update \"" + self.getVersionPointUpdate() + "\"")
        if self.candidate:
            print("Candidate: \"" + self.getCandidate() + "\"")
        if self.codeName:
            print("Code Name: \"" + self.getCodeName() + "\"")
        if not self.isValid():
            print("Not Valid")
    def parse(self):
        self.__parse()


class RedHatRelease(Release):

    def __init__(self):
        Release.__init__(self, "/etc/redhat-release")
        self.kernel = None
        self.arch = None
        self.kernelRPMName = None
        self.kernelCoreRPMName = None
        self.kernelModulesRPMName = None
        self.kernelDevelRPMName = None
        self.kernelFlavor = None
        self.debugKernel = False
        self.realtimeRelease = None
        self.kernelVersionRelease = None
        self.realtimeRelease = Release("/etc/mrg-realtime-release")

        if self.isValid():
            self.__getKernelInfo()

    def updateProductFromService(self):
        """
        It detects the installed product and gathers its info
        from PNS.
        Sets the product and its short name, returns True in positive
        scenario. Returns False in negative scenario.
        """
        pns = ProductNameService()

        # Using separate logic for RHEL and Real Time product detection.
        # Reason: The product names for RHEL and Real Time products returned by
        # 'subscription-manager list' command do not match with the names
        # given in PNS. These names can not be updated in PNS because it may
        # impact CWE.
        rhel_based_prod = {
          "architectures": {
            "x86_64": Constants.RHEL_PROD_ID,
          },
          "flavors": {
            "rt": Constants.RHELRT_PROD_ID
          }
        }

        # Real Time products: rhelmrg or rhelrt
        if rhel_based_prod['flavors'].get(self.kernelFlavor):
            product = pns.getProduct(rhel_based_prod['flavors'][self.kernelFlavor])
        # RHEL
        elif rhel_based_prod['architectures'].get(self.arch):
            product = pns.getProduct(rhel_based_prod['architectures'][self.arch])

        else:
            # Other products
            cmd = "subscription-manager list > installed_products; grep -n 'Product Name:' installed_products"
            installed_products = Command(cmd)
            try:
                installed_products.run()
            except HwCertCommandException as err:
                print("Warning: could not determine product via command: %s" % cmd)
                return False
            else:
                labels = {"High Touch": "", "Beta": ""}  # remove extra labels from product-name
                regexp = re.compile('|'.join(sorted(labels, key=len, reverse=True)))
                installed_products = [(regexp.sub(lambda match: labels[match.group(0)], prod.split(':')[-1])).strip()
                                      for prod in installed_products.output]
                pns_products = pns.getProgram(Programs.hardware)
                for pns_product in pns_products['products']:
                    for installed_product in installed_products:
                        if installed_product == pns_product['name']:
                            product = pns_product
                # If rhcert detected and pns product names do not match
                try:
                    product
                except NameError:
                    print("Warning: could not determine product. "
                          "None of the installed product: %s seems to belong to Red Hat." % str(installed_products))
                    return False

        self.product = product["name"]
        self.productShortName = product["shortName"]
        return True

    def __getKernelInfo(self):
        uname = os.popen('uname -r')
        self.kernel = uname.readline()[:-1]
        uname.close()
        uname = os.popen('uname -m')
        self.arch = uname.readline()[:-1]
        uname.close()
        uname = os.popen("uname -r")
        unameOutput = uname.readline()
        uname.close()
        if not unameOutput:
            print("Error: uname failed")
        else:
            unameOutput = unameOutput.strip()
        self.parseUnameOutput(unameOutput)

    def parseUnameOutput(self, unameOutput):
        self.kernelRPMName = None
        self.kernelCoreRPMName = None
        self.kernelModulesRPMName = None
        self.kernelDevelRPMName = "kernel-devel"
        self.kernelVersionRelease = None
        if "Red Hat" in self.getProduct() and self.getVersion() == 4:
            pattern = re.compile("(?P<versionRelease>[^a-z]*)(?P<name>smp|hugemem|largesmp)?$")
            match = pattern.search(unameOutput)
            if match:
                # if the release ends with ELsmp, ELhugemem, or ELlargesmp
                if match.group("name"):
                    self.kernelRPMName = "kernel-%s-%s" % (match.group("name"), match.group("versionRelease"))
                else:
                    self.kernelRPMName = "kernel-%s" %  match.group("versionRelease")
                self.kernelVersionRelease = match.group("versionRelease")
            if "debug" in unameOutput:
                self.debugKernel = True

        elif "Red Hat" in self.getProduct() and self.getVersion() >= 5:
            pattern = re.compile("(?P<versionRelease>.*?)\.{0,1}"
                                 "(?P<flavorSection>(?P<flavor>xen|PAE|rt).*){0,1}\."
                                 "(?P<product>(aa|ael|el)[0-9]+[_a-zA-Z0-9]{0,3})\.{0,1}(?P<arch>.*)$")

            match = pattern.search(unameOutput)
            if match:
                self.kernelDevelRPMName = "kernel-devel"
                if match.group("flavor"):
                    self.kernelFlavor = match.group("flavor")
                    suffix = "%s.%s.%s" % (match.group("versionRelease"),
                                           match.group("flavorSection"),
                                           match.group("product"))
                    self.kernelRPMName = "kernel-%s-%s" % (self.kernelFlavor, suffix)
                    self.kernelCoreRPMName = "kernel-%s-core-%s" % (self.kernelFlavor, suffix)
                    self.kernelModulesRPMName = "kernel-%s-modules-%s" % (self.kernelFlavor, suffix)
                    self.kernelDevelRPMName = "kernel-%s-devel" % self.kernelFlavor
                else:
                    suffix = "%s.%s" %(match.group("versionRelease"),
                                       match.group("product"))
                    self.kernelRPMName = "kernel-%s" % suffix
                    self.kernelCoreRPMName = "kernel-core-%s" % suffix
                    self.kernelModulesRPMName = "kernel-modules-%s" % suffix
                # RT kernel tags rt on the end of el5 through el6
                if match.group("flavor") == "rt" and self.getVersion() < 7:
                    self.kernelRPMName += "rt"
                self.kernelVersionRelease = match.group("versionRelease")
                if match.group("arch"):
                    if "debug" in match.group("arch"):
                        self.debugKernel = True

        elif 'Fedora' in self.getProduct():
            pattern = re.compile("(?P<versionRelease>.*)\.(?P<product>fc[0-9]+)(?P<arch>.*)$")
            match = pattern.search(unameOutput)
            if match:
                suffix = "%s.%s%s" % (match.group("versionRelease"),
                                      match.group("product"),
                                      match.group("arch"))
                self.kernelRPMName = "kernel-%s" % suffix
                self.kernelCoreRPMName = "kernel-core-%s" % suffix
                self.kernelModulesRPMName = "kernel-modules-%s" % suffix
                self.kernelVersionRelease = match.group("versionRelease")
                if "debug" in match.group("arch"):
                    self.debugKernel = True
        if not self.kernelRPMName:
            print("Error: could not determine kernel RPM name")
            self.kernelRPMName = None
            return False

        return True

    def getProductFromUname(self, unameOutput):
        pattern = re.compile("(?P<product>.ael|.el|.fc|.aa)(?P<version>[0-9]+[a-z]{0,1})")
        self.product = None
        self.version = None
        match = pattern.search(unameOutput)
        if match:
            if match.group("product") == ".fc":
                self.product = "Fedora"
            elif match.group("product").startswith(".aa") or unameOutput.endswith("arch64"):
                self.product = "Red Hat Server for ARM"
            elif match.group("product") == ".el":
                self.product = "Red Hat Enterprise Linux"
            elif match.group("product").startswith(".ael"):
                self.product = "Red Hat Asterisk Extension Logic"
            if match.group("version"):
                self.version = int(match.group("version")[0])
            return True
        else:
            return False

    def getKernelRPMName(self): return self.kernelRPMName
    def getKernelCoreRPMName(self): return self.kernelCoreRPMName
    def getKernelModulesRPMName(self): return self.kernelModulesRPMName
    def getKernelDevelRPMName(self): return self.kernelDevelRPMName
    def getKernel(self): return self.kernel
    def getKernelVersionRelease(self): return self.kernelVersionRelease
    def getArch(self): return self.arch
    def getKernelFlavor(self): return self.kernelFlavor
    def kernelIsDebug(self): return self.debugKernel
    def getRealTimeRelease(self):
        if self.realtimeRelease and self.realtimeRelease.isValid():
            return self.realtimeRelease
        # otherwise
        return None

    def dump(self):
        Release.dump(self)
        print("Kernel Version \"" + self.getKernelVersionRelease() + "\"")
        if self.getRealTimeRelease():
            print("Realtime:")
            self.getRealTimeRelease().dump()


def test_kernel_product_name(redHatRelease):
    """
    Unit test for updateProductFromService method
    """
    result = 0
    for flavor, architecture, prod in [(None, 'x86', 'RHEL'),
                                       ('rt', 'x86', 'RHELMRG'),
                                       (None, 'aarch64', 'RHELSA'),
                                       (None, 'ppc64le','RHELforPower'),
                                       # negative test
                                       (None, 'ppc64le','RHEL')]:
        try:
            print("\nExpected product is %s" % prod)
            redHatRelease.updateProductFromService()
            assert(redHatRelease.productShortName == prod)
        except AssertionError:
            print("Error: expected and actual product mismatched.")
            result = 1
    return result


def unit_test():
    PASSED = 0
    FAILED = 1
    result = PASSED
    redHatRelease = RedHatRelease()

    if not redHatRelease.isValid():
        print("NOT VALID")
        result = FAILED
    else:
        redHatRelease.dump()

    print("\nTest to determine product name and its short name:")
    result = test_kernel_product_name(redHatRelease)

    print("\nTesting some other /etc/redhatrelease strings")
    texts = ("Red Hat Enterprise Linux Server release 5.4 (Tikanga)",
             "Red Hat Enterprise Linux release 4 (Bogus Update 3)",
             "Red Hat Enterprise Linux Server release 5.5 Beta (Tikanga)",
             "Red Hat Enterprise Linux release 6.0 Beta (Santiago)",
             "Red Hat Enterprise Linux Server release 7.0 Alpha1 (Maipo)",
             "Red Hat MRG Realtime Linux release 2.3",
             "Red Hat Enterprise Linux Server for ARM (Development Preview release 1.5)"
             )
    redHatRelease.debug = True
    for text in texts:
        print("\n")
        redHatRelease.text = text
        redHatRelease.parse()

        if not redHatRelease.isValid():
            print("Error: text parsing failed")
            result = FAILED
        else:
            redHatRelease.dump()

    print("\nTest uname parsing on this system, and  number of other kernels")
    unameOutputs = (os.uname()[2],
         "3.10.0-327.rt56.183.el6rt.x86_64",
         "2.6.33.2-rt13.13.el5rt",
         "2.6.24.7-155.el5rt",
         "2.6.18-164.15.1.el5",
         "2.6.18-164.15.1.el5xen",
         "2.6.18-164.15.1.el5PAE",
         "2.6.32-17.el6.x86_64",
         "2.6.26.8-57.fc8",
         "2.6.27.25-78.2.56.fc9.i686",
         "2.6.30.10-105.2.23.fc11.x86_64",
         "2.6.32.11-99.fc12.x86_64",
         "2.6.34-0.38.rc5.git0.fc14.x86_64",
         "2.6.31.12-1.rt21.1.fc12.ccrma.x86_64",
         "2.6.32-220.el6.i686.debug",
         "3.10.0-229.rt56.141.el7.x86_64",
         "3.10.0-229.el7.x86_64",
         "3.10.0-229.ael7a.ppc64le",
         "3.19.0-0.rc5.66.aa7a.aarch64",
         "4.2.0-0.19.el7.aarch64")

    for unameOutput in unameOutputs:
        redHatRelease = RedHatRelease()
        print("\nuname: " + unameOutput)
        redHatRelease.getProductFromUname(unameOutput)
        print((redHatRelease.getProduct() + " %s") % redHatRelease.getVersion())
        if redHatRelease.parseUnameOutput(unameOutput):
            print("kernel RPM: " + redHatRelease.getKernelRPMName())
            print("kernel devel RPM: " + redHatRelease.getKernelDevelRPMName())
            if redHatRelease.getKernelFlavor():
                print("kernel flavor: " + redHatRelease.getKernelFlavor())
            if redHatRelease.kernelIsDebug():
                print("Warning: debug kernel")

        else:
            result = FAILED

    if result == FAILED:
        print("FAILED")
    else:
        print("PASSED")
    return result


if __name__ == "__main__":
    exit(unit_test())

