# Copyright (c) 2018 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Pooja Tiwari
#
# Overview:
#    The RemoteCommand class is a wrapper for shell commands that needs to be
#    run on remote system. Example usage can be seen in the
#    __main__ self test function at the end of this file.

import os
from rhcert.command import Command
try:
    import paramiko
except ImportError as err:
    if Command.isDebug():
        print("Warning: %s." % str(err))

from rhcert.model.commandxml import CommandXML
import time
from rhcert.environment import Environment

DEFAULT_PRIVATE_KEY = '/root/.ssh/id_rsa'

class RemoteCommand:

    def __init__(self, command, remote_ip, user, private_key=DEFAULT_PRIVATE_KEY, timeout=20):
        self.command = command
        self.remote_ip = remote_ip
        self.user = user
        self.private_key = private_key
        environment = Environment()
        self.timeout = timeout if timeout > environment.getSSHRetryTimelimit() else environment.getSSHRetryTimelimit()

    def make_connection(self):
        """
        Build SSH connection to a remote host.
        """
        if not os.path.isfile(self.private_key):
            print("Error: The key file (%s), required for remote host "
                "connection, does not exist on the system" % self.private_key)
            print("Corrective Action: Please create the missing file and re-execute the test.")
            return False
        self.client = paramiko.SSHClient()
        # Using private key for authentication
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        private_key_file = os.path.expanduser(self.private_key)
        private_key = paramiko.RSAKey.from_private_key_file(private_key_file)
        retry_interval = 5
        timeout = self.timeout
        timeout_start = time.time()
        while time.time() < timeout_start + timeout:
            time.sleep(retry_interval)
            try:
                # Make SSH connection
                self.client.connect(self.remote_ip, username = self.user, pkey = private_key, timeout=retry_interval)
                return True
            except paramiko.AuthenticationException as authException:
                print(("Error: Authentication failed, please verify your private key: %s" % authException))
                continue
            except paramiko.SSHException as sshException:
                print(("Error: Unable to establish SSH connection: %s" % sshException))
                continue
            except paramiko.BadHostKeyException as badHostKeyException:
                print(("Error: Unable to verify server's host key: %s" % badHostKeyException))
            except Exception as e:
                print("Error: Unable to build connection with %s due to %s issue." % (self.remote_ip, str(e)))
                continue
        print('Could not establish SSH connection after multiple retries')
        return False

    def run(self):
        """
        Gather command output from remote host.
        """
        output = str()
        if not self.make_connection():
            return output
        cmd  = 'sudo sh -c \'' + self.command + '\''
        try:
            stdin, stdout, stderr = self.client.exec_command(cmd)
            output = stdout.read()

        except Exception as e:
            print("Error: error in command execution '%s' on remote system '%s' due to issue: %s" % (cmd, self.remote_ip, str(e)))
        finally:
            # close SSH connection
            self.client.close()

        return output

def unitTest():
    try:
        # Please change below params based on your environment
        cmd = "uname -a"
        ip = "192.168.24.11"
        user = "heat-admin"
        # With default private key
        print(" *** Running with default private key: %s *** " % DEFAULT_PRIVATE_KEY)
        print(
            "Executing '{0}' on a remote host with ip: '{1}' and user:{2}.".format(
                cmd,
                ip,
                user))
        remoteCommand = RemoteCommand(cmd, ip, user)
        output = remoteCommand.run()
        if not output:
            print("Error: No output")
        else:
            print("Output: \n" + str(output))

        # With a different private key (change it based on your environment)
        pkey = "/home/stack/.ssh/id_rsa"
        print("\n *** Running with a different private key: %s *** " % pkey)
        print(
            "Executing '{0}' on a remote host with ip: '{1}' and user:{2}.".format(
                cmd,
                ip,
                user,
                pkey))
        remoteCommand = RemoteCommand(cmd, ip, user, pkey)
        output = remoteCommand.run()
        if not output:
            print("Error: No output")
        else:
            print("Output: \n" + str(output))

    except Exception as e:
        pass

if __name__ == "__main__":
    unitTest()
