# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
# The Report object packages and/or presents test results in text, rpm, or html form.
#

from rhcert.tags import Constants, Tags
from rhcert.controller import Controller
from rhcert.environment import Environment

class Report(Controller):

    def __init__(self, options, certification):
        self.certification = certification
        self.options = options
        self.Debugging = options.debug
        self.environment = Environment()
        self.horizontalRule = "------------------------------------------------------------------------------"

    def Write(self):
        if not self.certification:
            print("No test results or plan to print")
            return False
        #otherwise
        if self.certification.getNumberOfTestRuns() > 0:
            # no options - print the whole thing:
            # BTW - what about using XSLT to map the test into a document?
            print("\nRed Hat Hardware Certification test")
            print("--------------------------------------------")
            print("Test Suite:    %s    Release: %s" % (self.certification.getVersion(), self.certification.getRelease()))
            print("Plan Created:  %s" % (self.certification.getPlanTime()))
            print("Test Server:   %s" % self.certification.getTestServer())
            print("--------------------------------------------")
            runNumber = 0
            for n in range(self.certification.getNumberOfTestRuns()):
                runNumber = n+1
                counts = self.certification.getTestCounts(runNumber)
                print("\nRun: %u on %s" % (runNumber, counts['start']))
                print("--------------------------------------------")         
                print("Tests: %u planned,  %u run, %u passed, %u failed" % ( counts['count'], counts['run'], counts['passed'], counts['failed']))
                print("--------------------------------------------\n")
                tests = self.certification.getTests() # all of them
                self.WriteTestRun(tests, runNumber)

        else:
            self.WriteTestPlan()

        return True

    def WriteTestRun(self, tests, runNumber):

        print("\nTest Run %s" % runNumber)
        print("-----------------------------------------------------------------------------")
        for test in tests:
            run = test.getRun(runNumber)
            if run:
                summary = "- %s" % run.getSummary()
            elif test.isDisabled():
                summary = "- DISABLED"
            else:
                summary = "- "
            device =  test.getTestParameters().get(Constants.device)
            if device is None: device = ""
            udi = test.getTestParameters().get(Constants.udi)
            if udi is None: udi = "" 


            print("%-11s %-7s %-50s %s" % (test.getName(), device, udi, summary))
            continue

            print("Test: %s: %s" % (test.getLabel(), test.getName()))
            print("Tested UDI: %s" % test.getUDI())
            if run:
                print("Test Start: %s   Test End %s" % (run.getRunTime(), run.getEndTime()))
                print("Return Value: %s" % run.getReturnValue())
                print("Result: %s" % run.getSummary())
                if self.options.report == Constants.full:
                    print("System Log:")
                    print(run.getSystemLog())
                    print("Test Output:")
                    print(run.getTestOutput())
                    # print subcertifications for fv_ tests
                    subCertification = run.getSubCertification()
                    if subCertification:
                        self.WriteSubCertification(subCertification)

            elif test.isDisabled():
                print("DISABLED")
            else:
                print("not run")

    def WriteSubCertification(self, subCertification):
        #otherwise
            print("--------------------------------------------")
            print("Test Suite:    %s    Release: %s" % (subCertification.getVersion(), subCertification.getRelease()))
            print("OS:            %s %s (%s)" % (subCertification.getOS(Tags.product), subCertification.getOS(Tags.release), subCertification.getOS(Tags.name)))
            print("--------------------------------------------")
            runNumber = 0
            for n in range(subCertification.getNumberOfTestRuns()):
                runNumber = n+1
                tests = subCertification.getTests() # all of them
                self.WriteTestRun(tests, runNumber)

    def WriteTestPlan(self):
        print("\nTest Plan:")
        print(self.horizontalRule)
        tests = self.certification.getTests()
        for test in tests:        
            if not test.isDeleted():
                print(self.formatTest(test))

    def formatTest(self, test):
        disabled = ""
        if test.isDisabled():
            disabled = " - DISABLED"
        device = test.getTestParameters().get(Constants.device)
        if not device: device = ""
        udi = test.getTestParameters().get(Constants.udi)
        if not udi: 
            udi = ""
        elif len(udi) > 36:
            udi = "..." + udi[len(udi)-33:]
        return "%-30s %-10s %-36s %s" % (test.getPath(),  device,  udi, disabled)




