# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
# The Report object packages and/or presents test results in text, rpm, or html form.
#

import string, sys

from rhcert.tags import Constants, Tags
from rhcert.controller import Controller

class ReportHTML(Controller):

    def __init__(self, options, certification):
        self.certification = certification
        self.options = options
        self.Debugging = options.debug

    def Write(self):
        self.__WriteHeader()
        self.__WriteOverview()
        if self.certification:
            if self.options.report == Constants.device:
                self.__WriteDevice()
            else:
                if self.options.test:
                    self.__WriteTest()
                else:
                    self.__WriteSummary()
        else:
            print("\nThere are no test results or plan<br>\n")
        self.__WriteFooter()
        return True

    def __WriteHeader(self):
        title = "rhcert"
        if self.certification:
            title = "rhcert %s %s" % (self.certification.getHardware(Tags.vendor), self.certification.getHardware(Tags.model))

        header = """
        <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
        <html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">

        <head>
          <title>%s</title>
          <link rel="shortcut icon" href="/rhcert/images/favicon.ico">
          <link rel="stylesheet" href="/rhcert/css/rhcert.css">

        </head>

        <body id="mainone">

        <div id="wrap">
        <div id="innerwrap">

        <!-- HEADER -->
        <div id="header">
        <a href="/rhcert"><img src="/rhcert/images/logo_rh_home.png" alt="Red Hat Home" width="96" height="31" id="rhLogo" /></a>
        </div><!-- end header -->

        <!-- page body -->

        <!-- overview -->
        <h1>Hardware Certification Test Suite</h1>
        """ % title
        print(header)

    def __WriteOverview(self):
        beginTable =  """
        <table class="overview">
            <thead></thead>
        """
        print(beginTable)
        print("<tr><td>Vendor:</td><td>%s</td></tr>" % self.certification.getHardware(Tags.vendor))
        print("<tr><td>Model:</td><td>%s</td></tr>" % self.certification.getHardware(Tags.model))
        print("<tr><td>Arch:</td><td>%s</td></tr>" %  self.certification.getHardware(Tags.arch))
        print("<tr><td>OS:</td><td>%s %s</td></tr>" %  (self.certification.getOS(Tags.name), self.certification.getOS(Tags.release)))
        print("<tr><td>rhcert:</td><td>%s Release %s</td></tr>" % (self.certification.getVersion(), self.certification.getRelease()))
        print("<tr><td>Date:</td><td>%s</td></tr>" % self.certification.getPlanTime())    
        print("</table>")

        self.__WriteNavigation()

    def __WriteNavigation(self):
        nav = """
            <table class="navigation">
            <tr><td>See Also:</td>
                <td><a href=\"hwCertWeb.py\">test summary</a></td>
                <td><a href=\"hwCertWeb.py?report=device\">device details</a></td>
            </tr>
            </table>
        """
        print(nav)

    def __WriteSummary(self):
        header =  """
        <!-- results table -->
        <h2>Test Results:</h2>
        <table class="results" cellpadding="0" cellspacing="0" border="0">
            <thead>
                <tr><th>Run</th>
        """
        print(header)
        tests = self.certification.getTests()
        name = ""
        devices = list()
        colspan = 1
        for test in tests:    
            if name == "":
                name = test.getName()
            elif test.getName() != name:    
                colspanString = ""
                if colspan > 1:    
                    colspanString = "colspan=%u" % colspan     
                print("<th %s>%s</th>" % (colspanString, name))
                name = test.getName()
                colspan=1
            else:
                colspan = colspan + 1

            if test.getTestParameters().get(Constants.device) and test.getTestParameters().get(Constants.device) != "":
                devices.append(test.getTestParameters().get(Constants.device))
            else:
                devices.append("&nbsp;")

        #write the last test
        if name != "":
            coslpanString = ""
            if colspan > 1:    
                colspanString = "colspan=%u" % colspan     
            print("<th %s>%s</th>" % (colspanString, name))
        print("</tr><tr><th>&nbsp;</th>")

        # print the devices
        for device in devices:
            print("<th>%s</th>" % device)
        print("</tr></thead>")

        # print the results
        for n in range(self.certification.getNumberOfTestRuns()):
            runNumber = n+1
            print("<tr><td>%u</td>" % runNumber)
            for test in tests:    
                run = test.getRun(runNumber)
                if run and run.getSummary():
                    result = run.getSummary()
                    print("<td class=\"%s\"><a href=\"hwCertXml.py?test=%s&udi=%s&device=%s&run=%u\">%s</a></td>" 
                            % (result.lower(), test.getName(), test.getTestParameters().get(Constants.udi), test.getTestParameters().get(Constants.device), runNumber, result.lower()))
                else:
                    print("<td>&nbsp;</td>")
            print("</tr>")
        # add run links
        print("<tr><td>&nbsp;</td>")
        for test in tests:
            print("<td><a href=\"hwCertWeb.py?command=run&test=%s&udi=%s&device=%s\">run</a></td>" 
                    % ( test.getName(), test.getTestParameters().get(Constants.udi), test.getTestParameters().get(Constants.device)))
        print("</tr>")
        print("</table>")

    def __WriteDevice(self): 
        if self.options.device:
            print("NOT IMPLEMENTED")
        elif self.options.udi:
            for deviceClassName, deviceClass in self.certification.getDeviceClasses().items():
                for device in deviceClass.getDevices():
                    if self.options.udi in device.getUDI():
                        self.__WriteSingleDevice(device)
                        return True
            # otherwise 
            print("Device Not Found")
            return False
        else:
            self.__WriteDeviceClassTable()

    def __WriteDeviceClassTable(self):
        header = """
        <!-- device table -->
        <h2>Devices:</h2>
        <table class="devices" cellpadding="0" cellspacing="0" border="0">
        """
        print(header)
        try:
            for deviceClassName, deviceClass in self.certification.getDeviceClasses().items():
                print("<tr><td colspan=\"3\" style=\"font-weight: 700;\">%s</td></tr>" % deviceClassName)
                for device in deviceClass.getDevices():
                    print("<tr><td>&nbsp;</td><td><a href=\"hwCertWeb.py?report=device&udi=%s\">%s [%s]</a></td><td>%s</td></tr>" 
                            % (self.getShortUDI(device.getUDI()), device.getProduct(), device.getSource(), self.getShortUDI(device.getUDI())))
        except Exception as e:
            print("Error: ")
            print(e)
        print("</table>")

    def __WriteSingleDevice(self, device):
        deviceHeader = """
        <h2>Device:</h2>
        <table class="overview">
            <thead></thead>
        """
        print(deviceHeader)
        try:
            print("<tr><td>Product:</td><td colspan=\"2\">%s</td></tr>" % device.getProduct())
            print("<tr><td>UDI:</td><td colspan=\"2\">%s</td></tr>" % device.getUDI())
            productHeader = """
                </table>
                <hr size=\"1\">
                <table class="devices">
            """
            print(productHeader)
            attributes = device.getProperties().keys()
            attributes.sort()
            for attribute in attributes:
                value = device.getProperties()[attribute]
                if value:
                    try:
                        valueString = value.str()
                    except:
                        valueString = "%s" % value
                else:
                    valueString = "(None)"

                print("<tr><td colspan=\"2\">%s</td><td>%s</td></tr>" % (attribute, valueString))
        except Exception as e:
            print("Error: ")
            print(e)
        print("</table>")


    def __WriteFooter(self):
        footer = """
        </div></div>
        </body>
        </html>
        """
        print(footer)
    
    def __WriteTest(self):
      try:
        print("<a href=\"hwCertWeb.py\">Back to Summary</a><br/>")
        testNames = list()
        testNames.append(self.options.test)
        tests = self.certification.getTests(names=testNames)
        for test in tests:
            if test.getTestParameters().get(Constants.udi) == self.options.udi or test.getTestParameters().get(Constants.device) == self.options.device:
                break

        if test:
            print("<h2>%s</h2>" % test.getName())
            if test.getTestParameters().get(Constants.device):
                print("<strong>device:</strong>&nbsp;%s<br/>" % test.getTestParameters().get(Constants.device))
            if test.getTestParameters().get(Constants.udi) and self.getShortUDI(test.getTestParameters().get(Constants.udi)) != "":
                print("<strong>udi:</strong>&nbsp;%s<br/>" % self.getShortUDI(test.getTestParameters().get(Constants.udi)))
            if test.getDeviceClassName():
                print("<strong>class:</strong>&nbsp;%s<br/>" % test.getDeviceClassName())
            sys.stdout.flush()

            run = test.getRun(int(self.options.run))
            if run:
                print("<strong>run:</strong>&nbsp;%s<br/>" % self.options.run)
                print("<strong>start:</strong>&nbsp;%s<br/>" % run.getRunTime())
                print("<strong>end:</strong>&nbsp;&nbsp;&nbsp;%s<br/>" % run.getEndTime())
                if len(run.getRunTime()) > 0 and len(run.getEndTime()):
                    print("<strong>duration:</strong>&nbsp;%s<br/>" % run.duration(run.getRunTime(), run.getEndTime()))
                print("<strong>mode:</strong>&nbsp;%s<br/>" % run.getMode().lower())
                if run.getSummary().lower() == "fail":
                    print("<strong>result:</strong>&nbsp;<font class=\"error\">fail</font><br/>")
                else:
                    print("<strong>result:</strong>&nbsp;%s<br/>" % run.getSummary().lower())

                print("<h3>Log:</h3>\n")
                for line in run.getTestOutput().splitlines():
                    style = "normal"
                    if "Error: " in line:
                        style = "error"
                    elif "Warning" in line:
                        style = "warning"
                    print("<font class=\"%s\">%s</font><br>\n" % (style, line))

                logTextAreaFormat = "<h3>%s:</h3>\n<textarea id=\"%s\" rows=\"10\" cols=\"110\" readonly >\n%s\n</textarea>"
                print(logTextAreaFormat % ("System", "system",  run.getSystemLog()))
                print("<h3>%s:</h3>\n" % "Devices")
                print("<table class=\"devices\" cellpadding=\"0\" cellspacing=\"0\" border=\"0\">")
                for device in run.getDevices():
                    print("<tr><td>&nbsp;</td><td><a href=\"hwCertWeb.py?report=device&udi=%s\">%s</a></td><td>%s</td></tr>" 
                            % (self.getShortUDI(device.getUDI()), device.getProduct(), self.getShortUDI(device.getUDI())))
                print("</table>")
      except Exception as e:
          print("Error: ")
          print(e)


    def getShortUDI(self,  udi):       
        if not udi: 
            return ""
        # chop everything before and including /devices/
        prefix = "/devices/"
        if udi.index(prefix):
            udi = udi[udi.index(prefix)+len(prefix):]   
        # take the last 5 nodes
        udi = "/".join(udi.split("/")[-5:])
        return udi
