# Copyright (c) 2014 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols <gnichols@redhat.com>
#
import errno
import os
import pwd
import re
import shutil
import socket
import time

from rhcert.command import Command
from rhcert.environment import Environment
from rhcert.model.certificationdocument import CertificationDocument
from rhcert.model.imagePair import ImagePair
from rhcert.model.documentbase import makeEnumeration
from rhcert.model.resultsdocument import ResultsDocument
from rhcert.model.vendorproducts import VendorProductsDocument
from rhcert.model.serverSettings import ServerSettingsDocument
from rhcert.tags import Attributes, Programs, Constants, Tags, Regex

FileType = makeEnumeration("FileType",
    results = "results",
    certification = "certification",
    multihost = "multihost",
    product = "product"
)

class RedHatCertificationStore(object):
    """ RedHatCertificationStore provides certification file management and utilities for /var/www/rhcert/store

        filesystem:

        store
        |-- results
        |   |--<cert type>-<cert id>
        |       |-- redhat-certification-<type>-<id>.xml (certification documents)
        |       |-- rhcert-results-<host>-<timestamp>.xml (results for a certification)
        |       |-- rhcert-multi-<type>-<id>-<timestamp>.xml (multihost results for a certification)
        |-- systems
        |   |-- rhcert-results-<host>-<timestamp>.xml.gz (system registrations)
        |   |-- products.xml
        |-- log
    """

    def __init__(self, database=None):
        self.environment = Environment()
        self.database = database

        """ file name parsing """
        self.resultsFileRegex = re.compile(Regex.resultsFileName)
        self.containerDirectoryRegex = re.compile("^rhcert-container-(?P<id>[0-9]+)-(?P<hostname>.+)-(?P<timestamp>[0-9]+)$")
        self.containerFileRegex = re.compile("^rhcert-(?P<type>(manifest|image|description))-(?P<filename>.+)$")
        self.certificationFileRegex = re.compile("^redhat-certification-.*\.xml(.gz)?$")
        self.multiHostFileRegex = re.compile(Regex.multiHostFileName)
        self.socketDefaultTimeout = socket.getdefaulttimeout()
        self.xslFilterPath = os.path.join(self.environment.getServerDirectory(), "xsl/default.xsl")
        self.vendorProducts = None
        self.serverSettings = None
        self.clearErrors()
        self.productFilePath = os.path.join(self.getStoreDirectory(), Constants.systems, "products.xml")
        self.settingsFilePath = os.path.join(self.getStoreDirectory(), Constants.systems, "settings.xml")

    def getErrors(self):
        return self.errors
    def appendError(self, error):
        self.errors.append(error)
    def clearErrors(self):
        self.errors = list()

    def __makeDirs(self, path):
        try:
            apacheUID = pwd.getpwnam('apache')[2]
            uid = os.getuid()
            if uid != apacheUID:
                os.setuid(apacheUID)
            os.makedirs(path)
            if uid != apacheUID:
                os.setuid(uid)
        except OSError:
            # directory exists
            pass

    def __getCertificationDirectory(self, document):
        return "%s-%s" % (document.getCertificationProgram(), document.getCertificationID())

    def __getCertificationDirectoryPath(self, document):
        return  os.path.join(self.getStoreDirectory(), Constants.results,
                           self.__getCertificationDirectory(document))

    def getStoreDirectory(self):
        if self.database:
            return os.path.join(self.environment.getStoreDirectory(), self.database)
        return self.environment.getStoreDirectory()

    def getCertificationInfo(self, path, errors):
        """ returns (fileType, document)
            fileType = (results | certification | multihost)
            document = ResultsDocument for (results), CertificationDocument (for certification and image)
        """
        errors = None
        filename = os.path.basename(path)
        if self.resultsFileRegex.match(filename):
            try:
                resultsDocument = self.getResultsDocument(path)
                return (FileType.results, resultsDocument)
            except Exception as exception:
                errors = "Error: could not load test results:\n"
                errors += "%s\n" % exception
                return (FileType.results, None)

        # otherwise
        if self.multiHostFileRegex.match(filename):
            try:
                resultsDocument = ResultsDocument()
                resultsDocument.load(path)
                return (FileType.multihost, resultsDocument)
            except Exception as exception:
                errors = "Error: could not load multihost results:\n"
                errors += "%s\n" % exception
                return (FileType.multihost, None)

         # otherwise
        if self.certificationFileRegex.match(filename):
            try:
                certificationDocument = CertificationDocument()
                certificationDocument.load(path)
                return (FileType.certification, certificationDocument)
            except Exception as exception:
                errors = "Error: could not load certification:\n"
                errors += "%s\n" % exception
                return (FileType.certification, None)

        # otherwise
        errors = "Error: could not determine file type"
        return (None, None)

    def getEnvironment(self):
        return self.environment

    def makeContainerDirectoryPath(self, certificationID, hostname):
        directory = "rhcert-container-%s-%s-%s" % (certificationID, hostname, self.getCurrentUTCTimeStamp())
        directoryPath = os.path.join(self.getStoreDirectory(), self.environment.getServerTransfer(), directory)
        self.__makeDirs(directoryPath)
        return directoryPath

    def getContainerFileName(self, containerFile, filename):
        return "rhcert-%s-%s" % (containerFile.lower(), filename)

    def getXSLTParameters(self):
        xsltParameters = { "devices":"\'hide\'", "device-classes": "\'hide\'", "rpath":"\'%s\'" % self.environment.getResultsPath(), "server-email":"\'%s\'" % self.environment.getServerEMail()}
        return xsltParameters

    def getCurrentUTCTime(self):
        return time.gmtime(time.time())
    def timeStringToTimestamp(self, theTime):
        theTimeString = time.strftime(Constants.DATETIMEFORMAT, theTime)
        return theTimeString.replace(" ", "").replace(":", "").replace("-", "")
    def getCurrentUTCTimeStamp(self):
        return self.timeStringToTimestamp(self.getCurrentUTCTime())
    def timestampToDateTime(self, dateStamp):
        date = "%s-%s-%s" % (dateStamp[4:6], dateStamp[6:8], dateStamp[:4])
        time = "%s:%s:%s" % (dateStamp[8:10], dateStamp[10:12], dateStamp[12:])
        return (date, time)

    def __getResults(self, location=Constants.results, hostname=None, ipAddress=None):
        """ returns dict[path] = (host, date, time)
            hostname=None returns all
        """
        results = dict()
        localPath = os.path.join(self.getStoreDirectory(), location)
        for (root,_,files) in os.walk(localPath):
            if files:
                files = sorted(files)
                for aFile in files:
                    match = self.resultsFileRegex.search(aFile)
                    if match:
                        (date, time) = self.timestampToDateTime(match.group("date"))
                        if not (hostname and ipAddress) or hostname == match.group("hostname") or ipAddress == match.group("hostname"):
                            results[os.path.join(root, aFile)] = (match.group("hostname"), date, time)
        return results

    def getResultsDictionary(self):
        return self.__getResults()

    def getContainerDirectories(self, certificationID=-1):
        """ return container directories matching certificationID (or all if certificationID == -1)
            returns dict[path] = (certificationID, host, date, time)
        """
        containers = dict()
        localPath = os.path.join(self.getStoreDirectory(), self.environment.getServerTransfer())
        for (root,dirs,files) in os.walk(localPath):
            if dirs:
                dirs = sorted(dirs)
                for directory in dirs:
                    match = self.containerDirectoryRegex.search(directory)
                    if match:
                        if certificationID < 0 or (certificationID == match.group("id")):
                            (date, time) = self.timestampToDateTime(match.group("timestamp"))
                            containers[os.path.join(root, directory)] = (match.group("id"), match.group("hostname"), date, time)
        return containers

    def getContainerDirectoryInfo(self, directoryPath):
        """ return (certID, hostname, date, time) """
        match = self.containerDirectoryRegex.search(os.path.basename(directoryPath))
        if match:
            (date, time) = self.timestampToDateTime(match.group("timestamp"))
            return (match.group("id"), match.group("hostname"), date, time)
        else:
            return (None, None, None, None)

    def getContainerFiles(self, path):
        """ returns (manifestPath, imagePath, keyPath, descriptionPath) for the container path """
        manifestPath = None
        imagePath = None
        keyPath = None
        descriptionPath = None
        for (root,_,files) in os.walk(path):
            for containerFile in files:
                match = self.containerFileRegex.match(containerFile)
                if match:
                    if match.group("type") == "manifest":
                        manifestPath = os.path.join(root, containerFile)
                        continue
                    if match.group("type") == "image":
                        if os.path.splitext(containerFile)[1] == ".key":
                            keyPath = os.path.join(root, containerFile)
                        else:
                            imagePath = os.path.join(root, containerFile)
                        continue
                    if match.group("type") == "description":
                        descriptionPath = os.path.join(root, containerFile)
                        continue
        return (manifestPath, imagePath, keyPath, descriptionPath)

    def getContainerImageDescription(self, path):
        (_, _, _, descriptionPath) = self.getContainerFiles(path)
        if descriptionPath:
            image = ImagePair()
            image.load(descriptionPath)
            return image
        return None

    def getContainerFileInfo(self, path):
        """ return (image|manifest, filename) """
        match = self.containerFileRegex.match(os.path.basename(path))
        if match:
            return (match.group("type"), match.group("filename"))
        else:
            return (None, None)


    def getHosts(self):
        systems = dict()
        for (resultsFile, (host, date, time)) in self.__getResults().items():
            systems[host] = (resultsFile, date, time)
        return systems

    def getSystemsDocuments(self):
        systemsFiles = self.__getResults(Constants.systems)
        # clean out old documents
        systemsByHost = dict() # (systemFile, systemDocument)[hostname]
        for (systemFile, (hostname, _, _)) in systemsFiles.items():
            systemDocument = self.getResultsDocument(systemFile)
            if hostname not in systemsByHost:
                systemsByHost[hostname] = (systemFile, systemDocument)
                continue
            #otherwise, delete the older one
            existingFile, existingDocument = systemsByHost[hostname]
            if systemDocument.getPlanTime() > existingDocument.getPlanTime():
                systemsByHost[hostname] = (systemFile, systemDocument)
                os.remove(existingFile)
            else:
                os.remove(systemFile)
        systemsDocuments = dict()
        for (systemFile, systemDocument) in systemsByHost.values():
            systemsDocuments[systemFile] = systemDocument

        return systemsDocuments

    def getSystemPath(self, hostname):
        systemsFiles = self.__getResults(Constants.systems)
        for path in systemsFiles:
            if hostname in path: # note hostname can be a path
                return path
        return None

    def getSystemPathForHost(self, host):
        """ host Host """
        if host.getHostname():
            path = self.getSystemPath(host.getHostname())
            return path
        if host.getIPAddress():
            for path in self.__getResults(Constants.systems):
                results = self.getResultsDocument(path)
                if results.document and results.getHostOS().getIPAddress() == host.getIPAddress():
                    return path
        return None  # maybe None

    def removeSystem(self, hostname):
        systemsFiles = self.__getResults(Constants.systems)
        for path in systemsFiles:
            if hostname in path: # note hostname can be a path
                os.remove(path)
                return True
        return False

    def getResultsDocuments(self, location):
        return self.__getResultsDocuments(self.resultsFileRegex, location)

    def __getResultsDocuments(self, regex, location, timestamp=None):
        path = os.path.join(self.getStoreDirectory(), location)

        resultsDocuments = dict() # ResultsDocument[file path]
        for (root,_,files) in os.walk(path):
            if files:
                files = sorted(files)
                for aFile in files:
                    match = regex.search(aFile)
                    if match and (not timestamp or match.group("date") == timestamp):
                        filePath = os.path.join(root, aFile)
                        results = self.getResultsDocument(filePath, regex)
                        if results.document:
                            resultsDocuments[filePath] = results
        return resultsDocuments

    def getResultsDocument(self, path, regex=None):
        if not regex:
            if self.multiHostFileRegex.match(os.path.basename(path)):
                regex = self.multiHostFileRegex
            else:
                regex = self.resultsFileRegex

        if not regex.search(os.path.basename(path)):
            return None

        # otherwise
        results = ResultsDocument()
        results.loadFiltered(path, self.xslFilterPath, self.getXSLTParameters())
        if not results.document:
            return None
        return results

    def getLatestResultsDocument(self, host, certificationProgram, certificationID):
        """ host Host """
        if not certificationProgram or not certificationID:
            return None
        (_, certification) = self.getCertification(certificationProgram, certificationID)
        if certification:
            resultsDocuments = self.getResultsDocumentsFor(certification)
            if resultsDocuments:
                hostDocumentPaths = list()
                for (path, document) in resultsDocuments.items():
                    if host and (host.getHostname() == document.getHostname() or host.getIPAddress() == document.getOS(Tags.ip_address)):
                        hostDocumentPaths.append(path)
                if hostDocumentPaths:
                    # return the latest matching doc.
                    latestPath = sorted(hostDocumentPaths, reverse=True)[0]
                    return (resultsDocuments[latestPath], latestPath)

    def getMultiHostResultsDocuments(self, program, certificationID, timestamp=None):
        location=os.path.join(Constants.results, "%s-%s" % (program, certificationID))
        return self.__getResultsDocuments(self.multiHostFileRegex, location, timestamp)

    def getMultiHostResults(self, resultsDocument):
        timestamp = resultsDocument.timeStringToTimestamp(resultsDocument.getMultiHostRunTime())
        documents = self.getMultiHostResultsDocuments(
                            resultsDocument.getCertificationProgram(),
                            resultsDocument.getCertificationID(),
                            timestamp)
        # should either be none or one
        if not documents:
            return None
        if len(documents.values()) > 1:
            #raise RedHatCertificationStoreException(
            #    "Error: more than one multihost results file with the same timestamp")
            return None
        return documents.values()[0]

    def deleteResults(self, path):
        if self.getResultsDocument(path):
            os.remove(path)
            return True
        return False
    def deleteMultiHostResults(self, path):
        if self.multiHostFileRegex.match(os.path.basename(path)):
            os.remove(path)
            return True
        return False

    def deleteImagePair(self, path):
        if self.containerDirectoryRegex.match(os.path.basename(path)):
            shutil.rmtree(path)
            return True
        return False

    def deleteCertification(self, certificationProgram, certificationID):
        (_, document) = self.getCertification(certificationProgram, certificationID)
        if document:
            shutil.rmtree(self.__getCertificationDirectoryPath(document))
            return True
        return False

    def deleteSandbox(self, program, sandboxID):
        (_, document) = self.getCertification(program, sandboxID)
        if document and document.isSandbox():
            shutil.rmtree(self.__getCertificationDirectoryPath(document))
            return True
        return False


    def getResultsDocumentsFor(self, certification, skipEmpty=False, host=None):
        """ returns ResultsDocuments[path] host is a Host, host=None means all hosts"""
        documents = self.__getResults(location=os.path.join(Constants.results,
                                self.__getCertificationDirectory(certification)))
        matchingDocuments = dict()
        for (path) in documents.keys():
            document = self.getResultsDocument(path)
            if skipEmpty and not document.getRunTime():
                continue
            if host and host.getName() != document.getHostname() and host.getIPAddress() != document.getOS(Tags.ip_address):
                continue
            matchingDocuments[path] = document
        return matchingDocuments

    def getCertificationDocuments(self, location=""):
        path = os.path.join(self.getStoreDirectory(), Constants.results, location)
        certifications = dict() # CertificationDocument[file path]
        for (root,_,files) in os.walk(path):
            if files:
                files = sorted(files)
                for aFile in files:
                    match = self.certificationFileRegex.search(aFile)
                    if match:
                        filePath = os.path.join(root, aFile)
                        certification = CertificationDocument()
                        certification.load(filePath)
                        if certification.document:
                            certifications[filePath] = certification
        return certifications

    def getCertifications(self, vendorID=None, productID=None):
        certifications = list()
        for certification in self.getCertificationDocuments().values():
            if vendorID:
                for vendor in certification.getVendors():
                    if vendor.getID() == vendorID:
                        if (not productID) or (productID and vendor.getProduct(productID)):
                            certifications.append(certification)
            else:
                certifications.append(certification)
        return certifications

    def getCertification(self, certificationProgram, certificationID):
        certifications = self.getCertificationDocuments("%s-%s" % (certificationProgram, certificationID))
        for (path, certification) in certifications.items():
            if (certification.getCertificationID() == certificationID
                    and certification.getCertificationProgram() == certificationProgram):
                return (path, certification)
        return (None, None)

    def getCertificationWithBugID(self, certificationProgram, bugID):
        certifications = self.getCertificationDocuments()
        for (path, certification) in certifications.items():
            if (certification.getBugID() == bugID
                    and certification.getCertificationProgram() == certificationProgram):
                return (path, certification)
        return (None, None)

    def saveResultsDocument(self, resultsDocument):
        certificationDirectory = os.path.join(self.getStoreDirectory(), Constants.results,
                           self.__getCertificationDirectory(resultsDocument))
        self.__makeDirs(certificationDirectory)
        if not resultsDocument.isMultiHost():
            path = os.path.join(certificationDirectory, resultsDocument.getVerboseFileName())
        else:
            path = os.path.join(certificationDirectory, resultsDocument.getMultiHostFileName())
        return resultsDocument.save(path)

    def __saveResultsFile(self, resultsFileName, resultsFileContents, location):
        directory = os.path.realpath(os.path.join(self.getStoreDirectory(), location))
        path = os.path.realpath(os.path.join(directory, resultsFileName))
        # make sure the full path created is still in the store directory
        if not os.path.commonprefix([self.getStoreDirectory(), path]) == self.getStoreDirectory():
            self.appendError("Error: invalid results file name %s" % (path))
            return None

        print("saving to path %s " % path)
        try:
            self.__makeDirs(directory)
            fileToSave = open(path, "wb")
            fileToSave.write(resultsFileContents)
            fileToSave.close()
        except Exception as e:
            self.appendError("Error: could not save file %s to %s" % (resultsFileName, location))
            self.appendError("%s" % e)
            return None
        return path

    def saveResultsFileSimple(self, resultsFileName, resultsFileContents):
        return self.__saveResultsFile(resultsFileName, resultsFileContents, Constants.results)

    def saveResultsFile(self, resultsFileName, resultsFileContents):
        #if there are any matching results files (host, cert id, cert type) with less info (no test runs),
        # delete them
        newFilePath = self.__saveResultsFile(resultsFileName, resultsFileContents, Constants.results)
        if not newFilePath:
            return (False, None)

        try:
            newResultsDocument = self.getResultsDocument(newFilePath)
        except Exception as error:
            self.appendError("Error: %s" % (error))
            newResultsDocument = None

        if not newResultsDocument:
            self.appendError("Error: Could not import results file %s" % newFilePath)
            # attempt to delete the bad file:
            try:
                os.remove(newFilePath)
            except:
                pass
            return (False, None)

        # otherwise:
        # since release 5.11 we are setting certification ID as string
        if newResultsDocument.getCertificationID() == '0': # no cert info, done
            return (False, newFilePath)

        # otherwise
        newFilePath = self.__moveResultsFile(newResultsDocument)
        newResultsDocument = ResultsDocument()
        newResultsDocument.load(newFilePath)
        # if vendor and product aren't there, copy them in
        for vendor in newResultsDocument.getVendors(notRedHat=True):
            storedVendor = self.getVendorProducts().copyVendor(vendor)
            vendor.setID(storedVendor.getID())
        newResultsDocument.save() # vendor ID may have changed
        if self.vendorProducts:
            self.vendorProducts.save() # vendor/product data may have changed.

        # if there's no cert document there, make one
        (_, existingCertification) = self.getCertification(newResultsDocument.getCertificationProgram(), newResultsDocument.getCertificationID())
        if not existingCertification:
            certification = CertificationDocument()
            certification.copy(newResultsDocument)
            self.saveCertification(certification)
        hostname=newResultsDocument.getOS(Tags.hostname)
        ipAddress=newResultsDocument.getOS(Tags.ip_address)
        self.__deleteEmptyMatchingResults(newFilePath, newResultsDocument.getCertificationID(), hostname, ipAddress)

        # note: newResultsDocument is from before it was fully saved to newFilePath
        self.updateMultiHostResults(newResultsDocument, newFilePath)

        return (True, self.__makeResultsURL(newResultsDocument, newFilePath))

    def __makeResultsURL(self, resultsDocument, newFilePath):
        certificationType = "certification" if not resultsDocument.isSandbox() else "sandbox"
        if self.multiHostFileRegex.match(os.path.basename(newFilePath)):
            # will return url for multihost result file
            return "/" + os.path.join( certificationType,
                                        resultsDocument.getCertificationProgram(),
                                        resultsDocument.getCertificationID(),
                                        resultsDocument.getMultiHostRunTime(),
                                        "assign")
        else:
            return "/" + os.path.join( certificationType,
                                        resultsDocument.getCertificationProgram(),
                                        resultsDocument.getCertificationID(),
                                        resultsDocument.getHostname(),
                                        resultsDocument.getPlanTime(),
                                        "results")

    def __moveResultsFile(self, resultsDocument):
        newDirectory  = self.__getCertificationDirectoryPath(resultsDocument)
        self.__makeDirs(newDirectory)
        newFilePath = os.path.join(newDirectory, os.path.basename(resultsDocument.getLoadedFilePath()))
        if os.path.exists(newFilePath):
            os.remove(newFilePath)
        shutil.move(resultsDocument.getLoadedFilePath(), newDirectory)
        return newFilePath

    def fixResultsFileLocation(self, resultsDocument):
        print("moving %s to certification ID %s" % (resultsDocument.getLoadedFilePath(), resultsDocument.getCertificationID()))
        return self.__moveResultsFile(resultsDocument)

    def __deleteEmptyMatchingResults(self, resultsPath, certificationID, hostname, ipAddress):
        otherResults = self.__getResults(Constants.results, hostname, ipAddress)
        for path in otherResults.keys():
            if path == resultsPath:
                continue

            # otherwise
            otherResultsDocument = self.getResultsDocument(path)
            if otherResultsDocument and otherResultsDocument.getCertificationID() == certificationID:
                if otherResultsDocument.getNumberOfTestRuns() == 0:
                    self.deleteResults(path)

    def updateMultiHostResults(self, resultsDocument, resultsPath):

        # is this a multi-host results doc that is part of a multihost run?
        if not (resultsDocument.isMultiHost() and resultsDocument.getMultiHostRunTime()):
            return

        resultsDocument = ResultsDocument()
        resultsDocument.load(resultsPath) # don't filter
        # otherwise
        multiHostResults  = self.getMultiHostResults(resultsDocument)
        if not multiHostResults: # then this is the first one from the run
            self.newMultiHostResults(resultsDocument)
            return

        # otherwise, merge the results into the multihost
        # reload without filtering away attachments
        multiHostResults.reload()
        multiHostResults.mergeMultiHost(resultsDocument)
        multiHostResults.save()

    def newMultiHostResults(self, resultsDocument, path=None):
        # fix up actual host info
        resultsDocument.reload()
        resultsDocument.setActualHostRoles()
        path  = path if path else self.__getCertificationDirectoryPath(resultsDocument)
        resultsDocument.save(os.path.join(path, resultsDocument.getMultiHostFileName()))
        return resultsDocument

    def saveSystemFile(self, resultsFileName, resultsFileContents, path=Constants.systems):
        return self.__saveResultsFile(resultsFileName, resultsFileContents, path)

    def saveTransferFile(self, resultsFileName, resultsFileContents, ipAddress):
        location = os.path.join(Constants.transfer, ipAddress)
        return self.__saveResultsFile(resultsFileName, resultsFileContents, location)

    def saveCertification(self, certification, overwrite=False):
        """ certification: CertificationDocument """
        (_, existingCertification) = self.getCertification(certification.getCertificationProgram(), certification.getCertificationID())
        if not overwrite and existingCertification:
            print("Error: certification %s %s exists" % (certification.getCertificationProgram(),certification.getCertificationID()))
            return None
        directory = os.path.join(self.getStoreDirectory(), Constants.results, self.__getCertificationDirectory(certification))
        self.__makeDirs(directory)
        path = os.path.join(directory, certification.getVerboseFileName())
        certification.save(path)
        return path

    def updateCertification(self, certification):
        (_, localCertification) = self.getCertification(certification.getCertificationProgram(), certification.getCertificationID())
        path = None
        if not localCertification:
            path = self.saveCertification(certification)
        else:
            # otherwise, merge the update into the existing certification document
            localCertification.update(certification)
            path = self.saveCertification(localCertification, overwrite=True)

        if not path:
            return None # errors reported in saveCertification

        certification = CertificationDocument()
        certification.load(path)
        return certification

    def makeScratchCertificationID(self):
        nextScratchPath = os.path.join(self.getStoreDirectory(), ".nextScratchPath")
        id = "1"
        try:
            idFile = open(nextScratchPath)
            id = idFile.read()
            idFile.close()
        except:
            pass

        try:
            nextID = int(id)+1
            idFile = open(nextScratchPath, "w")
            idFile.write("%s" % nextID)
        except IOError:
            print("Error: could not write to scrath ID file %s" % nextScratchPath)
            return ""

        return id

    def getVendorProducts(self, reload=False):
        if reload or not self.vendorProducts:
            self.vendorProducts = VendorProductsDocument()
            if os.path.exists(self.productFilePath):
                self.vendorProducts.load(self.productFilePath)
            else:
                self.vendorProducts.new()
                self.__makeDirectoryPath(os.path.join(self.getStoreDirectory(), Constants.systems))
                self.vendorProducts.save(self.productFilePath)
        return self.vendorProducts

    def getServerSettings(self, reload=False):
        if reload or not self.serverSettings:
            self.serverSettings = ServerSettingsDocument()
            if os.path.exists(self.settingsFilePath):
                self.serverSettings.load(self.settingsFilePath)
            else:
                self.serverSettings.new()
                self.__makeDirectoryPath(os.path.join(self.getStoreDirectory(), Constants.systems))
                self.serverSettings.save(self.settingsFilePath)
        return self.serverSettings

    def __makeDirectoryPath(self, directory):
        """Emulate mkdir -p behavior: create the given directory, including all
        needed parent directories. Do not return an error if the directories
        already exist."""
        try:
            os.makedirs(directory)
        except OSError as e:
            if e.errno != errno.EEXIST: raise

class RedHatCertificationStoreException(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message

    # BaseException.message has been deprecated since Python 2.6.  To prevent
    # DeprecationWarning from popping up over this pre-existing attribute, use
    # a new property that takes lookup precedence.
    def _get_message(self): return self.__message
    def _set_message(self, value): self.__message = value
    message = property(_get_message, _set_message)

def unitTestMultiHost():
    from rhcert.commandLineUI import CommandLineUI
    store = RedHatCertificationStore()
    ui = CommandLineUI()
    program = ui.prompt("certification program: ", answers=["cloud", "hardware", "openstack"] )
    certID = str(ui.prompt("certification id: "))
    (_, certification) = store.getCertification(program, certID)
    if not certification:
        print("No certification found for %s %s" % (program, certID))
        return False
    results = store.getResultsDocumentsFor(certification)
    if not results:
        print("No results for certification %s %s" % (program, certID))
        return False
    for result in results.values():
        if result.isMultiHost():
            print("Found multihost results file " + result.getLoadedFilePath())
            store.updateMultiHostResults(result)
            multiHostResults = store.getMultiHostResults(result)
            print("Got multihost results: " + multiHostResults.getLoadedFilePath())
def unit():
    unitTestMultiHost()
    return

    errors = 0
    store = RedHatCertificationStore()
    results = store.getResultsDictionary()
    emptyResultsHost = None
    otherResultsPath = None
    emptyResultsID = None
    for key in results.keys():
        (hostname, date, time) = results[key]
        print("%s: %s %s %s" % (key, hostname, date, time))
        doc = store.getResultsDocument(key)
        if doc:
            print("certification id %s type %s" % (doc.getCertificationID(), doc.getCertificationProgram()))
            print("%s test runs" % doc.getNumberOfTestRuns())
            if doc.getNumberOfTestRuns() == 0 and doc.getCertificationID() != 0:
                emptyResultsHost = hostname
                emptyResultsID = doc.getCertificationID()
                print("found %s with empty results for cert %s" % (emptyResultsHost, emptyResultsID))
        else:
            print("Error: could not load document")
            errors += 1
        print("")

    for key in results.keys():
        (hostname, date, time) = results[key]
        print("%s: %s %s %s" % (key, hostname, date, time))
        doc = store.getResultsDocument(key)
        if doc:
            print("certification id %s type %s" % (doc.getCertificationID(), doc.getCertificationProgram()))
            print("%s test runs" % doc.getNumberOfTestRuns())
            if doc.getCertificationID() != 0 and doc.getCertificationID() == emptyResultsID:
                otherResultsPath = key
                print("found some other results for host %s" % emptyResultsHost)
                break
        else:
            print("Error: could not load document")
            errors += 1
        print("")

    if not otherResultsPath:
        print("no clean up needed")
    else:
        print("testing save for path " + otherResultsPath)
        file = open(otherResultsPath)
        contents = file.read()
        file.close()
        store.saveResultsFile(os.path.basename(otherResultsPath), contents)

    if errors:
        print("%s errors" % errors)
        return 1

    index = store.getServerInstallation()
    print(index.toXML())
    return 0

if __name__ == "__main__":
    exit(unit())
