# Copyright (c) 2016 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols

import sys
import os
import json
import requests
from rhcert.environment import Environment
from rhcert.services.microservice import MicroService

class Authorization(MicroService):

    def __init__(self, environment):
        MicroService.__init__(self, environment)
        # reset servicesURL to catalog host:
        self.servicesURL = self.environment.getSSOServer()
        self.error = None

    def getURL(self):
        return "auth/realms/redhat-external/protocol/openid-connect/token"

    def getError(self):
        return self.error

    def getCacheFileName(self):
        return None # no caching

    def call(self):
        """ returns a requests.response """
        self.__api_call()
        return self.response

    def __api_call(self):
        proxyURLs = self.environment.getProxyURLs()
        print(proxyURLs)
        try:
            url = os.path.join(self.servicesURL, self.getURL())
            self.response = requests.request("POST", url, data=self.getPayload(), headers=self.getHeaders(), timeout=self.timeout, proxies=proxyURLs)
            if self.response.status_code not in [200, 202]:
                print("Error: " + str(self.response.raise_for_status()))

        except requests.exceptions.HTTPError as HTTPError:
            self.error = "Error: Authentication HTTPError occurred: " + str(HTTPError)
            print(self.error)

        except Exception as err:
            self.error = "Error: Authentication: " + str(err)
            print(self.error)

    def getAccessToken(self):
        return self.__getResponse("access_token")
    def getRefreshToken(self):
        return self.__getResponse("refresh_token")
    def getTokens(self):
        return {"access_token": self.getAccessToken(), "refresh_token": self.getRefreshToken()}

    def __getResponse(self, attribute):
        if not self.response:
            return None
        try:
            responseDict = self.response.json()
            return responseDict.get(attribute)
        except:
            pass
        return None

class Login(Authorization):

    def __init__(self, username, password, environment):
        Authorization.__init__(self, environment)
        self.username = username
        self.password = password

    def getPayload(self):
        return {
                "username": self.username,
                "password": self.password,
                "grant_type": "password",
                "client_id": "rhcert-cwe"
        }

class Refresh(Authorization):

    def __init__(self, refresh_token, environment):
        Authorization.__init__(self, environment)
        self.refresh_token = refresh_token

    def getPayload(self):
        return {
                "refresh_token": self.refresh_token,
                "grant_type": "refresh_token",
                "client_id": "rhcert-cwe"
            }

def unit_test():
    environment = Environment()
    response = None
    print("\nlogin ------------------------------------------------------------------------------")
    login = Login("gnichols_qa", "redhat123", environment)
    print(login.getMethod())
    print(login.getPayload())
    try:
        response = login.call()
        if response:
            responseDict = response.json()
            print(json.dumps(responseDict, indent=1))
        else:
            print("Login Failed: response:  %s" % response)
    except Exception as error:
        print("Login failed:")
        print(error)
        return False

    print("\nrefresh ------------------------------------------------------------------------------")
    refresh = Refresh(login.getRefreshToken(), environment)
    try:
        response = refresh.call()
        responseDict = response.json()
        print(json.dumps(responseDict, indent=1))
    except Exception as error:
        print("Refresh failed:")
        print(error)
        return False

    return True

if __name__ == "__main__":
    if unit_test():
        sys.exit(0)
    sys.exit(1)
