# Copyright (c) 2018 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Pooja Tiwari
import sys
from rhcert.compatability import StringIO
from rhcert.tags import TestResult
from rhcert.services.microservice import MicroService


class ContainerImageValidation(MicroService):

    def __init__(
            self,
            container_image=None,
            container_name=None,
            container_image_layers=None,
            environment=None,
            deferredCall=None):
        MicroService.__init__(self, environment, deferredCall)
        self.container_image_layers = container_image_layers
        self.container_image = container_image
        self.container_name = container_name

    def getURL(self):
        return "rhc-cert-service/find-container-by-uncompressed-hashes"

    def getHeaders(self):
        return {"Content-Type": "application/json"}

    def getMethod(self):
        return "POST"

    def getPayload(self):
        if self.container_image_layers:
            return {"Layers": self.container_image_layers}
        return MicroService.getPayload(self)

    def getCacheFileName(self):
        return None  # no caching

    def getImage(self):
        if self.deferredCall:
            return self.deferredCall.getImage()
        return self.container_image

    def getContainer(self):
        if self.deferredCall:
            return self.deferredCall.getContainer()
        return self.container_name

    def update(self, outputDocument, runDocument):
        msg = ("\n\nContainer Name: {0}\nBuilt on image: {1}\nTrying to find "
               "out if this image is certified by Red Hat ...").format(
               self.getContainer(),
               self.getImage())
        print(msg)
        response = self.call()
        if not response:
            return False # can't update
        status = TestResult(outputDocument.getSummary())
        log = StringIO()
        log.write(msg)
        success_msg = ("\nContainer '%s' found in Red Hat certified container registry"
            " and therefore it is approved for use in certification.")
        response = response.json()
        is_certified = response.get("certified")
        if is_certified:
            log.write(success_msg % (self.getContainer()))
            certified_containers = dict()
            certified_containers[self.getContainer()] = (self.getImage(), self.getPayload()["Layers"])
            runDocument.setCertifiedContainers(certified_containers)
            project_id = response.get("unique_isv_pid")
            # when project_match_count is 1.
            if project_id:
                log.write("\nProject ID: %s" % project_id)
                # TODO: Add Red Hat URL for container reference.
                status = TestResult.PASS
            # when project_match_count > 1.
            else:
                if response.get('result') and response['result']:
                    log.write("\nNeeds Review: The container image is "
                        "associated with more than one projects. "
                        "Project IDs are as below:\n%s"
                        % '\n'.join(response['result'].keys()))
                else:
                    log.write("\nNeeds Review: no project id found.")
                status.combine(TestResult.REVIEW)
            if response.get('result') and response['result']:
                    runDocument.setContainerProjects(
                        {self.getContainer():response['result'].keys()})
            elif project_id:
                    runDocument.setContainerProjects(
                        {self.getContainer():[project_id]})

        elif is_certified == False:
            log.write(
                "\nError: The container image does not appear to be a Red Hat certified."
                "It should be certified to perform the functional testing.")
            log.write(
                "\nCorrective Action: Please utilize a Red Hat container image, or "
                "certify your container image first using Red Hat Connect "
                "(https://connect.redhat.com/).")
            status.combine(TestResult.FAIL)
        else:
            log.write("\nNeeds Review: container not found.")
            status.combine(TestResult.REVIEW)

        outputDocument.appendLog(log.getvalue())
        outputDocument.setSummary(status)
        return status

if __name__ == "__main__":
    sys.exit(1)

