# Copyright (c) 2016 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
import os
import sys
import json
import requests


from rhcert.environment import Environment


class MicroService(object):
    """ MicroService is a base class for service wrappers """
    def __init__(self, environment=None, deferredCall=None):
        self.environment = environment
        self.deferredCall = deferredCall
        if not self.environment:
            self.environment = Environment()
        self.servicesURL = self.environment.getServicesURL()
        # default to the one in site-packages/rhcert/info
        cacheDir = os.path.join("/", *(self.environment.getSuiteDirectory().split('/')[:-1]))
        self.cacheDir = os.path.join(cacheDir, "info")
        self.response = None
        self.cache = None
        self.timeout = 120 # sec

    def getURL(self):
        if self.deferredCall:
            return self.deferredCall.getURL()
        return None 

    def getHeaders(self):
        if self.deferredCall:
            return self.deferredCall.getHeaders()
        return None # default

    def getMethod(self):
        return "GET" # default

    def getPayload(self):
        if self.deferredCall:
            return self.deferredCall.getPayload()
        return None

    def getCacheFileName(self):
        return None # default:  no caching

    def call(self):
        """ returns a requests.response """
        self.response = self.__api_call()
        return self.response

    def getCache(self):
        """ returns a string buffer """
        if not self.getCacheFileName(): # no cache for this service
            return None
        path = os.path.join(self.getCacheDir(), self.getCacheFileName())
        try:
            with open(path, "r") as serviceFile:
                self.cache = serviceFile.read()
                return self.cache
        except IOError as error:
            print("Could not read cache:")
            print(error)
            return None


    def __api_call(self):
        """
        Makes a call to API and returns the response.
        """
        proxyURLs = self.environment.getProxyURLs()

        try:
            url = os.path.join(self.servicesURL, self.getURL())
            if self.getMethod() == 'GET':
                response = requests.request("GET", url, headers=self.getHeaders(), timeout=self.timeout, proxies=proxyURLs)
            elif self.getMethod() == 'POST':
                response = requests.request("POST", url, data=json.dumps(self.getPayload()), headers=self.getHeaders(), timeout=self.timeout, proxies=proxyURLs)
            elif self.getMethod() == 'DELETE':
                # add implementation when applicable
                pass
            elif self.getMethod() == 'PUT':
                # add implementation when applicable
                pass
            else:
                print("Error: Method not allowed.")
                return None

            if response.status_code not in [200, 202]:
                self.__printError(response.raise_for_status())
                return None
            return response

        except Exception as err:
            self.__printError(err)
            return None


    def __printError(self, error):
        msg = "Microservice at " + self.servicesURL + " is not accessible: " + str(error)
        if self.deferredCall:
            print("Error: %s" % msg)
            print("Please resolve the above issue to update the results.")
        else:
            print(msg)

        """
        # Kept it for future reference.
        if self.deferredCall:
            print("Method: " + self.getMethod())
            if self.no_service_url:
                print("End-Point: " + self.getURL())
            else:
                print("URI: " + os.path.join(self.servicesURL, self.getURL()))
            if self.getHeaders():
                print("Header: " + str(self.getHeaders()))
            if self.getPayload():
                print("Request Body: " + str(self.getPayload()))
        """

    def getCacheDir(self):
        return self.cacheDir
    def setCacheDir(self, path):
        self.cacheDir = path

    def update(self, outputDocument, runDocument):
        return None # subclasses must implement


if __name__ == "__main__":

    sys.exit(1)
