# Copyright (c) 2016 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
import os
import sys


from rhcert.environment import Environment
from rhcert.command import Command
from rhcert.tags import TestResult

from rhcert.services.validaterpms import ValidateRPMs
from rhcert.services.validkernels import ValidKernels
from rhcert.services.productnameservice import ProductNameService
from rhcert.services.rpmfreshness import RpmFreshness
from rhcert.services.container_image_validation import ContainerImageValidation

class MicroServices(object):
    """ mostly for backwards compatability to rhcert 4.0 """

    def __init__(self, environment=None):
        self.environment = environment
        self.cacheDir = None

    def getCacheDir(self):
        return self.cacheDir
    def setCacheDir(self, path):
        self.cacheDir = path

    def rpmFreshness(self, product, version, packages):
        return RpmFreshness(product, version, packages, environment=self.environment).call()
    def getProducts(self):
        return ProductNameService(environment=self.environment).call()
    def getValidKernels(self, product, version):
        return ValidKernels(product, version, environment=self.environment).call()
    def validate_rpms(self, product, version, packages):
        return ValidateRPMs(product, version, packages, environment=self.environment).call()

    def update(self, resultsDocument):
        """ attempt to re-run deferred calls to microservices """
        updated = False
        if not resultsDocument.getDeferredTests():
            print("Warning: Nothing to update. No deferred test found in supplied result.")
            return updated

        for runNumber, tests in resultsDocument.getDeferredTests().items():
            for (testDocument, outputDocument) in tests:
                print("INFO: Deferred test found: %s/%s" % (testDocument.getPath(), outputDocument.getName()))
                previousResult = TestResult(outputDocument.getSummary())
                print("INFO: Current test status: %s" % previousResult)
                print("INFO: Trying to update the result by calling service ...")
                deferredCalls = outputDocument.getDeferredCalls()
                for deferredCall in deferredCalls:
                    microService = self.getMicroService(deferredCall)
                    result = microService.update(outputDocument, testDocument.getRun(runNumber))
                    testDocument.getRun(runNumber).updateSummary()
                if result:
                    print("INFO: Result updated successfully!")
                    print("INFO: Tests status after update: %s" % outputDocument.getSummary())
                    outputDocument.removeDeferredNode(deferredCall.getName())
                    updated = True
        return updated

    def getMicroService(self, deferredCall):
        if deferredCall.getName() == "ContainerImageValidation":
            return ContainerImageValidation(environment=self.environment, deferredCall=deferredCall)
        if deferredCall.getName() == "RpmFreshness":
            return RpmFreshness(environment=self.environment, deferredCall=deferredCall)
        if deferredCall.getName() == "ProductNameService":
            return ProductNameService(environment=self.environment, deferredCall=deferredCall)
        if deferredCall.getName() == "ValidKernels":
            return ValidKernels(environment=self.environment, deferredCall=deferredCall)
        if deferredCall.getName() == "ValidateRPMs":
            return ValidateRPMs(environment=self.environment, deferredCall=deferredCall)

    def buildCache(self):
        """ build the cache of service data in case services are not available
            cache is in /usr/lib/pythonX.X/site-packages/rhcert/info
        """

        services = list()
        # Product Name Service
        services.append(ProductNameService(self.environment))

        # Valid Kernel Service
        valid_kernel = ValidKernels(environment=self.environment)
        valid_kernel.getAll = True
        services.append(valid_kernel)

        for service in services:
            if self.getCacheDir():
                service.setCacheDir(self.getCacheDir())
            print("calling %s: \n%s" % (service.__class__.__name__, os.path.join(self.environment.getServicesURL(), service.getURL())))
            response = service.call()
            if response:
                path = os.path.join(service.getCacheDir(), service.getCacheFileName())
                with open(path, "wb+") as serviceFile:
                    serviceFile.write(response.content)
                    print("wrote %s" % path)
def unit_test():
    environment = Environment()
    microServices = MicroServices(environment)
    print("Building Cache:")
    # microServices.buildCache()


    rpm = Command("rpm -qa")
    rpm.run()
    packages = rpm.output

    if False:
        print("Calling 4.0 API:")
        print("\nrpm freshness:")
        response = microServices.rpmFreshness("RHEL", "7.2", packages)
        if response:
            print(response.content)

        print("\nvalid kernels:")
        response = microServices.getValidKernels("RHEL", "7.2")
        print(response.content)

        print("\nproduct name service:")
        response = microServices.getProducts()
        if response:
            print(response.content)

        print("\nvalidate rpms:")
        response = microServices.ValidateRPMs("RHEL", "7.2", packages)
        if response:
            print(response.content)


    print("4.1 API:")
    services = list()
    services.append(RpmFreshness("RHEL", "7.2", packages, environment))
    services.append(ValidateRPMs("RHEL", "7.2", packages, environment))
    services.append(ValidKernels("RHEL", "7.2", environment))
    services.append(ProductNameService(environment))
    for service in services:
        print("calling (no caching) %s: \n%s" % (service.__class__.__name__, os.path.join(environment.getServicesURL(), service.getURL())))
        response = service.call()
        if response:
            print(response.content)
        else:
            print("Error: no response")

    print("\nforce caching:")
    servicesElement = environment.findOrCreateElement(environment._getURLsElement(), "services-url")
    environment.setTextNode(servicesElement, "http:bogus.redhat.com")
    for service in services:
        if service.getCacheFileName():
            print("calling (isolated, caching) %s: \n%s" % (service.__class__.__name__, os.path.join(environment.getServicesURL(), service.getURL())))
            response = service.getCache()
            if response:
                print(response)
            else:
                print("Error: No response")

if __name__ == "__main__":
    if unit_test():
        sys.exit(0)
    sys.exit(1)
