# Copyright (c) 2016 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols

import sys
import json

from rhcert.tags import Programs
from rhcert.services.microservice import MicroService

class ProductNameService(MicroService):

    def __init__(self, environment=None, deferredCall=None):
        MicroService.__init__(self, environment, deferredCall)
        self.programs = None
        self.current = False

    def getURL(self):
        return "rhproducts/all"
    def getHeaders(self):
        return  {"Accept": "application/json"}
    def getCacheFileName(self):
        return "redhat-products.json"

    def getAll(self):
        """ returns the entire product name dictionary """

        # already got it
        if self.programs:
            return self.programs

        # get it from the service
        self.programsJson = self.call()
        if self.programsJson:
            try:
                self.programs = self.programsJson.json()
                self.current = True
                return self.programs
            except Exception as err:
                print("Could not fetch product detail from service due to error: %s" % str(err))

        print("Using local cache for product detail")
        # get it from the cache
        self.programsJson = self.getCache()
        self.programs = json.loads(self.programsJson)
        return self.programs

    def getAllJson(self):
        self.getAll()
        return self.programsJson

    def getProgram(self, programName):
        self.getAll()
        for program in self.programs:
            if str(program["ecosystem"]) == str(programName):
                return program
        return None

    def getAllProducts(self):
        """ get all product certifications from all programs
            exceptions:
                - container program is ignored.
        """
        self.getAll()
        products = list()
        for program in self.programs:
            if str(program["ecosystem"]) == Programs.container:
                continue

            if str(program["ecosystem"]) == Programs.hardware:
                for product in program.get("products"):
                    if product.get("name") in ["Red Hat OpenStack Platform",
                                               "Red Hat Enterprise MRG"]:
                        continue
                    product["program"] = Programs.hardware
                    products.append(product)
                continue

            for product in program.get("products"):
                # Do not add duplicate products. Example: RHEL is part of cloud and hardware both.
                if product["id"] not in [_product["id"] for _product in products]:
                    product["program"] = str(program["ecosystem"])
                    products.append(product)
        return {"products": products }

    def getProductTree(self, programName, productID):
        program = self.getProgram(programName)
        if program:
            programCopy = program.copy()
            for product in programCopy["products"]:
                if int(product.get("id")) == int(productID):
                    programCopy["products"] = [product]
                    return programCopy
        return None


    def getProduct(self, productID):
        self.getAll()
        for program in self.programs:
            for product in program["products"]:
                if int(product.get("id")) == int(productID):
                    return product
        return None

    def getProductVersion(self, product, versionID):
        for productVersion in product["versions"]:
            if int(productVersion.get("id")) == int(versionID):
                return productVersion
        return None

    def getPlatform(self, productVersion, platformID):
        for platform in productVersion["platforms"]:
            if int(platform.get("id")) == int(platformID):
                return platform
        return None

    def getBaseVersion(self, productVersion, baseVersionID):
        if not (productVersion and productVersion.get("baseVersions")):
            return None
        for baseVersion in productVersion["baseVersions"]:
            if int(baseVersion.get("id")) == int(baseVersionID):
                return baseVersion
        return None

    def getProductPath(self, productID, productVersionID, baseVersionID):
        redhatProduct = self.getProduct(productID)
        redhatProductVersion = self.getProductVersion(redhatProduct, productVersionID)
        redhatBaseVersion = None
        if baseVersionID:
            redhatBaseVersion = self.getBaseVersion(redhatProductVersion, baseVersionID)
        return (redhatProduct, redhatProductVersion, redhatBaseVersion)

    def get_filtered_data(self, all_programs_data):
        """
        This method remove 'Red Hat Container Certification' data.
        """

        for program in all_programs_data:
            if program["ecosystem"] == Programs.container:
                all_programs_data.remove(program)
            else:
                pass
        return all_programs_data

    def getAllowedPrograms(self, redhatProductID):
        allowedPrograms = {
            1: { "programs": [Programs.hardware, Programs.cloud, Programs.software, Programs.openstack], "description": "CCSP, Software and Hardware products can be certified on Red Hat Enterprise Linux"},
            5: { "programs": [Programs.hardware], "description": "Hardware products can be certified on Red Hat Gluster if previously certified on RHEL"},
            6: { "programs": [Programs.hardware], "description": "Hardware products can be certified on Red Hat Enterprise Linux for Realtime if previously certified on RHEL"},
            7: { "programs": [Programs.hardware], "description": "Hardware products can be certified on Red Hat Enterprise Linux for ARM"},
            8: { "programs": [Programs.hardware], "description": "Hardware products can be certified on Red Hat Enterprise Linux Hyperconverged Infrastructure if previously certified on RHEL and Red Hat Gluster"},
            9: { "programs": [Programs.hardware], "description": "Hardware products can be certified on Red Hat Enterprise Linux for Power LE"},
            15: { "programs": [Programs.hardware], "description": "Hardware products can be certified on Red Hat Enterprise Linux for Power BE"},
            14: { "programs": [Programs.hardware], "description": "Hardware products can be certified on Red Hat Enterprise Linux for Power 9"},
            16: { "programs": [Programs.hardware], "description": "Hardware products can be certified on Red Hat Enterprise Linux for IBM z Systems"},
            17: { "programs": [Programs.hardware], "description": "Hardware products can be certified on Red Hat Enterprise Linux for IBM System z (Structure A)"},
            18: {"programs": [Programs.hardware], "description": "Hardware products can be certified on Red Hat OpenStack Platform for Real-Time Applications if previously certified on RHEL"},
            3: { "programs": [Programs.openstack, Programs.hardware], "description": "Hardware and OpenStack products can be certified on the Red Hat OpenStack Platform"},
            31: { "programs": [Programs.cloud], "description": "Cloud images can be certified on Red Hat Enterprise Linux High Availability - Update Services for SAP Solutions"},
        }
        return allowedPrograms.get(int(redhatProductID))

if __name__ == "__main__":
    import pprint
    pp = pprint.PrettyPrinter(indent=4)
    productNameService = ProductNameService()
    for product in productNameService.getAllProducts()["products"]:
        print("%s %s: %s" % (product.get("program"), product.get("id"), product.get("name")))
        for version in product.get("versions"):
            print("    %s: %s" % (version.get("id"), version.get("name")))
    # print(json.dumps(productNameService.getAllProducts(), sort_keys=True, indent=4, separators=(',', ': ')))
    sys.exit(0)