# Copyright (c) 2016 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
import sys

from rhcert.compatability import StringIO
from rhcert.tags import TestResult
from rhcert.services.microservice import MicroService


class RpmFreshness(MicroService):

    def __init__(self, product=None, version=None, packages=None, environment=None, deferredCall=None):
        MicroService.__init__(self, environment, deferredCall)
        self.product = product
        self.version = version
        self.packages = packages

    def getURL(self):
        return "playback/rpm-freshness"

    def getHeaders(self):
        return {"Content-Type": "application/json"}

    def getMethod(self):
        return "POST"

    def getPayload(self):
        if self.product and self.version and self.packages:
            product_version = "%s-%s" % (self.product, self.version)
            return {"operatingSystem": product_version,
                    "packages": list(filter(None, self.packages))}
        return MicroService.getPayload(self)

    def getCacheFileName(self):
        return None # no caching

    def update(self, outputDocument, runDocument):
        security_updates = self.call()
        if not security_updates:
            return False

        log = StringIO()
        log.write("\nUpdating Deferred Test: " + self.__class__.__name__ 
                + "\n--------------------------------------------\n")
        result = None
        security_updates = security_updates.json()
        if not security_updates['updates']:
            log.write("RPM Freshness successful.\n")
            result = TestResult.PASS
        else:
            _msg = str()
            # check for critical/important security vulnerabilities
            pkg_width = 20
            pkg_version_width = 30
            for package, update in security_updates['updates'].items():
                if update['severity'] in ['Critical', 'Important']:
                    _msg = (
                        _msg +
                        "\n|{}|{}|{}|").format(
                        package.ljust(pkg_width),
                        update['installed_version'].ljust(pkg_version_width),
                        update['version_available'].ljust(pkg_version_width))
            if not _msg:
                log.write("RPM Freshness successful.")
                result = TestResult.PASS
            else:
                log.write("Warning: Following packages needs to be updated:\n")
                separator = "\n+{}+{}+{}+".format("-" *
                                                pkg_width, "-" *
                                                pkg_version_width, "-" *
                                                pkg_version_width)
                log.write(separator)
                log.write("\n|{}|{}|{}|".format("Package".ljust(pkg_width),
                                          "Installed Version".ljust(pkg_version_width),
                                          "Available Version".ljust(pkg_version_width)))
                log.write(separator + _msg)
                log.write(separator + "\n")

                result = TestResult.WARN

        if result != None:
            outputDocument.appendLog(log.getvalue())
            outputDocument.setSummary(result)
        return result

if __name__ == "__main__":

    sys.exit(1)
