# Copyright (c) 2016 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols

import sys
import os
import json
import requests
import time
from rhcert.environment import Environment
from rhcert.services.microservice import MicroService

from rhcert.services.authorization import Login

class Strata(MicroService):

    def __init__(self, jwt, environment):
        MicroService.__init__(self, environment)
        self.jwt = jwt

    def getURL(self):
        return "/rs"

    def getCacheFileName(self):
        return None # no caching

    def call(self):
        """ returns a requests.response """
        startTime = time.time()
        self.response = self.__api_call()
        print(("Strata call time: %.1f sec." % (time.time()-startTime)))
        return self.response

    def __api_call(self):
        try:
            url =  self.environment.getCatalogServer() + self.getURL()
            print("\ncalling: " + url)
            response = requests.get(url, cookies={"rh_jwt": self.jwt}, headers=self.getHeaders(), timeout=self.timeout)
            if response.status_code not in [200, 202]:
                print("Error: " + str(response.raise_for_status()))
                return None
            return response

        except Exception as err:
            print("Error: authorization failed:")
            print(str(err))
            return None

class StrataVendorIDs(Strata):

    def getURL(self):
        return os.path.join(Strata.getURL(self), "ecosystem/vendors")
    def getHeaders(self):
        return  {"Accept": "application/vnd.redhat.vendors+json"}

class StrataVendor(Strata):

    def __init__(self, vendorID, jwt, environment):
        Strata.__init__(self, jwt, environment)
        self.vendorID = vendorID

    def getURL(self):
        return os.path.join(Strata.getURL(self), "ecosystem/vendors", str(self.vendorID))
    def getHeaders(self):
        return  {"Accept": "application/vnd.redhat.vendor+json"}

class StrataUserInfo(Strata):
    def __init__(self, jwt, environment):
        Strata.__init__(self, jwt, environment)

    def getURL(self):
        return os.path.join(Strata.getURL(self), "users/current")
    def getHeaders(self):
        return  {"Accept": "application/vnd.redhat.user+json"}


def unit_test():
    environment = Environment()
    print("Login\n ------------------------------------------------------------------------------")
    login = Login("cwetestuser", "redhat", environment)
    try:
        response = login.call()
        if not response:
            print("Login Failed.")
    except Exception as error:
        print("Login failed:")
        print(error)
        return False
    jwt = login.getAccessToken()
    print("StrataUserInfo\n ------------------------------------------------------------------------------")
    response = StrataUserInfo(jwt, environment).call()
    print(json.dumps(response.json(), indent=1))

    print("StrataVendorIDs\n ------------------------------------------------------------------------------")
    response = StrataVendorIDs(jwt, environment).call()
    print(json.dumps(response.json(), indent=1))

    for vendor in response.json().get('vendor'):
        response = StrataVendor(vendor.get("vendor_id"), jwt, environment).call()
        print("\nVendor ------------------------------------------------------------------------------")
        print(json.dumps(response.json(), indent=1))


    return True

if __name__ == "__main__":
    if unit_test():
        sys.exit(0)
    sys.exit(1)
