#!/usr/bin/env python
# Copyright (c) 2016 Red Hat, Inc. All rights reserved.
# This copyrighted material is made available to anyone
# wishing to use, modify, copy, or redistribute it subject
# to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.
#
# Author: Pooja Tiwari <ptiwari@redhat.com>
""" This module calls cloud configuration tests """
from __future__ import absolute_import
import sys
import socket
import os

try:  # python2
    from ConfigParser import ConfigParser
except ImportError:  # python3
    from configparser import ConfigParser
from rhcert.test import SubTest
from rhcert.redhatrelease import RedHatRelease
from rhcert.tags import Constants, Programs, TestResult, TestTag
from rhcert.command import Command, HwCertCommandException, HwCertCommandErrorReturned, HwCertCommandStringNotFound
from rhcert.configfile import ConfigFile
from rhcert.suites.rhcert.supportability import Utility


class OSRunLevelTest(SubTest):
    """
    OS Run Level Test class
    """

    def __init__(self, test):
        SubTest.__init__(self, test, "Default OS Runlevel",
                         "Check for system's runlevel.")
        self.desired_runlevel = self.test.policy["RUN_LEVELS"]
        self.actual_runlevel = str()

    def get_system_logs(self):
        """
        Get the system's runlevel.
        """

        try:
            runlevel = Utility().command_execute(
                "runlevel")
            self.actual_runlevel = runlevel.split(" ")[1].strip("\n")

        except Exception as err:
            print(err)

    def validate_logs(self):
        """
        Validate the runlevel.
        """
        try:
            if self.actual_runlevel in self.desired_runlevel:
                msg = ("OS Runlevel validation passed.\n" +
                       "Runlevel '{0}' found. \n").format(
                    self.actual_runlevel)
                print(msg)
                return True
            else:
                msg = ("Error: Runlevel '{0}' found.\n" +
                       "It should be one among '3', '4' " +
                       "and '5'").format(
                    self.actual_runlevel)
                print(msg)
                return False
        except Exception as err:
            print(err)

    def runSubTest(self):
        """
        OS RunLevel validation test.
        """

        try:
            self.get_system_logs()
            return self.validate_logs()
        except Exception as err:
            print(err)


class ServiceControlTest(SubTest):
    """
    Service Control Test Class
    """
    def __init__(self, test):
        SubTest.__init__(self, test,
                         "System Services",
                         "Check for Service Access")
        self.program = test.getSuiteName()
        self.release = RedHatRelease()

    def __serviceIsRunning(self, command, regex):
        statusCommand = Command(command)
        try:
            statusCommand.getString(regex=regex, singleLine=False)
            return True
        except HwCertCommandStringNotFound:
            # expected if the service is not running
            return False
        return False

    def __toggleService(self, toggle):
        toggle_cmd = Command(toggle)
        toggle_cmd.echo()

    def check_services_status_cloud_sap(self):
        """
        Check status of cloud_sap specific services:
        - 'chronyd' service is started and enabled and functional
        - 'uuidd' service is started and enabled
        """
        print("\n*** Checking status of required Cloud SAP services ***")
        result = TestResult()
        required_services = ["chronyd", "uuidd"]
        status_regex = "active \(running\)"
        for service in required_services:
            status_command = "systemctl status {}".format(service)

            if self.__serviceIsRunning(status_command, status_regex):
                print("\nSuccess: Service '{}' is active (running)".format(service))
                
                # Check if 'chronyd' is functional
                if service == 'chronyd':
                    try:
                        chronyc_tracking = Command("chronyc tracking")
                        chronyc_tracking.run()
                    except HwCertCommandErrorReturned as error:
                        result.combine(TestResult.FAIL)
                        print("\nError: Service 'chronyd'is not functional:")
                        print(error)
                    else:
                        print("\nSuccess: Service 'chronyd' is functional")
            else:
                result.combine(TestResult.FAIL)
                print("\nError: Service '{}'is not running:".format(service))
        return result

    def perform_service_control(self):
        """
        Tests for controlling a service by a root user
        The test will do following operations:
        * If service is running:
          * root user can stop a service then start a service
        * else service is inactive:
          * root user can start a service then stop a service.
        * Otherwise
          * failure message will be shown to user.
        """
        result = TestResult()
        pass_msg = "Root user can control a service successfully"
        print("\n*** Checking if root user can control a system service ***")
        # check if we can toggle the state of crond

        if self.release.getVersion() == 6:  # RHEL6.X is "service"
            self.system_commands = {"status_cmd": "service crond status",
                             "stop_cmd": "service crond stop",
                             "start_cmd": "service crond start",
                             "active_state_notifier": "running"}
        else: # RHEL7.X and beyond: systemd
            self.system_commands = {"status_cmd": "systemctl status crond",
                             "stop_cmd": "systemctl stop crond",
                             "start_cmd": "systemctl start crond",
                             "active_state_notifier": "active \(running\)"} # escaping needed for regex

        try:
            if self.__serviceIsRunning(self.system_commands["status_cmd"],
                                     self.system_commands["active_state_notifier"]):
                print("\nTrying to stop the crond service ...")
                self.__toggleService(self.system_commands["stop_cmd"])
                print("Service 'crond' stopped.\n")
                print(pass_msg)
                self.__toggleService(self.system_commands["start_cmd"])
                return result

            # otherwise, cron was initially stopped
            print("\nTrying to start the crond service ...")
            self.__toggleService(self.system_commands["start_cmd"])
            print("\n Service 'crond' started.")
            print(pass_msg)
            self.__toggleService(self.system_commands["stop_cmd"])
            return result

        except HwCertCommandException as error:
            print("Error: Could not control 'crond' service:")
            print(error)
            result.combine(TestResult.FAIL)

        return result

    def check_service_configured_vs_actual_state(self):
        """
        This method ensures that the current enabled/disabled state for
        each installed service matches it's current status.
        """

        print("\n*** Checking configured vs actual state of system services ***")

        result = TestResult()
        _platform = "el%s" % self.release.getVersion()

        service_info = {
            "el6": {
                "states": {
                    "enabled": {
                        "cmd": "chkconfig --list | grep '3:on'",
                        "services": list()
                    },
                    "disabled": {
                        "cmd": "chkconfig --list | grep '3:off'",
                        "services": list()
                    }
                },
                "status": "service %s status",
                "dead_state": ["not running", "stopped", "Stopped"]
            },
            "el7": {
                "states": {
                    "enabled": {
                        "cmd": "systemctl list-unit-files --type service --all | grep -w enabled",
                        "services": list()
                    },
                    "disabled": {
                        "cmd": "systemctl list-unit-files --type service --all | grep -w disabled",
                        "services": list()
                    }
                },
                "status": "systemctl status %s",
                "dead_state": ["inactive (dead)"]
            },
            "el8": {
                "states": {
                    "enabled": {
                        "cmd": "systemctl list-unit-files --type service --all | grep -w enabled",
                        "services": list()
                    },
                    "disabled": {
                        "cmd": "systemctl list-unit-files --type service --all | grep -w disabled",
                        "services": list()
                    }
                },
                "status": "systemctl status %s",
                "dead_state": ["inactive (dead)"]
            }
        }

        for state, info in service_info[_platform]["states"].items():
            print("\nFetching list of %s services." % state)
            print("--------------------------------------")
            try:
                cmd = Command(info["cmd"])
                cmd.echo()
                if _platform in ['el7', 'el8']:
                    for line in cmd.output:
                        for word in line.split(' '):
                            if word.endswith(".service"):
                                info["services"].append(word)
                                break
                else:
                    info["services"].extend([line.split(' ')[0]
                                             for line in cmd.output])
            except HwCertCommandException as error:
                print("\nError: Could not determine the list of %s services." % state)
                result.combine(TestResult.FAIL)

        print('\n')
        # Check if all enabled services are currently active
        for service in service_info[_platform]["states"]["enabled"]["services"]:
            status = Utility().command_execute(
                service_info[_platform]["status"] % service)
            for keyword in service_info[_platform]["dead_state"]:
                if keyword in status:
                    print("Error:  %s is enabled but not active." % service)
                    result.combine(TestResult.FAIL)
                    break

        # Check if all disabled services are currently inactive
        for service in service_info[_platform]["states"]["disabled"]["services"]:
            status = Utility().command_execute(
                service_info[_platform]["status"] % service)
            if "is running" in status:
                print("Error:  %s is disabled but active." % service)
                result.combine(TestResult.FAIL)

        if result == TestResult.FAIL:
            print("\nCorrective Action: Please keep the configured and actual state of services in sync.")

        return result

    def runSubTest(self):
        """
        Run service control tests
        """

        result = TestResult()
        subtest_list = [self.perform_service_control,
                      self.check_service_configured_vs_actual_state]
        if self.program == Programs.cloud_sap:
            subtest_list.append(self.check_services_status_cloud_sap)
        for subtest in subtest_list:
            result.combine(subtest())
        return result


class DefaultLoggingTest(SubTest):
    """
    Default logging Test Class
    """

    def __init__(self, test):
        SubTest.__init__(self, test, "Default System Logging",
                         "Check for Default System Logging")

    def runSubTest(self):
        """
        Validate Default logging information
        """

        if (os.path.exists('/var/log') and os.path.isfile("/var/log/messages")
            and os.stat("/var/log/messages").st_size > 0):
            print((
                "Default logging information automatically " +
                " goes to /var/log/messages file."))
            return True
        # otherwise
        print(("Error: Default logging information " +
               "does not go to /var/log/messages file."))

        return False


class NetworkAccessTest(SubTest):
    def __init__(self, test):
        SubTest.__init__(self, test, "Network Configuration",
                         "Check Firewall and Network Access")
        self.program = test.getSuiteName()
        self.allowed_open_ports = ["80", "8009", "22", "443"]
        self.rhel6_services = ["sshd", "iptables"]
        self.rhel6_service_dir = "/etc/init.d/"
        self.rhel7_services = ["sshd", "firewalld"]
        self.rhel7_service_dir = "/usr/lib/systemd/system/"
        self.rhel8_services = ["sshd", "nftables", "firewalld"]
        self.rhel8_service_dir = "/usr/lib/systemd/system/"

    def os_version(self):
        """
        Get the Operating system version
        """

        release = RedHatRelease()
        if not release.text:
            print("Error: Unsupported Operating System.")
            return False
        self.release = release.getVersion()
        return True

    def check_service_status(self):

        result = TestResult()
        validate_firewall_backend = False
        if self.release == 6:
            services = self.rhel6_services
            current_state_cmd = "service %s status"
            dead_states = ["not running", "stopped"]
            configured_state_cmd = "chkconfig --list %s"
            enabled_state = "3:on"
        elif self.release == 7:
            services = self.rhel7_services
            current_state_cmd = "systemctl status %s.service"
            dead_states = ["dead"]
            configured_state_cmd = "systemctl is-enabled %s.service"
            enabled_state = "enabled"
        elif self.release == 8:
            services = self.rhel8_services
            current_state_cmd = "systemctl status %s.service"
            dead_states = ["dead"]
            configured_state_cmd = "systemctl is-enabled %s.service"
            enabled_state = "enabled"
            validate_firewall_backend = True

        for service in services:
            print("\n*** Checking configured and current state of service: %s ***\n" % service)
            statusCommand = Command(current_state_cmd % service)
            is_service_dead = False
            for state in dead_states:
                try:
                    print(statusCommand.getString(state, singleLine=False, ignoreErrors=True))
                except HwCertCommandStringNotFound:
                    continue
                print("Error: %s service is inactive and dead." % service)
                print("Corrective Action: Please run %s service and rerun the test." % service)
                result.combine(TestResult.FAIL)
                is_service_dead = True
                break
            if not is_service_dead:
                configCommand = Command(configured_state_cmd % service)
                try:
                    configCommand.getString(enabled_state, singleLine=False, ignoreErrors=True)
                    print("%s service is enabled and running." % service)
                    result.combine(TestResult.PASS)
                except HwCertCommandStringNotFound:
                    print("Error: %s service is running but disabled." % service)
                    print("Corrective Action: Please enable %s service and rerun the test." % service)
                    result.combine(TestResult.FAIL)

        if validate_firewall_backend:
            config_path = '/etc/firewalld/firewalld.conf'
            config_parser = ConfigFile(filePath=config_path, separator='=')
            firewall_backend = config_parser.getParameter('FirewallBackend')
            if firewall_backend != 'nftables':
                msg = ("Error: Unsupported firewall backend implementation\n "
                       "Currently enabled backend: {0}\n"
                       "Corrective Action: Please set value for 'FirewallBackend' to 'nftables' in '{1}' and retry"
                       .format(firewall_backend, config_path))

                print(msg)
                result.combine(TestResult.FAIL)
        return result

    def get_service_source_package(self, service):

        if self.release == 6:
            service_dir = self.rhel6_service_dir + service
        elif self.release == 7:
            service_dir = self.rhel7_service_dir + service + "d.service"
        elif self.release == 8:
            service_dir = self.rhel8_service_dir + service + "d.service"

        cmd = "rpm -qf %s" % service_dir
        try:
            package = Command(cmd)
            package.run()
            package = package.output
            return package[0]
        except HwCertCommandException as error:
            return None

    def validate_hostname(self):
        print("\n*** Validating Hostname And FQDN ***\n")
        result = TestResult()
        hostname = Command("hostname")
        hoststr  = Command("hostname -s")
        hostfqdn = Command("hostname -f")
        try:
            if hostname.getString() == hoststr.getString():
                print ("Success: system hostname is not a FQDN and is set to short name")
            try:
                if socket.gethostbyname(hostfqdn.getString()):
                    print("Success: 'hostname -f' provides a FQDN for the host")
            except socket.gaierror as error:
                result.combine(TestResult.FAIL)
                print("Error: 'hostname -f' doens't provide a FQDN for the host")
                print(error)

        except HwCertCommandException as error:
            result.combine(TestResult.FAIL)
            print ("Error: could not determine hostname")
            print(error)
        return result

    def port_status(self):
        print("\n*** Checking Open Ports ***\n")
        result = TestResult()
        result.combine(TestResult.PASS)
        selected_ip = str()
        ips = Command("ip a | grep 'inet.*brd' | awk '{print $NF, $2 }'")
        try:
            ips.run()
            available_ips = list()
            for ip in ips.output:
                available_ips.append(ip[:-3])
            if len(available_ips) > 1:
                msg = ("Multiple interfaces are configured. "
                       "List of interfaces and associated ip addresses:\n%s\n"
                       "Please pick the System's IP address: " % (",\n".join(
                    available_ips)))
                available_ips = [ip.split(' ')[1] for ip in available_ips]
                if self.test.mode != Constants.auto:
                    if self.test.mode == Constants.server:
                        print(msg)
                    selected_ip = self.test.select(msg, available_ips)
                else:
                    print(msg)
                    # In auto/CI mode selecting the first IP address
                    selected_ip = available_ips[0]
                    print(("\nRunning nmap with %s. "
                           "To pick another IP address, please run this test "
                           "via rhcert-cli in normal mode." % selected_ip))
            else:
                selected_ip = available_ips[0].split(' ')[1]

            print("\nPicked %s\n" % selected_ip)
        except HwCertCommandException as error:
            print ("Error: could not check for IP address")
            print(error)
            return TestResult.FAIL

        nmap_cmd = "nmap -Pn %s" % selected_ip
        nmap = Command(nmap_cmd)
        try:
            nmap.run()
            output = nmap.output
            self.ports_metadata = {"ports": list()}
            error_msg = list()
            review_msg = list()
            flag = False

            for line in output:
                if not flag:
                    if not line.startswith("PORT"):
                        continue
                    else:
                        flag = True
                else:
                    if len(line) != 0:
                        params = list(filter(None, line.split(' ')))
                        if len(params) == 3:
                            self.ports_metadata["ports"].append({
                                "port": params[0].split('/')[0],
                                "state": params[1],
                                "service": params[2],
                                "package": str()})

            for port in self.ports_metadata["ports"]:
                if port["service"] == "unknown":
                    error_msg.append(
                        "\nError: Port %s is opened but no service is running on it." %
                        port["port"])
                    result.combine(TestResult.FAIL)
                else:
                    service_package = self.get_service_source_package(
                        port["service"])
                    if service_package:
                        if os.system("rpm -q %s --qf '%%{buildhost}\n\' | grep -q redhat.com" %
                                     service_package) is not 0:
                            port["package"] = service_package + \
                                              " (non-RH package)"
                            review_msg.append(
                                ("Needs Review: Service '{0}' is running "
                                 "on port '{1}'. The source rpm of service '{2}' is '{3}' "
                                 "which is non-Red Hat rpm. Please supply an explanation "
                                 "on why this service is required: ").format(
                                    port["service"], port["port"], port["service"], service_package))
                            result.combine(TestResult.REVIEW)
                        else:
                            port["package"] = service_package + " (RH package)"
                    else:
                        if port["port"] not in self.allowed_open_ports:
                            review_msg.append(
                                ("Needs Review: Service '{0}' is running "
                                 "on port '{1}'. Not able to find the source rpm of "
                                 "service '{2}'. Please supply an explanation why this "
                                 "service is required: ").format(
                                    port["service"], port["port"], port["service"]))
                            result.combine(TestResult.REVIEW)

            self.print_port_info()
            if error_msg:
                for error in error_msg:
                    print(error)
            if review_msg:
                if self.test.mode != Constants.auto:
                    for review in review_msg:
                        if self.test.mode == Constants.server:
                            print(review)
                        response = self.test.prompt(review)
                        while not response:
                            response = self.test.prompt(
                                "Please supply an explanation: ")
                else:
                    # For Constants.auto mode
                    for review in review_msg:
                        print(review)
            return result
        except HwCertCommandException as error:
            print ("Error: could not check for open ports")
            print('.'.join(error.message))
            return TestResult.FAIL

    def print_port_info(self):
        width = 15
        pkg_width = 50
        _msg = str()
        print("Displaying Open Ports Information:")
        print("NOTE: Allowed Open Ports: %s\n" % str(self.allowed_open_ports))
        separator = "+{0}+{1}+{2}+".format("-" * width,
                                           "-" * width,
                                           "-" * pkg_width)

        print(separator)
        print("|{0}|{1}|{2}|".format("Open Port".ljust(width),
                                     "Service".ljust(width),
                                     "Package(source of service)".ljust(pkg_width)))
        for port in self.ports_metadata["ports"]:
            _msg = (
                "\n|{0}|{1}|{2}|").format(
                port["port"].ljust(width),
                port["service"].ljust(width),
                port["package"].ljust(pkg_width))
            print(separator + _msg)
        print(separator + "\n")

    def runSubTest(self):
        """
        Run network access tests
        """
        result = TestResult()
        if not self.os_version():
            return result.FAIL
        subtest_list = [self.check_service_status,
                        self.port_status]
        if self.program == Programs.cloud_sap:
            subtest_list.append(self.validate_hostname)
        for subtest in subtest_list:
            result.combine(subtest())
        return result


class SubscriptionServiceTest(SubTest):
    """
    Subscription Service Test class
    """

    def __init__(self, test):
        SubTest.__init__(self, test, "Subscription Services",
                         "The root user must be able to download and " +
                         "install an rpm from the configured " +
                         "subscription management service.")
        self.pkg = "zsh"

    def install_package(self):
        """
        Installs package.
        """
        print("Installing '%s' package..." % self.pkg)
        if 'Installed' and 'Complete!' in Utility().command_execute(
                        "yum install -y " + self.pkg):
            print("Success: '%s' package is successfully installed." % self.pkg)
            return True
        else:
            print("Error: Failed to install '%s' package." % self.pkg)
            return False

    def remove_package(self):
        """
        Removes package.
        """
        print("Removing '%s' package." % self.pkg)
        if 'Removed' and 'Complete!' in Utility().command_execute(
                        "yum remove -y " + self.pkg):
            print("Success:'%s' package is successfully removed." % self.pkg)
            return True
        else:
            print("Error: Failed to remove '%s' package." % self.pkg)
            return False

    def check_yum_functionality(self):
        """
        Check yum operations.
        """
        print("*** Checking Yum Functionality ***")
        try:
            if 'No matching Packages to list' in Utility().command_execute(
                            "yum list installed " + self.pkg):
                if not self.install_package():
                    return TestResult.FAIL
                if not self.remove_package():
                    return TestResult.FAIL
            else:
                if not self.remove_package():
                    return TestResult.FAIL
                if not self.install_package():
                    return TestResult.FAIL
            print("\nSuccess: Successfully executed yum install/remove operations")
            return TestResult.PASS
        except Exception as error:
            print("\nError: Failed to execute yum install/remove operations.")
            return TestResult.FAIL

    def check_subscription(self):
        """
        Check if subscribed with rhui, if yes then print rhui repos,
        otherwise check status of subscription manager.
        Give WARNING if not subscribed with any of these.
        """
        print("\n*** Check Channel Subscriptions ***")
        result = TestResult()
        try:
            print("\nChecking for RHUI Subscription...")
            # check that there are RHUI package(s) installed
            Command("rpm -qa | fgrep -i rhui").echo()
            print("\nRHUI packages are installed. Looking for RHUI repositories...")
            configParser = ConfigParser()
            yum_repo_path = '/etc/yum.repos.d'
            rhui_repos = dict()
            files = [file for file in os.listdir(yum_repo_path)
                     if file.endswith('.repo')]

            for file in files:
                configParser.read(os.path.join(yum_repo_path, file))

                if configParser.sections():
                    url = str()
                    for section in configParser.sections():
                        if configParser.has_option(section, 'baseurl'):
                            url = configParser.get(section, 'baseurl')
                        elif configParser.has_option(section, 'mirrorlist'):
                            url = configParser.get(section, 'mirrorlist')
                        if url and 'rhui' in url.lower():
                            rhui_repos.update({section: url})
            if rhui_repos:
                print("\nSuccess: Subscribed with RHUI. Found following RHUI repositories:")
                for count, repo in enumerate(rhui_repos.items(), 1):
                    print(count, repo[0], "(", repo[1], ")")
                return result
            else:
                print("\nError: No RHUI repositories configured.")
        except HwCertCommandErrorReturned:
            print("\nError: No RHUI packages installed.")
        
        print("\nChecking for Satellite Subscription...")
        try:
            # check if katello-ca-consumer package(s) installed
            Command("rpm -qa | fgrep -i katello-ca-consumer").echo()
            print("\n'katello-ca-consumer' package is installed. Verifying subscription...")
            base_url = Command("subscription-manager config | grep baseurl")
            base_url.run()
            if base_url.output:
                if "redhat.com" not in base_url.output[0]:
                    print("\nSuccess: Subscribed with Satellite.")
                    return result
                else:
                    print("\nError: Not subscribed with Satellite. The baseurl does not look like of Satellite.")
            else:
                print("\nError: Failed to determine the baseurl for subscription manager.")
        except HwCertCommandErrorReturned:
            print("\nError: 'katello-ca-consumer' package not installed.")

        result.combine(TestResult.FAIL)
        print("Error: Neither subscribed with RHUI nor Satellite")
        print("Corrective Action: Please attach to required subscriptions.")
        return result

    def runSubTest(self):
        """
        Run Subscription service tests
        """
        result = TestResult()
        for subtest in [self.check_yum_functionality(),
                        self.check_subscription()]:
            result.combine(subtest)

        return result

