#!/usr/bin/env python
# Copyright (c) 2018 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Pooja Tiwari
#

import sys
from rhcert.tags import Constants, TestResult, Programs
from rhcert.command import Command, HwCertCommandException, HwCertCommandErrorReturned


SEPARATOR = '-' * 50


class ContainerTest(object):

    def __init__(self, container_util):
        """
        :container_util would be podman in case of CCSP
        and docker in case of OSP.
        """
        self.container_util = container_util
        self.image_name_id_mapping = dict()

    def log_available_images(self):
        """
        Log container images available on SUT.
        """
        try:
            print("\nAvailable container images:\n%s" % SEPARATOR)
            images = Command(self.container_util + " images")
            images.run()
            if images.output:
                for image in images.output[1:]:
                    image = image.split()
                    self.image_name_id_mapping[image[0] + ':' + image[1]] = image[2]
        except HwCertCommandException:
            print("Warning: could not determine list of available images.")
            return TestResult.WARN

        return TestResult.PASS

    def log_available_containers(self, program):
        """
        Log running/stopped containers  available on SUT.
        """
        available_containers = list()
        print("\nAvailable containers:\n%s" % SEPARATOR)
        try:
            cmd = self.container_util + " ps -a --format \"{{.ID}}          {{.Image}}          {{.Names}}\""
            containers = Command(cmd)
            containers.run()
        except HwCertCommandException:
            print("Error: could not determine list of available containers.")
            return available_containers, TestResult.FAIL
        if not containers.output:
            msg = "No container exists for testing"
            if program == Programs.openstack:
                print("Error: %s" % msg)
                return available_containers, TestResult.FAIL
            else:
                print(msg)
                return available_containers, TestResult.PASS
        else:
            available_containers = containers.output
            return available_containers, TestResult.PASS

    def log_image_inspect(self, available_containers):
        """
        Log inspect output for all containers images.
        """
        containers_and_images_info = list()
        if not available_containers:
            return containers_and_images_info, TestResult.FAIL
        result = TestResult()
        err_msg = "Error: could not determine configuration info for image: %s."
        print(
            "\nLogging $%s inspect for all containers images:\n%s" %
            (self.container_util, SEPARATOR))
        # look for container uuid, image and name
        for container_data in available_containers:
            container = container_data.split()
            if container:
                if container[0].isalpha():
                    continue
                '''
                container[1] stores the image identifier. It could be
                in one of the following format:
                1. id
                2. name:tag
                3. name
                Please note that id or name, both can hold an alphanumeric value
                '''
                image_id = container[1]
                # in case its not an id
                if container[1] not in self.image_name_id_mapping.values():
                    # 3rd format i.e name without tag,
                    # adding 'latest' as default tag.
                    if ':' not in container[1]:
                        container[1] = container[1] + ":latest"
                    orphan_container = True
                    if self.image_name_id_mapping.get(container[1]):
                        orphan_container = False
                        image_id = self.image_name_id_mapping[container[1]]
                try:
                    failure = False
                    _python = 'python'
                    if sys.version.startswith("3"):
                        _python = 'python3'
                    cmd = (
                        self.container_util + " inspect %s | %s -m json.tool" %
                        (image_id, _python))
                    inspect_out = Command(cmd)
                    inspect_out.run()
                except Exception as err:
                    failure = True
                    result.combine(TestResult.FAIL)
                    print(err_msg % container[1])
                    print(str(err))
                    if orphan_container:
                        print("Corrective Action: The container with id '%s' belongs to a non-existent "
                              "image on this system. Such containers can not be verified. "
                              "Please resolve this problem and re-run the test."
                              % container[0])
                    continue
                if not failure:
                    inspect_output = inspect_out.get_json_output()[0]
                    vendor = str()
                    authoritative_url = str()
                    url = str()
                    container_config = inspect_output.get('ContainerConfig') or inspect_output.get('Config')
                    if container_config.get('Labels'):
                        labels = container_config.get('Labels')
                        authoritative_url = labels.get('authoritative-source-url') or labels.get('Authoritative_Registry')
                        vendor = labels.get('vendor')
                        url = labels.get('url')
                        # url value looks something like this
                        # 'https://access.redhat.com/containers/#/registry.access.redhat.com/rhosp16/openstack-manila-share/images/16.1-47'
                        if not authoritative_url and url:
                            try:
                                authoritative_url = [ registry for registry in url.split("/") if registry.startswith("registry")][0]
                            except IndexError:
                                authoritative_url = ""

                    rootfs_layers = inspect_output.get('RootFS', {}).get('Layers') or inspect_output.get('Rootfs')
                    containers_and_images_info.append(
                        {
                            "container": container[2],
                            "image": container[1],
                            "image_metadata": {
                                'url': url or str(),
                                "registry": str(),
                                "authoritative-source-url": authoritative_url or str(),
                                "vendor": vendor or str(),
                                "RootFSLayers": rootfs_layers or list()}})

        if not containers_and_images_info or result == TestResult.FAIL:
            print(
                "Error: Failed to determine the required information to proceed further.")
            return containers_and_images_info, TestResult.FAIL
        else:
            self.container_table = str()
            width = 30
            name_width = 35
            image_width = 80
            msg = str()
            separator = "+{0}+{1}+{2}+{3}+".format("-" * name_width,
                                                   "-" * image_width,
                                                   "-" * width,
                                                   "-" * width)
            self.container_table = self.container_table + separator + '\n' + (
                "|{0}|{1}|{2}|{3}|".format(
                    "Container".ljust(name_width),
                    "Image".ljust(image_width),
                    "Authoritative Source Url".ljust(width),
                    "Vendor".ljust(width)))
            for info in containers_and_images_info:
                self.container_table = self.container_table + '\n' + separator + (
                    "\n|{0}|{1}|{2}|{3}|").format(
                    info["container"].ljust(name_width),
                    info["image"].ljust(image_width),
                    info["image_metadata"]["authoritative-source-url"].ljust(
                        width),
                    info["image_metadata"]["vendor"].ljust(width))
            self.container_table = self.container_table + '\n' + separator + "\n"
            return containers_and_images_info, TestResult.PASS

    def detect_nonRH_containers(self, containers_and_images_info, test):
        result = TestResult()
        rh_containers = list()
        non_rh_containers = list()
        print("\n\nContainers and their images, authoritative-source-url and vendor:")
        print(self.container_table)
        container_registries = ['registry.access.redhat.com', 'registry.redhat.io']
        # look for container uuid, image and name
        for container_image_info in containers_and_images_info:
            # Check if container image belongs to Red Hat
            if any(source_url in container_image_info["image_metadata"]["authoritative-source-url"]
                   for source_url in container_registries):
                # catalog/connect filtering is required for Openstack only
                if test.getSuiteName() != Programs.openstack:
                    rh_containers.append(container_image_info)
                    continue
                # check if container belongs to catalog
                if 'Red Hat' in container_image_info[
                        "image_metadata"]["vendor"]:
                    container_image_info["image_metadata"][
                        "registry"] = "catalog"
                    rh_containers.append(container_image_info)
                # check if container belongs to connect
                else:
                    container_image_info["image_metadata"][
                        "registry"] = "connect"
                    rh_containers.append(container_image_info)
            # non-RH container
            else:
                non_rh_containers.append(container_image_info)

        if non_rh_containers:
            msg = "\nNeeds Review: Following containers do not appear to be from Red Hat.:\n%s\nPlease provide explanation on their need:\n" % (
                ",\n".join([container["container"] for container in non_rh_containers]))

            if test.mode != Constants.auto:
                if test.mode == Constants.server:
                    print(msg)
                response = str()
                while not response:
                    response = test.prompt(msg)
                result.combine(TestResult.REVIEW)
            else:
                print(msg)
                result.combine(TestResult.REVIEW)

        if not rh_containers:
            print (
                "\nError: none of the container appears to be Red Hat provided."
                "Only Red Hat recognized containers can be considered for testing.")
            return rh_containers, TestResult.FAIL

        print(
            "\nFollowing containers have been recognized by Red Hat:\n%s" %
            SEPARATOR)
        count = 1
        for container in rh_containers:
            print(count, container["container"])
            count += 1

        return rh_containers, result


