#!/usr/bin/env python
# Copyright (c) 2020 Red Hat, Inc. All rights reserved.
# This copyrighted material is made available to anyone
# wishing to use, modify, copy, or redistribute it subject
# to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.
#
# Author: Zahid Ajaz <zajaz@redhat.com>

import os
import re

from rhcert.tags import TestResult
from rhcert.test import SubTest
from rhcert.command import Command, HwCertCommandException, HwCertCommandStringNotFound
from rhcert.tags import Constants


class FingerprintTest(SubTest):

    def __init__(self, test):
        SubTest.__init__(self, test, "Fingerprint",
                         "Generate Application Software Fingerprint")

    # def update_service_status(self, services):
    #     services_status_updated = list()
    #     for service in services:
    #         state = Command("systemctl is-enabled {}".format(service)).getIgnoreErrors()
    #         sub = Command("systemctl is-active {}".format(service)).getIgnoreErrors()
    #         services_status_updated.append({'name': service,
    #                                         'state': state[0] if state else None,
    #                                         'sub': sub[0] if sub else None})
    #     return services_status_updated
    def collect_feature_data(self):
        feature_data = [{"suspect_services": self.suspect_services,
                        "confirmed_services": self.confirmed_services,
                        "suspect_rpms": self.suspect_process_rpms,
                        "confirmed_rpms": self.confirmed_process_rpms,
                        "suspect_processes": self.suspect_process_executables,
                        "confirmed_processes": self.confirmed_processes
                        }]
        self.logFeatureData("ISV Fingerprint" , feature_data)
 
    def redhat_provided(self, executable_meta):
        if "redhat.com" in executable_meta:
            return True
        return False

    def get_exectuable_meta(self, executable):
        cmd = Command(
            "rpm -qf %s --qf '%%{RPMTAG_NAME},%%{RPMTAG_BUILDHOST}\n'" % executable)
        try:
            return cmd.getString()
        except HwCertCommandException as error:
            return None

    def get_alternative_executable(self, path):
        try:
            cmd = Command("alternatives --display {} | grep '^/'".format(os.path.basename(path)))
            cmd.run()
            return [line.split(" - ")[0] for line in cmd.output or []][0]
        except HwCertCommandException as error:
            return

    def get_process_from_pid(self, pname):
        try:
            pidof = Command("pidof {}".format(pname)).getString()
            pid = pidof.split()[0]
            proc = Command("file /proc/{}/exe".format(pid)).getString()
            return re.search('[\/\w]+$', proc).group(0)
        except HwCertCommandStringNotFound as error:
            return

    def get_executable(self, pname):
        try:
            return Command("which {}".format(pname)).getString()
        except HwCertCommandStringNotFound as error:
            return self.get_process_from_pid(pname)

    def special_process(self, process):
        """
        Identify kernel/special processes
        """
        if re.match(r"^\[[\w&.\/\-\+\:]*\]$", process) or re.match(r"^\([\w&.\/\-\+\:]*\)$", process):
            return True
        return False

    def find_suspect_processes(self):
        """
        This method isolates processes spun by partner application software
        """
        print("\nInspecting running processes ['ps aux']...")
        processes = list()
        try:
            plist = Command("ps aux |sed '1d'| awk '{print $11}'")
            plist.run()
            for process in plist.output:
                process = re.sub(r"(^\-)|(\:$)", "", process)
                if not self.special_process(process):
                    processes.append(process)
        except HwCertCommandException as error:
            print("\nError: Could not determine the list of processes. {}".format(error))
        self.suspect_process_rpms = list()
        self.suspect_process_executables = list()
        for process in set(processes):
            if process.startswith("/"):
                if not os.path.exists(process):
                    continue
                executable = process
            else:
                executable = self.get_executable(process)

            executable_meta = self.get_exectuable_meta(executable) if executable else None
            if executable_meta:
                if self.redhat_provided(executable_meta):
                    continue
                else:
                    self.suspect_process_rpms.append(executable_meta)
            else:
                alternative_executable = self.get_alternative_executable(process)
                executable_meta = self.get_exectuable_meta(alternative_executable) if alternative_executable else None
                if executable_meta:
                    if self.redhat_provided(executable_meta):
                        continue
                    else:
                        self.suspect_process_rpms.append(executable_meta)
                else:
                    self.suspect_process_executables.append(process)

        self.confirmed_processes = self.test.multiSelect("Please select the processes/executables that belong to your application software", [
                                                  executable for executable in self.suspect_process_executables]) if self.suspect_process_executables else None
        
        self.confirmed_process_rpms = self.test.multiSelect("Please select the rpm(s) that belong to your application software", [
                                                  meta.split(",")[0] for meta in self.suspect_process_rpms]) if self.suspect_process_rpms else None

    def suspect_service(self, service):
        service = re.sub("@", "@1", service)  # indexing unit services
        fragment_path = Command(
            "systemctl show {} | grep 'FragmentPath'".format(service)).getIgnoreErrors()
        exec_path = re.sub("^FragmentPath=", "", fragment_path[0]) if fragment_path else None
        if exec_path and exec_path != "/dev/null":
            executable_meta = self.get_exectuable_meta(exec_path) or ""
            if not self.redhat_provided(executable_meta):
                return True
        return False
    
    def find_suspect_system_services(self):
        """
        This method isolates system services run by partner application software
        """
        print("\nInspecting system services...")
        try:
            cmd = Command("systemctl list-unit-files --type service --all|sed '1d'| awk '{print $1}'")
            cmd.run()
        except HwCertCommandException as error:
            print("\nError: Could not determine the list of system services.")
            return
        self.suspect_services = list()
        for line in cmd.output:
            if line.endswith(".service") and self.suspect_service(line):
                self.suspect_services.append(line)
        self.confirmed_services = self.test.multiSelect("Please select the services that belong to your application software", [
                                                  service for service in self.suspect_services]) if self.suspect_services else None

    def runSubTest(self):
        """
        Cases with their return values:
        """
        result = TestResult()
        service_identified = False
        process_identified = False
        if self.test.promptConfirm("\nPlease ensure that your application software is running:"):
            self.find_suspect_system_services()
            if self.confirmed_services:
                msg = "\nPlease confirm if following command is the correct way to determine if the application under test is running:\n'systemctl status {}'".format(
                        " ".join(self.confirmed_services))
                if self.test.ui.promptConfirm(msg):
                    service_identified = True

            self.find_suspect_processes()
            if self.confirmed_process_rpms:
                msg = "\nPlease confirm if following command is the correct way to determine if your application is installed:\n'rpm -q {}'".format(
                        " ".join(self.confirmed_process_rpms))
                if self.test.ui.promptConfirm(msg):
                    process_identified = True

            if self.confirmed_processes and not service_identified:
                msg = "\nPlease confirm if following is the correct process to identify if your application is running:\n'ps aux | grep {}'".format(
                        " ".join(self.confirmed_processes))
                if self.test.ui.promptConfirm(msg):
                    process_identified = True
            self.collect_feature_data()
        if not (service_identified or process_identified):
            explanation = self.test.prompt("\nHow can a customer confirm that their software is installed and running ?[Please enter your response below]\n")
            if explanation:
                print("Needs Review: {}".format(explanation))
                result.combine(TestResult.REVIEW)
            else:
                print("Error: No explanation provided")
                result.combine(TestResult.FAIL)
        return result
