#!/usr/bin/env python
# Copyright (c) 2016 Red Hat, Inc. All rights reserved.
# This copyrighted material is made available to anyone
# wishing to use, modify, copy, or redistribute it subject
# to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.
#
# Author: Pooja Tiwari <ptiwari@redhat.com>
import os
from copy import deepcopy

from rhcert.tags import Constants, TestResult, Programs
from rhcert.test import SubTest
from rhcert.command import Command, HwCertCommandErrorReturned, HwCertCommandErrorOutput, HwCertCommandException
from rhcert.suites.rhcert.rpm_verification import RPMVerification
from rhcert.suites.rhcert.rpm_validation import RPMValidation

SEPARATOR = "-"*50


class InstalledRPMs(SubTest):

    def __init__(self, test, resultSeverityLimit=TestResult.FAIL):
        SubTest.__init__(self, test, "Installed RPMs",
                         "Check Red Hat Software RPMs")
        self.packages = list()
        self.non_rh_pkgs = list()
        self.program = test.getSuiteName()
        self.result = TestResult()
        self.output_directory = self.test.getOutputDirectory()
        self.resultSeverityLimit = TestResult(resultSeverityLimit)
        self.rpm_error_logs = list()

    def get_installed_packages(self):
        """
        Get details of installed packages.
        """

        rpm = Command("rpm -qa")
        rpm.run()
        packages = rpm.output
        self.packages = deepcopy(packages)

    def verify_rh_packages(self):
        """
        Verify each Red Hat package with command 'rpm -V'.
        """
        result = TestResult()
        result.combine(TestResult.PASS)
        print("\n*** Red Hat Package Verification ***")
        print("""\nINFO: Test package integrity. If anything has changed, it outputs with following symbols:
            S file Size differs
            M Mode differs (includes permissions and file type)
            5 MD5 sum differs
            D Device major/minor number mismatch
            L readLink(2) path mismatch
            U User ownership differs
            G Group ownership differs
            T mTime differs
            """)
        del RPMVerification.rpm_verificaton_log[:]
        for package in self.packages:
            if package in self.non_rh_pkgs:
                continue
            rpm_verify_silent = Command(
                "rpm -V > /dev/null " + package)
            rpm_verify = Command("rpm -V " + package)
            try:
                rpm_verify_silent.run()
            except HwCertCommandErrorReturned:
                result.combine(self.report_rpm_verification_failures(rpm_verify,
                                                                     package))
            except HwCertCommandErrorOutput as stderr:
                result.combine(self.report_rpm_verification_failures(rpm_verify,
                                                                     package,
                                                                     stderr._get_message()))
            except HwCertCommandException as err:
                print("Could not perform verification on package '%s'" %package)
                print(err)
                result.combine(TestResult.FAIL)

        if result == TestResult.PASS:
            print("\nVerified all packages.\n")
        else:
            print("\n%s Some rpm verification failures have been encountered during testing. \n"
                  % (TestResult(result).getMessagePrefix()))
            RPMVerification.store_failures_into_a_file(self.output_directory, self.rpm_error_logs)
            print("\nCorrective Action: Please re-install these packages.\n")
        return result

    def report_rpm_verification_failures(self, rpm_verify, package, stderr=None):
        result = TestResult()
        result.combine(TestResult.PASS)

        errors = rpm_verify.getIgnoreErrors()
        for error in errors[:]:
            if error.startswith("missing") or error.endswith(".pyc") or error.endswith(".pyo"):
                errors.remove(error)
        if errors:
            RPMVerification.rpm_verificaton_log.append(str(SEPARATOR))
            RPMVerification.rpm_verificaton_log.append(str("\nVerification failed for '%s'" % (package)))
            RPMVerification.rpm_verificaton_log.append(str(SEPARATOR))
            verification_test = RPMVerification(self.program, errors, self.output_directory)
            result.combine(verification_test.rpm_verify())
            if stderr:
                result.combine(TestResult.REVIEW)
            if result == TestResult.PASS:
                RPMVerification.rpm_verificaton_log.append(str("INFO: No critical failure occurred."))
            elif stderr:
                for line in stderr:
                    RPMVerification.rpm_verificaton_log.append(str("Needs Review: (stderr) %s" %line))

            # removing blank failures from the run log
            if len(RPMVerification.rpm_verificaton_log) > 3:
                self.rpm_error_logs.extend(RPMVerification.rpm_verificaton_log)
            RPMVerification.rpm_verificaton_log = []
        # impose results severity limit
        return result if result.better(self.resultSeverityLimit) else self.resultSeverityLimit

    def check_non_rh_packages(self):
        """
        Test to check if non redhat packages are present.
        Returns the test status.
        """
        print("\n*** Test presence for non-Red Hat packages ***\n")
        validation_test = RPMValidation(self)        
        print("\n1. Checking rpm build host information to isolate non-redhat packages.\n")
        self.non_rh_pkgs = validation_test.get_non_rh_packages_via_build_host()
        if self.non_rh_pkgs:
            msg = ("Needs Review: Following non-Red Hat packages are installed:\n%s\n"
                   "Please describe the need for these packages: " % (
                       "\n".join(self.non_rh_pkgs)))

            if self.test.mode != Constants.auto:
                if self.test.mode == Constants.server:
                    print(msg)
                response = self.test.prompt(msg)
                while not response:
                    response = self.test.prompt("Please describe the need for these packages: ")
            else:
                # For Constants.auto mode
                print(msg)
            self.result.combine(TestResult.REVIEW)
        
        print("\n2. Check that the installed RPMs are from the RedHat product(s) available in the offering.\n")
        packages, result, message = validation_test.get_non_rh_packages_via_offerings()
        if not packages:
            print(message)
            self.result.combine(result)

        if packages.get('non_red_hat_packages'):
            if set(packages['non_red_hat_packages']).issubset(self.non_rh_pkgs):
                msg = ("\nWarning: Non Red Hat packages found:\n%s\n" % ("\n".join(packages['non_red_hat_packages'])))
            else:
                extra_non_rh_pkgs = list(set(self.non_rh_pkgs).symmetric_difference
                                         (set(packages['non_red_hat_packages'])))
                msg = ("\nNeeds Review: Following packages are not available from the "
                       "RedHat product(s) in the offering:\n%s\n" % ("\n".join(extra_non_rh_pkgs)))
            print(msg)
            self.non_rh_pkgs = list(set().union(*[self.non_rh_pkgs, packages['non_red_hat_packages']]))
            self.result.combine(TestResult.REVIEW)
        else:
            print("RedHat product service did not report any non-redhat packages")
        
        if not self.non_rh_pkgs:
            print("Passed: All installed packages are from Red Hat product.")
            self.result.combine(TestResult.PASS)

        else:
            print("\n*** Listing RPM metadata for non-redhat packages***\n""")
            for pkg in self.non_rh_pkgs:
                print(SEPARATOR)
                print("\nMetadata for package %s" % pkg)
                print(SEPARATOR)
                rpm_meta_data = Command("rpm -qli --scripts " + pkg)
                try:
                    rpm_meta_data.run()
                except Exception as e:
                    print("Failed to generate metadata for package %s" % pkg)
        
        return self.result

    def collect_feature_data(self):
        installed_packages_NVRA = self.get_NVRA_for_installed_packages()
        if installed_packages_NVRA:
            feature_data = list()
            for package in installed_packages_NVRA:
                package_components = package.split(' ')
                feature_data.append(
                    {"name": package_components[0],
                     "version": package_components[1],
                     "release": package_components[2],
                     "arch": package_components[3]})
            self.logFeatureData("Installed RPMs" , feature_data)

    def get_NVRA_for_installed_packages(self):
        try:
            installed_packages = Command('rpm --queryformat "%{NAME} %{VERSION} %{RELEASE} %{ARCH}\n" -qa')
            installed_packages.run()
            installed_packages_NVRA = installed_packages.output
            return installed_packages_NVRA
        except HwCertCommandException as err:
            print("Warning: Could not determine NVRA for installed packages")

    def runSubTest(self):
        """
        RPM validation and verificaion.
        """
        result = TestResult()
        self.get_installed_packages()
        for subtest in [self.check_non_rh_packages(),
                        self.verify_rh_packages()]:
            result.combine(subtest)

        self.collect_feature_data() 
        return result

def unit_test():
    pass
