#!/usr/bin/env python
# Copyright (c) 2018 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
# PeerDemo test - demonstrate one SUT calling another peer SUT to run a test, then getting the result.
"""
   SUT->SUT Peer Test Demo

   Outline:
   1) Request Peer SUT hostname/IP from user
   2) Check and wait if it's ready
   3) request a test run (rhcert/self_check in this demo)
   4) Check and wait for it to be ready
   5) Request the Peer's results document and check the result
   6) Save the Peer results to the log so they're merged in
"""
import os
import time

from rhcert.environment import Environment
from rhcert.tags import Constants, TestTag, TestResult, Programs, Tags
from rhcert.command import Command, HwCertCommandException, HwCertCommandErrorReturned, HwCertCommandErrorOutput
from rhcert.test import Test, SubTest
from rhcert.client.client import RHCertClient
from rhcert.model.resultsdocument import ResultsDocument
from rhcert.host import Host


class PeerTest(Test):
    """ base class for running tests from one SUT to another peer SUT """

    def __init__(self):
        Test.__init__(self, "rhcert/peer")
        self.environment = Environment()
        self.timeout = 120 # sec to wait for a peer run

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification]

    def plan(self, planner):
        tests = list()
        tests.append(self.makeCopy())
        return tests

    def run(self):
        result = TestResult()
        hostname = self.prompt("Please enter the hostname or IP address of the peer: ")
        self.peerHost = Host(hostname)
        self.peerProxy = RHCertClient(self.peerHost)

        for subtest in [RequestTestRun(self, "rhcert/self_check"), GetResults(self)]:
            result.combine(subtest.run())

        return result

    def runCLI(self, rhcertCLICommand):
        if not self.peerProxy.waitUntilReady():
            return TestResult.FAIL
        result = self.peerProxy.runCLI(rhcertCLICommand + " --debug low --mode auto")
        if not result:
            print("Error: requesting CLI command!")
            return TestResult.FAIL
        return TestResult.PASS


class RequestTestRun(SubTest):
    def __init__(self, test, testPath):
        SubTest.__init__(self, test, "Request Test Run", "Request the peer host run a test")
        self.testPath = testPath

    def runSubTest(self):
        result = TestResult()
        print("clearing peer test results")
        if self.test.runCLI("rhcert-cli clean") == TestResult.FAIL:
            result.combine(TestResult.WARN)
        print("Requesting test run")
        result.combine(self.test.runCLI("rhcert-cli run --test " + self.testPath))

class GetResults(SubTest):
    def __init__(self, test):
        SubTest.__init__(self, test, "Get Peer Results", "Get results from a peer")

    def runSubTest(self):
        if not self.test.peerProxy.waitUntilReady():
            return TestResult.FAIL
        print("Getting Peer Results")
        peerResultsDocument = self.test.peerProxy.getResults()
        if not peerResultsDocument:
            print("Error: no results document")
            return TestResult.FAIL
        selfCheckTest = peerResultsDocument.getTest("rhcert/self_check")
        if not selfCheckTest:
            print("Error: could not find peer test run")
            return TestResult.FAIL
        run = selfCheckTest.getLastRun()
        if not run:
            print("Error: test didn\'t run")
            return TestResult.FAIL
        print("Peer test run result: %s" % run.getSummary())
        # save results to log directory - they'll get automatically merged in
        peerResultsPath = os.path.join(self.test.getOutputDirectory(),
                                    peerResultsDocument.getVerboseFileName())
        print("saving results to %s" % peerResultsPath)
        peerResultsDocument.save(peerResultsPath)
        return TestResult(run.getSummary())

