#!/usr/bin/env python
# Copyright (c) 2016 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
import os, sys, string, shutil, re, json, rpm

from rhcert.controller import Controller
from rhcert.tags import Constants, TestTag, VersionType, TestResult, Programs
from rhcert.command import Command, HwCertCommandException, HwCertCommandErrorReturned
from rhcert.redhatrelease import RedHatRelease
from rhcert.environment import Environment
from rhcert.test import Test, SubTest
from rhcert.services.microservices import ValidKernels

import rhcert.version


class KernelTest(Test):

    def __init__(self, path):
        Test.__init__(self, path)
        self.mandatory = Constants.post # required at the end of each run
        # what RHEL are we testing?
        self.redHatRelease = RedHatRelease()
        self.priority = 1000 # run last
        self.techPreviewModuleWhitelist = ["fscache"]
        self.environment = Environment()
        # which program are we testing?
        self.program = self.getSuiteName()

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification]

    def plan(self, planner):
        tests = list()
        tests.append(self.makeCopy())
        return tests

    def __getTaintedModules(self, sign):
        pattern = re.compile("^(?P<mod_name>\w+)[\s\S]+\((?P<signs>[A-Z]+)\)")
        procModules = open("/proc/modules")
        modules = list()
        for line in procModules.readlines():
            match = pattern.match(line)
            if match:
                if sign in match.group("signs"):
                    modules.append(match.group("mod_name"))
        procModules.close()
        return modules

    def checkKernelPackageVersion(self):
        print("\n*** Check that the running kernel is released by Red Hat for use with the RHEL version  ***\n")
        release = self.redHatRelease
        # re-check via uname and PNS
        if self.redHatRelease.updateProductFromService():
            print ("Product Information:")
            print ("-------------------------------")
            print("Name: %s" % self.redHatRelease.getProduct())
            print("ShortName: %s" % self.redHatRelease.getProductShortName())
            print("Version: %s %s" % (self.redHatRelease.getVersionPointUpdate(), self.redHatRelease.getLabel()))
        else:
            print("Needs Review: Need to verify if installed kernel: %s belongs to Red Hat." % release.getKernelRPMName())
            return TestResult.REVIEW
        # if kernel is RT
        if self.redHatRelease.getKernelFlavor() == Constants.rt:
            if self.redHatRelease.getVersion() < 7:
                if self.redHatRelease.getRealTimeRelease():
                    release = self.redHatRelease.getRealTimeRelease()
                # else, warn - disable check
                else:
                    print("Warning: the running kernel is a realtime kernel, but product release information could not be found")
                return TestResult.PASS

        # try to call home
        return self.__checkKernelPackageVersionViaService(self.redHatRelease.getProductShortName(), release)


    def kernel_test_status(self, installed_kernel, kernel_data, via_service = True):
        """
        This method decides the test status based on program and kernel-type
        and returns it.

        installed_kernel: installed kernel package NVR.
        kernel_data: list of ga/candidate/update kernel packages for a given release.
        via_service: source of kernel_data. Default value is True. If data
                     comes from cache then it should be set to False.
        """

        # Installed_kernel is GA. PASS for every program.
        if installed_kernel in kernel_data["ga"]:
            print(("The installed kernel package, {0}, is a GA.").format(installed_kernel))
            return TestResult.PASS

        # Installed_kernel is a candidate.
        if installed_kernel in kernel_data["candidate"]:
            msg = ("The installed kernel package, {0}, is a candidate.").format(
                installed_kernel)
            # REVIEW for every program if GA does not exist for the release.
            if not kernel_data["ga"]:
                print(("Needs Review: {0} GA does not exist for the release.").format(msg))
                return TestResult.REVIEW
            # GA exists for the release.
            else:
                _msg = "It does not meet the certification requirements."
                # FAIL for every program if source is service.
                if via_service:
                    print(("Error: {0} {1}").format(msg, _msg))
                    return TestResult.FAIL
                # REVIEW for every program if source is cache.
                else:
                    print(("Needs Review: {0}(found via cache). {1} " \
                           "Please verify it via service.").format(msg, _msg))
                    return TestResult.REVIEW

        # Installed kernel is an update.
        msg = ("The installed kernel package, {0}, is an update").format(installed_kernel)
        # Installed kernel belongs to Z stream.
        if installed_kernel in kernel_data["update"]["Z"]:
            # REVIEW for hardware.
            if self.program in [Programs.hardware, Programs._hardware]:
                print(("Needs Review: {0}(Z).").format(msg))
                return TestResult.REVIEW
            # PASS for others
            else:
                print(("{0}(Z).").format(msg))
                return TestResult.PASS

        # Installed kernel is not associated with any stream,
        # but belongs to Red Hat. Only true for RHELMRG 5/6
        elif installed_kernel in kernel_data["update"]["Other"]:
            print(msg)
            return TestResult.PASS

        else:
            stream = str()
            # Installed kernel belongs to LL stream.
            if installed_kernel in kernel_data["update"]["LL"]: stream = "LL"
            # Installed kernel belongs to AUS stream.
            elif installed_kernel in kernel_data["update"]["AUS"]: stream = "AUS"
            # Installed kernel belongs to EUS stream.
            elif installed_kernel in kernel_data["update"]["EUS"]: stream = "EUS"

            if stream:
                # FAIL for openstack and hardware
                if self.program in [Programs.hardware, Programs._hardware, Programs.openstack, Programs.baremetal]:
                    print(("Error: {0}({1}). It does not meet certification requirements").format(msg, stream))
                    return TestResult.FAIL
                # REVIEW for others
                else:
                    print(("Needs Review: {0}({1}).").format(msg, stream))
                    return TestResult.REVIEW

        # Installed kernel does not belong to Red Hat.
        msg = ("The installed kernel package, {0}, does not belong to Red Hat. " \
               "It does not appear to be a valid kernel for certification.").format(installed_kernel)
        # FAIL for every program if source is service.
        if via_service:
            print(("Error: {0}").format(msg))
            return TestResult.FAIL
        # REVIEW for every program if source is cache.
        else:
            print(("Needs Review: {0}(found via cache). " \
                   "Please verify it via service.").format(msg))
            return TestResult.REVIEW

    def __checkKernelPackageVersionViaService(self, rhelShortName, release):
        majorDotMinorVersion = ("%s.%s %s" % (release.getVersion(), release.getUpdate(), release.getLabel())).strip()
        product = "%s-%s" % (rhelShortName, majorDotMinorVersion)
        installed_kernel = release.getKernelRPMName()
        print("Installed Kernel Package Version: '%s'." % installed_kernel)
        print("Fetching Versions for Red Hat Kernel Packages ... ")
        validKernelService = ValidKernels(product, self.environment)
        response = validKernelService.call()
        via_service = True

        if not response or len(response.json()) != 3:
            via_service = False
            msg = ("Needs Review: Needs to verify if running kernel, %s ,is released by Red Hat for use with the '%s'"
                   % (installed_kernel, product))
            kernel_json = validKernelService.getCache()
            if kernel_json:
                # Using cached kernel version info
                print("Using local kernel information ... ")
                kernelVersions = json.loads(kernel_json)
                productName = rhelShortName.upper()
                productVersion = (majorDotMinorVersion.replace(release.getLabel(), "")).strip()  # true version
                if not kernelVersions.get(productName):
                    print(("Could not find kernel information for %s.\n%s" % (productName, msg)))
                    return TestResult.REVIEW
                if not kernelVersions[productName].get(productVersion):
                    print(("Could not find kernel information for %s.\n%s" %(product, msg)))
                    return TestResult.REVIEW
                kernelVersions = kernelVersions[productName][productVersion]
            else:
                print(msg)
                return TestResult.REVIEW
        else:
            kernelVersions = response.json()

        return self.kernel_test_status(installed_kernel,
                                       kernelVersions,
                                       via_service)


    def __printKernelVersionError(self, installedKernel, kernels):
        """ try reverse-map installed to kernelType for better message """
        kernelType = None
        for kernelType in ["ga", VersionType.update, VersionType.candidate]:
            if installedKernel in kernels.get(kernelType):
                break
        if kernelType:
            print("Error: installed %s is a %s kernel, and does not meet certification requirements" % (installedKernel, kernelType))
        else:
            print("Error: installed kernel does not appear to be a valid kernel version for certification.")


    def __checkModuleVenderAndBuildHost(self, module):
        result = True
        goodVendorList = ["Red Hat", "Red Hat, Inc."]
        warnVendorList = ["Fedora", "Fedora Project"]
        try:
            print("Module '%s' is not from running kernel package. Checking in Red Hat Driver Update" % str(module))
            moduleFile = self.__getModuleFile(module)
            if not moduleFile:
                return False

            vendor = Command("rpm -qf %s --qf %%{VENDOR}" % moduleFile).getString()
            if vendor in goodVendorList:
                pass
            elif vendor in warnVendorList:
                print("Warning: '%s' is '%s' module not Red Hat." % (module, vendor))
            else:
                print("Error: Non-Red Hat driver vendor '%s' for module '%s'" % (vendor, module))
                result = False

            buildhost = Command("rpm -qf %s --qf %%{BUILDHOST}" % moduleFile).getString()
            if "redhat.com" not in buildhost:
                print("Error: Driver module '%s' is built on '%s' not built at Red Hat." % (module, buildhost))
                result = False
        except HwCertCommandException as exception:
                print("Error: could not determine driver packaging info for module %s" % module)
                print(exception)
                result = False

        return result

    def __readABIWhitelist(self, whitelist):
        '''Read in the ABI whitelist'''
        symbols = list()

        if not os.path.isfile(whitelist):
            print("Error: Cannot read whitelist file")
            return None

        whitelistFile = open(whitelist,"r")

        while True:
            line = whitelistFile.readline()

            if line == "":
                break
            if line == "\n":
                continue
            line.split()
            if line[0] == '[':
                group=line[1:-2]
                continue
            symbol=line.strip()
            symbols.append(symbol)

        return symbols

    def __readSymbols(self, symbols):

        moduleSymbols = list()

        for symbol in symbols:
            foo,moduleSymbol = symbol.split()
            moduleSymbols.append(moduleSymbol)

        return moduleSymbols

    def __readModule(self, module):
        '''Read in symbols from a module'''

        symbols = list()
        moduleFile = self.__getModuleFile(module)

        if not moduleFile:
            print("Error: Can not find module file for %s" % module)
            return None
        # otherwise
        if not os.path.isfile(moduleFile):
            print("Error: Cannot read module file %s" % moduleFile)
            return None

        if moduleFile[-2:] == "ko":
            nm = os.popen('modprobe --dump-modversions ' + moduleFile)
        else:
            nm = open(moduleFile,"r")

        while True:
            line = nm.readline()
            if line == "":
                break

            symbols.append(line)
        nm.close()

        return self.__readSymbols(symbols)

    def __getModuleFile(self, module):
        try:
            modInfoSrcVersion = Command("modinfo -F srcversion %s" % module).getString()
            sysSrcVersion = open("/sys/module/%s/srcversion" % module).readline().strip()
            if sysSrcVersion == modInfoSrcVersion:
                moduleFile =Command("modinfo -F filename %s" % module).getString()
                if os.path.islink(moduleFile):
                    moduleFile = os.readlink(moduleFile)
                return moduleFile
            else:
                print("Error: module source version mismatch: ")
                print("   /sys/modules/%s/srcversion: %s" % (module, sysSrcVersion))
                print("   modinfo -F srcversion: %s" % modInfoSrcVersion)

        except HwCertCommandException as e:
            print("Error: could no find module file for %s:" % module)
            print(e)
            e.command.printErrors()

        return None


    def __abiCheckSymbols(self, whitelistSymbols, moduleSymbols):
        '''Check ABI compliance of a set of symbols.'''

        extraSymbols = list()

        for symbol in moduleSymbols:
            if symbol not in whitelistSymbols:
                extraSymbols.append(symbol)

        return extraSymbols


    def __abiCheck(self, module):
        # skip abi checks for rt kernel
        if self.redHatRelease.getKernelFlavor() == Constants.rt:
            return True
        whitelistLocations = [("/lib/modules/kabi-current/kabi_whitelist_" + self.redHatRelease.getArch()), # RHEL 6.3
                              ("/lib/modules/kabi/kabi_whitelist_" + self.redHatRelease.getArch()), # RHEL 6.2 ?
                              ("/usr/src/kernels/%s-%s/kabi_whitelist" % (self.redHatRelease.getKernel(), self.redHatRelease.getArch())) # ?
                            ]

        for whitelist in whitelistLocations:
            if os.path.exists(whitelist):
                break

        if not os.path.exists(whitelist):
            print("Error: could not find whitelist file in any of the following locations:")
            print("\n".join(whitelistLocations))
            return False

        print("Red Hat Enterprise Linux %u ABI" % self.redHatRelease.getVersion())
        print("--------------------------------")
        print("")
        print("Module: " + module)
        print("Kernel: " + self.redHatRelease.getKernel())
        print("Whitelist: " + whitelist)
        print("")

        whitelistSymbols = self.__readABIWhitelist(whitelist)
        if not whitelistSymbols:
            return False
        moduleSymbols = self.__readModule(module)
        if not moduleSymbols:
            return False
        extraSymbols = self.__abiCheckSymbols(whitelistSymbols, moduleSymbols)

        # check for a DUP greylist
        if extraSymbols != []:
            greylist = "/usr/share/doc/kmod-%s/greylist.txt" % module
            if os.path.exists(greylist):
                print("checking greylist for %s" % module)
                greylistSymbols = self.__readABIWhitelist(greylist)
                extraSymbols = self.__abiCheckSymbols(greylistSymbols, extraSymbols)

            elif self.getDebug() != Constants.off:
                print("%s has no greylist %s" % (module, greylist))

        if extraSymbols != []:
            message = "Error: The following symbols are used by %s are not on the ABI whitelist" % module
            if os.path.exists(greylist):
                message += " or DUP greylist."
            else:
                message += "."
            print(message)
            print("")
            for symbol in extraSymbols:
                print(symbol)
            return False
        else:
            print("Note: %s seems to use only official ABI." % module)

        print("")
        return True


    def is_kernel_tainted(self):
        """
        Returns False/TestResult.FAIL, if tainted.
        """

        print("\n*** Check if kernel is tainted ***\n")
        try:
            with open("/proc/sys/kernel/tainted", "r") as taintedFile:
                tainted = int(taintedFile.readline())

            if tainted == 0:
                print("Test Passed: kernel is not tainted.\n")
                return TestResult.PASS

            else:
                print("Kernel Taint value: %s" % tainted)
                # only tainted via unsigned modules - just a warning
                if tainted == Constants.TAINT_UNSIGNED_MODULE:
                    print("Warning: Following unsigned modules are loaded:\n")
                    modules = self.__getTaintedModules("U")
                    for module in modules:
                        print(module)

                    return TestResult.WARN

                if tainted == Constants.TAINT_PROPRIETARY_MODULE:
                    print("Error: Following modules with non-GPL license have been loaded (this includes modules with no license). Please remove these modules:")
                    modules = self.__getTaintedModules("P")
                    for module in modules:
                        print(module)

                return TestResult.FAIL

        except Exception as error:
            print("Error: could not determine if kernel is tainted\n")
            print("%s" % error)
            return TestResult.FAIL


    def is_kernel_from_red_hat(self):
        """
        Returns False/TestResult.FAIL, if kernel does not belong to Red Hat.
        """

        result = TestResult.PASS
        # Required for hardware certification
        print("+ rpm -ql %s" % self.redHatRelease.getKernelRPMName())

        if self.redHatRelease.kernelIsDebug():
            print("Error: debug kernel")
            result = TestResult.FAIL

        print("\n*** Check that the installed kernel rpm belongs to Red Hat ***\n")

        installed_kernel = self.redHatRelease.getKernelRPMName()
        print("Installed Kernel package version: '%s'." % installed_kernel)
        print("1. Check the kernel rpm buildhost")
        kernel_verify = Command("rpm -q %s --qf '%%{buildhost}\n\' | grep -q redhat.com" % installed_kernel)
        try:
            kernel_verify.run(ignoreWarnings=True)
            print("Test Passed: Installed kernel rpm belongs to Red Hat.")
        except HwCertCommandErrorReturned:
            print("\nError: This kernel (%s) is not built by Red Hat." % installed_kernel)
            result = TestResult.FAIL


        print("2. Check if kernel package modified")
        kernel_verify = Command("rpm -V --nomtime --nomode --nocontexts %s" % installed_kernel)
        try:
            kernel_verify.run(ignoreWarnings=True)
            print("Test Passed: Kernel package is consistent.")
        except HwCertCommandErrorReturned:
            print("\nError: files from '%s' have been modified." % (installed_kernel))
            kernel_verify.echoIgnoreErrors()
            result = TestResult.FAIL
        return result


    def is_kernel_module_tech_preview(self):
        """
        Returns False/TestResult.FAIL, if loaded kernel modules
        have identified themselves as Technology Preview.
        """

        result = TestResult.PASS
        print("\n*** Check if loaded kernel modules have identified themselves as Technology Preview ***\n")
        modules = self.__getTaintedModules("T")
        if not modules:
            print("Test Passed: kernel modules have not been identified themselves as Technology Preview.")
        else:
            badModules = list()
            for module in modules:
                if module not in self.techPreviewModuleWhitelist:
                    badModules.append(module)
            if badModules:
                print("Error: Following Technology Preview modules found:\n")
                for module in badModules:
                    print(module)
                result = TestResult.FAIL
            else:
                print("Test Passed: kernel modules have not been identified themselves as Technology Preview.")
        return result


    def check_kernel_modules_from_red_hat(self):

        """ check 3rd-party modules """
        print("\n*** Checking that kernel modules are from Red Hat ***\n")
        error_message = "contains no files"
        sys.stdout.flush()
        kernelRPMModuleList = list()
        kernelRPMModuleList = Command("rpm -ql %s" % self.redHatRelease.getKernelRPMName()).getStringList(regex=".*", ignoreErrors=True)
        if len(kernelRPMModuleList) == 1 and error_message in kernelRPMModuleList[0]:
            kernelRPMModuleList = Command("rpm -ql %s" % self.redHatRelease.getKernelCoreRPMName()).getStringList(regex=".*", ignoreErrors=True)
            kernelRPMModuleList.extend(Command("rpm -ql %s" % self.redHatRelease.getKernelModulesRPMName()).getStringList(regex=".*", ignoreErrors=True))
        procModules = open("/proc/modules")
        success = True
        while True:
            line = procModules.readline()
            if not line:
                break
            module = line.split()[0]
            moduleFile = self.__getModuleFile(module)

            if not moduleFile:
                success = False
                continue

            # check that module is from the kernel RPM.
            if moduleFile not in kernelRPMModuleList:
                # RHEL4 does not include kmod
                if "Red Hat" in self.redHatRelease.getProduct() and self.redHatRelease.getVersion() == 4:
                    print("Error: Non-Red Hat kernel module %s" % module)
                    success = False
                # RH kmods are allowed
                elif self.__checkModuleVenderAndBuildHost(module):
                    try:
                        Command("rpm -Vf %s" % moduleFile).run()
                        print("kmod package: %s" % Command("rpm -qf %s" % moduleFile).getString())
                        success = self.__abiCheck(module)
                    except Exception as e:
                        print("Error: verification failed for kmod %s" % moduleFile)
                        print(e)
                        success = False
                # Check non-RH modules ABI; but still fail
                else:
                    self.__abiCheck(module)
                    success = False

        if success:
            print("Test Passed: All kernel modules are from Red Hat.\n")
        return success


    def test_kernel(self):
        """
        Verifies following:
        1. The running kernel is a Red Hat kernel.
        2. The running kernel is released by Red Hat for use with the RHEL version.
        3. The running kernel is not tainted.
        Returns True if above three criterian match.
        """

        result = TestResult()

        release = self.redHatRelease
        for subtest in [self.is_kernel_from_red_hat(),
                        self.checkKernelPackageVersion(),
                        self.is_kernel_tainted()]:
            result.combine(subtest)
        return result


    def test_kernel_modules(self):
        """
        Verifies following:
        1. The loaded kernel modules have not identified themselves as Technology Preview.
        2. The loaded kernel modules are from Red Hat.
        """
        result = TestResult()

        for subtest in [self.is_kernel_module_tech_preview(),
                        self.check_kernel_modules_from_red_hat()]:
            result.combine(subtest)
        return result

    def test_boot_parameters(self):
        """ show boot parameters """
        try:
            params = Command("cat /proc/cmdline").getString()
            print("Boot Parameters: %s" % params)
        except HwCertCommandException as exception:
            print("Warning: could not determine boot parameters")
            print(exception)
            return TestResult.WARN

        return TestResult.PASS

    def run(self):

        result = TestResult()

        result.combine(self.runSubTest(self.test_kernel, name="Kernel", description="Verify installed kernel package"))

        result.combine(self.runSubTest(self.test_kernel_modules, name="Kernel Modules", description="Verify loaded kernel modules."))

        result.combine(self.runSubTest(self.test_boot_parameters, name="Boot Parameters", description="Show boot parameters"))

        return result


class CheckSELinux(SubTest):
    def __init__(self, test):
        SubTest.__init__(self, test, "SELinux Status", "Check SELinux configuration on the system.")
        self.program = test.getSuiteName()
        self.expectedStatus = 'enabled'
        self.expectedModes = self.test.policy["SELINUX"]
        self.expectedConfigModes = self.test.policy["SELINUX"]

    def runSubTest(self):
        result = TestResult.PASS
        try:
            selinuxParams = Command("/usr/sbin/sestatus")
            selinuxStatus = selinuxParams.getString(regex="SELinux status:[ \t]+(?P<selinuxStatus>[a-zA-Z]+)",
                                                    regexGroup="selinuxStatus", singleLine=False)
            if selinuxStatus == self.expectedStatus:
                selinuxMode = selinuxParams.getString(regex="Current mode:[ \t]+(?P<selinuxMode>[a-zA-Z]+)",
                                                      regexGroup="selinuxMode", singleLine=False)

                selinuxConfigMode = selinuxParams.getString(
                    regex="Mode from config file:[ \t]+(?P<selinuxConfigMode>[a-zA-Z]+)",
                    regexGroup="selinuxConfigMode", singleLine=False)

                if selinuxMode in self.expectedModes and selinuxConfigMode in self.expectedConfigModes:
                    if self.program == Programs.cloud_sap and selinuxMode == "enforcing":
                        result = TestResult.WARN
                    print("{0} SELinux running in \"{1}\" mode.\n".format(TestResult(result).getMessagePrefix(),
                                                                          selinuxMode))

                if selinuxMode == "disabled" or selinuxConfigMode == "disabled":
                    result = TestResult.FAIL

                    print("{0} SELinux is in {1} mode - it must be set to \"{2}\" for certification testing.".format(
                        TestResult(result).getMessagePrefix(), selinuxMode, ' or '.join(self.expectedModes)))
            else:
                result = TestResult.FAIL
                print("{0} SELinux is {1} - it must be \"enabled\" for certification testing.".format(
                    TestResult(result).getMessagePrefix(), selinuxStatus))

            if self.getDebug() != Constants.off:
                print("\nDetected SELinux configuration \n")
                print("------------------------------\n")
                selinuxParams.printOutput()

        except HwCertCommandException as exception:
            print("Error: could not determine selinux parameters\n")
            print(exception)
            result = TestResult.FAIL
        return result


class LogVersions(SubTest):
    def __init__(self, test):
        SubTest.__init__(self, test, "Log Versions", "Log rhcert and OS version and release")

    def runSubTest(self):
        self.redHatRelease = RedHatRelease()
        print("Tested OS: %s %s (%s)" % (self.redHatRelease.getProduct(), self.redHatRelease.getVersionPointUpdate(), self.redHatRelease.getCodeName()))

        print("Kernel RPM: %s" % self.redHatRelease.getKernelRPMName())
        # YK: to show hwcert version as well
        print("%s version %s, release %s" % (rhcert.version.name, rhcert.version.version, rhcert.version.release))

        if self.test.getDebug() != Constants.off:
            if self.test.getResultsDocument():
                for (name, value) in self.test.getResultsDocument().getCertificationParameters().getAll().items():
                    print("certification parameter: %s %s" % (name, value))
            else:
                print("no results document set")

        return True


class LogDmidecode(SubTest):
    def __init__(self, test):
        SubTest.__init__(self, test, "dmidecode", "Log dmidecode output")
        self.dmidecode = "/usr/sbin/dmidecode"

    def run(self):
        """ override run wapper to skip sub-logging """
        if os.path.exists(self.dmidecode):
            return SubTest.run(self)
        return True

    def runSubTest(self):
        if os.path.exists(self.dmidecode):
            try:
                Command(self.dmidecode).echo()
            except HwCertCommandException as e:
                print("Error: %s failed." % self.dmidecode)
                print(e)
                return False
        # otherwise
        return True


class GenerateSystemReport(SubTest):

    def __init__(self, test):
        SubTest.__init__(self, test, "System Report", "Generates a system report (sosreport)")

    def getSosVersion(self):
        sos_version = Command('rpm -qa | grep sos')
        sos_version.run()
        self.sos_version = '.'.join(sos_version.raw_output.split('.')[:-2])

    def checkMinVersion(self):
        min_version = 'sos-3.4-6'
        match = rpm.labelCompare(('0', 'sos', self.sos_version.partition('sos-')[-1]),
                                 ('0', 'sos', min_version.partition('sos-')[-1]))
        if match >= 0:
            return True
        else:
            return False

    def compareSosVersions(self):
        self.getSosVersion()
        accepted_versions = [
            'sos-3.3-5.el7_3.bz1415972',
            'sos-3.3-5.bz1415972.el7_3']
        print(
            "Info: Checking sos version to see if required information will be collected.")
        if (self.sos_version in accepted_versions) or self.checkMinVersion():
            print("Info: sos version is in the accepted versions.")
            return TestResult.PASS
        else:
            print("Needs Review: sos version is not in accepted versions. Please Review")
            return TestResult.REVIEW

    def check_sos_plugins(self, plugins):
        cmd = "sosreport -l"
        plugin_list = plugins.split("|")
        print("Checking if '%s' plugins are active in sos ..." % (','.join(plugin_list)))
        try:
            output = Command(cmd)
            output.run()
            self.sos_plugins = output.raw_output.strip()
            # openstack will pass manila|cinder|sahara|neutron and baremetal will pass ironic
            expr1 = 'openstack_(?:%s)\s+(?:inactive)\s+Open.*' % plugins
            expr2 = 'openstack_(?:%s)\s+Open.*' % plugins
            match_active = re.findall(expr2, self.sos_plugins, re.IGNORECASE)
            match_inactive = re.findall(expr1, self.sos_plugins, re.IGNORECASE)
            if len(match_active) == len(plugin_list):
                print("Success: Found all the required plugins as active")
                return TestResult.PASS
            else:
                print(
                    "Needs Review: Some of the plugins are not active. Please check sos tarball.")
                return TestResult.REVIEW
        except HwCertCommandException as err:
            print("Needs Review: Failed to determine the status of plugins. "
                  "The command '%s' execution failed with error: '%s'." %
                  (cmd, str(err)))
            return TestResult.REVIEW

    def check_container_plugin(self, container_utility):
        print("Checking if '%s' plugins are active in sos ..." % (container_utility))
        if not(hasattr(self, 'sos_plugins') and self.sos_plugins):
            print("Needs Review: Failed to determine the status of plugins.")
            return TestResult.REVIEW

        expr1 = '%s\s+(?:inactive).*\s+%s.*' %(container_utility, container_utility.capitalize())
        expr2 = '%s\s+%s.*' %(container_utility, container_utility.capitalize())
        match_active = re.findall(expr2, self.sos_plugins, re.IGNORECASE)
        match_inactive = re.findall(expr1, self.sos_plugins, re.IGNORECASE)
        if len(match_active) == 1:
            print("Success: Found %s plugin as active in sos utility" % container_utility)
            return TestResult.PASS
        else:
            print("Needs Review: %s plugin is inactive in sos utility." % container_utility)
            return TestResult.REVIEW

    def runSubTest(self):

        print("Generating System Report(sosreport). It may take a while. Please wait ...\n")

        # limit the size of collected logs
        log_size = int(self.test.environment.getMaximumSosPluginSize()/1024/1024) # bytes to MB conversion

        result = self._processSystemReport("sosreport --batch -n selinux -n logs --log-size {0}".format(log_size))
        return result

    def _processSystemReport(self, reportCommand):
        pipe = os.popen("echo -e '\n\n' | %s -k rpm.rpmva=off 2>&1" % reportCommand)
        result = False
        while True:
            line = pipe.readline()
            if line:
                sys.stdout.flush()
                tarFile = None
                if "generated and saved in" in line:
                    tarFile = pipe.readline().strip()
                    print(tarFile)
                    sys.stdout.flush()

                if tarFile:
                    if os.path.getsize(tarFile) > self.test.environment.getMaximumAttachmentSize():
                        print(("Error: sosreport is %s MB, and too large to attach (Max size: %s MB)"
                                 % (int(os.path.getsize(tarFile)/Constants.MB),
                                    int(self.test.environment.getMaximumAttachmentSize()/Constants.MB))))
                        return False
                    shutil.copy(tarFile, self.test.getOutputDirectory())
                    print("Copied %s %s to %s" % (reportCommand, tarFile, self.test.getOutputDirectory()))
                    sys.stdout.flush()
                    result = True
                    break
            else:
                break
        pipe.close()
        return result


class CheckHardwareHealth(SubTest):
    def __init__(self, test, ram_per_core_policy=TestResult.WARN, vulnerability_check=TestResult.WARN):
        SubTest.__init__(self, test, "Hardware Health", "Checks the health of the System Hardware")
        self.ram_per_core_policy = ram_per_core_policy
        self.vulnerability_check = vulnerability_check

    def check_unsupported_hardware(self):
        """Checks message log for any reported unsupported hardware."""
        print("Checking Unsupported Hardware ...")
        result = TestResult()
        controller = Controller(debug=self.getDebug())
        for line in controller.getSystemLogSinceBoot():
            if 'this hardware has not undergone testing' in line:
                result.combine(TestResult.FAIL)
                print("%s The kernel has identified unsupported hardware." % TestResult(result).getMessagePrefix())
                print(self._messagesWarningToError(line))
                break
        else:
            print("%s kernel has identified supported hardware" % TestResult(result).getMessagePrefix())
        return result

    def _messagesWarningToError(self, warning):
        """
        Removes everything in the string from the beginning through the first
        instance of 'Warning: ' and replaces it with 'Error: '
        """
        warning_snippet = "Warning: "
        warning_index = warning.find(warning_snippet)
        starting_at = warning_index+len(warning_snippet)
        issue = warning[starting_at:].splitlines()[0]
        return "Error: %s" % issue

    def ram_per_core(self):
        result = TestResult()
        threshold = 1.5 if int(RedHatRelease().getVersion()) >= 8 else 1
        try:
            system_ram = Command("free -g | awk '/^Mem:/{print $2}'").getInteger() # System RAM in GBs
            cpu_count = Command("lscpu | awk '/^CPU\(s\):/{print $2}'").getInteger()
            if (system_ram/cpu_count) < threshold:
                result.combine(self.ram_per_core_policy)
                print("{0}System RAM ({1} GiB) per CPU ({2} logical {3}) ratio is lower than the "
                      "required threshold ({4})".format(result.getMessagePrefix(), system_ram, cpu_count,
                                                        'cores' if cpu_count > 1 else 'core', threshold))
                return False
            else:
                result.combine(TestResult.PASS)
                print("{0}System RAM ({1} GiB) per CPU ({2} logical {3}) ratio is above the "
                      "required threshold ({4})".format(result.getMessagePrefix(), system_ram, cpu_count,
                                                        'cores' if cpu_count > 1 else 'core', threshold))
        except Exception as error:
            print("Error: Failure in test execution\n")
            print(error)
            return False
        return True

    def check_ram_per_core(self):
        print("\nChecking RAM per Core ratio ...")
        result = TestResult()
        if self.ram_per_core(): # separated the method for re-usability
            result.combine(TestResult.PASS)
        else:
            result.combine(TestResult.WARN)
        return result

    def check_known_vulnerabilities(self):
        print("\nChecking Known System Vulnerabilities ...")
        get_vul = " grep '' /sys/devices/system/cpu/vulnerabilities/*"
        output_vul = []
        result = TestResult.PASS
        try:
            output_vul = Command(get_vul).getStringList(regex=".*")
        except HwCertCommandException:
            pass

        for val in output_vul:
            if "Vulnerable" in val:
                result = self.vulnerability_check
                file_message = val.split(':', 1)
                print("{0}Known Vulnerability found: {1}".format(TestResult(result).getMessagePrefix(),
                                                                        file_message[0]))
                print("{0}\n".format(file_message[1]))

        return result

    def check_ht_offline_cpus(self):
        print("\nChecking Hyper-Threading ...")
        check_ht = "lscpu | grep Thread | awk -F ' ' '{print $NF}'"
        result = TestResult()
        try:
            ht_val = Command(check_ht).getInteger(regex=".*")
            if ht_val >= 2:
                print("Warning: Hyper-Threading is enabled in the system")
                result.combine(TestResult.WARN)
        except HwCertCommandException:
            print("Warning: Unable to get the Hyper-Threading details")
            result.combine(TestResult.WARN)

        print("\nChecking Offline CPU(s) ...")
        check_offline_cpus = "lscpu | grep Off-line | awk -F ' ' '{print $NF}'"
        try:
            offline_cpus = Command(check_offline_cpus).getString(regex=".*")
            print("Warning: Following Offline CPU(s) found : {0}".format(offline_cpus))
            result.combine(TestResult.WARN)
        except HwCertCommandException:
            print("No Offline CPUs found.")
            result.combine(TestResult.PASS)
        return result

    def runSubTest(self):
        result = TestResult()
        result.combine(self.check_unsupported_hardware())
        result.combine(self.check_ram_per_core())
        result.combine(self.check_ht_offline_cpus())
        result.combine(self.check_known_vulnerabilities())
        return result


class LogVersionsTest(SubTest):
    """
    Logs installed kernel version and detected RHEL release info.
    For OpenStack it also compares RHEL certificate and detected version.
    """
    def __init__(self, test):
        self.program = test.getSuiteName()
        description = "Logs the installed kernel version and detected RHEL release info."
        if self.program == 'openstack':
            description = description + "Also compares RHEL certificate and detected version."
        SubTest.__init__(self, test, "Log Versions", description)
        self.release = RedHatRelease()

    def runSubTest(self):
        kernel = self.release.getKernelRPMName()
        rhel = self.release.text
        result = TestResult.PASS
        if rhel:
            print("\nInstalled RHEL version: %s" % rhel)
        else:
            print("Error: Could not determine the RHEL version.")
            result = TestResult.FAIL
        if kernel:
            print("Installed kernel version: %s" % kernel)
        else:
            print("Error: Could not determine the kernel version.")
            result = TestResult.FAIL

        if self.program == 'openstack':
            cert_ID = self.test.getResultsDocument().getCertificationID()
            detected_version = self.release.getVersionPointUpdate()
            if int(cert_ID) != 0:
                cert_version = self.test.getResultsDocument().getOSValues()['release']
                print("The certification requires RHEL %s" % cert_version)
            else:
                print("RHEL %s is installed" % detected_version)
                print("Needs Review: Could not determine the base RHEL version from the certification")
                cert_version = None
                return TestResult.REVIEW
            print("RHEL %s is installed" % detected_version)
            if cert_version == detected_version:
                print("Certificate version matches the detected RHEL version.")
                return TestResult.PASS
            else:
                print("Needs Review: The RHEL version does not match the base RHEL version required by the certification")
                return TestResult.REVIEW
        else:
            return result


def test_kernel_test_status():
    """
    Unit test for method kernel_test_status
    """

    kernel_data = {
        "candidate": [
            "kernel-1",
            "kernel-2"
        ],
        "ga": [
            "kernel-0"
        ],
        "update": {
            "AUS": ["kernel-aus"],
            "EUS": ["kernel-eus"],
            "LL": ["kernel-ll"],
            "Other": ["kernel-other"],
            "Z": ["kernel-z"]
      }
    }

    for kernel_obj in [KernelTest("baremetal/"),
                       KernelTest("cloud/"),
                       KernelTest("openstack/"),
                       KernelTest("hardware/"),
                       KernelTest("hwcert/")]:

        print("\nProgram: %s" % kernel_obj.program)
        print("------------------------------------------")
        print("Passing GA")
        print(kernel_obj.kernel_test_status("kernel-0", kernel_data))

        print("Passing candidate (via service)")
        print(kernel_obj.kernel_test_status("kernel-1", kernel_data))

        print("Passing candidate (via cache)")
        print(kernel_obj.kernel_test_status("kernel-1", kernel_data,
            via_service = False))

        print("Passing update Z")
        print(kernel_obj.kernel_test_status("kernel-z", kernel_data))

        print("Passing update AUS")
        print(kernel_obj.kernel_test_status("kernel-aus", kernel_data))

        print("Passing update LL")
        print(kernel_obj.kernel_test_status("kernel-ll", kernel_data))

        print("Passing update EUS")
        print(kernel_obj.kernel_test_status("kernel-eus", kernel_data))

        print("Passing update Other")
        print(kernel_obj.kernel_test_status("kernel-other", kernel_data))

        print("Passing invalid (via service)")
        print(kernel_obj.kernel_test_status("kernel-invalid", kernel_data))

        print("Passing invalid (via cache)")
        print(kernel_obj.kernel_test_status("kernel-invalid", kernel_data,
            via_service = False))


def unit_test():
    test_kernel_test_status()
    test = KernelTest("unit/kernel")
    return test.do(sys.argv)

if __name__ == "__main__":
    sys.exit(unit_test())

