#!/usr/bin/env python
# Copyright (c) 2020 Red Hat, Inc. All rights reserved.
# This copyrighted material is made available to anyone
# wishing to use, modify, copy, or redistribute it subject
# to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.
#
# Author: Zahid Ajaz <zajaz@redhat.com>

import os, re
from rhcert.tags import TestResult
from rhcert.test import SubTest
from rhcert.command import Command, HwCertCommandException
from rhcert.tags import Constants


class RPMDependencyTrackingTest(SubTest):

    def __init__(self, test):
        SubTest.__init__(self, test, "RPM Dependency Tracking",
                         "Test RPM Dependencies")

    def get_whatprovides(self, name_path):
        cmd = Command("rpm -q --whatprovides '%s'" % name_path)
        return cmd.getIgnoreErrors()

    def rpmlib_showrc(self, rpmlib_package):
        cmd= Command("rpm --showrc | grep '{}'".format(rpmlib_package))
        return cmd.getIgnoreErrors()

    def file_binary(self, path):
        """
        Check if a file is a ELF binary
        """
        cmd = Command("file %s" % path)
        file_info = cmd.getIgnoreErrors()[0]
        regex = '^%s:\sELF\s[0-9]+\-bit\sLSB\sshared\sobject' % path
        if re.match(regex, file_info, re.IGNORECASE):
            return True
        return False
    
    def get_shared_binaries(self, path):
        cmd = Command("ldd %s" % path)
        regex = r'(\/.*?[\w\.]+)'
        binaries = self.test.get_string_list(cmd, regex)
        return [binary for binary in binaries if not 'vdso' in binary]

    def inspect_rpm_binaries(self, rpm):
        result = TestResult()
        cmd = Command("rpm -ql %s" % rpm)
        cmd.run()
        regex = re.compile(r'\(contains no files\)')
        files = [file_name for file_name in cmd.output if not regex.match(file_name)]
        if not files:
            print("\nError: RPM '%s' doesn't report any files\n" % rpm)
            return TestResult.FAIL

        for path in files:
            if self.file_binary(path):
                for binary in self.get_shared_binaries(path):
                    nothing_provides = 'no package provides %s' % binary
                    if any(whatprovides == nothing_provides for whatprovides in self.get_whatprovides(binary)):
                        print("\nError: RPM '%s' doesn't list whatprovides '%s' \
                        which is a shared binary for %s\n" % (rpm, binary, path))
                        return TestResult.FAIL
        return result

    def inspect_rpm_dupes(self):
        result = TestResult()
        cmd = Command("rpm -qa --dupes")
        dupes = cmd.getIgnoreErrors()
        for duplicate in dupes or []:
            if duplicate in self.test.foreign_rpms:
                print("\nError: Duplicate package '%s' found, which is a Non Redhat pacakge" % duplicate)
                return TestResult.FAIL
        return result

    def inspect_rpm_obsoletes(self, rpm):
        result = TestResult()
        cmd = Command("rpm -q --obsoletes '%s'" % rpm)
        obsoleted_pkgs = cmd.getIgnoreErrors()
        if not obsoleted_pkgs:
            return result
        for pkg in obsoleted_pkgs:
            pkg_unversioned = re.split(r'([\s\<\=]+)', pkg)[0]
            if set(self.get_whatprovides(pkg_unversioned)).issubset(set(self.test.redhat_rpms)):
                print("\nError: RPM '%s' obsoletes '%s' which is a RedHat package" %(rpm, pkg))
                return TestResult.FAIL
        return result

    def inspect_rpm_requires(self, rpm):
        result = TestResult()
        cmd = Command("rpm -q --requires %s" % rpm)
        cmd.run()
        required_pkgs = cmd.output
        if not required_pkgs:
            print("\nError: RPM '%s' doesn't list any dependencies" %rpm)
            return TestResult.FAIL

        for pkg in required_pkgs:
            required_rpm = re.split(r'([\s\<\=]+)', pkg)[0]
            if required_rpm.startswith('rpmlib'):
                if self.rpmlib_showrc(required_rpm):
                    continue
                else:
                    print("\nError: RPM '%s' requires '%s' which coudn't be found in rpmlib section by showrc option." %(rpm, required_rpm))
                    return TestResult.FAIL
            if set(self.get_whatprovides(required_rpm)).issubset(set(self.test.redhat_rpms)):
                return TestResult.PASS
            elif required_rpm in self.test.redhat_rpms and 'el8' not in required_rpm:
                print("\nError: RPM '%s' requires package '%s' that is not supported by RedHat on RHEL8" %(rpm, pkg))
                return TestResult.FAIL
            else:
                result.combine(TestResult.FAIL)
        if result == TestResult.FAIL:
            print("\nError: %s doesn't list any Redhat package in dependencies" %rpm)
        return result

    def validate_rpm_dependencies(self):
        result = TestResult()
        result.combine(self.inspect_rpm_dupes())
        for frpm in self.test.foreign_rpms:
            result.combine(self.inspect_rpm_requires(frpm))
            result.combine(self.inspect_rpm_obsoletes(frpm))            
            result.combine(self.inspect_rpm_binaries(frpm))
        return result

    def validate_rpm_handler(self):
        print("\nValidating dependencies of partner-listed RPM(s)\n")
        if self.test.promptConfirm(
            "Please confirm whether all dependencies of your application are handled by RPM, DNF, or Software Modules"):
                return TestResult.PASS
        else:
            self.partner_tools = self.test.prompt("Please specify what tool(s) handle your application dependencies:\n")
            print("Needs Review: Following application dependency handling tool(s) have been listed by partner:\n",
                  self.partner_tools)
            return TestResult.REVIEW

    def runSubTest(self):
        """
        Cases with their return values:
        """
        result = TestResult()
        if not self.test.user_rpms:
            print("\nError: No RPM(s) specified as application software")
            return TestResult.FAIL
        for check in [self.validate_rpm_handler(),
                        self.validate_rpm_dependencies()]:
            result.combine(check)
        return result
