#!/usr/bin/env python
# Copyright (c) 2020 Red Hat, Inc. All rights reserved.
# This copyrighted material is made available to anyone
# wishing to use, modify, copy, or redistribute it subject
# to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.
#
# Author: Zahid Ajaz <zajaz@redhat.com>

import os, re

from rhcert.tags import TestResult
from rhcert.test import SubTest
from rhcert.command import Command
from rhcert.tags import Constants


class RPMProvenanceTest(SubTest):

    def __init__(self, test):
        SubTest.__init__(self, test, "RPM Provenance",
                         "Test RPM Provenance")

    def pretty_print_meta_data(self, rpm , meta_data):
        print("Meta-data for the RPM '%s' is inconsistent" % rpm)
        print("-"*20)
        for k, v in meta_data.items():
            print("{:<10} {:<15}".format(k, v))

    def check_gpg_keys(self):
        result = TestResult()
        cmd = Command("rpm -qa gpg-pubkey*")
        self.gpg_keys = cmd.getIgnoreErrors()
        if not self.gpg_keys:
            print("Error: No GPG keys installed.")
            print("Corrective Action: Make sure that you have correct GPG keys installed used to sign the rpms\n")
            result.combine(TestResult.FAIL)
        return result

    def redhat_signed_rpm(self, rpm, sign):
        sign_label = re.split(r'Key\sID\s', sign)[-1][-8:]
        for key in self.gpg_keys or []:
            key_label = key.split("-")[2]
            if sign_label == key_label:
                meta = self.test.get_rpm_meta(key, ['PACKAGER'])
                if self.identified_rht_meta(meta['PACKAGER']):
                    return True
            continue
        return False

    def identified_rht_meta(self, name):
        """
        Validate RPM packager/vendor information
        """
        redhat_identifiers = [ "RedHat", "Red Hat", "Red Hat, Inc.", "RHT", "Fedora", "CentOS", "JBOSS"]
        if any(identifier.lower() in name.lower() for identifier in redhat_identifiers):
            return True
        else: False

    def check_unowned_dir(self, path):
        check_str = "file %s is not owned by any package" % path
        check_cmd = Command("rpm -qf %s" % path)
        check_output = check_cmd.getIgnoreErrors()
        if check_output:
            if check_str in check_output[0]:
                return TestResult.FAIL
            else:
                return TestResult.PASS
        return TestResult.FAIL

    def check_rpm_conflicts(self):
        """
        Identify conflicting foreign/partner RPM(s)
        """
        self.identified_rpms = list(set(self.test.foreign_rpms) - set(self.test.user_rpms))
        meta_list  = ['NAME', 'VENDOR', 'PACKAGER', 'SIGPGP:pgpsig']
        for rpm in self.test.user_rpms:
            for irpm in self.identified_rpms:
                if len(dict(self.test.get_rpm_meta(rpm, meta_list).items() & self.test.get_rpm_meta(irpm, meta_list).items())) > 0:
                    return True
        return False

    def inspect_rpm_files(self, rpm):
        """
        Inspect whether a foreign/partner RPM reports files or not
        """
        # Check if RPM lists files
        list_files_cmd = Command("rpm -ql %s" % rpm)
        list_files_cmd.run()
        
        if not list_files_cmd.output:
            print("\nError: RPM '%s' doesn't report any files\n" % rpm)
            return TestResult.FAIL
        result = TestResult()
        # Check for unowned dir(s) in the package files
        for path in list_files_cmd.output:
            if os.path.isdir(path):
                if self.check_unowned_dir(path) == TestResult.FAIL:
                    print("Error: %s is not owned by RPM %s" %(path, rpm))
                    result.combine(TestResult.FAIL)
        # Check if RPM reports document files
        list_doc_files = Command("rpm -qd %s" % rpm)
        list_doc_files.run()
        if not list_doc_files.output:
            print("\nWarning: RPM '%s' doesn't report any documentation files\n" % rpm)
            result.combine(TestResult.WARN)
        # Check if RPM reports config files 
        list_config_files = Command("rpm -qc %s" % rpm)
        list_config_files.run()
        if not list_config_files.output:
            print("\nWarning: The RPM '%s' doesn't report any config files\n" % rpm)
            result.combine(TestResult.WARN)

        return result

    def verify_rpm_files(self, rpm):
        """
        Verify a foreign/partner RPM for file inconsistencies
        """
        result = TestResult()
        rpm_verify = Command("rpm -V " + rpm)
        errors = rpm_verify.getIgnoreErrors()
        if errors:
            print("\nWarning: RPM '%s' failed to verify cleanly\n" % rpm)
            result.combine(TestResult.WARN)        
            for error in errors[:]:
                error = error.split()
                path = error[-1]
                result.combine(self.check_unowned_dir(path))
                if len(error)==3:
                    if error[1]=='g':    
                        result.combine(TestResult.FAIL)
                    if error[1]=='c' and path.lower().startswith('/etc'):
                        result.combine(TestResult.WARN)
                elif any(symbol in error[0] for symbol in ["M", "G", "5"]) and path.lower().startswith('/var/log' or '/var/run'):
                        result.combine(TestResult.WARN)
                else:
                    result.combine(TestResult.FAIL)
            if result == TestResult.FAIL:
                print("\nError: RPM '%s' has reported file modifications\n" % rpm)
        return result

    def check_rpm_file_tracking(self):
        result = TestResult()
        for frpm in self.test.foreign_rpms:
            result.combine(self.inspect_rpm_files(frpm))
            result.combine(self.verify_rpm_files(frpm))
        return result

    def check_rpm_metadata(self):
        """
        Check the RPM(s) for meta-data inconsistencies
        """
        print("\nChecking and Validating RPM meta-data\n")
        result = TestResult()
        redhat_rpm_names =  [self.test.get_rpm_meta(name, ['NAME'])['NAME'] for name in self.test.redhat_rpms]
        meta_list  = ['NAME', 'VENDOR', 'PACKAGER', 'SIGPGP:pgpsig']
        result.combine(self.check_gpg_keys())
        rpm_meta_data = list()
        for frpm in self.test.foreign_rpms:
            meta_data = self.test.get_rpm_meta(frpm, meta_list)
            if not(meta_data['NAME'] and meta_data['VENDOR'] and meta_data['PACKAGER'] and meta_data['SIGPGP:pgpsig']):
                result.combine(TestResult.FAIL)
                self.pretty_print_meta_data(frpm, meta_data)
            else:
                if self.redhat_signed_rpm(frpm, meta_data['SIGPGP:pgpsig']):
                    print("\nError: RPM '%s' has been signed using a Redhat GPG key" % frpm)
                    result.combine(TestResult.FAIL)
                if meta_data['NAME'] in redhat_rpm_names:
                    print("\nError: RPM 'NAME' in meta data of '%s' conflicts with an existing RedHat package" % frpm)
                    result.combine(TestResult.FAIL)
                if self.identified_rht_meta(meta_data['VENDOR']) or self.identified_rht_meta(meta_data['PACKAGER']):
                    print("\nError: RPM meta data for '%s' identifies it as a RedHat packaged rpm" % frpm)
                    result.combine(TestResult.FAIL)
                rpm_meta_data.append(meta_data)
        self.logFeatureData("ISV RPM Meta" , rpm_meta_data)
        return result

    def check_listed_software_rpms(self):
        """
        Looking for Partner application RPM package(s)
        """
        self.test.user_rpms = list()
        msg = "\nThe following RPM(s) are not recognized as provided by Red Hat:\n%s" % (",\n".join(
            [rpm for rpm in self.test.foreign_rpms]))
        if self.test.mode != Constants.auto:
            if self.test.mode == Constants.server:
                print(msg)
            self.test.user_rpms = self.test.multiSelect(
                "\nPlease select the RPM(s) that are a part of your application", [
                    rpm for rpm in self.test.foreign_rpms])
        if not self.test.user_rpms:
            print("\nError: No RPM(s) specified as application software")
            return TestResult.FAIL

        if self.check_rpm_conflicts():
            msg = ("\nFollowing additional RPM(s) identical to your software RPM(s) have been found:\n%s\n"
            "Are you sure there is no additional RPM(s) that is a part of your software?" % ','.join(self.identified_rpms))
            if self.test.promptConfirm(msg):
                return TestResult.REVIEW
            else:
                self.check_listed_software_rpms()
        if self.identified_rpms:
            msg = ("\nNeeds Review: Following additional non-Red Hat packages are installed:\n%s\n" % ','.join(self.identified_rpms))
            return TestResult.REVIEW

        return TestResult.PASS

    def runSubTest(self):
        """
        Cases with their return values:
        """
        result = TestResult()
        print("\nLooking for Partner application RPM package(s)\n")
        result.combine(self.check_listed_software_rpms())
        if result == TestResult.FAIL:
            return result
        for check in [self.check_rpm_metadata(),
                        self.check_rpm_file_tracking()
                    ]:
            result.combine(check)
        return result