#!/usr/bin/env python
# Copyright (c) 2020 Red Hat, Inc. All rights reserved.
# This copyrighted material is made available to anyone
# wishing to use, modify, copy, or redistribute it subject
# to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.
#
# Author: Zahid Ajaz <zajaz@redhat.com>

import os

from rhcert.redhatrelease import RedHatRelease
from rhcert.tags import TestResult, Programs
from rhcert.services.validaterpms import ValidateRPMs
from rhcert.command import Command, HwCertCommandException, HwCertCommandErrorOutput


class RPMValidation:
    """Validate installed RPMs whether or not they belong to RedHat"""

    def __init__(self, subtest):
        self.release = RedHatRelease()
        self.subtest = subtest
        self.whitelisted_pkgs = list()

    def provisional_whitelisting(self):
        """
        White lists packages based on specific filters
        """
        whitelisted_labels = ['rhui', 'satellite']
        for package in self.subtest.packages:
            # Filter 'gpg' packages
            if package.startswith(('gpg-pubkey', 'shim')):
                self.whitelisted_pkgs.append(package)

            # Whitelist packages installed to manage ca-certs for cloud program.
            # Also a default criteria for any new program
            if self.subtest.program not in [Programs.hardware, Programs.openstack, Programs.baremetal, Programs._hardware]:
                if any(label in package.lower() for label in whitelisted_labels) and \
                                os.system("rpm -qli %s | grep '/etc/pki/*'" % package) is 0:
                    self.whitelisted_pkgs.append(package)

    def get_non_rh_packages_via_build_host(self):
        """
        Check for the non-redhat packages from the
        build host information. Returns the list.
        """
        non_rh_pkgs = list()
        for package in self.subtest.packages:
            if os.system(
                "rpm -q %s --qf '%%{buildhost}\n\' | grep -q redhat.com" %
                    package) is not 0:
                    non_rh_pkgs.append(package)
        self.provisional_whitelisting()
        return (list(set(non_rh_pkgs) - set(self.whitelisted_pkgs)))

    def get_non_rh_packages_via_offerings(self):
        """
        Check that the installed RPMs are from the
        Red Hat product(s) available in the offering.
        Returns the list.
        """
        if not self.release.text:
            message = "Error: Unsupported Operating System."
            return ({} , TestResult.FAIL, message)
        
        major_dot_minor_version = ("%s.%s %s" % (self.release.getVersion(),
                                                        self.release.getUpdate(),
                                                        self.release.getLabel())).strip()
        # TODO: product short name should come from 'product name service'
        # product_short_name = MicroServices().product_shortname(self.release.text)
        product_short_name = "RHEL"
        packages = list(set(self.subtest.packages) - (set(self.whitelisted_pkgs)))
        validateRPMsService = ValidateRPMs(product_short_name,
                                           major_dot_minor_version,
                                           packages)
        response = validateRPMsService.call()

        # RedHat product service is not accessible
        if not response:
            message = "Needs Review: Check if the installed packages belong to RedHat Products."
            self.subtest.deferredServices.append(validateRPMsService)
            return ({}, TestResult.REVIEW, message)

        return (response.json(), None, None)
