#!/usr/bin/env python
# Copyright (c) 2017 Red Hat, Inc. All rights reserved.
# This copyrighted material is made available to anyone
# wishing to use, modify, copy, or redistribute it subject
# to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.
#
# Author: Pooja Tiwari <ptiwari@redhat.com>
import os
from rhcert.tags import TestResult, Programs
from rhcert.command import Command, HwCertCommandException, HwCertCommandErrorOutput, HwCertCommandErrorReturned
from rhcert.redhatrelease import RedHatRelease

IGNORE_SELINUX = '/var/lib/selinux/'
LSM_ALLOWED_VERSION = ['8.0', '8.1']


class RPMVerification:
    rpm_verificaton_log = list()

    def __init__(self, program, errors=None, output_directory=None):
        self.program = program
        self.errors = errors
        self.result = TestResult()
        self.result.combine(TestResult.PASS)
        self.output_directory = output_directory
        self.redHatRelease = RedHatRelease()
        self.checks = {
            'S': self.size_check,
            'M': self.mode_check,
            '5': self.md5_check,
            'D': self.device_check,
            'L': self.readlink_check,
            'U': self.user_ownership_check,
            'G': self.group_ownership_check,
            'T': self.time_check,
            'P': self.capability_check
        }
    
    def rpm_availability(self, package):
        """
        Check if requested packages are installed
        """
        command = Command("rpm -q %s" % package)
        try:
           command.run()
        except HwCertCommandErrorReturned as error:
            return (False, error.message)
        else:
            return (True, command.raw_output)

    def rpm_verify(self):
        """
        wrapper function. It combines the result of all
        checks and send back to test.
        """
        # ignoring Selinux errors
        self.errors = [path for path in self.errors if IGNORE_SELINUX not in path]
        for error in self.errors:
            err = error.split()
            for err_code, action in self.checks.items():
                if err_code in err[0]:
                    self.result.combine(action(err))
        return self.result

    def preliminary_check(self, error, check=None):
        """
        Preliminary Check Function:
        - white listing config files
        - conditional handling for different check flags
        """
        path = error[-1]
        if check in ['S', 'D', '5', 'L']:
            is_config_file = False
            if len(error) == 3:
                '''Files marked with "c" are config files by definition
                regardless of where they are.
                '''
                is_config_file = error[1] == 'c'
            if is_config_file:
                return None, TestResult.PASS
            else:
                '''Files in /etc but not marked "c" by the RPM are likely config
                files and these should exit REVIEW. The review team can then pass
                when the files are confirmed to be config files.'''
                if path.startswith("/etc"):
                    return "Needs Review", TestResult.REVIEW

        if path.startswith("/var"):
            '''Files under /var need special handling
            /var failures - REVIEW (probably bad packaging but we need to look)
            /var/log/gdm 'M','G' failures - PASS (white list)
            /var/log '5' failures - WARN (bad packaging).'''

            if path == '/var/log/gdm' and check in ["M", "G"]:
                return None, TestResult.PASS
            elif path == '/var/log' and check == "5":
                return "Warning", TestResult.WARN
            else:
                return "Needs Review", TestResult.REVIEW

        if path.startswith("/run/lsm"):
            '''Files under /var/lsm 'G' failures should be in allowlist 
               This be allowed on RHEL-[7.6, 8.1] and older versions 
            '''
            rhel_version = self.redHatRelease.getVersionPointUpdate()
            if check in ['G'] and float(rhel_version) < 7.7 or rhel_version in LSM_ALLOWED_VERSION:
                return None, TestResult.PASS

        if self.program in [Programs.openstack, Programs.baremetal]:
            '''Bypass auto-generated config files'''
            if path.startswith("/usr/share/openstack-dashboard/static/"):
                return None, TestResult.PASS

        if path.startswith("/boot/grub2/grubenv"):
            '''Bypass grub2 config files'''
            return None, TestResult.PASS

        return "Error", TestResult.FAIL

    def size_check(self, error):
        """
        S file Size differs
        FAIL if not bypassed in preliminary check(all programs)
        """
        msg = "found differed file size(S): {0}".format(error[-1])
        prefix, status = self.preliminary_check(error, "S")
        if prefix is not None:
            if self.output_directory:
                RPMVerification.rpm_verificaton_log.append(str("{0}: {1}".format(prefix, msg)))
            else:
                print(str("{0}: {1}".format(prefix, msg)))
        return status

    def md5_check(self, error):
        """
        5 digest (formerly MD5 sum) differs
        FAIL if not bypassed in preliminary check(all programs)
        """
        msg = "found differed MD5 sum(5): {0}".format(error[-1])
        prefix, status = self.preliminary_check(error, "5")
        if prefix is not None:
            if self.output_directory:
                RPMVerification.rpm_verificaton_log.append(str("{0}: {1}".format(prefix, msg)))
            else:
                print(str("{0}: {1}".format(prefix, msg)))
        return status

    def device_check(self, error):
        """
        D Device major/minor number mismatch
        FAIL if not bypassed in preliminary check(all programs)
        """
        msg = "found device major/minor number mismatch(D): {0}".format(error[-1])
        prefix, status = self.preliminary_check(error, "D")
        if prefix is not None:
            if self.output_directory:
                RPMVerification.rpm_verificaton_log.append(str("{0}: {1}".format(prefix, msg)))
            else:
                print(str("{0}: {1}".format(prefix, msg)))
        return status

    def readlink_check(self, error):
        """
        L readLink(2) path mismatch
        FAIL if not bypassed in preliminary check(all programs)
        """
        msg = "found readLink path mismatch(L): {0}".format(error[-1])
        prefix, status = self.preliminary_check(error, "L")
        if prefix is not None:
            if self.output_directory:
                RPMVerification.rpm_verificaton_log.append(str("{0}: {1}".format(prefix, msg)))
            else:
                print(str("{0}: {1}".format(prefix, msg)))
        return status

    def time_check(self, error):
        """
        T mTime differs
        WARN if not bypassed in preliminary check(all programs)
        """
        msg = "found differed modification time(T): {0}".format(error[-1])
        prefix, status = self.preliminary_check(error, "T")
        if prefix is not None:
            if prefix == "Error":
                prefix = "Warning"
                status = TestResult.WARN
            if self.output_directory:
                RPMVerification.rpm_verificaton_log.append(str("{0}: {1}".format(prefix, msg)))
            else:
                print(str("{0}: {1}".format(prefix, msg)))
        return status

    def user_ownership_check(self, error):
        """
        U User ownership differs
        If not bypassed in preliminary check(all programs):
        CCSP, HWCERT: FAIL
        RHOSP, BAREMETAL: WARN
        """
        msg = "found differed user ownership(U): {0}".format(error[-1])
        prefix, status = self.preliminary_check(error, "U")
        if prefix is not None:
            if self.program in [Programs.openstack, Programs.baremetal]:
                prefix = "Warning"
                status = TestResult.WARN
                if self.output_directory:
                    RPMVerification.rpm_verificaton_log.append(str("{0}: {1}".format(prefix, msg)))
                else:
                    print(str("{0}: {1}".format(prefix, msg)))
        return status

    def group_ownership_check(self, error):
        """
        G group ownership differs
        If not bypassed in preliminary check(all programs):
        CCSP, HWCERT: FAIL
        RHOSP, BAREMETAL: WARN
        """
        msg = "found differed group ownership(G): {0}".format(error[-1])
        prefix, status = self.preliminary_check(error, "G")
        if prefix is not None:
            if self.program in [Programs.openstack, Programs.baremetal]:
                prefix = "Warning"
                status = TestResult.WARN
            if self.output_directory:
                RPMVerification.rpm_verificaton_log.append(str("{0}: {1}".format(prefix, msg)))
            else:
                print(str("{0}: {1}".format(prefix, msg)))
        return status

    def capability_check(self, error):
        """
        P caPabilities differ
        If not bypassed in preliminary check(all programs):
        CCSP, HWCERT: FAIL
        RHOSP, BAREMETAL: WARN
        """
        msg = "found differed capabilities(P): {0}".format(error[-1])
        prefix, status = self.preliminary_check(error, "P")
        if prefix is not None:
            if self.program in [Programs.openstack, Programs.baremetal]:
                prefix = "Warning"
                status = TestResult.WARN
            if self.output_directory:
                RPMVerification.rpm_verificaton_log.append(str("{0}: {1}".format(prefix, msg)))
            else:
                print(str("{0}: {1}".format(prefix, msg)))
        return status

    def mode_check(self, error):
        """
        M Mode differs (includes permissions and file type)
        If not bypassed in preliminary check(all programs):
        RHOSP, BAREMETAL: WARN
        HWCERT: FAIL except for EFI partition (which will be vfat) then WARN
        CCSP: PASS for GPT partition, WARN for EFI partition (which will be vfat) else FAIL
        """
        msg = "found differed mode(M): {0}".format(error[-1])
        prefix, status = self.preliminary_check(error, "M")
        if prefix is not None:
            if self.program in [Programs.openstack, Programs.baremetal]:
                if self.output_directory:
                    RPMVerification.rpm_verificaton_log.append(str("Warning: %s" % msg))
                else:
                    print(str("Warning: %s" % msg))
                status = TestResult.WARN

            # For cloud program. Also a default criteria for any new program
            elif self.program not in [Programs.hardware, Programs._hardware]:
                # Check if GPT partition (hex code: ee) exists
                if self.is_partition(['ee', 'GPT']):
                    if self.output_directory:
                        RPMVerification.rpm_verificaton_log.append(str("INFO: Ignoring the error '%s' since GPT partition exists." % msg))
                    else:
                        print(str("INFO: Ignoring the error '%s' since GPT partition exists." % msg))
                    status = TestResult.PASS

            # Check if EFI partition (hex code: ef) exists
            elif self.is_partition(['ef', 'EFI']):
                if self.output_directory:
                    RPMVerification.rpm_verificaton_log.append(str("Warning: %s [EFI partition exists]" % msg))
                else:
                    print(str("Warning: %s [EFI partition exists]" % msg))
                status = TestResult.WARN

            else:
                if self.output_directory:
                    RPMVerification.rpm_verificaton_log.append(str("{0}: {1}".format(prefix, msg)))
                else:
                    print(str("{0}: {1}".format(prefix, msg)))
        return status

    @staticmethod
    def store_failures_into_a_file(output_directory, rpm_error_logs):
        log_file = os.path.join(output_directory, "rpm_verification_report.log")
        rpm_verification_log_file = open(log_file, "w+")
        # failures dumping to the file
        for error in rpm_error_logs:
            rpm_verification_log_file.write('\n' + str(error))
        rpm_verification_log_file.close()
        print("Please check the attached rpm_verification_report.log file for more details.")

    def is_partition(self, partion_type):
        """
        Checks if given partition exists
        Returns True, if yes.
        """
        cmd = Command("fdisk -l")
        try:
            cmd.run()
        except HwCertCommandErrorOutput:
            pass # ignore warning in stderr
        except HwCertCommandException as error:
            # non-zero return?
            print(error)
            return False
        file_data = cmd.output
        if not file_data:
            return False
        found = False
        for data in file_data:
            if not found:
                if 'Device Boot' and 'System' in data:
                    found = True
            else:
                if data and len(data.split()) > 1:
                    if data.split()[-2] in partion_type:
                        return True
                else:
                    found = False
        return False


def test_openstack_baremetal():
    print("Checking for OpenStack..")
    seperator = "-" * 40
    print(seperator)
    print("Error: Verification failed for 'openstack-tempest-10.0.0-3.b4a056dgit.el7ost.noarch'")
    print(seperator)
    errors = [
        "S.5....T. /usr/share/openstack-tempest-10.0.0/tempest/_init_.py",
        "S.L.5.. c /usr/share/openstack-tempest-10.0.0/tempest/api/_init_.py",
        "L.S.5.. /etc/openstack",
        "M.D.L..  /usr/share/openstack-tempest-10.0.0/tempest/api/baremetal/admin/test_ports_negative.py",
        "P.U.G.. /usr/share/openstack-tempest-10.0.0/tempest/api/compute/admin/test_agents.py"]
    verification_test = RPMVerification('openstack', errors)
    verification_test = RPMVerification('baremetal', errors)
    verification_test.rpm_verify()


def test_cloud_hardware():
    print("Checking for cloud/hardware..")
    seperator = "-" * 40
    print(seperator)
    print("Error: Verification failed for 'python-requests-2.6.0-1.el7_1.noarch'")
    print(seperator)
    errors = [
        "S.5....T. /usr/lib/python2.7/site-packages/requests/api.py",
        "S.L.5 c /usr/lib/python2.7/site-packages/requests/apiconf",
        "S..5.L. /etc/someconf",
        "M.D.L..  /usr/lib/python2.7/site-packages/requests/auth.py",
        "P.U.G.. /usr/lib/python2.7/site-packages/requests/adapters.py"]
    verification_test = RPMVerification('cloud', errors)
    verification_test.rpm_verify()


if __name__ == "__main__":
    test_openstack_baremetal()
    test_cloud_hardware()
