#!/usr/bin/env python
# Copyright (c) 2020 Red Hat, Inc. All rights reserved.
# This copyrighted material is made available to anyone
# wishing to use, modify, copy, or redistribute it subject
# to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.
#
# Author: Zahid Ajaz <zajaz@redhat.com>

import os, re
from rhcert.tags import TestResult
from rhcert.test import SubTest
from rhcert.command import Command, HwCertCommandException
from rhcert.tags import Constants


class RPMVersionHandlingTest(SubTest):

    def __init__(self, test):
        SubTest.__init__(self, test, "RPM Version Handling",
                         "Test RPM Versions")

    def inspect_rpm_changelog(self, rpm):
        result = TestResult()
        cmd = Command("rpm -q --changelog %s" % rpm)
        regex = r"^([\*\w\s\.-]*)(\<*[\w\.-]+@[\w\.-]+(\.[\w]+)+\>*)([\s\w.\-]+)"
        cmd_output = self.test.get_string_list(cmd, regex)
        if not cmd_output:
            print("\nError: RPM '%s' doesn't show any changelog" %rpm)
            result.combine(TestResult.FAIL)
            return result
        meta = self.test.get_rpm_meta(rpm, ['VERSION'])
        if len(cmd_output)>1:
            if not any(meta['VERSION'] in vstr for vstr in cmd_output):
                print("\nError: The current NVRE '%s' for '%s' isn't reflected changelog" %(meta['VERSION'],rpm))
                result.combine(TestResult.FAIL)
        elif len(cmd_output)==1 and meta['VERSION'] in cmd_output[0]:
            print("\nWarning: RPM '%s' doesn't list any other versions in changelog" %rpm)
            result.combine(TestResult.WARN)
        else:
            pass
        return result

    def inspect_rpm_version(self, rpm):
        result = TestResult()
        rpm_meta_list = ['NAME', 'VERSION', 'RELEASE' , 'ARCH']
        rpm_data = self.test.get_rpm_meta(rpm, rpm_meta_list)
        
        # Check if package provides version, relase information
        if not (rpm_data['VERSION'] and rpm_data['RELEASE']):
            print("\nError: RPM '%s' version/release information missing" %rpm)
            result.combine(TestResult.FAIL)

        # Check if package provides $NAME = $VERSION(long) , $NAME{$ARCH} = $VERSION(long)
        cmd = Command("rpm -q --provides %s" % rpm)
        if rpm_data['NAME']:
            regex = r"^(%s)([(\w\d\s\-)]*)\=([\w\d.\s\-]+)"%rpm_data['NAME']
            cmd_output =self.test.get_string_list(cmd, regex)
            if not cmd_output:
                print("\nError: RPM '%s' couldn't output to 'rpm -q --provides'" %rpm)
                result.combine(TestResult.FAIL)
            elif len(cmd_output)<2:
                print("\nError: RPM '%s' doesn't list significant version/architecture info" %rpm)
                result.combine(TestResult.FAIL)
            else:
                pass
        return result

    def runSubTest(self):
        """
        Cases with their return values:

        """
        result = TestResult()
        if not self.test.user_rpms:
            print("\nError: No RPM(s) specified as application software")
            return TestResult.FAIL
        for frpm in self.test.foreign_rpms:
            result.combine(self.inspect_rpm_version(frpm))
            result.combine(self.inspect_rpm_changelog(frpm))  
        return result
