#!/usr/bin/env python
# Copyright (c) 2016 Red Hat, Inc. All rights reserved.
# This copyrighted material is made available to anyone
# wishing to use, modify, copy, or redistribute it subject
# to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.
#
# Author: Pooja Tiwari <ptiwari@redhat.com>

""" This module calls cloud security tests. """

import sys
from rhcert.test import SubTest
from rhcert.tags import TestResult
from rhcert.command import Command, HwCertCommandException
from rhcert.suites.rhcert.platform import CheckSELinux
from rhcert.suites.rhcert.rpm_freshness import RPMFreshnessTest


class PamValidationTest(SubTest):
    """
    Pam Validation Test Class
    """

    def __init__(self, test):
        SubTest.__init__(self, test, "Password Configuration",
                         "Check for Valid Password Hash")
        self.expected_hash = self.test.policy["PASSWORD_HASH"]

    def runSubTest(self):
        """
        PAM validation test.
        """
        # assume RHEL8/authselect, fail-over to RHEL6/7 authconfig
        authSelect = Command("authselect list")
        try:
            authSelect.run()
        except HwCertCommandException:
            return self.__testAuthConfig()
        return self.__testAuthSelect()

    def __testAuthConfig(self):
        try:
            configured_hash = Command("authconfig --test|grep hashing")
            configured_hash.run()
            configured_hash = configured_hash.output[0].split()[-1]
        except HwCertCommandException as error:
            print("Error: Could not determine configured password hashing algorithm.")
            print("%s" % error)
            return TestResult.FAIL 

        if configured_hash.lower() in self.expected_hash:
            print("PAM validation passed.\nSupported password hash '%s' found.\n" % configured_hash)
            return TestResult.PASS 
        else:
            print("Error: Unsupported password hash '%s' found.\nRed Hat supports 'SHA512'." % configured_hash)
            return TestResult.FAIL

    def __testAuthSelect(self):
        try:
            authSelectCurrent = Command("authselect current")
            authSelectCurrent.echo()
            regex = "^Profile ID:\s+(?P<id>.*)$"
            currentProfile = authSelectCurrent.getString(regex=regex, regexGroup='id', singleLine=False)
        except HwCertCommandException as error:
            print("Error: Could not determine the configured authentication profile.")
            print("%s\n" % error)
            print("Corrective Action: Please select and activate appropriate authentication profile."
                  "See 'authselect --help'")
            return TestResult.FAIL

        try:
            authSelectTest = Command("authselect test %s -p" % currentProfile.strip())
            regex = "^password\s+sufficient\s+(?P<values>.*)$"
            values = authSelectTest.getString(regex=regex, regexGroup="values", singleLine=False)
        except HwCertCommandException as error:
            print("Error: Could not determine the configured password hashing algorithm.")
            print("%s" % error)
            return TestResult.FAIL

        for algorithm in self.expected_hash:
            if algorithm in values.split():
                print("PAM validation passed.\nSupported password hash '%s' found.\n" % algorithm)
                return TestResult.PASS
        else:
            print("Error: Red Hat supported algorithm(s) '%s' not found." % (", ".join(self.expected_hash)))
            return TestResult.FAIL


def unit_test():
    test = PamValidationTest()
    return test.do(sys.argv)
