#!/usr/bin/env python
# Copyright (c) 2016 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
import os, sys, string, shutil, re

from rhcert.tags import Constants, TestTag, TestResult
from rhcert.command import Command, HwCertCommandException, HwCertCommandErrorReturned, HwCertCommandErrorOutput
from rhcert.suites.rhcert.rpm_verification import RPMVerification
from rhcert.environment import Environment
from rhcert.test import Test, SubTest


class SelfCheck(Test):

    def __init__(self):
        Test.__init__(self, "rhcert/self_check")
        self.mandatory = Constants.pre # required at the beginning of each run
        # what RHEL are we testing?
        self.priority = 1 # run first
        self.environment = Environment()

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification]

    def plan(self, planner):
        tests = list()
        tests.append(self.makeCopy())
        return tests

    def run(self):
        result = TestResult()

        for subtest in [CheckRHCertPackages(self)]:
            result.combine(subtest.run())

        return result


class CheckRHCertPackages(SubTest):
    def __init__(self, test):
        SubTest.__init__(self, test, "Verify Test Suite", "Verify the redhat-certification packages")
        self.program = test.getSuiteName()

    def runSubTest(self):
        result = TestResult()
        result.combine(TestResult.PASS)
        rpmSearch = Command("rpm -qa | fgrep redhat-certification")
        try:
            rpmSearch.run()
            packages = rpmSearch.getStringList("[^\s]+")
        except HwCertCommandException as error:
            print(error)
            return TestResult.FAIL

        # otherwise
        for package in packages:
            print("\nChecking %s:" % package)
            rpm_verify_silent = Command(
                "rpm -V > /dev/null " + package)
            rpm_verify = Command("rpm -V " + package)
            try:
                rpm_verify_silent.run()
            except HwCertCommandErrorReturned:
                result.combine(self.report_rpm_verification_failures(rpm_verify,
                                                                     package))
            except HwCertCommandErrorOutput as stderr:
                result.combine(self.report_rpm_verification_failures(rpm_verify,
                                                                     package,
                                                                     stderr._get_message()))
            except HwCertCommandException as err:
                print("Could not perform verification on package '%s'" %package)
                print(err)
                result.combine(TestResult.FAIL)

        if result == TestResult.PASS:
            print("\nVerified all packages.\n")
        else:
            print("\nCorrective Action: Please re-install these packages.\n")

        self.check_configuration()
        return result


    def report_rpm_verification_failures(self, rpm_verify, package, stderr=None):
        result = TestResult()
        result.combine(TestResult.PASS)
        seperator = '-' * 50
        errors = rpm_verify.getIgnoreErrors()
        for error in errors[:]:
            if error.startswith("missing") or error.endswith(".pyc") or error.endswith(".pyo"):
                errors.remove(error)
        if errors:
            print(seperator)
            print("\nVerification failed for '%s'" % (package))
            print(seperator)
            verification_test = RPMVerification(self.program, errors)
            result.combine(verification_test.rpm_verify())
            if stderr:
                result.combine(TestResult.REVIEW)
            if result == TestResult.PASS:
                print("INFO: No critical failure occurred.")
            elif stderr:
                for line in stderr:
                    print("Needs Review: (stderr) %s" %line)
            result.combine(self.save_modified_test_files(errors))
        return result


    def save_modified_test_files(self, errors):
        # 451074: FEAT: save any modified test files
        pyFilePattern= re.compile(".*\.py$")
        found = False
        result = TestResult()
        for line in errors:
            tokens = line.split()
            if len(tokens) > 1 and pyFilePattern.match(tokens[1]):
                if not found:
                    found = True
                    print("\nSaving following modified source files:")
                print(tokens[1])
                try:
                    shutil.copy(tokens[1], self.test.getOutputDirectory())
                except IOError as error:
                    print("Error: could not save modified (or missing?) file:")
                    print(error)
                    result.combine(TestResult.FAIL)
        return result

    def check_configuration(self):
        print("\nChecking redhat-certification configuration:")
        if not self.test.environment.showChangesFromDefault():
            print("Using defaults.")
        else:
            print("Using customized.")

def unit_test():
    test = SelfCheck()
    return test.do(sys.argv)

if __name__ == "__main__":
    sys.exit(unit_test())

