#!/usr/bin/env python
# Copyright (c) 2019 Red Hat, Inc. All rights reserved.
# This copyrighted material is made available to anyone
# wishing to use, modify, copy, or redistribute it subject
# to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.
#
# Author: Pooja Tiwari <ptiwari@redhat.com>
#         Chandan Kumar <chkumar@redhat.com>
#         Javed Alam <jalam@redhat.com>

""" This module calls supportibility tests. """
from __future__ import absolute_import
from rhcert.redhatrelease import RedHatRelease
import re
import os

try:  # python2
    from ConfigParser import ConfigParser
except ImportError:  # python3
    from configparser import ConfigParser
import math

from rhcert.compatability import urllib_urlopen

from rhcert.tags import Constants, Programs, TestResult
from rhcert.test import SubTest
from rhcert.command import HwCertCommandErrorReturned, HwCertCommandException
from rhcert.suites.rhcert.container_test import ContainerTest
from rhcert.suites.rhcert.rpm_verification import RPMVerification
from rhcert.command import Command, HwCertCommandException as RHCertCommandException
from rhcert.compatability import rhcert_make_unicode



def check_hypervisor():
    """
    Get the output of virt-what to
    check the underlined hypervisor.
    """

    virt_type = Command("virt-what")
    try:
        virt_type.run()
        if not virt_type.output:
            return ["baremetal"]
        else:
            return virt_type.output
    except HwCertCommandException as err:
        print("Error: Could not determine the underlined hypervisor: %s" % err)
        return None


class FileSystemTest(SubTest):
    """
    File Type Validation Test class
    """

    def __init__(self, test):
        SubTest.__init__(self, test,
                         "Filesystem Layout",
                         "Check for Red Hat Supported File Type and Size")
        self.supported_file_type = self.test.policy["FILE_TYPE"]
        self.desired_disk_size = self.test.policy["FILESYS_SIZE"]
        self.expected_boot_partition_size = self.test.policy["BOOT_PARTITION_SIZE"]
        self.expected_root_partition_size = self.test.policy["ROOT_PARTITION_SIZE"]
        self.system_logs = str()
        self.file_type = str()
        self.actual_disk_size = dict()
        self.release = RedHatRelease()

    def get_file_type(self):
        """
        Parse mount command output to get filetype
        """
        mountCommand = Command("mount")
        try:
            self.file_type = mountCommand.run()
            mount_data = mountCommand.output
        except Exception as error:
            print("Error: mount command failed:")
            print(error)
            return None

        file_data = []
        keys = ['filesystem', 'mountpoint', 'type', 'arguments']
        for files in mount_data[:-1]:
            data = re.split(r' on | type | ', files)
            file_type = dict(zip(keys, data))
            file_data.append(file_type)

        for filetype in file_data:
            if filetype['mountpoint'] == '/':
                return filetype['type']

    def check_file_type(self):
        """
        Validate File type
        """
        print("\n*** Checking File Type ***\n")
        rhel_version = "RHEL-%s" % self.release.getVersion()
        actual_file_type = self.get_file_type()
        supported_file_type = self.supported_file_type[rhel_version]

        if actual_file_type in supported_file_type:
            msg = ("File Type Validation passed.\n" +
                   "Supported File Type '{0}' found.\n"
                   ).format(actual_file_type)
            print(msg)
            return True
        else:
            msg = ("Error: Unsupported File Type '{0}' found.\n" +
                   "Red Hat supported file types for {1}: {2}.").format(
                actual_file_type,
                rhel_version,
                ", ".join(supported_file_type))
            print(msg)
            return False

    def get_filesys_data(self, command):
        """
        Collect the output of command `df`
        """
        df = Command(command)
        try:
            df.run()
        except Exception as error:
            print("'%s' command execution failed" % command)
            print(error)
            return []

        df_output = list(filter(None, df.output))
        file_list = []
        new_list = []
        for content in df_output[1:]:
            if re.search(r'^\W', df_output[df_output.index(content) - 1]) and re.search(r'^\s', content):
                value = df_output[df_output.index(content) - 1] + content
                new_list.append(value)
                new_list.remove(df_output[df_output.index(content) - 1])
            else:
                new_list.append(content)

        keys = re.split(r'\s+', df_output[0], maxsplit=5)
        for data in new_list:
            contents = re.split(r'\s+', data, maxsplit=5)
            df_output = dict(zip(keys, contents))
            file_list.append(df_output)
        return file_list

    def check_size(self, filesys_data):
        """
        Get used and total disk size
        """
        available, used = 0, 0
        for data in filesys_data:
            if 'Available' in data.keys() and data['Available'].isdigit():
                available += float(data['Available'])
            else:
                available += 0

            if 'Used' in data.keys() and data['Used'].isdigit():
                used += float(data['Used'])
            else:
                used += 0

        total = available + used

        return {'available': available / 1024, 'total': total / 1024}

    def check_disk_size(self):
        """
        Validate the Disk Size.
        """

        print("\n*** Checking Disk Size ***\n")
        try:
            self.actual_disk_size = self.check_size(self.get_filesys_data("df -al"))
            if self.actual_disk_size['total'] >= self.desired_disk_size[0]:
                msg = ("Disk size validation passed.\n" +
                       "Found following data:\n" +
                       "------------------------------\n" +
                       "Total space: {0} MB\n" +
                       "Available space: {1} MB").format(
                    int(math.ceil(self.actual_disk_size['total'])),
                    int(math.ceil(self.actual_disk_size['available'])))
                print(msg)
                return True
            else:
                msg = ("Error: Total space: '{0} MB' found. " +
                       "\nSize should not be less than '{1} MB'.").format(
                    int(math.ceil(self.actual_disk_size['total'])),
                    str(self.desired_disk_size[0]))
                print(msg)
                return False
        except Exception as err:
            print(err)

    def check_partitions(self):
        """
        Check file system layout and partition size
        """
        result = TestResult()
        boot_partition_size = root_partition_size = 0
        valid_boot_partition = valid_root_partition = False
        try:
            print("\n*** Checking File System Layout ***\n")

            blocks = self.get_filesys_data("df --block-size=GB")

            for block in blocks:
                block_size = int(block['1GB-blocks'].replace('GB', ''))

                if block['Mounted on'] == '/':
                    root_partition_size = block_size
                    valid_root_partition = True if block_size >= self.expected_root_partition_size else False

                if block['Mounted on'] == '/boot':
                    boot_partition_size = block_size
                    valid_boot_partition = True if block_size >= self.expected_boot_partition_size else False

            partition_details = ("\nRoot '/' partition size: {0} GB\n" +
                                 "Boot '/boot' partition size: {1} GB\n").format(root_partition_size,
                                                                                 boot_partition_size)

            if valid_boot_partition and valid_root_partition:
                result.combine(TestResult.PASS)
                msg = ("%s File system layout conforms with recommended standards\n"
                       % TestResult(result).getMessagePrefix())
                print(msg + '-' * len(msg) + partition_details)
            else:
                result.combine(TestResult.FAIL)
                msg = ("%s File system layout NOT in conformation with recommended standards\n"
                       % TestResult(result).getMessagePrefix())
                print(msg + '-' * len(msg) + partition_details)

        except Exception as error:
            result.combine(TestResult.FAIL)
            print("%s Failure in test execution\n" % TestResult(result).getMessagePrefix())
            print(error)

        return result

    def runSubTest(self):
        """
        Disk size validation test.
        """

        result = TestResult()
        subtests = [self.check_file_type()]

        # RHEL-8 specific sub-tests
        if int(self.release.getVersion()) >= 8:
            subtests.append(self.check_partitions())  # Disk size validation for release >= 8
        else:
            subtests.append(self.check_disk_size())  # Disk size validation for release < 8
        for subtest in subtests:
            result.combine(subtest)

        return result


class OSArchitectureTest(SubTest):
    """
    OS architecture Test class
    """

    def __init__(self, test):
        SubTest.__init__(self, test, "Hypervisor/Partitioning",
                         "Validates the system's architecture over the underlined hypervisor.")
        self.desired_arch = self.test.policy["OS_ARCH"]
        self.actual_arch = str()
        self.result = TestResult()
        self.release = RedHatRelease()

    def get_system_logs(self):
        """
        Get the system's architecture.
        """

        try:
            self.actual_arch = Command("uname -m").getString()
        except Exception as error:
            print("Error: uname command failed:")
            print(error)

    def validate_logs(self):
        """
        Validate the OS Architecture.
        """

        rhel_release = "el%s" % self.release.getVersion()

        virt_types = check_hypervisor()
        if not virt_types:
            return TestResult.FAIL

        if self.actual_arch in self.desired_arch[rhel_release].keys():
            for virt_type in virt_types:
                if virt_type.lower() in self.desired_arch[rhel_release][self.actual_arch]:
                    msg = ("The architecture '{0}' on '{1}' hypervisor is supported."
                           "\n".format(self.actual_arch, virt_type))
                    print(msg)
                    self.result.combine(TestResult.PASS)
                else:
                    if self.test.getSuiteName() == Programs.cloud and \
                        virt_type.lower() == 'baremetal' and self.actual_arch in ['x86', 'x86_64']:
                            self.result.combine(TestResult.REVIEW)
                    else:
                        self.result.combine(TestResult.FAIL)
                    msg = ("{0}: The architecture '{1}' on '{2}' hypervisor is not supported. " \
                        "It is supported over {3} hypervisors.\n").format(
                            self.result.getMessagePrefix(),
                            self.actual_arch, virt_type,
                            ", ".join(self.desired_arch[rhel_release][self.actual_arch]))
                    print(msg)

        else:
            msg = ("Error: Unsupported architecture '{0}' found." +
                   "\nRed Hat supported architectures: {1}.").format(
                   self.actual_arch,
                   ", ".join(self.desired_arch[rhel_release].keys()))
            print(msg)
            self.result.combine(TestResult.FAIL)
        return self.result

    def runSubTest(self):
        """
        OS Architecture validation test.
        """

        self.get_system_logs()
        return self.validate_logs()


class SoftwareRepositories(SubTest):
    """
    Software Repositories Test class
    """

    def __init__(self, test):
        SubTest.__init__(self, test, "Software Repositories",
                         "Check Configured Yum Repositories")
        self.rh_repos = dict()
        self.non_rh_repos = dict()
        self.release = RedHatRelease()

    def get_repos(self):
        """
        Get yum repos with their baseurl and mirrorlist
        """

        configParser = ConfigParser()
        yum_repo_path = '/etc/yum.repos.d'
        repos = {}
        files = [file for file in os.listdir(yum_repo_path)
                 if file.endswith('.repo')]

        for file in files:
            configParser.read(os.path.join(yum_repo_path, file))

            if configParser.sections():
                for section in configParser.sections():
                    url = ""
                    name = configParser.get(section, 'name')
                    status = configParser.get(section, 'enabled')
                    if configParser.has_option(section, 'baseurl'):
                        url = configParser.get(section, 'baseurl')
                    elif configParser.has_option(section, 'mirrorlist'):
                        url = configParser.get(section, 'mirrorlist')
                    repos.update({section: {"name": name, "url": url, "status": status}})
                    if configParser.has_option(section, 'gpgcheck'):
                        gpgcheck = configParser.get(section, 'gpgcheck')
                        repos[section].update({'gpgcheck': gpgcheck})
                    if configParser.has_option(section, 'gpgkey'):
                        gpgkey = configParser.get(section, 'gpgkey')
                        repos[section].update({'gpgkey': gpgkey})

        if repos:
            for repo, data in repos.items():
                url = self.is_redhat_repo(data["url"])
                if not url:
                    self.non_rh_repos.update({repo: data})
                else:
                    data["url"] = url
                    self.rh_repos.update({repo: data})

    def is_redhat_repo(self, url):
        """
        Check for valid yum repos
        """
        for redhathost in self.test.policy['RH_URLS']:
            if redhathost in url:
                return url.replace(redhathost, redhathost.upper())
        return None

    def check_redhat_repos(self):
        """
        Check if Red Hat repos are configured.
        Return False if no.
        """
        if not self.rh_repos:
            print("Error: No Red Hat repository configured.")
            print("Corrective Action: Configure and enable required Red Hat repositories.")
            return False
        msg = "Following Red Hat repositories were found"
        self.print_repos(self.rh_repos, msg)
        return True

    def check_non_redhat_repos(self):
        """
        Check if non Red Hat repos configured.
        ask for review/explanation for the presence of the repo.
        """
        if self.non_rh_repos:
            flag = "Non-Red Hat repositories"
            if any('redhat' in data['url'].lower() for repo, data in self.non_rh_repos.items()):
                flag = "Unsupported repositories"
            msg = "Following {0} have been found:".format(flag)
            self.print_repos(self.non_rh_repos, msg)
            _msg = ("\nNeeds Review: list of {0} repositories:\n{1}\n"
                    "Please describe the need for these repositories:".format(
                        flag, "\n".join(self.non_rh_repos.keys())))
            if self.test.mode != Constants.auto:
                if self.test.mode == Constants.server:
                    print(_msg)
                response = self.test.prompt(_msg)
                while not response:
                    response = self.test.prompt("Please describe the need for these repositories: ")
            else:
                # For Constants.auto mode
                print(_msg)

            return TestResult.REVIEW

        return TestResult.PASS

    def gpg_check_for_repos(self, repos, check_for='Red Hat repos'):
        """
        Environment GPG Keys test.
        """

        status = 'Error'
        failed_status = TestResult.FAIL
        if check_for == 'Non Red Hat repos':
            status = 'Warning'
            failed_status = TestResult.WARN

        print("\n**** Performing GPG Check for %s ****" % check_for)

        result = TestResult()
        result.combine(TestResult.PASS)
        missing_gpgkeys_repos = dict()
        invalid_gpgkeys_repos = dict()
        disabled_gpgcheck_repos = dict()
        for repo, data in repos.items():
            if not data.get("gpgcheck") or data["gpgcheck"] == '0':
                disabled_gpgcheck_repos.update({repo: data})
            if not data.get("gpgkey"):
                missing_gpgkeys_repos.update({repo: data})
            elif not data["gpgkey"]:
                missing_gpgkeys_repos.update({repo: data})
            else:
                keys = data["gpgkey"].split(',')
                for key in keys:
                    if key.startswith("file:"):
                        if not os.path.isfile(key.split(":")[1]):
                            invalid_gpgkeys_repos.update({repo: data})
                    elif key.startswith("ftp:") or key.startswith("http:") or key.startswith("https:"):
                        try:
                            if not urllib_urlopen(key).code == 200:
                                invalid_gpgkeys_repos.update({repo: data})
                        except:
                            invalid_gpgkeys_repos.update({repo: data})
                    else:
                        if not os.path.isfile(key):
                            invalid_gpgkeys_repos.update({repo: data})
        if disabled_gpgcheck_repos:
            msg = "%s: GPG key check is disabled for following repos" % status
            self.print_repos(disabled_gpgcheck_repos, msg)
            result.combine(failed_status)

        if missing_gpgkeys_repos:
            msg = "%s: GPG keys are missing for following repos" % status
            self.print_repos(missing_gpgkeys_repos, msg)
            result.combine(failed_status)

        if invalid_gpgkeys_repos:
            msg = "%s: Invalid gpg key for following repos" % status
            self.print_repos(invalid_gpgkeys_repos, msg)
            result.combine(failed_status)

        if result == failed_status:
            print("\nCorrective Action: Please enable GPG checks and configure valid keys for the mentioned repos.\n")
        return result

    def get_rhel_sap_support_extended(self):
        """
        Provides RHEL for SAP extended support info
        """
        if self.test.getSuiteName() == Programs.cloud_sap:
            rhel_version = self.release.getVersionPointUpdate()
            if rhel_version == "7.5":
                return ['eus']
            if float(rhel_version) < 8.0:
                return ['eus', 'e4s']
            if float(rhel_version) >= 8.0:
                return ['e4s']
        return

    def check_required_repos_enabled(self, repos):
        """
        Check if required RHEL repositories are enabled
        """
        result = TestResult()
        version_label = "RHEL-{0}".format(self.release.getVersion())
        repo_data = self.test.policy["RH_REPOS"]
        required_repos = repo_data["REQUIRED"][version_label]
        if required_repos:
            print("\n**** Verify required RHEL repositories are enabled ****")
            # Additional Checks for RHEL SAP Repos
            extended_support =  self.get_rhel_sap_support_extended() if repo_data.get("EXTENDED_SUPPORT") else []
            # Check whether required repos are enabled
            enabled_repos = set()
            for repository, data in repos.items():
                repo_url = data.get('url').lower()
                repo_name = data.get('name').lower()
                repo_status = int(data.get('status'))
                for repo in required_repos:
                    if extended_support:
                        if any(support_type in repo_url for support_type in extended_support) and \
                            (any(repo.lower() in repo_info for repo_info in [repo_name, repo_url])) and \
                                (repo_status == 1):
                                    enabled_repos.add(repo)
                    else:
                        if ((any(repo.lower() in repo_info for repo_info in [repo_name, repo_url]))) and \
                            (repo_status == 1):
                                enabled_repos.add(repo)
            disabled_repos = list(set(required_repos) - enabled_repos)
            if disabled_repos:
                print("\nError: Following required Red Hat repositories are not enabled: %s" % ', '.join(disabled_repos))
                print("Corrective Action: Please enable the above listed repositories")
                if extended_support:
                    print("Ensure that enabled repositories are %s supported" % ",".join(extended_support))
                result.combine(TestResult.FAIL)
            else:
                print("\nSuccess: Required RHEL repositories are enabled")
        return result

    def print_repos(self, repos, msg):
        print(msg)
        id_width = 60
        name_width = 90
        status_width = 11
        gpgcheck_width = 11
        gpgkey_width = 100
        baseurl_width = 100
        _msg = str()
        feature_items = list()
        separator = "+{0}+{1}+{2}+{3}+{4}+{5}+".format("-" * id_width,
                                                       "-" * name_width,
                                                       "-" * status_width,
                                                       "-" * gpgcheck_width,
                                                       "-" * gpgkey_width,
                                                       "-" * baseurl_width)

        print(separator)
        print("|{0}|{1}|{2}|{3}|{4}|{5}|".format("Repo ID".ljust(id_width),
                                                 "Repo Name".ljust(name_width),
                                                 "Is Enabled?".ljust(status_width),
                                                 "GPG Status".ljust(gpgcheck_width),
                                                 "GPG Key".ljust(gpgkey_width),
                                                 "Base URL".ljust(baseurl_width)))
        for repo_id, repo_data in repos.items():
            feature_item = dict()
            feature_item["id"] = repo_id
            feature_item["name"] = repo_data["name"]
            feature_item["status"] = "enabled" if repo_data.get("status") == '1' else "disabled"
            feature_items.append(feature_item)
            _msg = (
                "\n|{0}|{1}|{2}|{3}|{4}|{5}|").format(
                repo_id.ljust(id_width),
                repo_data["name"].ljust(name_width),
                "Yes".ljust(status_width) if repo_data.get("status") == '1' else "No".ljust(status_width),
                repo_data['gpgcheck'].ljust(gpgcheck_width) if repo_data.get("gpgcheck") else ''.ljust(gpgcheck_width),
                repo_data['gpgkey'].ljust(gpgkey_width) if repo_data.get("gpgkey") else ''.ljust(gpgkey_width),
                repo_data["url"].ljust(baseurl_width))
            print(separator + _msg)
        print(separator + "\n")
        self.logFeatureData("Software Repositories", feature_items)

    def runSubTest(self):
        """
        Validate repos.
        """
        result = TestResult()
        self.get_repos()
        if self.check_redhat_repos():
            result.combine(self.gpg_check_for_repos(self.rh_repos))
        else:
            result.combine(TestResult.FAIL)
        result.combine(self.check_non_redhat_repos())
        result.combine(self.check_required_repos_enabled(self.rh_repos))
        if self.non_rh_repos:
            result.combine(self.gpg_check_for_repos(self.non_rh_repos, check_for="Non Red Hat repos"))
        return result


class SoftwareContainers(SubTest):
    """
    Software Containers Test class
    """

    def __init__(self, test):
        SubTest.__init__(self, test, "Software Containers",
                         "Validate Software Containers")
        self.container_tool = 'podman'
        self.rh_container_registries = self.test.policy["RH_CONTAINER_REGISTRIES"]
        self.container_test = ContainerTest(self.container_tool)

    def required_packages(self):
        print("\n*** Checking for required container tool package '%s' ***\n" % self.container_tool)
        command = Command("rpm -q %s" % self.container_tool)
        try:
            command.run()
        except HwCertCommandErrorReturned:
            print("Error: container tool package '%s' not installed" % self.container_tool)
            print("Corrective Action: Please install the required container tool package '%s'" % self.container_tool)
            return False
        else:
            print("Installed version of '%s': %s\n" % (self.container_tool, command.raw_output))
        return True

    def validate_configured_registries(self):
        print("\n*** Checking for configured container registries***\n")
        result = TestResult()
        config = ConfigParser()
        container_registries_config = '/etc/containers/registries.conf'
        config.read(container_registries_config)
        container_registries = list()
        if config.sections():
            container_registries = config.get('registries.search', 'registries')
        else:
            print("Error: invalid container registries config.\n")
            result.combine(TestResult.FAIL)
        if container_registries:
            if self.rh_container_registries["PRIMARY"] not in container_registries:
                print("Error: required container registry '%s' is not enabled" % self.rh_container_registries["PRIMARY"])
                result.combine(TestResult.FAIL)
            if self.rh_container_registries["OPTIONAL"] not in container_registries:
                result.combine(TestResult.WARN)
                print("Warning: container registry '%s' for third-party products is not enabled" %
                      self.rh_container_registries["OPTIONAL"])
        if result == TestResult.FAIL:
            print("Corrective Action: Please ensure that required container registry '%s' is enabled" %
                  self.rh_container_registries["PRIMARY"])
        return result

    def verify_registry_accessibility(self):
        # TODO: This check is likely to be enabled in future
        print("\n*** Verifying operability of container registries ***\n")
        result = TestResult()
        namespace = "rhel8-beta"  # test container image namespace
        image = os.path.join(self.rh_container_registries["PRIMARY"], namespace)
        pull = Command("%s pull %s 2>&1 >/dev/null" % (self.container_tool, image))
        try:
            pull.run()
        except HwCertCommandException as error:
            print("Error: failed to pull container image '%s' from registry" % image)
            print("\nCorrective Action: Please make sure that you are authenticated to access container registries"
                  "\nInfo: Use '%s login' command to authenticate" % self.container_tool)
            result.combine(TestResult.FAIL)
        else:
            print("Success: '%s' was able to pull container image '%s' from registry %s"
                  % (self.container_tool, image, self.rh_container_registries["PRIMARY"]))
        return result

    def runSubTest(self):
        """
        Validate containers.
        """
        result = TestResult()

        # check-0
        # required container_tool is installed
        # required registries are configured and enabled
        if not self.required_packages():
            result.combine(TestResult.FAIL)
            return result
        result.combine(self.validate_configured_registries())

        # check-[1, 2]
        result.combine(self.container_test.log_available_images())
        self.available_containers, return_value_1 = self.container_test.log_available_containers(
            self.test.getSuiteName())
        result.combine(return_value_1)
        if not self.available_containers:
            return result
        self.containers_and_images_info, return_value_2 = self.container_test.log_image_inspect(
            self.available_containers)
        result.combine(return_value_2)
        if return_value_2 == TestResult.FAIL:
            return result

        # check-3
        self.rh_containers, return_value_3 = self.container_test.detect_nonRH_containers(
            self.containers_and_images_info,
            self.test)
        result.combine(return_value_3)

        return result


class SoftwareModules(SoftwareRepositories):
    """
    Software Modules Test class
    """

    def __init__(self, test):
        super(SoftwareModules, self).__init__(test)
        SubTest.__init__(self, test, "Software Modules",
                         "Check And Validate Installed Software Modules")
        self.non_rht_modules = list()
        self.unidentified_modules = list()

    def get_modules(self):
        """
        Get installed modules
        """
        print("\n*** Checking for installed Software Modules ***\n")
        modules = list()
        keys = ['Name', 'Stream', 'Profiles', 'Summary']
        regex_start = "\s+".join(keys)
        regex_end = "Hint:"
        get_module_list = Command("yum module list | awk 'BEGIN{logger=0} /%s/{logger=1} "
                                  "/%s/{logger=0} {if (logger) print }'" % (regex_start, regex_end))
        try:
            get_module_list.run(ignoreWarnings=True)
        except HwCertCommandException as e:
            print("Error: failed to list the installed software modules\n")
            print(e)
            return None
        # Get redhat/non-redhat repos information
        self.get_repos()
        self.redhat_repos = list(self.rh_repos.keys())
        self.non_redhat_repos = list(self.non_rh_repos.keys())
        repo_names = [repo['name'] for repo in self.rh_repos.values()] + [repo['name'] for repo in self.non_rh_repos.values()]
        # Label filters for get_module_list command output
        filters = ['Name', 'Red Hat Enterprise Linux'] + self.redhat_repos + self.non_redhat_repos + repo_names

        for line in get_module_list.output or []:
            if line.startswith(tuple(filters)):
                continue
            else:
                values = re.split(r'\s{2,}', line)
                modules.append(dict(zip(keys, values)))
        return modules

    def get_module_source_repos(self, module):
        """
        Returns module meta data (source-repo information)
        TODO: Use libmodulemd to get module metadata instead of command output
        """
        source_repos = set()
        info = Command("yum module info {0} | awk '/Repo\s+:/'".format(module))
        try:
            info.run(ignoreWarnings=True)
        except HwCertCommandException as e:
            print("Error: failed to determine module metadata for %s\n" % module)
            print(e)
            return source_repos

        for source_repo in info.output or []:
            source_repos.add(source_repo.split(":")[-1].strip())
        return list(filter(None, source_repos))

    def validate_source_repos(self, repos):
        """
        Validate if a module is provided from a valid RHT repo
        """
        for repo in repos:
            if repo not in self.redhat_repos:
                return False
            continue
        return True

    def check_non_redhat_modules(self):
        """
        Check if Non-RedHat modules are present.
        ask for review/explanation for the presence of the module.
        """
        modules = self.get_modules()
        result = TestResult()
        if modules:
            for module in modules:
                if module['Name']:
                    source_repos = self.get_module_source_repos(module['Name'])
                    if source_repos:
                        if self.validate_source_repos(source_repos):
                            continue
                        else:
                            module['Repos'] = source_repos
                            self.non_rht_modules.append(module)
                            result.combine(TestResult.FAIL)
                    else:
                        self.unidentified_modules.append(module)
                        result.combine(TestResult.REVIEW)
        else:
            result.combine(TestResult.FAIL)
            print("%s Not able to determine and validate the software modules" % TestResult(result).getMessagePrefix())
            return result
        if self.non_rht_modules or self.unidentified_modules:
            msg = "%s Following non-redhat modules were found:" % TestResult(result).getMessagePrefix()
            corrective_action = "Corrective Action: Please ensure that each software module is provided" \
                                " from a valid redhat repo"
            self.print_error_info(msg, corrective_action)
        else:  # TestResult.PASS
            print("%s No non-redhat software modules found" % TestResult(result).getMessagePrefix())
        return result

    def print_error_info(self, msg, corrective_action):
        separator = "-" * 80
        print("{0}\n{1}".format(msg, separator))
        print("{0:<35} {1:<20} {2:<50}\n{3}".format('Module', 'Stream', 'Source Repository', separator))
        modules = self.non_rht_modules + self.unidentified_modules
        for module in modules:
            print("{0:<35} {1:<20} {2:<50}".format(module.get('Name') or "", module.get('Stream') or "", ', '.join(module.get('Repos') or [])))
        print("\n{0}".format(corrective_action))

    def runSubTest(self):
        """
        Validate Modules.
        """
        return self.check_non_redhat_modules()

class InsightsSubTest(SubTest):
    """
    Verify RHT insights-client is installed in the images
    """

    def __init__(self, test):
        SubTest.__init__(self, test, "Insights",
                         "Confirm Availability Of Insights Client")
        self.insights_client = 'insights-client'

    def runSubTest(self):
        """
        Validate if the package is installed.
        """
        result = TestResult()
        print("\nChecking for availability of {0}:".format(self.insights_client))
        verification = RPMVerification(self.test.getSuiteName())
        (installed, response) = verification.rpm_availability(self.insights_client)
        if installed:
            print("\n{0}: Installed version '{1}'".format(result.getMessagePrefix(), response))
        else:
            result.combine(TestResult.FAIL)
            print("\n{0}: package '{1}' not installed".format(result.getMessagePrefix(), self.insights_client))
            print("Corrective Action: Please install the required package '{0}'".format(self.insights_client))
        return result


class Utility():
    """
    Class with common utility methods.
    """

    def command_execute(self, command):
        """
        TODO: replace calls to this with rhcert.command calls adding error reporting/handling

        Note: rhcert.command may raise some exceptions
        """
        commandObject = Command(command)
        try:
            commandObject.run()
        except RHCertCommandException:
            # ignore any errors
            pass
        result = u''
        if commandObject.output:
            for line in commandObject.output:
                result += u'\n' + rhcert_make_unicode(line)
        if commandObject.errors:  # append errors at the end
            for error in commandObject.errors:
                result += u'\n' + rhcert_make_unicode(error)
        return result + u"\n"
