# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
import copy, json, os, requests, string, sys, time
from six import with_metaclass

from rhcert.tags import Constants, TestTag, TestResult
from rhcert.testUI import TestUI
from rhcert.command import Command, HwCertCommandException
from rhcert.model.testparameters import TestParameters
from rhcert.model.outputdocument import DeferredCall
from rhcert.model.outputdocument import FeatureData

class TestMeta(type):
    def __init__(cls, name, bases, nmspc):
        super(TestMeta, cls).__init__(name, bases, nmspc)
        if not hasattr(cls, 'registry'):
            cls.registry = list()
        if ".." not in cls.__module__: # dont register classes via relative path
            source = os.path.dirname(sys.modules[cls.__module__].__file__)
            cls.registry.append((cls, source))
            # print "registered " + cls.__name__ + " from " + cls.__module__

    def getClasses(cls):
        return cls.registry


class Test(with_metaclass(TestMeta, TestUI)):

    # __metaclass__ = TestMeta

    def __init__(self, path):
        TestUI.__init__(self, echoResponses=True)
        self.path = path
        """ test name path: <suite name>/.../<test name>
            example: hwcert/network/Ethernet/1GigEthernet
        """
        self.testParameters = TestParameters()
        self.sourceDirectory = None # source directory for the test
        self.owner = "Cerification Partner Experience Team <certpx@redhat.com>"
        self.description = ""
        self.priority = 999
        self.serverMode = False
        self.mandatory = None # either  pre or post for mandatory tests (for each run)
        self.mode = None
        self.incomplete = False

        self.returnValue = 1
        self.debug = Constants.off # Set default value
        self.result = Constants.ABORT
        self.marking = False # is <output> sub-section currently active?
        self.setMode(Constants.normal)
        self.resultsDocument = None

    def setResultsDocument(self, resultsDocument):
        self.resultsDocument = resultsDocument

    def getResultsDocument(self):
        return self.resultsDocument

    """
    client abstract interface
    ------------------------------------------------------------------
    """
    def Key(self):
        """ Key used in uniqueness comparisons
            must match key in TestDocument """
        return "%s %s" % (self.path, self.testParameters)

    def tags(self):
        # default implementation - test subclasses override
        return [TestTag.interactive, TestTag.certification]

    def plan(self, planner):
        """
        called to generate a test plan.
        returns a list of test instances.
        This default implementation does nothing
        """
        if self.debug != Constants.off:
            print("default Test.plan called on test %s" % self.getPath())
        tests = list()
        return tests

    def getRequiredRPMs(self):
        """Test should implement this method if they have package
        requirements dependant on hardware or RHEL release
        returns a list of package names"""
        return list()

    def getHarmfulRPMs(self):
        """Test should implement this method if they have package
        that should be removed on hardware or RHEL release.
        returns a list of package names"""
        return list()

    def verify(self):
        """ verify it test is ready """
        if self.debug != Constants.off:
            print("default Test.verify called on test %s" % self.getPath())
        return True

    def start(self):
        """ performs any initialization before the test is to be run """

    def run(self):
        """ run the test  - must return True for passing, False for failing"""
        if self.debug != Constants.off:
            print("default Test.run called on test %s" % self.getPath())
        return False

    def finish(self, planner, runDocument):
        """ called after test run """
        if self.debug != Constants.off:
            print("default Test.finish called on test %s" % self.getPath())

    def add(self, planner):
        """ called when a test is manually (interactively) added to the plan
            subclasses should define this when more information is needed from the tester """
        pass


    """
    server abstract interface
    ------------------------------------------------------------------
    """
    def startServer(self):
        """ start server-end of the test """
        pass

    def stopServer(self):
        """ stop the server-end of the test """
        pass

    def statusServer(self):
        """ display status information of server-end test services """
        pass

    """
    access
    ------------------------------------------------------------------
    """
    def getName(self):
        return self.path.split('/')[-1]
    def getSuiteName(self):
        """
        Child class should override this.
        Keeping the path logic as a fallback approach to figure out the suite name
        """
        return self.path.split('/')[0]
    def getPath(self):
        return self.path
    def getSourceDirectory(self):
        return self.sourceDirectory
    def setSourceDirectory(self, directory):
        self.sourceDirectory = directory
    def getTestParameters(self):
        return self.testParameters
    def setTestParameters(self, testParameters):
        self.testParameters.copy(testParameters)
    def getSource(self):
        return self.testParameters.get(Constants.source)
    def setSource(self, value):
        self.testParameters.set(Constants.source, value)
    def isIncomplete(self):
        return self.incomplete
    def setIncomplete(self, value):
        self.incomplete = value

    def _compare(self, other):
        """ returns -1 self < other, 0 self == other, 1 self > other """
        # 1) pre-run mandatory
        if self.getMandatory() == Constants.pre and other.getMandatory() != Constants.pre:
            return -1
        if self.getMandatory() != Constants.pre and other.getMandatory() == Constants.pre:
            return 1

        # 2) interactive
        if TestTag.interactive in self.tags() and not TestTag.interactive in other.tags():
            return -1
        if not TestTag.interactive in self.tags() and  TestTag.interactive in other.tags():
            return 1

        # 3) post-run mandatory
        if self.getMandatory() != Constants.post and other.getMandatory() == Constants.post:
            return -1
        if self.getMandatory() == Constants.post and other.getMandatory() != Constants.post:
            return 1

        # if we got this far, tests are equivalent based on interactivity/non and mandatory/pre/post
        # so we sort by priority, test suite, then path

        # 4) priority
        # low means run before high
        if self.priority < other.priority:
            return -1
        if self.priority > other.priority:
            return 1

        # 5) test suite
        if self.getSuiteName() < other.getSuiteName():
            return -1
        if self.getSuiteName() > other.getSuiteName():
            return 1

        # 6) - alphabetical by path
        if self.getPath() < other.getPath():
            return -1
        if self.getPath() > other.getPath():
            return 1

        # at this point, the two had better be instances of the same test, perhaps with
        # different parameters (example, hardware devices).
        return 0

    def __lt__(self, other):
        return self._compare(other) == -1

    def __le__(self, other):
        return self._compare(other) in [-1, 0]

    def __eq__(self, other):
        return self._compare(other) == 0

    def __ge__(self, other):
        return self._compare(other) in [0, 1]

    def __gt__(self, other):
        return self._compare(other) == 1

    def __ne__(self, other):
        return self._compare(other) != 0

    def __str__(self):
        return self.path

    def __hash__(self):
        return hash(self.path)

    def makeCopy(self):
        testCopy = copy.copy(self)
        testCopy.testParameters = TestParameters()
        return testCopy

    def getOwner(self):
        return self.owner
    def getVersion(self):
        return self.version
    def getDescription(self):
        return self.description
    def getTestTime(self):
        return self.testTime
    def getInteractive(self):
        if TestTag.interactive in self.tags():
            return Constants.yes
        return Constants.no
    def getMandatory(self):
        return self.mandatory
    def getServerMode(self):
        if self.serverMode: return Constants.yes
        return Constants.no
    def getPriority(self):
        return ("%u" % self.priority)
    def getOutputFile(self):
        return self.testParameters.get(Constants.outputfile)
    def getOutputDirectory(self):
        return os.path.dirname(self.getOutputFile())
    def getDebug(self):
        return self.debug
    def setDebug(self, debug):
        self.debug = debug
    def getTestServer(self):
        return self.testParameters.get(Constants.testserver)


    # utilities
    def collabCall(self, command_request):
        test_server = self.getTestServer()
        endpoint = "http://%s/collab/" % test_server
        message_body = {}
        message_body["test_name"] = command_request.test_name
        message_body["function_name"] = command_request.function_name
        message_body["test_package_trail"] = command_request.test_package_trail
        message_body["requires_root"] = command_request.requires_root
        message_body["args"] = command_request.args
        message_body["kwargs"] = command_request.kwargs

        response = requests.post(endpoint, json=message_body)
        status_code = response.status_code
        reason = response.reason
        if status_code == requests.codes.bad_request:
            raise ValueError(reason)
        elif status_code == requests.codes.ok:
            return response.content
        else:
            response.raise_for_status()

    def printPipe(self, pipe):
        while 1:
            line = pipe.readline()
            if line:
                print(line)
            else:
                return pipe.close()

    def removeDirectory(self, directory):
        "Remove a directory (and all its contents)"
        try:
            for (root,dirs,files) in os.walk(directory, topdown=False):
                for f in files: os.unlink(os.path.join(root,f))
                for d in dirs: os.rmdir(os.path.join(root,d))
                os.rmdir(directory)
        except OSError:
            return

    def checkNFSRootFileSystem(self):
        self.isNFSRootSystem = False
        mount = os.popen("mount")
        while True:
            line = mount.readline()
            if not line:
                break
            words = line.split()
            # line should be <source> on <destination> type <mounttype> <other stuff>
            if words and len(words) >= 5 and words[2] == "/" and words[4] == "nfs":
                self.isNFSRootSystem = True
                break

        return self.isNFSRootSystem

    def loadKernelModule(self, module):
        # is it already running?
        if os.system("lsmod | fgrep -q %s" % module) == 0:
            print("Module %s is already loaded" % module)
            return True
        # otherwise, does it exist?
        if os.system("modinfo %s" % module) == 0:
            if self.promptConfirm("Probe Module %s?" % module):
                os.system("modprobe %s" % module)
                if os.system("lsmod | fgrep -q %s" % module) == 0:
                    print("Module %s is loaded" % module)
                    return True
                # otherwise
                print("Error: module would not load %s" % module)
                return False

            # otherwise - cancelled
            return False

        # otherwise:
        print("Error: no module %s" % module)
        return False

    def checkKernelModule(self, module):
        # is it already running?
        lsmod = Command("lsmod")
        try:
            lsmod.getString(module, singleLine=False)
            print("Module %s is loaded" % module)
            return True
        except:
            pass
        # otherwise, does it exist?
        modinfo = Command("modinfo %s" % module)
        modprobe = Command("modprobe %s" % module)

        try:
            modinfo.run()
            modprobe.run()
            lsmod.getString(module, singleLine=False)
            print("Loaded module %s" % module)
            return True
        except:
            print("Warning: could not load module %s" % module)

        return False

    def markOutput(self, name, description=None):
        self.closeOutput()
        if description:
            sys.stdout.logOnly("<output name=\"%s\" description=\"%s\">\n" % (name, description))
            sys.stdout.terminal.write("\nSubtest: %s - %s\n" % (name, description))
        else:
            sys.stdout.logOnly("<output name=\"%s\">\n" % (name))
            sys.stdout.terminal.write("\nSubtest %s:\n" % name)

        self.marking = True

    def closeOutput(self):
        if self.marking:
            sys.stdout.logOnly("</output>\n")
            self.marking = False

    def markSummary(self, summary):
        summary = str(TestResult(summary))
        sys.stdout.logOnly("<summary data-value=\"%s\">%s</summary>\n" % (summary, summary))
        sys.stdout.terminal.write(summary + "\n")

    def runSubTest(self, subtestFunction, name, description=None):
        self.markOutput(name, description)
        sys.stdout.flush()
        sys.stderr.flush()
        result = subtestFunction()
        sys.stdout.flush()
        sys.stderr.flush()
        self.markSummary(result)
        self.closeOutput()
        return result

    def runOnServer(self, args, outFile):
        outFile.write("Server side test for %s not implemented\n" % self.Name())

    def getLoadAvg(self):
        loadAvgFile = "/proc/loadavg"
        if os.path.exists(loadAvgFile):
            file = open(loadAvgFile)
            line = file.readline()
            if len(line.strip()) > 0:
                params = line.strip().split()
                loadAvg = float(params[0])
                print("Current load average: %s" % loadAvg)
                sys.stdout.flush()
                return loadAvg
        else:
            return 0

    def syncDisks(self):
        try:
            print("Syncing disks")
            sys.stdout.flush()
            command = Command("/bin/sync")
            command.echo()
        except HwCertCommandException as e:
            print("Warning: rhcert attempt to sync failed")
            print(e)

    def waitForLull(self):
        waitTime = 5
        retryLimit = 20
        self.syncDisks()

        print("Waiting for low load...")
        sys.stdout.flush()
        retryCount = 0
        while self.getLoadAvg() > 1 and retryCount < retryLimit:
            time.sleep(waitTime)
            retryCount = retryCount + 1
        print("Done waiting")
        sys.stdout.flush()

    def getMemoryInfo(self):
        self.checkNFSRootFileSystem()
        procMeminfo = open("/proc/meminfo", "r")
        self.freeMemory = 0
        self.systemMemory = 0
        self.swapMemory = 0
        while True:
            line = procMeminfo.readline()
            if line:
                tokens = line.split()
                if len(tokens) == 3:
                    if "MemTotal:" == tokens[0].strip():
                        self.systemMemory = int(tokens[1].strip())/1024
                    elif tokens[0].strip() in ["MemFree:", "Cached:", "Buffers:"]:
                        self.freeMemory +=  int(tokens[1].strip())
                    elif "SwapTotal:" == tokens[0].strip():
                        self.swapMemory = int(tokens[1].strip())/1024
            else:
                break
        procMeminfo.close()
        self.freeMemory = self.freeMemory/1024
        print("System Memory: %u MB" % self.systemMemory)
        print("Free Memory: %u MB" % self.freeMemory)
        print("Swap Memory: %u MB" % self.swapMemory)

        if self.systemMemory == 0:
            print("Error: could not determine system RAM")
            return False

        # Process Memory
        self.processMemory = self.freeMemory
        self.isProcessLimited = False
        try:
            arch = Command("uname -i").getString()
            if (arch in [Constants.i386, Constants.i686, Constants.s390])and self.freeMemory > 1024:
                self.isProcessLimited = True
                self.processMemory = 1024 # MB, due to 32-bit address space
                print("%s arch, Limiting Process Memory: %u" % (arch, self.processMemory))
        # others?  what about PAE kernel?
        except HwCertCommandException as exception:
            print("Error: could not determine system architecture via uname -i")
            print(exception)
            return False
        sys.stdout.flush()

        #otherwise
        return True

    def parseRecords(self, command, keyNames):
        """ parse output into records separated by blank lines, with name:value pairs.
        command: rhcert.Command
        keyNames: list(string)
        returns: dict[keyName][propertyName] ( dict by keyname of dict by property name of values )"""
        records = dict()
        properties = dict()
        command.start()
        while True:
            line = command.readline()
            # blank or empty line (EOF)
            if not line or line.strip() == "":
                # save the record
                if len(properties) > 0:
                    for keyName in keyNames:
                        if keyName in properties:
                            records[properties[keyName]] = properties
                    properties = dict()
                if not line:
                    break
                else:
                    continue
            tokens = line.split(":", 1)
            if len(tokens) == 2:
                properties[tokens[0].strip()] = tokens[1].strip()
            continue
        return records

class SubTest(object):
    """ SubTests are simple test functions that log output, and refer to their enclosing test for settings """

    def __init__(self, test, name, description):
        self.test = test
        self.name = name
        self.description = description
        self.result = False
        self.deferredServices = list()
        self.featureData = dict()

    def getName(self):
        return self.name
    def getDescription(self):
        return self.description
    def getDebug(self):
        return self.test.getDebug()
    def getResult(self):
        return self.result

    def logFeatureData(self, name, item_attributes):
        '''
        Usable for storing granularity data.
        Params:
        name: feature name (string)
        item_attributes: feature data (list of dictionaries)
        The self.featureData stores data in following manner:
        Example1:
        {
             "name": "software repositorires",
             "items": [
                 {"id": "1234", "name": "Fedora", "status": "enabled"},
                 {"id": "5678", "name": "RHEL", "status": "enabled"},
                 {"id": "91011", "name": "CentOS", "status": "disabled"},
                 ...
                 ]
        }
        Example2:
        {
             "name": "installed rpms",
             "items": [
                 {"name": "popt", "version": "1.13", "release": "16.el7", "arch": "x86_64"},
                 {"name": "libcgroup", "version": "0.41", "release": "21.el7", "arch": "x86_64"},
                 ...
                 ]
        }
        Usage: Sub-test has to call this method with required feature data
        '''
        self.featureData["name"] = name
        self.featureData["items"] = list()
        for attribute in item_attributes:
            self.featureData["items"].append(attribute)

    def runSubTest(self):
        print("Error: abstract call to subtest.")
        return False

    def run(self):
        self.test.markOutput(self.name, self.description)
        sys.stdout.flush()
        sys.stderr.flush()
        self.result = self.runSubTest()
        sys.stdout.flush()
        sys.stderr.flush()
        if self.featureData:
            self.markFeature()
        self.markSummary(self.result)
        self.test.closeOutput()
        return self.result

    def markFeature(self):
        feature = FeatureData()
        feature.loadService(self.featureData)
        sys.stdout.logOnly(feature.toXML())

    def markSummary(self, summary):
        summary = TestResult(summary)
        sys.stdout.logOnly("<summary data-value=\"%s\">%s" % (summary, summary))
        if self.deferredServices:
            for deferredService in self.deferredServices:
                deferred = DeferredCall()
                deferred.loadService(deferredService)
                sys.stdout.logOnly(deferred.toXML())
        sys.stdout.logOnly("</summary>\n")
        sys.stdout.terminal.write(str(TestResult(summary)) + "\n")

class UnitTest(Test):
    def __init__(self, path, tags, mandatory, priority):
        Test.__init__(self, path)
        self.sometags = tags
        self.mandatory = mandatory
        self.priority = priority

    def tags(self):
        return self.sometags

def unitTest():
    # unit test sorting
    tests = list()
    paths = ["unit/a", "unit/b"]
    tags = [TestTag.interactive, TestTag.noninteractive]
    mandatories = [Constants.pre, None, Constants.post]
    priorities = [1,2,3]
    for path in paths:
        for tag in tags:
            for mandatory in mandatories:
                for priority in priorities:
                    tests.append(UnitTest(path, [tag], mandatory, priority))
    print("tests:")
    for test in tests:
        print("%s, %s, %s, %s" % (test.getPath(), test.tags(), test.getMandatory(), test.getPriority()))
    print("sorted:")
    for test in sorted(tests):
        print("%s, %s, %s, %s" % (test.getPath(), test.tags(), test.getMandatory(), test.getPriority()))

    return True

if __name__ == "__main__":
    if not unitTest():
        print("test.py unit test FAILED")
        exit(1)
    print("test.py unit test passed")
    exit(0)


