# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
from rhcert.compatability import rhcert_unicode
from rhcert.tags import Constants
from rhcert.commandLineUI import CommandLineUI
from rhcert.listener.serverUI import ServerUI

class TestUI(object):
    """ TestUI provides user prompting test's interaction UI

        The TestUI class defines the abstract interface
        implemented by ServerUI and CommandLineUI classes

        The Test class inherits from TestUI class to prompt for
        input and receive responses.   Depending on the test
        mode, TestUI either instantiates the ServerUI class
        (server mode), or the CommandLineUI class (normal or
        auto modes).   Both ServerUI and TestUI implement the
        same set of methods.
    """

    def __init__(self, echoResponses=True):
        self.echo = echoResponses
        self.mode = Constants.normal
        self.ui = CommandLineUI(self.echo)

    def getMode(self):
        return self.mode
    def setMode(self, mode):
        """ setMode sets the mode: normal or auto uses
            command-line IO (CommandLineUI), server uses IO via
            the web server (ServerUI).
        """
        self.mode = mode
        if self.mode in [Constants.normal, Constants.auto]:
            self.ui = CommandLineUI(self.echo)
        elif self.mode == Constants.server:
            self.ui = ServerUI(self)

    def printPipe(self, pipe):
        return self.ui.printPipe(pipe)

    def statusMessage(self, message):
        return self.ui.statusMessage(rhcert_unicode(message))

    def promptConfirm(self, message):
        return self.ui.promptConfirm(message)

    def prompt(self, message, validation=None, default=None, answers=None):
        """presents the message.  sets defaults as the answer, allows editing,
           applies validation to answer(s), returns them."""
        # answers is deprecated, use select()
        if answers:
            return self.ui.select(message, answers, selected=None)
        return self.ui.prompt(message, validation=validation, default=default)

    def select(self, message, answers, selected=None):
        return self.ui.select(message, answers, selected)

    def multiSelect(self, message, answers, selected=[]):
        return self.ui.multiSelect(message, answers, selected)

    def editText(self, message, default):
        """ same as prompt, may be presented as an editable text area """
        return self.ui.editText(message, default)

    def promptPassword(self, message):
        return self.ui.promptPassword(message)


    def promptFile(self, message):
        """ prompt for the user to upload a file to the server transfer area """
        return self.ui.promptFile(message)

    def promptInteger(self, message, default=None):
        return (self.ui.promptInteger(message, default))


    # deprecated - Use promptConfirm
    def promptContinue(self, message):
        return self.promptConfirm(message=("%s - then continue? " % message))

    # deprecated - Use prompt or select.
    def promptEdit(self, message, default, answers=None):
        # editable string?
        if not answers:
            return self.ui.prompt(message, validation=None, default=default)
        # select from a list
        else:
            return self.ui.select(message, answers, selected=default)

