# Copyright (c) 2017 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
import sys
import os
import time
import datetime
import io

from rhcert.tags import Constants, Tags, Programs
from rhcert.model.resultsdocument import ResultsDocument
from rhcert.model.certificationdocument import CertificationDocument
from rhcert.workflow.workflow import Workflow
from rhcert.services.productnameservice import ProductNameService
from rhcert.compatability import rhcert_base64_encode



class Certification(Workflow):

    def __init__(self, config):
        Workflow.__init__(self, config)
        self.productNameService = ProductNameService()

    def getCurrentUserVendors(self):
        """ Get vendors that loggedin user is a member of """
        parameters = dict()
        response = self.callProceedure(parameters, "Cert.getCurrentUserVendors")
        if self.getErrors():
            self.pushError("Error: Could not fetch associated vendors for your account")
        return response

    def getCertifications(self, productID=None, program=None, dateRange=None, status=None):
        """ returns dict(vendor, product, make) """

        self.clearErrors()
        params = dict()
        # filtering options:
        if productID:
            params["product_id"] = productID
        if program:
            params["program"] = program

        # NOTE: When Filter is passed to API as ['all', 'normal', 'dustbin']. By
        # default is 'all' and will return all certs including dustbin certs,
        # 'normal' means to return certs that are not in dustbin and 'dustbin'
        # means dustbin only.

        # TODO: Remove after completing investigation of CERTPX-2286. Till then
        # we will fetch all certificates that are 'In Progress' or 'Certified'.
        params['filter'] = status if status else 'normal'

        (startDate, endDate) = self.__getDateRange(dateRange)
        if startDate:
            params["start_date"] = startDate
        if endDate:
            params["end_date"] = endDate

        response = self.callProceedure(params, "Cert.getCertifications_7_4")
        if self.getErrors():
            self.pushError("Error: could not download certifications")

        return response
        # return self.__decodeValues(response) - comment out attempts to use Binary data.

    def __decodeValues(self, responseList):
        for parameters in responseList:
            for (name, value) in parameters.items():
                if isinstance(value, XmlrpcBinary):
                    parameters[name] = value.data
        return responseList

    def __getDateRange(self, dateRange):
        if not dateRange or dateRange == "all":
            return (None, None)
        try:
            dateRange = int(dateRange)
        except:
            self.debugPrint("Error: date range %s is not an integer (days)" % dateRange)
            return (None, None)

        "YYYY-mm-dd HH:MM:SS"
        # now = time.gmtime(time.time())
        end = datetime.datetime.now()
        start = end + datetime.timedelta(-dateRange)
        return (start.strftime(Constants.DATETIMEFORMAT), end.strftime(Constants.DATETIMEFORMAT))

    def makeCertification(self, certificationData, program=None):
        """ make a certification object from a CWE dict produced by Cert.getCertifications

        Hardware Example:
        {'category': 'Server',
                         'cert_id': 31712,
                         'cert_nid': 2695213,
                         'classification': 'Layered Product',
                         'creation_time': '2016-11-13 22:18:42',
                         'is_draft': 0,
                         'platform': 'x86_64',
                         'program': 'hardware',
                         'rh_product': 'RHOSP 8.0',
                         'rh_product_id': 3,
                         'rh_product_version_id': 30,
                         'status': 'Certified',
                         'vendor': 'IBM',
                         'vendor_id': 5,
                         'vendor_product_id': 6408,
                         'vendor_product_make': 'BladeCenter',
                         'vendor_product_name': 'needinfo'},

        OpenStack Example:
        {'cert_id': 35213,
                           'cert_nid': 3057233,
                           'component': 'certification-blockstorage-plugin',
                           'creation_time': '2017-03-01 09:19:57',
                           'is_draft': 0,
                           'platform': 'x86_64',
                           'program': 'openstack',
                           'rh_base_version': 'RHEL 7.3',
                           'rh_base_version_id': 49,
                           'rh_product': 'RHOSP 10.0',
                           'rh_product_id': 3,
                           'rh_product_version_id': 52,
                           'status': 'In Progress',
                           'vendor': 'Abbee',
                           'vendor_id': 412,
                           'vendor_product_id': 2731,
                           'vendor_product_name': 'rohit-osp-manila'}

        """
        certification = CertificationDocument()
        certification.new()
        # these are absolutely required:
        if program:
            certification.setCertificationProgram(program)
        else:
            certification.setCertificationProgram(certificationData["program"])
        for attribute in ["cert_id", "certification_id"]:
            if certificationData.get(attribute):
                certification.setCertificationID(certificationData[attribute])
                break
        if not certification.getCertificationProgram() and certification.getCertificationID():
            self.appendError("Error: could not determine certification ID and/or program:\n")
            return None
        if certificationData.get("cert_nid"):
            certification.setNodeID(certificationData.get("cert_nid"))
        certification.setCreationTime(certificationData.get("creation_time"))
        certification.setUpdateTime(certificationData.get("update_time"))

        path = certification.getProductCertification()
        redhatVendorDocument = path.get(Tags.vendor)
        try:
            redhatProductID = int(certificationData["rh_product_id"])
            redhatProductVersionID = int(certificationData["rh_product_version_id"])
            redhatProductVersionFormatName = certificationData.get("rh_product_version_format_name")
            redhatBaseVersionID = 0
            if certificationData.get("rh_base_version_id"):
                redhatBaseVersionID = int(certificationData.get("rh_base_version_id"))
            (redhatProduct, redhatProductVersion, redhatBaseVersion) = self.productNameService.getProductPath(
                   redhatProductID,
                   redhatProductVersionID,
                   redhatBaseVersionID)
            if not (redhatProduct and redhatProductVersion):
                self.appendError("Error: Could not determine Red Hat Product ID %s version ID %s for certification %s" % (
                    redhatProductID, redhatProductVersionID, certification.getCertificationID()))
                return None
        except Exception as error:
            self.appendError("Error: Could not determine Red Hat Product from IDs for cert %s:\n%s" % (certification.getCertificationID(), error))
            return None

        redhatProductDocument = redhatVendorDocument.createProduct(redhatProduct.get("name"), redhatProduct.get("id"))
        if redhatProduct.get("shortName"):
            redhatProductDocument.setShortName(redhatProduct.get("shortName"))
        productVersionDocument = redhatProductDocument.createVersion()
        if redhatProductVersionFormatName:
            productVersionDocument.setVersion(redhatProductVersion.get("name") + ' ' + redhatProductVersionFormatName)
        else:
            productVersionDocument.setVersion(redhatProductVersion.get("name"))
        productVersionDocument.setID(redhatProductVersion.get("id"))
        for archAttribute in ["platform", "architecture"]:
            if certificationData.get(archAttribute):
                productVersionDocument.setPlatform(certificationData.get(archAttribute))
                break
        if not redhatBaseVersion:
            certification.getRequiredOS().setProduct(redhatProductDocument.getName())
            certification.getRequiredOS().setRelease(productVersionDocument.getVersion())
        else:
            certification.getRequiredOS().setProduct(Constants.Red_Hat_Enterprise_Linux)
            certification.getRequiredOS().setRelease(redhatBaseVersion.get("name"))

        # store the entire dictionary from CWE into generic parameters
        certification.getCertificationParameters().replace(certificationData)

        return certification

    def updateCertification(self, certification, certificationData):
        # these are absolutely required:
        try:
            if (certification.getCertificationProgram() != certificationData["program"]
                or int(certification.getCertificationID()) != int(certificationData["cert_id"])):
                    self.debugPrint("Error: certification program and ID do not match data")
                    return False
        except Exception as error:
            self.debugPrint("Error: Could not match certification ID and/or program:\n%s" % error)
            return False

        certification.setUpdateTime(certificationData.get("update_time"))
        # update the entire dictionary from CWE into generic parameters
        certification.getCertificationParameters().replace(certificationData)

        return True

    def get(self, certificationProgram, certificationID):
        params = None
        if certificationProgram in [Programs.cloud, Programs.cloud_sap]:
            params = self.getCloud(certificationID)
        if certificationProgram == Programs.software:
            params = self.getSoftware(certificationID)
        if certificationProgram == Programs.hardware:
            params = self.getHardware(certificationID)
            # might move this to line before return statement to cover all programs..
            self.__translateRedHatProduct(certificationProgram, params)
        if certificationProgram == Programs.openstack:
            params = self.getOpenStack(certificationID)
        return params


    def getHardware(self, certificationID):
        self.clearErrors()
        parameters = {
            "certification_id": certificationID,
        }
        response = self.callProceedure(parameters, "Cert.getHardware_4_10")
        return response

    def getCloud(self, certificationID):
        self.clearErrors()
        parameters = {
            "certification_id": certificationID,
        }
        response = self.callProceedure(parameters, "Cert.getCloud_4_7")
        return response

    def getSoftware(self, certificationID):
        self.clearErrors()
        parameters = {
            "certification_id": certificationID,
        }
        response = self.callProceedure(parameters, "Cert.getSoftware")
        return response

    def getOpenStack(self, certificationID):
        self.clearErrors()
        parameters = {
            "certification_id": certificationID,
        }
        response = self.callProceedure(parameters, "Cert.getOpenStack_4_7")
        return response

    def update(self, certificationProgram, certificationID, parameters):
        if certificationProgram in [Programs.cloud, Programs.cloud_sap]:
            return self.updateCloud(certificationID, parameters)
        elif certificationProgram == Programs.openstack:
            return self.updateOpenstack(certificationID, parameters)
        elif certificationProgram == Programs.hardware:
            return self.updateHardware(certificationID, parameters)
        return None

    def updateHardware(self, certificationID, parameters):
        self.clearErrors()
        args = {
            "cert_id": certificationID
        }
        args.update(self.filterNone(parameters))
        response = self.callProceedure(args, "Cert.updateHardware_4_7")
        return response

    def updateCloud(self, certificationID, parameters):
        self.clearErrors()
        args = {
            "cert_id": certificationID,
            "build": parameters["build"],
        }
        if "image_name" in parameters:
            args["image_name"] = parameters["image_name"]
        response = self.callProceedure(args, "Cert.updateCloud_4_7")
        return response

    def updateOpenstack(self, certificationID, parameters):
        self.clearErrors()
        args = { "cert_id": certificationID }
        args.update(self.filterNone(parameters))
        response = self.callProceedure(args, "Cert.updateOpenstack_4_7")
        return response

    def createCertification(self, vendorProduct, program, parameters):
        if program == Programs.openstack:
            return self.createOpenstackCertification(vendorProduct, parameters)

        # otherwise
        self.clearErrors()
        self.__translateRedHatProduct(program, parameters)
        params = {
                    "product_nid": vendorProduct.getNodeID(),
                    "program": program,
        }
        params.update(self.filterNone(parameters))
        response = self.callProceedure(params, "Cert.createCert")
        if not response:
            self.pushError("Error: could not create certification:")
            self.appendError("response: %s" % response)
            return None

        # otherwise
        return response

    def __translateRedHatProduct(self, program, parameters):
        if parameters is None:
            return None
        product = productVersion = platform = None
        try:
            product = self.productNameService.getProduct(parameters.get("rh_product_id"))
            if product and parameters.get("rh_product_version_id"):
                productVersion = self.productNameService.getProductVersion(product, parameters.get("rh_product_version_id"))
                if productVersion and parameters.get("architecture"):
                    # putting in this hack because architecture
                    # is a return param for get hardware cert 4_7
                    if parameters.get("architecture").isdigit():
                        platform = self.productNameService.getPlatform(productVersion, parameters.get("architecture"))

        except Exception as error:
            self.appendError("Error: could not determine product information:")
            self.appendError(error)
            return False

        if not product:
            self.appendError("Error: could not find Red Hat %s product %s version %s platform %s" %(
                                                    program,
                                                    parameters.get("rh_product_id"),
                                                    parameters.get("rh_product_version_id"),
                                                    parameters.get("architecture")))
            return False

        if program == Programs.openstack:
            parameters["architecture"] = platform["name"]
            return True

        if program == Programs.hardware:
            if product:
                parameters["product"] = product["name"]
            if productVersion:
                parameters["version"] = productVersion["name"]
            if platform:
                parameters["platform"] = platform["name"]
            return True

        # otherwise: cloud:
        parameters["program_version"] = productVersion["name"]
        parameters["architecture"] = platform["name"]
        return True

    def checkDHSS(self, supported_features):
        if "dhss_true" not in supported_features:
            supported_features.append(u'dhss_false')

    def createOpenstackCertification(self, vendorProduct, parameters):
        self.clearErrors()
        self.__translateRedHatProduct(Programs.openstack, parameters)
        params = {"product_nid": vendorProduct.getNodeID()}
        # For Cinder/Manila - (protocol-feature-widget.html)
        if parameters.get("component-done") in ["blockstorage", "filestorage"]:
            parameters['supported_features'] = dict()
            if parameters.get("protocol-supported"):  # If no pre-defined but only other protocol is selected
                protocol_selected = parameters["protocol-supported"]
                if not isinstance(protocol_selected, list):
                    protocol_selected = [protocol_selected]
                for protocol in protocol_selected:
                    feature_supported = parameters.get(protocol+"-feature-supported")
                    parameters['supported_features'][protocol] = [feature_supported] if not isinstance(
                        feature_supported, list) else feature_supported
                    # Add dhss_false if dhss_true was not found for filestorage
                    if parameters.get("component-done") == "filestorage":
                        self.checkDHSS(parameters['supported_features'][protocol])

            # make list of other protocol, if any
            if parameters["other-protocol-name"]:
                parameters["new_protocols"] = [parameters["other-protocol-name"]]
                feature_supported = parameters["Other-feature-supported"]
                parameters["supported_features"][parameters["other-protocol-name"]] = [feature_supported] if not isinstance(
                    feature_supported, list) else feature_supported
                # Add dhss_false if dhss_true was not found for filestorage
                if parameters.get("component-done") == "filestorage":
                    self.checkDHSS(parameters["supported_features"][parameters["other-protocol-name"]])
        else:
            if parameters.has_key("supported_features"):
                # For Neutron (feature-widget.html)
                if isinstance(parameters.get("supported_features"), list):
                    parameters['supported_features'] = {"default": parameters["supported_features"]}
                else:
                    parameters['supported_features'] = {"default": [parameters["supported_features"]]}
        params.update(self.filterNone(parameters))
        response = self.callProceedure(params, "Cert.createOpenstack_4_7")
        if not response:
            self.pushError("Error: could not create certification:")
            self.appendError("response: %s" % response)
            return None

        # otherwise
        return response

    def getComments(self, certification, productComments=False):
        self.clearErrors()
        params = { "id": certification.getNodeID() }
        if productComments:
            _, product = certification.getPartnerAndProduct()
            if product and product.getNodeID():
                params = { "id": product.getNodeID() }
            else:
                self.appendError("Error: could not determine product node ID for comments")
                return None
        response = self.callProceedure(params, "Cert.getCommentsOfCert")
        if not response:
            self.pushError("Error: could not get certification comments")
            self.appendError("response: %s" % response)
            return None

        # otherwise
        return response

    def getCommentsByID(self, certificationID):
        """ as of 5.8, certificationID is actually NodeID """
        self.clearErrors()
        params = { "id": certificationID }
        response = self.callProceedure(params, "Cert.getCommentsOfCert")
        if not response:
            self.pushError("Error: could not get certification comments")
            self.appendError("response: %s" % response)
            return None

        # otherwise
        return response


    def addComment(self, certification, comment, productComments=False):

        self.clearErrors()
        params = {
            "id": certification.getNodeID(),
            "comment": comment,
            "is_private": False,
            "is_highlight": False
        }
        if productComments:
            _, product = certification.getPartnerAndProduct()
            if product and product.getNodeID():
                params["id"] = product.getNodeID()
            else:
                self.appendError("Error: could not determine product node ID for comments")
                return None

        response = self.callProceedure(params, "Cert.addCommentForCert")
        # ignore response - seems to be always 0
        return response

    def getTestPlan(self, certificationID):
        self.clearErrors()
        params = { "certid": certificationID }
        response = self.callProceedure(params, "Cert.getOpenTestsInfo_7_4")
        return response

    def uploadTestResults(self, resultsFile, resultsDocument):
        self.clearErrors()
        try:
            certificationData = rhcert_base64_encode(io.open(resultsFile, 'rb').read())
        except Exception as error:
            self.pushError("Error: could not read and/or encode results")
            self.pushError("%s" % error)
            return None

        params = {
                  "id":resultsDocument.getNodeID(),
                  "filename":os.path.basename(resultsFile),
                  "description":"test results",
                  "data":certificationData
        }
        if resultsDocument.getCertificationProgram():
            params["program"] = resultsDocument.getCertificationProgram()

        response = self.callProceedure(params, "Cert.uploadTestLog")
        if response:
            # reload the document to get the whole thing
            currentFilePath = resultsDocument.getLoadedFilePath()
            resultsDocument = ResultsDocument()
            resultsDocument.load(currentFilePath)
            resultsDocument.setSubmitTime(self.getCurrentUTCTime())
            resultsDocument.setState(Constants.closed)
            if response.get("job_id"):
                resultsDocument.setValidationJobID(response.get("job_id"))
            if response.get("bug_id"):
                resultsDocument.setBugID(response.get("bug_id"))
            if response.get("cert_id"):
                resultsDocument.setCertificationID(response.get("cert_id"))
            resultsDocument.save()

        if not response or self.getErrors():
            self.pushError("Error: could not submit certification results")

        return response

    def getSupportCases(self, certificationID):
        self.clearErrors()
        params = {
            "cert_id": certificationID
        }
        response = self.callProceedure(params, "Cert.getSupportCases_4_7")
        if not response:
            self.pushError("Error: could not get certification support cases")
            self.appendError("response: %s" % response)
            return None

        # otherwise
        return response

    def addSupportCase(self, certificationID, supportCaseID):
        self.clearErrors()
        params = {
            "cert_id": certificationID,
            "support_case_id": supportCaseID
        }
        response = self.callProceedure(params, "Cert.addSupportCase_4_7")
        if not response:
            self.pushError("Error: could not add a support case")
            self.appendError("response: %s" % response)
            return None

        # otherwise
        return response

    def getKBases(self, certificationID):
        self.clearErrors()
        params = {
            "cert_id": certificationID
        }
        response = self.callProceedure(params, "Cert.getKbases_4_7")
        if not response:
            self.pushError("Error: could not get certification kbases")
            self.appendError("response: %s" % response)
            return None

        # otherwise
        return response

    def addKBase(self, certificationID, kbaseID, kbaseType):
        """ kbaseType = kbase_article or kbase_solution """
        self.clearErrors()
        params = {
            "cert_id": certificationID,
            "kbase_id": kbaseID,
            "kbase_type": kbaseType
        }
        response = self.callProceedure(params, "Cert.addKbase_4_7")
        if not response:
            self.pushError("Error: could not add kbase")
            self.appendError("response: %s" % response)
            return None

        # otherwise
        return response

    def checkSupported(self, packageNVR):
        self.clearErrors()
        params = {"NVR": packageNVR}
        response = self.callProceedure(params, "Cert.checkNVRSupported")
        return response

    def submitTestRunNoFileSave(self, resultsTestName, resultsDocument, resultsFileData):
        self.clearErrors()
        if not self.isLoggedIn():
            self.errors.append("Error: missing or invalid login/password")
            return None

        try:
            (resultsTestName, compressedFileData) = resultsDocument.compressNoSave(resultsTestName, resultsFileData)
        except Exception as error:
            self.pushError("Error: could not compress %s" % resultsTestName)
            self.pushError("%s" % error)
            return None

        try:
            certificationData = rhcert_base64_encode(compressedFileData) 
        except Exception as error:
            self.pushError("Error: could not read and/or encode results")
            self.pushError("%s" % error)
            return None

        params = {
                  "id":resultsDocument.getNodeID(),
                  "filename":os.path.basename(resultsTestName),
                  "description":"test results",
                  "data":certificationData
        }

        if resultsDocument.getCertificationProgram():
            params["program"] = resultsDocument.getCertificationProgram()

        response = self.callProceedure(params, "Cert.uploadTestLog")

        if not response:
            self.pushError("Error: could not submit test run")
            self.appendError("response: %s" % response)
            return None
            
        return response

    def submitTestRun(self, resultsPath, resultsDocument):
        self.clearErrors()
        if not self.isLoggedIn():
            self.errors.append("Error: missing or invalid login/password")
            return None

        # compress the results file if not already compressed
        try:
            if os.path.splitext(resultsPath)[1] == ".xml":
                resultsPath = resultsDocument.compressToFile(resultsPath)
        except Exception as error:
            self.pushError("Error: could not compress %s" % resultsPath)
            self.pushError("%s" % error)
            return None

        try:
            certificationData = rhcert_base64_encode(io.open(resultsPath, 'rb').read())
        except Exception as error:
            self.pushError("Error: could not read and/or encode results")
            self.pushError("%s" % error)
            return None

        params = {
            "cert_id": resultsDocument.getCertificationID(),
            "filename": os.path.basename(resultsPath),
            "description": "test results",
            "data": certificationData
        }

        if resultsDocument.getCertificationProgram():
            params["program"] = resultsDocument.getCertificationProgram()

        response = self.callProceedure(params, "Cert.uploadTestLog_7_14")

        if not response:
            self.pushError("Error: could not submit test run")
            self.appendError("response: %s" % response)
            return None

        # reload the document to get the whole thing
        resultsDocument = ResultsDocument()
        resultsDocument.load(resultsPath)
        resultsDocument.setSubmitTime(self.getCurrentUTCTime())
        resultsDocument.setState(Constants.closed)
        if response.get("cert_id"):
            resultsDocument.setCertificationID(response.get("cert_id"))
        resultsDocument.save()
        return response

    def getAttachments(self, certID):
        """ first added to rhcert for 4.9 """
        self.clearErrors()
        parameters = { "cert_id": certID }

        response = self.callProceedure(parameters, "Cert.attachments")
        return response

    def getLayered(self, rh_product_id, rh_product_version_id):
        self.clearErrors()
        parameters = {"rh_product_id": rh_product_id,
                      "rh_product_version_id": rh_product_version_id
        }
        response = self.callProceedure(parameters, "Cert.getLayered_4_11")
        return response

    def getLayeredProducts(self, certID):
        """ first added to rhcert for 4.9 (unit tests)"""
        self.clearErrors()
        parameters = { "cert_id": certID }

        response = self.callProceedure(parameters, "Cert.getAvaliableLayeredProductVersion")
        return response

    def createLayeredCertification(self, certID, redHatProductID, redHatProductVersionID, layered_product_format=None):
        """ first added to rhcert for 4.9 (unit tests)"""
        self.clearErrors()
        parameters = {
            "cert_id": certID,
            "rh_product_id": redHatProductID,
            "rh_product_version_id": redHatProductVersionID
        }
        if layered_product_format:
            parameters["layered_product_format"] = layered_product_format

        response = self.callProceedure(parameters, "Cert.createLayerCert")
        return response

    def getPublicVendors(self, program):
        self.clearErrors()
        parameters = {
            "program": program
        }

        response = self.callProceedure(parameters, "Cert.getPublicVendors_5_10")
        return response

    def getMakesByVendorLegacy(self, vendor_name):
        self.clearErrors()
        parameters = {
            "vendor_name": vendor_name
        }

        response = self.callProceedure(parameters, "Cert.getMakesByVendor")
        return response

    def getMakesByVendor(self, vendor_id):
        self.clearErrors()
        parameters = {
            "vendor_id": vendor_id
        }

        response = self.callProceedure(parameters, "Cert.getMakesByVendor_5_7")
        return response

    def createSupplimentalCertification(self, certID):
        """ first added to rhcert for 4.9 (unit tests)"""
        self.clearErrors()
        parameters = {
            "cert_id": certID
        }
        response = self.callProceedure(parameters, "Cert.createSupplementalCert")
        return response

    def createPassThroughSpecification(self, certificationID, parameters):
        """ first added to rhcert for 4.9 (unit tests)"""
        self.clearErrors()
        args = {
             "cert_id": certificationID
        }
        args.update(self.filterNone(parameters))
        response = self.callProceedure(args, "Cert.createPassThroughSpec")
        return response

    def createPassThroughCertification(self, passThroughSpecificationID, originalCertID):
        """ first added to rhcert for 4.9 (unit tests)"""
        self.clearErrors()
        parameters = {
            "product_id": passThroughSpecificationID,
            "original_cert_id": originalCertID
        }

        response = self.callProceedure(parameters, "Cert.createPassThroughCert")
        return response

    """
    /Cert.getHardwareRelatedCerts_4_9:
    get:
      description: Called by RHCERT to get related certs of one hardware cert
      parameters:
        - name: username
          required: true
          type: string
          in: query
        - name: password
          required: true
          type: string
          in: query
        - name: certification_id
          required: true
          description: the id of cert
          type: integer
          in: query
      responses:
        "200":
          description: Success
          schema:
            $ref: '#/definitions/HardwareRelatedCertsDetail'
    """
    def getRelatedCertifications(self, certID):
        """ first added to rhcert for 4.9 (unit tests)"""
        self.clearErrors()
        parameters = {
            "certification_id": certID
        }

        response = self.callProceedure(parameters, "Cert.getRelatedCerts_5_9")

        return response

    def getSubmittedLog(self, cert_nid):
        """ *Deprecated* first added in rhcert 5.0 """
        self.clearErrors()
        parameters = {
            "id": cert_nid
        }

        response = self.callProceedure(parameters, "Cert.getSubmittedLog")

        return response

    def getSubmittedResults(self, certificationID):
        """ first added in rhcert 5.4 """
        self.clearErrors()
        parameters = {
            "cert_id": certificationID
        }

        response = self.callProceedure(parameters, "Cert.listTestResult_5_4")

        return response

    def downloadSubmittedResultsFile(self, certificationID, fileID):
        """ first added in rhcert 5.4 """
        self.clearErrors()
        parameters = {
            "cert_id": certificationID,
            "file_id": fileID
        }

        response = self.callProceedure(parameters, "Cert.downloadTestResult_5_4")

        return response

    def applyComponentLeverage(self, fromCertification, fromComponent, toCertificationID, toComponentID):
        self.clearErrors()
        if fromComponent: # leverage of a component test from a system certification
            print("Apply: from cert %s component %s test %s to cert %s component %s" 
                    % (fromCertification.getCertificationID(), fromComponent.getID(), fromComponent.getTestID(), toCertificationID, toComponentID))
            # fixup bits - we get "32|64bits" from CWE, need to clear off the "bits"
            bits = fromComponent.getBits() if "bits" not in fromComponent.getBits() else fromComponent.getBits()[0:2]
            parameters = {
              "cert_id": toCertificationID,
              "testplan_id": toComponentID,
              "bit": bits, # 32 or 64
              "leverage_type": "system", #  (source is system cert or component cert)
              "leverage_id": fromComponent.getTestID(), # Test ID or Cert ID
            }
        else: # leverage of a component cert pass-through
            print("Apply: from cert %s to cert %s component %s" % (fromCertification.getCertificationID(), toCertificationID, toComponentID))
            parameters = {
              "cert_id": toCertificationID,
              "testplan_id": toComponentID,
              "leverage_type": "component", #  (source is system cert or component cert)
              "leverage_id": fromCertification.getCertificationID(), # Test ID or Cert ID
            }

        response = self.callProceedure(parameters, "Cert.leverageTestPlan_4_10")
        return response

    def submitLeverage(self, leverageType, leverageID, toCertificationID, toComponentID):
        self.clearErrors()
        parameters = {
            "cert_id": toCertificationID,
            "testplan_id": toComponentID,
            "leverage_type": "component",  # leverage of a component cert pass-through
            "leverage_id": leverageID
        }

        if leverageType == "testid":  # leverage of a component test from a system certification
            parameters["leverage_type"] = "system"

        response = self.callProceedure(parameters, "Cert.leverageTestPlan_4_10")
        return response

    def get_banner_message(self):
        self.clearErrors()
        parameters = None
        response = self.callProceedure(parameters, "Cert.getBannerMessage_5_3", login_needed=False)
        return response
