# Copyright (c) 2017 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#

from rhcert.tags import Constants
from rhcert.workflow.workflow import Workflow



class VendorProduct(Workflow):

    def __init__(self, config):
        Workflow.__init__(self, config)


    def list(self):
        self.clearErrors()
        params = {}
        response = self.callProceedure(params, "VendorProduct.list_7_4")
        if self.getErrors():
            self.pushError("Error: could not download vendor products")
        return response

    def createProduct(self, vendor, program, parameters):

        self.clearErrors()

        params = {
            "program": program,
            "vendor": vendor.getName(),
            "vendor_id": vendor.getID()
        }
        # copy form variables into the params
        for (name, value) in parameters.items():
            if value == None:
                continue
            params[name] = value
            if name == "product_name":
                params["model"] = value

        response = self.callProceedure(params, "VendorProduct.create_5_4")
        if not response:
            self.errors.append("Error: could not create vendor product")
            self.errors.append("response: %s" % response)

        return response


    def updateProduct(self, vendor, program, parameters):
        self.clearErrors()

        args = { "vendor": vendor.getName() }

        for (name, value) in parameters.items():
            if value == None:
                continue
            if name == "productID":
                args["id"] = value
            elif name == "supported_region":
                if isinstance(parameters["supported_region"], list):
                    # For updating product of Cloud program,
                    # convert the supported regions from List to unicode if more than regions selected
                    args["supported_regions"] = ','.join(parameters["supported_region"])
                else:
                    args["supported_regions"] = value
            else:
                args[name] = value

        if not args.get("supported_regions"):
            args["supported_regions"] = ''

        response = self.callProceedure(args, "VendorProduct.update")
        return response

    def getLegacy(self, program, vendorID, vendorProductID):
        self.clearErrors()
        parameters = {
            "program": program,
            "id": vendorProductID,
            "vendor_id": vendorID
        }
        response = self.callProceedure(parameters, "VendorProduct.get_4_7")
        return response

    def get(self, program, vendorID, vendorProductID):
        self.clearErrors()
        parameters = {
            "program": program,
            "id": vendorProductID,
            "vendor_id": vendorID
        }
        response = self.callProceedure(parameters, "VendorProduct.get_5_4")
        return response

    def makeProduct(self, vendor, productData):
        """ make a ProductDocument from the response to get() """
        # hardware product?
        if productData.get("model"):
            product = vendor.createProduct(productData.get("model"), productData.get("product_nid"))
            product.setMake(productData.get("make"))
        else: # software
            product = vendor.createProduct(productData.get("product_name"), productData.get("product_nid"))

        if productData.get("bug-id"):
            product.setCertificationID(productData.get("bug-id"))
        product.setNodeID(productData.get("product_nid"))
        product.setProgram(productData.get("program"))
        product.setWorkflow(Constants.cwe)
        product.setBaseProductID(productData.get("base_product_id"))
        product.setPassThroughProductIDs(productData.get("passthrough_product_ids"))
        return product

    def uploadProductLogo(self, program, product_nid, content_type, logo_data):
        self.clearErrors()
        parameters = {
            "program": program,
            "product_nid": product_nid,
            "content_type": content_type,
            "logo_data": logo_data.encode("base64")
        }
        response = self.callProceedure(parameters, "VendorProduct.uploadProductLogo_4_10")
        return response

    def getProductLogoURL(self, program, product_nid):
        self.clearErrors()
        parameters = {
            "program": program,
            "product_nid": product_nid
        } 
        response = self.callProceedure(parameters, "VendorProduct.getProductLogoUrl_4_10")
        return response

    def updateVendorProducts(self, vendorProductsList, vendorProductsDocument, passThroughVendor=False):

        for catalogProduct in vendorProductsList:
            if not catalogProduct.get("vendor_id"):
                self.appendError("Error: no vendor_id set in product from Catalog.")
                continue
            if not catalogProduct.get("product_id"):
                self.appendError("Error: no product_id set in product from Catalog.")
                continue
            try:
                (vendor, product) = vendorProductsDocument.getVendorProductByID(catalogProduct.get("vendor_id"),catalogProduct.get("product_id"))
            except ValueError as error:
                self.appendError("Error: Non-integer ID values in product information")
                self.appendError(str(error))
                continue

            if not vendor:
                # self.log(u"Vendor %s %s not found, creating." % (catalogProduct.get("vendor"), catalogProduct.get("vendor_id")))
                vendor = vendorProductsDocument.createVendor(catalogProduct.get("vendor"), catalogProduct.get("vendor_id"))
                if passThroughVendor:
                    vendor.setPassThough()
                vendor.setWorkflow(Constants.cwe)
            if not product:
                # self.log(u"Product not found, creating.")
                product = vendor.createProduct(catalogProduct.get("product_name"), catalogProduct.get("product_id"))
                product.setNodeID(catalogProduct.get("product_nid"))
                product.setProgram(catalogProduct.get("program"))
                product.setWorkflow(Constants.cwe)
                product.setBaseProductID(catalogProduct.get("base_product_id"))
                product.setPassThroughProductIDs(catalogProduct.get("passthrough_product_ids"))
            for setAttribute, attribute in [(product.setName, "product_name"), (product.setMake, "make"), (product.setCategory, "category")]:
                if catalogProduct.get(attribute):
                    setAttribute(catalogProduct.get(attribute))


class FakeVendor(object):
    def __init__(self, name, id):
        self.name = name
        self.id = id
    def getName(self):
        return self.name
    def getID(self):
        return self.id




