# Copyright (c) 2012 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
# Workflow is the base class for proxies to CWE

import sys
import re
import time
from rhcert.compatability import urlparse, UrllibRequest, UrllibProxyHandler, urllib_build_opener
from rhcert.compatability import XmlrpcTransport, XmlrpcServerProxy, XmlrpcFault, XmlrpcProtocolError
from rhcert.tags import Constants
from rhcert.services.authorization import Login, Refresh
from rhcert.workflow.credentials import WorkflowCredentials

class Workflow(object):

    def __init__(self, config):
        self.config = config
        self.clearErrors()
        self.clearLogin()
        self.retryingRPC = False
        self.tokenSource = None
        self.callTime = 0

    def filterNone(self, parameters):
        notNone = dict()
        for (name, value) in parameters.items():
            if value == None:
                continue
            notNone[name] = value
        return notNone

    def getCurrentUTCTime(self):
        return time.gmtime(time.time())

    def clearErrors(self):
        self.errors = list()

    def getErrors(self):
        return self.errors

    def appendError(self, error):
        self.errors.append(error)

    def pushError(self, error):
        self.errors.insert(0, error)

    def debugPrint(self, message, parameters=None):
        if self.config.debug != Constants.off:
            self.config.log(message, Constants.low)
            if not parameters:
                return
            if type(parameters) is not dict:
                self.config.log("parameters: %s" % parameters, Constants.low)
                return
            for (key, value) in parameters.items():
                if "password" in key or "token" in key:
                    self.config.log("    %s: <-hidden-> (type: %s)" % (key, type(value)), Constants.low)
                    continue
                if not "data" == key:
                    self.config.log("    %s: %s (type: %s)" % (key, value, type(value)), Constants.low)

    def getCallTime(self):
        return self.callTime

    def getServerProxy(self):
        # the serverProxy is just the proxy object to call the XMLRPC server
        access_token = None
        if self.credentials and self.credentials.getAccessToken():
            access_token = self.credentials.getAccessToken()

        # if http proxy is set in rhcert's settings (/etc/rhcert.xml), use a proxy transport
        if self.config.environment.getProxyURLs():
            transport = WorkflowProxyTransport(self.config.environment.getCatalogURL(),
                                               self.config.environment.getProxyURLs(),
                                               access_token,
                                               self.config)
        else:
            url = urlparse(self.config.environment.getCatalogURL())
            https = False
            if url.scheme == "https":
                https = True
            transport = WorkflowTransport(opener=None, access_token=access_token, https=https, config=self.config)

        self.serverProxy = XmlrpcServerProxy(self.config.environment.getCatalogURL(),
                                                 transport=transport,
                                                 allow_none=True)

        return self.serverProxy

    def getServer(self):
        pattern = re.compile("(?P<protocol>[^:]+)://(?P<host>[^/^\?]+)")
        match = pattern.match(self.config.environment.getCatalogURL())
        if match and match.group("host") and match.group("protocol"):
            return "%s://%s" % (match.group("protocol"), match.group("host"))
        # otherwise
        self.errors.append("Error: could not parse server hostname from catalog URL \"%s\"" % self.config.environment.getCatalogURL())
        return None

    def isReachable(self):
        return self.ping()

    def ping(self):
        self.clearErrors()
        params = {
        }
        self.debugPrint("Calling Util.ping:", params)
        response = self.call(params, self.getServerProxy().Util.ping, auth=False)
        if self.getErrors():
            self.pushError("Error: could not reach workflow server at %s:" % self.config.environment.getCatalogURL())
        return (response == "pong")

    def clearLogin(self):
        self.credentials = None
        self.serverProxy = None

    def setLogin(self, tokenSource):
        """ tokens is a dict of auth tokens
        """
        self.credentials = WorkflowCredentials(tokenSource)
        self.tokenSource = tokenSource

    def getCredentials(self):
        if not self.credentials or not self.credentials.isLoggedIn():
            return {}
        return self.credentials.getParameters()

    def getAccessCredentials(self):
        if not self.credentials.isLoggedIn():
            raise Exception("login required")
        usernameAndAccessToken = self.getCredentials()
        usernameAndAccessToken.pop("refresh_token")
        return usernameAndAccessToken

    def updateCredentials(self, tokens):
        if not isinstance(tokens, dict):
            return False
        if not self.credentials:
            self.credentials = WorkflowCredentials(tokens)
        else:
            self.credentials.update(tokens)
            try:
                self.tokenSource["access_token"] = self.credentials.access_token
            except Exception as error:
                self.debugPrint("Error: attempting access_token refresh to session")
                self.debugPrint(str(error))
        return True

    def isLoggedIn(self):
        if not self.credentials:
            return False
        return self.credentials.isLoggedIn()

    def login(self, username, password):
        # try JWT first
        try:
            if self.loginJWT(username, password):
                return True
            # otherwise
            self.debugPrint("loginJWT failed - returned False")

        except Exception as error:
            self.debugPrint("loginJWT failed: %s" % error)

        return False

    def loginJWT(self, username, password):
        self.clearLogin()
        self.debugPrint("Calling loginJWT for %s" % username)
        login = Login(username, password, self.config.environment)

        loginResponse = login.call()
        if login.getError():
            self.errors.append(login.getError())
            return False
        elif not loginResponse:
            self.errors.append("Error: empty login response")
            return False
        if loginResponse.status_code == 200:
            tokens = {"username": username}
            tokens.update(login.getTokens())
            self.setLogin(tokens)
            return self.isLoggedIn()

        elif loginResponse.status_code == 401:
            self.errors.append('Error: Authentication Failed')
            return False

        # NOTE: Generic fallback in case we get any other 4XX or 5XX
        else:
            message = str(loginResponse.status_code) + ':' + str(loginResponse.message)
            self.errors.append(message)
            return False

    def refreshJWT(self):
        self.debugPrint("Calling refreshJWT")
        if not self.credentials:
            self.debugPrint("Not logged in: no refresh")
            return False
        if not self.credentials.getRefreshToken():
            self.debugPrint("No refresh token")
            return False
        refresh = Refresh(self.credentials.getRefreshToken(), self.config.environment)
        if refresh.call():
            self.debugPrint("Success:", refresh.getTokens())
            if self.updateCredentials(refresh.getTokens()):
                return True

        self.debugPrint("Could no refresh token - clearing login!")
        self.clearLogin()
        return False

    # This method has been deprecated with CERTPX-4771
    def logout(self, username):
        try:
            self.getServerProxy().User.logout({"username": username})
            self.clearLogin()
            return True

        except XmlrpcFault as error:
            self.errors.append("Error: logout failed:")
            self.handleRPCException(error)
        except XmlrpcProtocolError as error:
            self.errors.append("Error: logout failed:")
            self.errors.append(" code: %s" % error.errcode)
            self.errors.append(" error: %s" % error.errmsg)
            self.errors.append(" message: %s" % error.message)
            self.errors.append(" url: %s" % error.url)
        except Exception as e:
            self.errors.append("Error: logout failed: (%s)" % e)

        return False

    def handleRPCException(self, rpcException):
        """ determine error message and action - returns True if call should retry """
        if rpcException.faultCode == 300: # bad login/password
            self.errors.append("Error: login failed:")
            self.clearLogin()
            return False
        if rpcException.faultCode == 401:
            if not self.retryingRPC: # access token timeout
                self.debugPrint("refresh  timeout, retrying: %s" % rpcException.faultString)
                if self.tryRefeshJWT():
                    self.debugPrint("got a new token re-trying call")
                    self.retryingRPC = True
                    return True
        errorMessage = "Error: %s (%s)" % (rpcException.faultString, rpcException.faultCode)
        self.errors.append(errorMessage)
        self.debugPrint(errorMessage)
        return False

    def handleProxyException(self, workflowProxyException):
        """ determine error message and action - returns True if call should retry """
        if hasattr(workflowProxyException.error, "faultCode") and workflowProxyException.error.faultCode == 401:
            if not self.retryingRPC: # access token timeout
                self.debugPrint("proxy refresh timeout, retrying: %s" % workflowProxyException.error)
                if self.tryRefeshJWT():
                    self.debugPrint("got a new token re-trying call")
                    self.retryingRPC = True
                    return True
        elif hasattr(workflowProxyException.error, "faultString") and hasattr(workflowProxyException.error, "faultCode"):
            self.errors.append("Error: %s (%s)" % (workflowProxyException.error.faultString, workflowProxyException.error.faultCode))
            self.debugPrint("Error: %s (%s)" % (workflowProxyException.error.faultString, workflowProxyException.error.faultCode))
        else:
            self.errors.append(workflowProxyException.error.__repr__())
            self.debugPrint("WorkflowProxyException: %s" % workflowProxyException.error.__repr__())
        return False

    def call(self, params, xmlRPC, auth=True):
        """ generic wrapper to catch XMLRPC exceptions"""
        Munger.undo(params)
        startTime = time.time()
        try:
            if auth:
                params.update(self.getAccessCredentials())
            startTime = time.time()
            response = xmlRPC(params)
            Munger.do(response)
            self.retryingRPC = False
            return response

        except XmlrpcFault as error:
            if self.handleRPCException(error):
                # retry
                return self.call(params, xmlRPC)
        except WorkflowProxyException as error:
            if self.handleProxyException(error):
                # retry
                return self.call(params, xmlRPC)
        except XmlrpcProtocolError as error:
            self.errors.append("CWE URL: %s" % self.config.environment.getCatalogURL())
            self.errors.append("Error code: %s" % error.errcode)
            self.errors.append("Error message: %s" % error.errmsg)
        except Exception as error:
            self.errors.append("Error: %s" % error)

        finally:
            self.debugPrint("CWE call time: %.1f sec." % (time.time()-startTime))

        return None

    def callProceedure(self, params, xmlRPCName, login_needed=True):
        """ generic wrapper to catch XMLRPC exceptions
            - includes debug print wrappers so you don't have to."""
        Munger.undo(params)
        startTime = time.time()
        try:
            if login_needed:
                params.update(self.getAccessCredentials())
            self.debugPrint("\nRHCert Calling %s" % xmlRPCName, params)
            proceedure = getattr(self.getServerProxy(), xmlRPCName)
            startTime = time.time()
            if params:
                response = proceedure(params)
            else:
                response = proceedure()
            Munger.do(response)
            self.retryingRPC = False
            if response:
                self.debugPrint("CWE response:", response)
            else:
                self.debugPrint("CWE response is empty")

            if response and type(response) is dict and 'resp_errors' in response.keys():
                for error in response['resp_errors']:
                    self.debugPrint("Error in response: %s" % error)
                    self.errors.append(error)
            try:
                return response['resp_data']
            except (TypeError, KeyError):
                return response

        except XmlrpcFault as error:
            if self.handleRPCException(error):
                # retry
                return self.callProceedure(params, xmlRPCName)
        except WorkflowProxyException as error:
            if self.handleProxyException(error):
                # retry
                return self.callProceedure(params, xmlRPCName)
        except XmlrpcProtocolError as error:
            self.errors.append("CWE URL: %s" % self.config.environment.getCatalogURL())
            self.errors.append("Error code: %s" % error.errcode)
            self.errors.append("Error message: %s" % error.errmsg)
            self.debugPrint("Error code: %s" % error.errcode)
            self.debugPrint("Error message: %s" % error.errmsg)
        except Exception as error:
            self.errors.append("Error: %s" % error)
            self.debugPrint("CWE error: %s" % error)
        finally:
            self.callTime = time.time()-startTime
            self.debugPrint("CWE call time: %s %.1f sec." % (xmlRPCName, self.callTime))

        return None

    def tryRefeshJWT(self):
        if self.credentials:
            try:
                return self.refreshJWT()
            except Exception as error:
                self.errors.append("Error: Could not refresh token: %s" % error)
        return False


class WorkflowTransport(XmlrpcTransport):
    def __init__(self, opener, https,  access_token, config):
        XmlrpcTransport.__init__(self, use_datetime=0)
        self.opener = opener or urllib_build_opener()
        self.https = https
        self.config = config
        self.access_token = access_token

    def request(self, host, handler, request_body, verbose=0):
        proto = ('http', 'https')[bool(self.https)]
        req = UrllibRequest('%s://%s%s' % (proto, host, handler), request_body)
        self.config.log("request: {0}://{1}{2}".format(proto, host, handler), Constants.low)
        self.config.log("body:", Constants.high)
        self.config.log(request_body, Constants.high)
        req.add_header('User-agent', self.user_agent)
        req.add_header('Content-Type', "text/xml")
        if self.access_token:
            req.add_header("Authorization", "Bearer %s" % self.access_token)
        self.verbose = verbose
        try:
            response = self.parse_response(self.opener.open(req))
        except Exception as error:
            message = "Error: raising WorkflowProxyError for %s  on %s://%s%s" % (error, proto, host, handler)
            print(message)
            self.config.log(message)
            raise WorkflowProxyException(error, '%s://%s%s' % (proto, host, handler))

        self.config.log("response:", Constants.high)
        self.config.log(response, Constants.high)
        return response


class WorkflowProxyTransport(WorkflowTransport):
    def __init__(self, catalogURL, proxies,  access_token, config):
        if config.debug != Constants.off:
            config.log("Contacting catalog via proxy:", Constants.low)
            for protocol, url in proxies.items():
                config.log("proxy: %s %s" % (protocol, url), Constants.low)
        url = urlparse(catalogURL)
        https = False
        if url.scheme == "https":
            https = True
        opener = urllib_build_opener(UrllibProxyHandler(proxies))
        WorkflowTransport.__init__(self, opener, https, access_token, config)

class WorkflowProxyException(Exception):
    def __init__(self, error, url):
        self.error = error
        self.url = url

    def __str__(self):
        return "Workflow HTTP Proxy exception: %s for request %s" % (self.error, self.url)

class Munger(object):
    """ Munger is a test class for inserting strings in ID values to test support for non-integer ID strings  """
    mungeIDParameters = [
        "id",
        "certification_id",
        "cert_id",
        "certid",
        "cert_nid",
        "product_id",
        "product_nid",
        "vendor_product_id",
        "vendor_product_portal_id",
        "vendor_id",
        "vendor_nid",
    ]
    mungeString = "mXXXm-"

    @staticmethod
    def do(parameters):
        # disabled
        return
        try:
            if not parameters:
                return
            if type(parameters) == list:
                for item in parameters:
                    Munger.do(item)
            elif type(parameters) == dict:
                for (parameter, value) in parameters.items():
                    if type(value) in [list, dict]:
                        Munger.do(value)
                        continue
                    if parameter in Munger.mungeIDParameters and not str(value).startswith(Munger.mungeString) :
                        parameters[parameter] = Munger.mungeString + str(value)
        except Exception as error:
            print("Munger Error: ")
            print(error)
            print("type(parameters): %s" % type(parameters))

    @staticmethod
    def undo(parameters):
        # disabled
        return
        try:
            if not parameters:
                return
            if type(parameters) == list:
                for item in parameters:
                    Munger.undo(item)
            elif type(parameters) == dict:
                for (parameter, value) in parameters.items():
                    if type(value) in [list, dict]:
                        Munger.undo(value)
                        continue
                    if parameter in Munger.mungeIDParameters and str(value).startswith(Munger.mungeString):
                        parameters[parameter] = parameters[parameter][len(Munger.mungeString):]
                        print("XXX set to %s" % parameters[parameter])
                        try:
                            parameters[parameter] = int(parameters[parameter])
                            print("XXX set to %s" % type(parameters[parameter]))
                        except:
                            pass
        except Exception as error:
            print("Munger Error: ")
            print(error)
            print("type(parameters): %s" % type(parameters))
